#ifndef AVCODEC_H
#define AVCODEC_H

/**
 * @file avcodec.h
 * external api header.
 */

#include "common.h"
#include "Types.h"
#include "Tf.h"
#include "ImageProcessingApi.h"

#ifdef __cplusplus
extern "C" {
#endif

#define LIBAVCODEC_VERSION     "0.4.6"

///////////////////////////////////////////////////////////////////////////////
// Stream errors
///////////////////////////////////////////////////////////////////////////////
 
#define AVCODEC_ENCODER_SUCCESS                 0

#define AVCODEC_DECODER_SUCCESS                 0
#define AVCODEC_DECODER_FATAL_ERROR             1
#define AVCODEC_DECODER_UNSUPPORTED_STREAM      2
#define AVCODEC_DECODER_VIDEO_STREAM_ERROR      3
#define AVCODEC_DECODER_SLOW_RATE               4
#define AVCODEC_DECODER_AUDIO_STREAM_ERROR      5

///////////////////////////////////////////////////////////////////////////////
// Supported codecs
///////////////////////////////////////////////////////////////////////////////

// :ZZZ: The unused codec Id's are not removed yet, because they are used by Avf.

enum CodecID {
    CODEC_ID_NONE, 
    CODEC_ID_MPEG1VIDEO,
    CODEC_ID_H263,
    CODEC_ID_RV10,
    CODEC_ID_MP2,
    CODEC_ID_MP3LAME,
    CODEC_ID_VORBIS,
    CODEC_ID_AC3,
    CODEC_ID_MJPEG,
    CODEC_ID_MJPEGB,
    CODEC_ID_MPEG4,
    // :ZZZ: HW-based JPEG codec
    CODEC_ID_JPEGZ,
    // :ZZZ: C7 codec
    CODEC_ID_MPEG4Z,
    // :ZZZ: C8 codec
    CODEC_ID_MPEG2Z,
    // :ZZZ: C8 codec
    CODEC_ID_WMVZ,    
    CODEC_ID_WMVMZ,    
    // :ZZZ: C10 codec
    CODEC_ID_H264Z,
    CODEC_ID_MPEG1Z,

    CODEC_ID_RAWVIDEO,
    CODEC_ID_MSMPEG4V1,
    CODEC_ID_MSMPEG4V2,
    CODEC_ID_MSMPEG4V3,
    CODEC_ID_WMV1,
    CODEC_ID_WMV2,
    CODEC_ID_H263P,
    CODEC_ID_H263I,
    CODEC_ID_SVQ1,
    CODEC_ID_SVQ3,
    CODEC_ID_DVVIDEO,
    CODEC_ID_DVAUDIO,
    CODEC_ID_MACE3,
    CODEC_ID_MACE6,
    CODEC_ID_HUFFYUV,
    CODEC_ID_CYUV,
    CODEC_ID_H264,
    CODEC_ID_INDEO3,
    CODEC_ID_VP3,
    CODEC_ID_AAC,
    CODEC_ID_MPEG4AAC,
    // :ZZZ:
    CODEC_ID_BYPASS,
    CODEC_ID_RAWZ,
    CODEC_ID_WMAV8Z,
    CODEC_ID_AACZ,

    /* various pcm "codecs" */
    CODEC_ID_PCM_S16LE,
    CODEC_ID_PCM_S16BE,
    CODEC_ID_PCM_U16LE,
    CODEC_ID_PCM_U16BE,
    CODEC_ID_PCM_S8,
    CODEC_ID_PCM_U8,
    CODEC_ID_PCM_MULAW,
    CODEC_ID_PCM_ALAW,

    /* various adpcm codecs */
    CODEC_ID_ADPCM_IMA_QT,
    CODEC_ID_ADPCM_IMA_WAV,
    CODEC_ID_ADPCM_MS,
    CODEC_ID_ADPCM_4XM,
    CODEC_ID_ADPCM_G726,
    CODEC_ID_ADPCM_IMA_DK4,
    CODEC_ID_ADPCM_IMA_DK3,
    CODEC_ID_ADPCM_IMA_WS,
    CODEC_ID_ADPCM_XA,
    CODEC_ID_ADPCM_EA,
    CODEC_ID_ADPCM_IMA_SMJPEG,
    CODEC_ID_ADPCM_ADX,

	/* AMR */
    CODEC_ID_AMR_NB,
 };

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

enum CodecType {
    CODEC_TYPE_UNKNOWN = -1,
    CODEC_TYPE_VIDEO,
    CODEC_TYPE_AUDIO,
    CODEC_TYPE_TEXT,
};

///////////////////////////////////////////////////////////////////////////////
// Pixel format
///////////////////////////////////////////////////////////////////////////////

enum PixelFormat
{
    PIX_FMT_YUV420P, // Y, U, V plane
    PIX_FMT_Y,
    PIX_FMT_YUV422Z,   ///< YUYV
    PIX_FMT_YUV420Z,   // Y plane YU interleave UVUVUVUV 
    PIX_FMT_RAW8Z,     ///< 8 bit Bayer grid data
    PIX_FMT_RAW12Z,     ///< 12 bit Bayer grid data    
    PIX_FMT_RAW16Z,    ///< 16 bit Bayer grid data
    PIX_FMT_RGB24,     ///< 3 bytes, R is first 
    PIX_FMT_BGR24,     ///< 3 bytes, B is first 
    PIX_FMT_YUV444,    ///< 3 bytes interleaved YUVYUVYUV
    PIX_FMT_YUV422P,
    PIX_FMT_YUV444P,
    PIX_FMT_YUVJ420P,  ///< YUV full scale (jpeg)
    PIX_FMT_YUVJ422P,  ///< YUV full scale (jpeg)
    PIX_FMT_YUVJ444P,  ///< YUV full scale (jpeg)
    PIX_FMT_NB,
};
typedef enum PixelFormat PixelFormat;



typedef enum
{
    CODEC_OM_REAL_TIME,
    CODEC_OM_TRANSCODING,

} ECodecOperationMode;


///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

// in bytes 
#define AVCODEC_MAX_AUDIO_FRAME_SIZE 131072


// Required number of zero bytes at the end of the input bitstream for decoding.
// to avoid overreading (and possibly segfaulting)
#define FF_INPUT_BUFFER_PADDING_SIZE 256

///////////////////////////////////////////////////////////////////////////////
// Those flags can be passed in AVCodecContext.flags before initing 
///////////////////////////////////////////////////////////////////////////////

// No header will be generated
#define CODEC_FLAG_NO_HEADERS                   0x00000001

// Disable maching criteria mechanism
#define CODEC_FLAG_NO_MC                        0x00000002

// Quantizer can not be changed within the frame (no d-quant)
#define CODEC_FLAG_NO_DQUANT                    0x00000004

// Disable skip option for P-frames which are used as backward
// reference for B-frames.
#define CODEC_FLAG_BACKWARD_REF_DISABLE_SKIP    0x00000008

// Const type assignment will be used
#define CODEC_FLAG_CONST_TYPE                   0x00000010

// If this flag is set, I frame will be set at each GOP size interval
#define CODEC_FLAG_FORCE_I_GOP                  0x00000020

// B frame can not be last in gop (I cannot be backward reference of B)
#define CODEC_FLAG_B_FRAME_NOT_LAST_IN_GOP      0x00000040

// Notify if direct mode is enabled. Used for mpeg4.
#define CODEC_FLAG_DIRECT_MODE                  0x00000080

// Notify if SAD is available for each 2 consecutive frames. 
#define CODEC_FLAG_CLOSE_FRAMES_SAD_EXIST       0x00000100

#define CODEC_FLAG_BRC_TOOL_LPF_ENABLE          0x00000200  

// Notify if unrestricted search is enabled.
#define CODEC_FLAG_UNRESTRICTED_ENABLE          0x00000400

// Notify if 4mv encoding is enabled.
#define CODEC_FLAG_4MV_ENABLE                   0x00000800

// Notify if CABAC or VLC is used.
#define CODEC_FLAG_CABAC                        0x00001000

// Notify if Deblocking is used.
#define CODEC_FLAG_DEBLOCKING                   0x00002000

// Header changes for MP4 to pass checker tools
#define CODEC_FLAG_MP4_SUPPORT                  0x00004000

// Vui present flag
#define CODEC_FLAG_VUI_PRESENT_FLAG             0x00008000


// input bitstream might be truncated at a random location instead of only
// at frame boundaries  
#define CODEC_FLAG_TRUNCATED                    0x00010000

// Notify Normal DRAM Access Optimization Level
#define CODEC_FLAG_NORMAL_OPTIMIZATION          0x00020000

// Notify Max DRAM Access Optimization Level
#define CODEC_FLAG_MAX_OPTIMIZATION             0x00040000

// force low delay / will fail on b frames 
#define CODEC_FLAG_LOW_DELAY                    0x00080000

// Disable maching criteria mechanism
#define CODEC_FLAG_NO_MC_FS                     0x00100000

#define CODEC_FLAG_TF_ENABLE					0x00200000

// place global headers in extradata instead of every keyframe 
#define CODEC_FLAG_GLOBAL_HEADER                0x00400000

// Select pixel format as yuv420 for decoder
#define CODEC_FLAG_PIX_FORMAT_YUV420            0x00800000

//
#define CODEC_FLAG_I_GOP_FRAME_PRESERVE         0x01000000

//
#define CODEC_FLAG_CHECK_FRAME_END              0x02000000

#define CODEC_FLAG_WITHOUT_ZERO_AFTER_USER_DATA 0x04000000

#define CODEC_FLAG_NO_PADDING_AFTER_USER_DATA   0x08000000

// Notify codec to work in internal dram mode
#define CODEC_FLAG_INTERNAL_DRAM_MODE           0x10000000

// Notify whether Y/U/V range convertion is enabled
#define CODEC_FLAG_YUV_RANGE_CONVERSION         0x20000000

// Nothify codec that we record in interlaced mode
#define CODEC_FLAG_INTERLACED_MODE              0x40000000

// :ZZZ: Only VOP header is coded
#define CODEC_FLAG_ONLY_VOP                     0x80000000

///////////////////////////////////////////////////////////////////////////////
// 
///////////////////////////////////////////////////////////////////////////////

#define CODEC_CAP_TRUNCATED       0x0008

///////////////////////////////////////////////////////////////////////////////
// Frame types
///////////////////////////////////////////////////////////////////////////////

#define FF_I_TYPE           1 // Intra
#define FF_P_TYPE           2 // Predicted
#define FF_B_TYPE           3 // Bi-dir predicted
#define FF_SAMPLE_I4_TYPE_0 ( 0xF0 )
#define FF_SAMPLE_I4_TYPE_1 ( FF_SAMPLE_I4_TYPE_0 + 1 )
#define FF_SAMPLE_I4_TYPE_2 ( FF_SAMPLE_I4_TYPE_0 + 2 )
#define FF_SAMPLE_I4_TYPE_3 ( FF_SAMPLE_I4_TYPE_0 + 3 )

#define FF_SAMPLE_P_TYPE            0xFD // For Initial Quantiser, special frame, encoded as P/B
#define FF_SAMPLE_I_TYPE            0xFC // For Initial Quantiser, special frame, encoded as I
#define FF_SAMPLE_I_TYPE_SIMPLE     0xFB // For Initial Quantiser, special frame, encoded as I                                 

#define FF_B_TYPE_SINGLE    0    // Single B frame
#define FF_B_TYPE_DOUBLE_1  1    // 1st B frame out of 2
#define FF_B_TYPE_DOUBLE_2  2    // 2nd B frame out of 2

///////////////////////////////////////////////////////////////////////////////
// Aspect ratio
///////////////////////////////////////////////////////////////////////////////

#define FF_ASPECT_SQUARE 1
#define FF_ASPECT_4_3_625 2
#define FF_ASPECT_4_3_525 3
#define FF_ASPECT_16_9_625 4
#define FF_ASPECT_16_9_525 5
#define FF_ASPECT_EXTENDED 15

///////////////////////////////////////////////////////////////////////////////
// brc_mode - brc possible modes.
///////////////////////////////////////////////////////////////////////////////

#define AV_BRC_MODE_V1                  1
#define AV_BRC_MODE_CONST_QUALITY       2
#define AV_BRC_MODE_CONST_QSCALE        3
#define AV_BRC_MODE_CBR                 4


///////////////////////////////////////////////////////////////////////////////
// brc_profile - brc profiles, Controls the tradeoff between bitrate convergence 
// and video quality in transition between complex/simlpe scenes
///////////////////////////////////////////////////////////////////////////////

#define AV_BRC_PROFILE_NORMAL               0
#define AV_BRC_PROFILE_FAST_TRANSITION_C2S  1
#define AV_BRC_PROFILE_FAST_TRANSITION_ALL  2

///////////////////////////////////////////////////////////////////////////////
// block_buffer - buffers with variable block geometry
///////////////////////////////////////////////////////////////////////////////

#define AV_BLOCK_BUFFER_IMAGE           0
#define AV_BLOCK_BUFFER_IMAGE_C         1
#define AV_BLOCK_BUFFER_Y               2
#define AV_BLOCK_BUFFER_C               3
#define AV_BLOCK_BUFFER_NUM             4

/////////////////////////////////////////////////////
// H264 Profiles 
/////////////////////////////////////////////////////

#define H264_MAIN_PROFILE        77
#define H264_BASELINE_PROFILE    66

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////
#define FF_COMMON_FRAME \
    /**\
     * pointer to the picture planes.\
     * this might be different from the first allocated byte\
     * - encoding: \
     * - decoding: \
     */\
    uint8_t *data[10];\
    int linesize[10];\
    /**\
     * pointer to the first allocated byte of the picture. can be used in get_buffer/release_buffer\
     * this isnt used by lavc unless the default get/release_buffer() is used\
     * - encoding: \
     * - decoding: \
     */\
    uint8_t *base[10];\
    /**\
     * 1 -> keyframe, 0-> not\
     * - encoding: set by lavc\
     * - decoding: set by lavc\
     */\
    int key_frame;\
\
    /**\
     * picture type of the frame, see ?_TYPE below.\
     * - encoding: set by lavc for coded_picture (and set by user for input)\
     * - decoding: set by lavc\
     */\
    int pict_type;\
\
    /**\
     * frame B type of the frame, Used for encoder decision.\
     */\
    int frame_b_type;\
\
    /**\
     * original picture type of the frame, see ?_TYPE below.\
     * Different then pict_type for initial quantizer frames.\
     * - encoding: set by lavc for coded_picture (and set by user for input)\
     * - decoding: set by lavc\
     */\
    int original_pict_type;\
\
    /**\
     * presentation timestamp in micro seconds (time when frame should be shown to user)\
     * if 0 then the frame_rate will be used as reference\
     * - encoding: MUST be set by user\
     * - decoding: set by lavc\
     */\
    int64_t pts;\
\
    /**\
     * picture number in bitstream order.\
     * - encoding: set by\
     * - decoding: set by lavc\
     */\
    int coded_picture_number;\
    /**\
     * picture number in display order.\
     * - encoding: set by\
     * - decoding: set by lavc\
     */\
    int display_picture_number;\
    /**\
     * is this picture used as reference\
     * - encoding: unused\
     * - decoding: set by lavc (before get_buffer() call))\
     */\
    int reference;\
\
    /**\
     * frame average acitivity. Used for bit rate control.\
     */\
    int activity;\
    \
    /**\
     * frame accumulated Qscale (on line base). Used for bit rate control.\
     */\
    int acc_qscale;\
    \
    /**\
     * Maximum motion estimation search range in pixel units. The search size\
     * is for single side of macro block.\
    */\
    int me_search_x;\
    int me_search_y;\
    \
    /**\    
     * Scene cut flag: 1 if scene cut occured, 0 otherwise\
    */\
    int scene_cut;\
    \
    /**\    
     * Last frame in gop flag: 1 if last, 0 otherwise\
    */\
    int last_in_gop;\
    \
    /**\
     * frame window luma
    */\
    int frame_width;\
    int frame_height;\
    int left;\
    int top;\
    int width;\
    int height;\
    \
    /**\
     * frame window chroma
    */\
    int chroma_frame_width;\
    int chroma_frame_height;\
    int chroma_left;\
    int chroma_top;\
    int chroma_width;\
    int chroma_height;\
    \
    /**\
     * Level of brc tools
     */\
    int brc_tool_level;\
    \
    /**\
      * Skipped frame flag
      */\
    int skip_frame;\
    \
    /**\
      * Lfp index
      */\
    int lpf_index;\
    \
    /**\
      * Sad of 2 close frames
      */\
    int close_frames_sad;\
    \
    /**\
      * Length of extra headers
      */\
    int extra_header_length;\
	\
	/**\
	  * If frame was originally skipped
	  */\
	int originally_skipped;\
	\
	/**\
	  * reference context
	  */\
	void* ref_context;\
    \
    /**\
      * buf_allocated
      */\
    int buf_allocated;\
	\
	/**\
	  * flags
	  */\
	int flags;\
    \
   	/**\
	  * timestamp
	  */\
	int64_t timestamp;\
    \
    /**\
      * duration
      */\
    unsigned int duration;\
    \
    /**\
      * segment
      */\
    unsigned int segment;\
    \
	/**\
	  * code buffer
	  */\
	unsigned char *code_buf;\
    \
    /**\
      * Current frame is IDR frame
      */\
    int idr_frame;\
    \
    /**\
      * nal_ref_idc
      */\
    int nal_ref_idc;\
    \
    /**\
      * frame_num
      */\
    int frame_num;\
    \
    /**\
      * cpb_removal_delay
      */\
    int cpb_removal_delay;\
	\
    /**\
      * prevention_count
      */\
    int prevention_count;\
	\
	/**\
	  * replace_by_skip
	  */\
	int replace_by_skip;\
	\
	/**\
	  * Gop start
	  */\
	int gop_start;\
	\
	/**\
	  * IDR SEI message written for this frame?
	  */\
	int idr_sei_message_written;\
    \
    /**\ 
    *for color conversion on encoding 
    */\
	int target_ratio_called;\
    \
    SIPFCCOMPLETERC12 sFccData;\
    PutBitContext pb;\
	\
	/**\TF Data Structure*/\
	STFCONFIGTABLE tf_table;\
    SHPCONFIGTABLE hp_table;\
    SMAGNETTABLE magnet_table;\
    \
    int last_in_encoded_gop;\
    int snapshot;\
    int start_of_last_gop;\
    int code_buf_size;\
    int use_gmv; \
    int gmv_x; \
    int gmv_y; \
    int gmv_fwd_x; \
    int gmv_fwd_y; \
    int gmv_bwd_x; \
    int gmv_bwd_y; \
    unsigned int gmv_confidence; \
    unsigned int gmv_skip_cost_ratio; \
    \
    bool dummy_gop;\
    \
    /* frame in gop in display order*/ \
    int frame_in_gop;\


/**
 * Frame block buffers
 */
typedef struct AVBlockBuf {
    int frame_width;
    int frame_height;
    int width;
    int height;
    int left;
    int top;
    int bpp;
    int size;
    int alignment;
} AVBlockBuf;


/**
 * Audio Video Frame.
 */
typedef struct AVFrame {
    FF_COMMON_FRAME;
	struct AVFrame *linked_frame_ptr;
	struct AVFrame *pic_arg;
} AVFrame;


/**
 * four components are given, that's all.
 * the last component is alpha
 */
typedef struct AVPicture
{
    uint8_t *data[10];
    int linesize[10];
} AVPicture;

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

// :ZZZ:
typedef struct tagSMBINFO
{    
    // Indicates that skip can not be applied for current macro block
    // of the final stage. Relevant for P frame only.
    USHORT uwNoSkip;

    USHORT uwReserved;

}   SMBINFO, *PSMBINFO;

///////////////////////////////////////////////////////////////////////////////
// main external api structure
///////////////////////////////////////////////////////////////////////////////

// 
// Debug alternatives.

#define FF_DEBUG_PICT_INFO 1
#define FF_DEBUG_BITSTREAM 4
#define FF_DEBUG_MB_TYPE   8
#define FF_DEBUG_QP        16
#define FF_DEBUG_MV        32
#define FF_DEBUG_SKIP      0x00000080
#define FF_DEBUG_STARTCODE 0x00000100
#define FF_DEBUG_PTS       0x00000200

// :ZZZ: New visuality
#define FF_DEBUG_VIS_MB_TYPE    0x80000000

typedef int (*AVErrorCallbackFnc) ( int err_id );
typedef struct AVCodecContext {

    //
    // Common
    //     

    struct AVCodec *codec;
    
    void *priv_data;

    /**
     * CODEC_FLAG_*.
     * - encoding: set by user.
     * - decoding: set by user.
     */
    int flags;

    /**
     * some codecs need / can use extra-data like huffman tables.
     * mjpeg: huffman tables
     * rv10: additional flags
     * mpeg4: global headers (they can be in the bitstream or here)
     * - encoding: set/allocated/freed by lavc.
     * - decoding: set/allocated/freed by user.
     */
    void *extradata;
    int extradata_size;

    char codec_name[32];
    enum CodecType codec_type; /* see CODEC_TYPE_xxx */
    enum CodecID codec_id; /* see CODEC_ID_xxx */
    
    /**
     * fourcc (LSB first, so "ABCD" -> ('D'<<24) + ('C'<<16) + ('B'<<8) + 'A').
     * this is used to workaround some encoder bugs
     * - encoding: unused
     * - decoding: set by user, will be converted to upper case by lavc during init
     */
    unsigned int codec_tag;

    /**
     * internal_buffers. 
     * Dont touch, used by lavc default_get_buffer()
     */
    void *internal_buffer;

    /**
     * 
     * 
     */
    void *hPool;

    /**
     * called at the beginning of each frame to get a buffer for it.
     * if pic.reference is set then the frame will be read later by lavc
     * - encoding: unused
     * - decoding: set by lavc, user can override
     */
    int (*get_buffer)(struct AVCodecContext *c, AVFrame *pic);
    
    /**
     * called to release buffers which where allocated with get_buffer.
     * a released buffer can be reused in get_buffer()
     * pic.data[*] must be set to NULL
     * - encoding: unused
     * - decoding: set by lavc, user can override
     */
    void (*release_buffer)(struct AVCodecContext *c, AVFrame *pic);

    /**
     * the picture in the bitstream.
     * - encoding: set by lavc
     * - decoding: set by lavc
     */
    AVFrame *coded_frame;

    /**
     * the average bitrate.
     * - encoding: set by user. unused for constant quantizer encoding
     * - decoding: set by lavc. 0 or some bitrate if this info is available in the stream 
     */
    int bit_rate;

    /**
     * :ZZZ: compration ratio for clip
     * - encoding: set by user. 
     * - decoding: not in use 
     */
    int cmp_ratio;

    //
    // Audio only
    //     
            
    /* audio only */
    int sample_rate; ///< samples per sec 
    int sample_rate_in; ///< Input sample rate for encoders that support resampling
    int channels;
    int channels_in;
    int sample_fmt;  ///< sample format, currenly unused 

    /* the following data should not be initialized */
    int frame_size;     ///< in samples, initialized when calling 'init' 
    int frame_number;   ///< audio or video frame number 
    
    /*
     * used by some WAV based audio codecs 
    */
    int block_align;        

    //
    // Video only
    //     

    /**
     * pixel format, see PIX_FMT_xxx.
     * - encoding: unused
     * - decoding: set by lavc.
     */
    PixelFormat pix_fmt;

    /* video only */
    /**
     * frames per sec multiplied by frame_rate_base.
     * for variable fps this is the precission, so if the timestamps 
     * can be specified in msec precssion then this is 1000*frame_rate_base
     * - encoding: MUST be set by user
     * - decoding: set by lavc. 0 or the frame_rate if available
     */
    int frame_rate;
    
    /**
     * width / height.
     * - encoding: MUST be set by user. 
     * - decoding: set by user, some codecs might override / change it during playback
     */
    int physical_width;
    int physical_height;

    int logical_width;
    int logical_height;

    // Block buffers geometry 
    AVBlockBuf block_buf[AV_BLOCK_BUFFER_NUM];

    /**
     * the number of pictures in a group of pitures, or 0 for intra_only.
     * - encoding: set by user.
     * - decoding: unused
     */
    int gop_size;
        
    /**
     * the number of pictures in the first group of pitures. 
     * Needed for intialize TF algorithm. 
     * - encoding: set by user.
     * - decoding: unused
     */
    int dummy_gop_size;

    /**
     * maximum number of b frames between non b frames.
     * note: the output will be delayed by max_b_frames+1 relative to the input
     * - encoding: set by user.
     * - decoding: unused
     */
    int max_b_frames;
    
    /**
     * hurry up amount.
     * - encoding: unused
     * - decoding: set by user. 1-> skip b frames, 2-> skip idct/dequant too, 5-> skip everything except header
     */
    int hurry_up;

     /**
     * :ZZZ:
     * Maximum motion estimation search range in pixel units. The search size
     * is for single side of macro block.
     * 
     * - encoding: set by user.
     * - decoding: unused.
     */
    int me_p_search_x;
    int me_p_search_y;
    int me_b_search_x;
    int me_b_search_y;
    int me_b_near_search_x;
    int me_b_near_search_y;
    int me_b_far_search_x;
    int me_b_far_search_y;

    /**
     * :ZZZ:
     * Bitrate mode - one of the following modes is possible.
     * constant bitrate cbr (new) - AV_BRC_MODE_CBR
     * constant bitrate vbr (old) - AV_BRC_MODE_V1
     * constant quality   - AV_BRC_MODE_CONST_QUALITY
     * constatn quantizer - AV_BRC_MODE_CONST_QSCALE    
     * 
     * - encoding: set by user.
     * - decoding: unused.
     */   
    int brc_mode;

    /**
     * :ZZZ:
     * CBR algorithm profile - one of the following profiles is 
     * possible.  
     * AV_BRC_PROFILE_NORMAL - normal cbr mode 
     * AV_BRC_PROFILE_FAST_TRANSITION_C2S - fast transition from 
     * complex to simple scene. 
     * AV_BRC_PROFILE_FAST_TRANSITION_ALL - fast transition from 
     * comples to simple and from simple to complex scenes 
     * - encoding: set by user.
     * - decoding: unused.
     */   
    int brc_profile;

    int brc_qscale_max;
    int brc_qscale_min;

    /**
     * :ZZZ:
     * Quantization values for constant quality / constant quantizers.
     * 
     * - encoding: set by user.
     * - decoding: unused.
     */
    int qscale[3];

    /**
     * :ZZZ:
     * Average activity value for constant quality.
     * 
     * - encoding: set by user.
     * - decoding: unused.
     */
    int iAvgActivity;

    /**
     * :ZZZ:
     * Quantization values for quantizer's limits
     * 
     * - encoding: set by user.
     * - decoding: unused.
     */
    int qscale_min[3];
    int qscale_max[3];

    /**
     * :ZZZ:
     *
     *  All coefficients above the index will be zeroed (luma and chroma)
     *
     * - encoding: Set by user
     * - decoding: Unused
     */
    int zero_coeff;
    int zero_coeff_chroma;
    int force_chroma_zero_coeff;

    /**
     * :ZZZ:
     *
     *  First level, at which the brc tools are activated.
     *
     */
    int brc_active_tool_level;

    /**
     * :ZZZ:
     *
     *  Wmv parameters - ac coding set
     *  Expected values: 0, 1, 2
     *
     * - encoding: Set by user
     * - decoding: Unused
     */
    int ac_coding_set;

    /**
     * :ZZZ:
     *
     *  Wmv parameters - dc table index
     *  Expected values: 0, 1
     *
     * - encoding: Set by user
     * - decoding: Unused
     */
    int dc_table;

    /**
     * :ZZZ:
     *
     *  Wmv parameters - cpb table index for p and b frames
     *  Expected values: 0, 1, 2, 3
     *
     * - encoding: Set by user
     * - decoding: Unused
     */
    int cbp_table;

    /**
     * :ZZZ:
     *
     *  Wmv parameters - mvc table index for p and b frames
     *  Expected values: 0, 1, 2, 3
     *
     * - encoding: Set by user
     * - decoding: Unused
     */
    int mvd_table;

    //
    // Format
    // 
                
    /**
     * private data of the user, can be used to carry app specific stuff.
     * - encoding: set by user
     * - decoding: set by user
     */
    void *opaque;
                                                        
    /**
     * bits per sample/pixel from the demuxer (needed for huffyuv).
     * - encoding: set by lavc
     * - decoding: set by user
     */
     int bits_per_sample;
                
    /**
     * frame_rate_base.
     * for variable fps this is 1
     * - encoding: set by user.
     * - decoding: set by lavc.
     * @todo move this after frame_rate
     */

    int frame_rate_base;

    /**
     * color table ID.
     * - encoding: unused.
     * - decoding: which clrtable should be used for 8bit RGB images
     *             table have to be stored somewhere FIXME
     */
    int color_table_id;

    //
    // Debug
    //
    
    /**
     * debug.
     * - encoding: set by user.
     * - decoding: set by user.
     */
    int debug;
    
    /**
     * called at the beginning of each frame to get a buffer for it.
     * if pic.reference is set then the frame will be read later by lavc
     * - encoding: unused
     * - decoding: set by lavc, user can override
     */
    int (*draw_debug)(struct AVCodecContext *c, AVFrame *pic);

    /**
     * preserve_context
     * - encoding: Set by user while opening and closing MPEG codec 
     *             in order to switch to JPEG compression on >= C8
     * - decoding: Unused
     */
    char preserve_context;    

    /**
     * Codec header type
     * - encoding: Set by user 
     * - decoding: Not used 
     */
    int header_type;

    /** HRD parameters */
    /* HRD Buffer size */
    unsigned int hrd_buffer_size;

    /* HRD Input Rate */
    unsigned int hrd_input_rate;

    int hrd_enable;

    int hrd_ignore_overflow;

    //User custom tags for AVI format
    char *strn_video;
    char *strn_audio;

    // Custom user_data
    char* user_data;

    // Syntax elements flag for MOV/H264
    unsigned int syntax_elements_flag;

	//TF Data Structure
	STFCONFIGTABLE tf_table;
    SHPCONFIGTABLE hp_table;
    SMAGNETTABLE   magnet_table;

    // Specify error type the codec has been encountered.
    unsigned int error_code;

    // H264 profile and level
    int profile;
    int level;

    // Video Codec context
    void* stream_ctx;

    int num_of_codecs;

	int last_p_skipped;

	int cbr_flag;

	int max_initial_cpb_removal_delay;

	void (*pfn_set_tf_callback)(void*);

	// For decoder - last decoded frame size
	int last_decoded_frame_size;

    int main_rotation_angle;
    int display_rotation_angle;
    int dual_stream_1080;

    int pic_init_qp;
    int crop_top;
    int crop_left;

    unsigned short *puwColrAtom;
    int is_intra_only;

} AVCodecContext;

///////////////////////////////////////////////////////////////////////////////
// AVCodec.
///////////////////////////////////////////////////////////////////////////////

typedef struct AVCodec 
{
    const char *name;
    enum CodecType type;
    int id;
    int priv_data_size;
    int (*init)(AVCodecContext *);
    int (*encode)(AVCodecContext *, uint8_t *buf, int buf_size, void *data);
    int (*close)(AVCodecContext *);
    int (*decode)(AVCodecContext *, void *outdata, int *outdata_size,
                  uint8_t *buf, int buf_size);
    int capabilities;
    struct AVCodec *next;

} AVCodec;

///////////////////////////////////////////////////////////////////////////////
// codecs defines.
///////////////////////////////////////////////////////////////////////////////

extern AVCodec mp2_encoder;
extern AVCodec mjpeg_encoder;
// :ZZZ: Encoder
extern AVCodec mpeg4z_encoder;
extern AVCodec pcm_mulaw_encoder;
extern AVCodec h264z_encoder;
extern AVCodec mpeg1z_encoder;

extern AVCodec pcm_mulaw_decoder;
extern AVCodec mjpeg_decoder;
extern AVCodec mp2_decoder;
// :ZZZ: Decoder
extern AVCodec mpeg4z_decoder;
extern AVCodec jpegz_decoder;
extern AVCodec rawz_encoder;
extern AVCodec rawz_decoder;
extern AVCodec h264z_decoder;
extern AVCodec mpeg1z_decoder;
extern AVCodec mp3_decoder;

extern AVCodec aacz_decoder;
extern AVCodec aacz_encoder;

/* pcm codecs */
#define PCM_CODEC(id, name) \
extern AVCodec name ## _decoder; \
extern AVCodec name ## _encoder

PCM_CODEC(CODEC_ID_PCM_S16LE, pcm_s16le);
PCM_CODEC(CODEC_ID_PCM_S16BE, pcm_s16be);
PCM_CODEC(CODEC_ID_PCM_U16LE, pcm_u16le);
PCM_CODEC(CODEC_ID_PCM_U16BE, pcm_u16be);
PCM_CODEC(CODEC_ID_PCM_S8, pcm_s8);
PCM_CODEC(CODEC_ID_PCM_U8, pcm_u8);
PCM_CODEC(CODEC_ID_PCM_ALAW, pcm_alaw);
PCM_CODEC(CODEC_ID_PCM_MULAW, pcm_mulaw);

/* adpcm codecs */

PCM_CODEC(CODEC_ID_ADPCM_IMA_QT, adpcm_ima_qt);
PCM_CODEC(CODEC_ID_ADPCM_IMA_WAV, adpcm_ima_wav);
PCM_CODEC(CODEC_ID_ADPCM_MS, adpcm_ms);
PCM_CODEC(CODEC_ID_ADPCM_4XM, adpcm_4xm);
PCM_CODEC(CODEC_ID_ADPCM_G726, adpcm_g726);

#undef PCM_CODEC

/* :ZZZ: bypass picture */
extern AVCodec bypass_encoder;
extern AVCodec bypass_decoder;

///////////////////////////////////////////////////////////////////////////////
// Component global variables.
///////////////////////////////////////////////////////////////////////////////

//extern AVCodec *first_avcodec;

///////////////////////////////////////////////////////////////////////////////
// Api functions
///////////////////////////////////////////////////////////////////////////////

void register_avcodec(AVCodec *format);

AVCodecContext *avcodec_alloc_context(void);

AVFrame *avcodec_alloc_frame(void);

AVFrame *avcodec_alloc_frame(void);

int avcodec_open(AVCodecContext *avctx, AVCodec *codec);

int avcodec_encode_audio(	AVCodecContext *avctx, 
								uint8_t *buf, 
								int buf_size, 
								const short *samples);

int avcodec_encode_video(	AVCodecContext *avctx, 
								uint8_t *buf, 
								int buf_size, 
								const AVFrame *pict);


int avcodec_decode_video(	AVCodecContext *avctx, 
								AVFrame *picture, 
		                        int *got_picture_ptr,
		                        uint8_t *buf, int buf_size);

int avcodec_decode_audio(	AVCodecContext *avctx, 
								int16_t *samples, 
								int *frame_size_ptr,
								uint8_t *buf, 
								int buf_size);

int avcodec_close(AVCodecContext *avctx);

AVCodec *avcodec_find_encoder(enum CodecID id);

AVCodec *avcodec_find_encoder_by_name(const char *name);

AVCodec *avcodec_find_decoder(enum CodecID id);

AVCodec *avcodec_find_decoder_by_name(const char *name);

void avcodec_get_context_defaults(AVCodecContext *s);

VOID
CodecSetPriorities(INT              priority);

VOID
CodecSetOperationMode( ECodecOperationMode mode );

///////////////////////////////////////////////////////////////////////////////
// Component common functions.
///////////////////////////////////////////////////////////////////////////////

int
Av_avpicture_fill(AVPicture *picture,
                  uint8_t *ptr,
                  int pix_fmt,
                  int width,
                  int height);

int
avpicture_get_size(int pix_fmt, int width, int height);

void 
avcodec_string(char *buf, int buf_size, AVCodecContext *enc, int encode);

/**
 * reduce a fraction.
 * this is usefull for framerate calculations
 * @param max the maximum allowed for dst_nom & dst_den
 * @return 1 if exact, 0 otherwise
 */
int av_reduce(int *dst_nom, int *dst_den, int64_t nom, int64_t den, int64_t max);

/**
 * rescale a 64bit integer.
 * a simple a*b/c isnt possible as it can overflow
 */
int64_t 
av_rescale(int64_t a, int b, int c);

/* memory */
void 
__av_freep(void **ptr);

#define av_freep(p) __av_freep((void **)(p))

void 
*av_fast_realloc(void *ptr, unsigned int *size, unsigned int min_size);

/* for static data only */
/* call av_free_static to release all staticaly allocated tables */
void 
av_free_static(void);

void*
__av_mallocz_static(void** location, unsigned int size);

#define av_mallocz_static(p, s) __av_mallocz_static((void **)(p), s)
#define av_malloc(s)    av_malloc_ex(s, __FILE__, __LINE__)
#define av_mallocz(s)   av_mallocz_ex(s, __FILE__, __LINE__)
#define av_realloc(p,s) av_realloc_ex(p,s, __FILE__, __LINE__)
#define av_free(p)      av_free_ex(p, __FILE__, __LINE__)

void*
av_malloc_ex(unsigned int size,char* file, int line);

void*
av_mallocz_ex(unsigned int size,char* file, int line);

void*
av_realloc_ex(void *ptr, unsigned int size,char* file, int line);

void 
av_free_ex(void *ptr,char* file, int line);

void 
av_mem_lock(void *ptr);

///////////////////////////////////////////////////////////////////////////////
// Frame buffers mapping for mpeg4z codec.
///////////////////////////////////////////////////////////////////////////////

// External buffers.
#define AV_BUF_YUV_IMAGE            0
#define AV_BUF_YUV_IMAGE_C          1

// Internal buffers.
#define AV_BUF_REF_LUMA             2
#define AV_BUF_REF_CHROMA           3
#define AV_BUF_REF_DECIMATED        4
#define AV_BUF_FIFO_WRITE           5
#define AV_BUF_FIFO_TEMPLATE        6

// Buffers index.
#define AV_BUF_EXTERNAL_START       AV_BUF_YUV_IMAGE
#define AV_BUF_EXTERNAL_END         AV_BUF_YUV_IMAGE_C
#define AV_BUF_INTERNAL_START       AV_BUF_REF_LUMA
#define AV_BUF_INTERNAL_END         AV_BUF_FIFO_TEMPLATE

// Flag, which motifies if the image is allocated. Not a real buffer!!!
#define AV_BUF_ALLOCATED            7

// Total buffers number
#define AV_BUFFERS                  AV_BUF_ALLOCATED

#define AV_BUF_FREE             0x00000000
#define AV_BUF_REFERENCE_USED   0x00000001
#define AV_BUF_PICTURE_USED     0x00000002


BOOL
av_mep_set_code_buf(UINT uiId, PULONG pulCodeBuf, UINT uiSize);

BOOL
av_cop_set_code_buf(UINT uiId, PBYTE pbyCodeBuf, ULONG ulSize);

///////////////////////////////////////////////////////////////////////////////
// Audio Header for aacencz
///////////////////////////////////////////////////////////////////////////////
#define AUDIO_HEADER_AAC_RAW       0x0
#define AUDIO_HEADER_AAC_ADTS      0x1
#define AUDIO_HEADER_AAC_ADIF      0x2

#ifdef __cplusplus
}
#endif

#endif /* AVCODEC_H */

