/*
 * Raw Video Codec
 * Copyright (c) 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
/**
 * @file rawz.c
 * Bypass Video Codec
 */
 
#include "avcodec.h"
#include "av.h"
#include "AvApi.h"
#include "Allocator.h"

EXTERN BOOL
DmaBlockCopy(PVOID pvDst,
             PVOID pvSrc,
             UINT  uiLineSize,
             UINT  uiDstLineJump,
             UINT  uiSrcLineJump,
             UINT  uiLines);


static int rawz_init_decoder(AVCodecContext *avctx)
{
    avctx->coded_frame = avcodec_alloc_frame();
    avctx->coded_frame->key_frame = 1;
    avctx->coded_frame->pict_type = FF_I_TYPE;
 
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].width        = avctx->physical_width;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].height       = avctx->physical_height;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].left         = 0;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].top          = 0;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].frame_width  = avctx->physical_width; 
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].frame_height = avctx->physical_height;

    if ( avctx->codec_tag == MKTAG('B', 'Y', 'R', '2') ) 
    {
        avctx->pix_fmt = PIX_FMT_RAW16Z;
    }
    else if ( avctx->codec_tag == MKTAG('I', '4', '2', '0') ) 
    {
        avctx->pix_fmt = PIX_FMT_YUV420P;
    }
    else
    {
        avctx->pix_fmt = PIX_FMT_YUV422Z;
    }

    return 0;
}

static int rawz_decode(AVCodecContext *avctx,
                       void *data,
                       int *data_size,
                       uint8_t *buf,
                       int buf_size)
{
    AVPicture* picture = (AVPicture*)data;

    *data_size = 0;

    if ( buf == NULL )
        return buf_size;
          
    av_mem_lock(buf);

    if ( ( avctx->pix_fmt == PIX_FMT_YUV420P ) || ( avctx->pix_fmt == PIX_FMT_YUV420Z ) )
    {
        picture->data[AV_BUF_YUV_IMAGE_C] = av_malloc( avctx->physical_width *avctx->physical_height / 2 );
    }

    AvApi_avpicture_fill(picture, 
                          buf,
                          avctx->pix_fmt, 
                          avctx->logical_width, 
                          avctx->logical_height);
        
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].width        = avctx->physical_width;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].height       = avctx->physical_height;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].left         = 0;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].top          = 0;
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].frame_width  = avctx->physical_width; 
    avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].frame_height = avctx->physical_height;

    picture->data[AV_BUF_FIFO_WRITE] = NULL;

    // Free code buffer
    PoolFree( NULL, buf );

    *data_size = sizeof(AVPicture);

    return buf_size;
}

static int rawz_close_decoder(AVCodecContext *avctx)
{
    av_freep(&avctx->coded_frame);
    return 0;
}

static int rawz_init_encoder(AVCodecContext *avctx)
{    
    avctx->coded_frame = avcodec_alloc_frame();
    avctx->coded_frame->key_frame = 1;
    avctx->coded_frame->pict_type = FF_I_TYPE;

    return 0;
}

static int rawz_encode(AVCodecContext *avctx,
                       unsigned char *frame,
                       int buf_size,
                       void *data)
{
    AVFrame* pic_arg = data;
    INT i, j;
    UINT uiBytesPerSample = avctx->bits_per_sample / 8;

    if ( frame == NULL || data == NULL )
    {
        return 0;
    }
    
    pic_arg->code_buf = frame;
     
    switch ( avctx->pix_fmt )
    {
        default:
            DmaBlockCopy(frame, pic_arg->data[AV_BUF_YUV_IMAGE],
                         pic_arg->width * uiBytesPerSample,
                         pic_arg->width * uiBytesPerSample,
                         pic_arg->frame_width * uiBytesPerSample,
                         pic_arg->height);
    }

    // convert interleaved UV (expected from scaler) into planar U,V
    if ( pic_arg->data[AV_BUF_YUV_IMAGE_C] != NULL )
    {
        // pointers to U,V planes
        unsigned char* puchChroma = pic_arg->data[AV_BUF_YUV_IMAGE_C];
        unsigned char* puchUPlane = frame + pic_arg->width * pic_arg->height * uiBytesPerSample;
        unsigned char* puchVPlane = puchUPlane + pic_arg->width * pic_arg->height / 4;

        INT iChromaSrcLineJump = pic_arg->frame_width;
        INT iChromaDestLineJump = pic_arg->width / 2;

        // copy to UV into planes
        for ( i = 0; i < pic_arg->height / 2; i++ )
        {
            for ( j = 0; j < pic_arg->width / 2; j++ )
            {
                puchUPlane[j] = puchChroma[j * 2];
                puchVPlane[j] = puchChroma[j * 2 + 1];
            }

            puchChroma += iChromaSrcLineJump;
            puchUPlane += iChromaDestLineJump;
            puchVPlane += iChromaDestLineJump;
        }
    }

    // free the data

    // Release all references memories.
    for ( i = AV_BUF_EXTERNAL_START; i <= AV_BUF_EXTERNAL_END; i++ )
    {
        PoolFree( NULL, pic_arg->data[i] );
        pic_arg->data[i] = NULL;
    }

    return buf_size;
}

static int rawz_close_encoder(AVCodecContext *avctx)
{
    av_freep(&avctx->coded_frame);
    return 0;
}

AVCodec rawz_encoder = {
    "rawz",
    CODEC_TYPE_VIDEO,
    CODEC_ID_RAWZ,
    0,
    rawz_init_encoder,
    rawz_encode,
    rawz_close_encoder,
    NULL
};

AVCodec rawz_decoder = {
    "rawz",
    CODEC_TYPE_VIDEO,
    CODEC_ID_RAWZ,
    0,
    rawz_init_decoder,
    NULL,
    rawz_close_decoder,
    rawz_decode,
};
