#ifndef AVFORMAT_H
#define AVFORMAT_H

#ifdef __cplusplus
extern "C" {
#endif

#define LIBAVFORMAT_VERSION_INT 0x000406  
#define LIBAVFORMAT_VERSION     "0.4.6"
#define LIBAVFORMAT_BUILD       4605

#include "../Av/Include/avcodec.h"

#include "avio.h"
#include <time.h>

#define avf_mallocz(s)	avf_mallocz_ex((s), __FILE__, __LINE__);

#define avf_malloc(s)	avf_malloc_ex((s), __FILE__, __LINE__);

#define avf_freep(p) __avf_freep((void **)(p))

#define avf_free(p) avf_free_ex(p, __FILE__, __LINE__)

#define avf_realloc(p,s) avf_realloc_ex(p, (s), __FILE__, __LINE__)

/* packet functions */

// :ZZZ: Avoid using zero as no pts value
#define AV_NOPTS_VALUE  ((int64_t)0xFFFFFFFFFFFFFFFFLL)

// :ZZZ: Set this flag when reading header, if the index table/s mustn't be loaded
#define AV_READ_HEADER_FLAG_NO_INDEX                0x00000001
#define AV_READ_HEADER_FLAG_MP3_SUPPORT_MPEG25      0x00000002
#define AV_FAST_FILE_OPEN                           0x00000004
#define AV_NO_DRM                                   0x00000008
#define AV_READ_HEADER_FLAG_INDEX_FIRST_ENTRY_ONLY  0x00000010

#define AV_SEEK_CURR_POS_WRITE(s) s->pos + (s->buf_ptr - s->buffer)
#define AV_SEEK_CURR_POS_READ(s)  s->pos - (s->buf_end - s->buffer) + (s->buf_ptr - s->buffer)

//:ZZZ: Field IDs for content field parameter in AvSetStreamContentInfo() fn. 
typedef enum
{
    CONTENT_INFO_TITLE,
    CONTENT_INFO_AUTHOR,
    CONTENT_INFO_COPYRIGHT,
    CONTENT_INFO_COMMENTS
} CONTENT_INFO_TYPE;

typedef enum 
{
    AV_PARAMID_CONTEXT,
    AV_PARAMID_RESET_IO,
    AV_PARAMID_MOV_ATOM_MODIFY,
    AV_PARAMID_CURRENT_POS,
    AV_PARAMID_SET_CONTEXT_FLAGS
} AV_PARAMID;


typedef struct AVPacket {
    int64_t pts; /* presentation time stamp in stream units (set av_set_pts_info) */
    uint8_t *data;
    int   size;
    int   stream_index;
    int   flags;
    int   duration;
    void  (*destruct)(struct AVPacket *);
    void  *priv;
} AVPacket; 
#define PKT_FLAG_KEY   0x0001

static inline void av_init_packet(AVPacket *pkt)
{
    pkt->pts   = AV_NOPTS_VALUE;
    pkt->flags = 0;
    pkt->stream_index = 0;
}

int av_new_packet(AVPacket *pkt, int size);

/**
 * Free a packet
 *
 * @param pkt packet to free
 */
void Avf_free_packet(AVPacket *pkt);

/*************************************************/
/* fractional numbers for exact pts handling */

/* the exact value of the fractional number is: 'val + num / den'. num
   is assumed to be such as 0 <= num < den */
typedef struct AVFrac {
    int64_t val, num, den; 
} AVFrac;

void av_frac_init(AVFrac *f, int64_t val, int64_t num, int64_t den);
void av_frac_add(AVFrac *f, int64_t incr);
void av_frac_set(AVFrac *f, int64_t val);

/*************************************************/
/* input/output formats */

struct AVFormatContext;

typedef HCE_GENERIC_CALLBACK GenericCallbackFn;

/* this structure contains the data a format has to probe a file */
typedef struct AVProbeData {
    const char *filename;
    unsigned char *buf;
    int buf_size;
    GenericCallbackFn handle_error;
    GenericCallbackFn custom_probe;
} AVProbeData;

// These have to be shadowed in Zcam/Api/Api_Func.h
#define AVI_RIFF_ERROR_ID                0
#define AVI_AVI_ERROR_ID                 1 
#define AV_MODIFY_CREATION_TIME          2
#define AV_SEAMLESS_RECORDING            3
#define AV_MODIFY_FTYP_ATOM              4
#define AV_MODIFY_LANGUAGE_CODE          5
#define AV_MODIFY_UUID_PROF              6
#define AV_MODIFY_UDTA_ATOM              7
#define AV_MODIFY_FREE_ATOM              8
#define AV_VIDEO_FRAME_COUNT             9
#define AV_MODIFY_UUID_PROF_SECONDARY    10
#define AV_THUMBNAILS_SIZE               11
#define AV_VIDEO_BIT_RATE                12


#define AV_PRIMARY_STREAM   0
#define AV_SECONDARY_STREAM 1

#define AVI_ERR_FAIL    0
#define AVI_ERR_OK      1

#define AVPROBE_SCORE_MAX 100
#define AVPROBE_BUF_SIZE 2048

typedef struct AVStreamFrameInfo {
    PBYTE  buffer_ptr;
    UINT   buffer_length;
    UINT   type;
    UINT   flags;
    UINT   framenumber;    
    UINT   timestamphi;
    UINT   timestamplow;
} AVStreamFrameInfo;

typedef struct AVFormatParameters {
    int frame_rate;
    int frame_rate_base;
    int sample_rate;
    int channels;
    int width;
    int height;
    PixelFormat pix_fmt;
    struct AVImageFormat *image_format;
    int channel; /* used to select dv channel */
    const char *device; /* video4linux, audio or DV device */
    int flags;
} AVFormatParameters;

#define AVFMT_NOFILE        0x0001 /* no file should be opened */
#define AVFMT_NEEDNUMBER    0x0002 /* needs '%d' in filename */ 
#define AVFMT_NOHEADER      0x0004 /* signal that no header is present
                                      (streams are added dynamically) */
#define AVFMT_SHOW_IDS      0x0008 /* show format stream IDs numbers */
#define AVFMT_RAWPICTURE    0x0020 /* format wants AVPicture structure for
                                      raw picture data */
#define AVFMT_MODIFY_APPEND 0x0040 /*   */

typedef struct AVOutputFormat {
    const char *name;
    const char *long_name;
    const char *mime_type;
    const char *extensions; /* comma separated extensions */
    /* size of private data so that it can be allocated in the wrapper */
    int priv_data_size;
    /* output support */
    enum CodecID audio_codec; /* default audio codec */
    enum CodecID video_codec; /* default video codec */
    int (*write_header)(struct AVFormatContext *);
    int (*write_packet)(struct AVFormatContext *, 
                        int stream_index,
                        unsigned char *buf, int size, int64_t force_pts);
    int (*write_trailer)(struct AVFormatContext *);
    /* can use flags: AVFMT_NOFILE, AVFMT_NEEDNUMBER */
    int flags;
    /* currently only used to set pixel format if not YUV420P */
    int (*set_parameters)(struct AVFormatContext *, 
    				      AVFormatParameters *, 
    				      void *pval);
    /* private fields */
    struct AVOutputFormat *next;
    /* Return the esimated size of the trailer that will be written (index table, etc.) */
    int (*estimate_trailer_size)(struct AVFormatContext *);
   
 /* the function will parse private context of first parameter
       and prepare the context of second parameter accordingly */
    int (*parse_private_context)(struct AVFormatContext *, struct AVFormatContext *);

    int (*delete_packet)(struct AVFormatContext *, unsigned long flags);

    int (*add_stream)(struct AVFormatContext *, int id);

    int (*read_text_stream)(struct AVFormatContext *s, 
                            int first, 
                            int total,
                            char** text,
                            int* indexes,
                            int stream_id );

    int (*video_to_text)(struct AVFormatContext *s, int video, int text_stream);
    /* external creation time function*/
    GenericCallbackFn generic_callback;


} AVOutputFormat;

typedef struct AVInputFormat_tag {
    const char *name;
    const char *long_name;
    /* size of private data so that it can be allocated in the wrapper */
    int priv_data_size;
    /* tell if a given file has a chance of being parsing by this format */
    int (*read_probe)(AVProbeData *);
    /* read the format header and initialize the AVFormatContext
       structure. Return 0 if OK. 'ap' if non NULL contains
       additionnal paramters. Only used in raw format right
       now. 'av_new_stream' should be called to create new streams.  */
    int (*read_header)(struct AVFormatContext *,
                       AVFormatParameters *ap);
    /* read one packet and put it in 'pkt'. pts and flags are also
       set. 'av_new_stream' can be called only if the flag
       AVFMT_NOHEADER is used. */
    int (*read_packet)(struct AVFormatContext *, AVPacket *pkt);
    /* close the stream. The AVFormatContext and AVStreams are not
       freed by this function */
    int (*read_close)(struct AVFormatContext *);
    /* seek at or before a given pts (given in microsecond). The pts
       origin is defined by the stream */
    int (*read_seek)(struct AVFormatContext *, int64_t pts);
    /* :ZZZ: seek by index */
    int (*read_seek_idx)(struct AVFormatContext *s, 
                         int stream_id, 
                         int pos, 
                         int whence);
    
    int (*read_text_stream)(struct AVFormatContext *s, 
                            int first, 
                            int total,
                            char** text,
                            int* indexes,
                            int stream_id );
    /* :ZZZ: check if the given frame is key frame */
    int (*read_is_key_frame)(struct AVFormatContext *s, 
                             int stream_id, 
                             int frame);

    /* can use flags: AVFMT_NOFILE, AVFMT_NEEDNUMBER, AVFMT_NOHEADER */
    int flags;
    /* if extensions are defined, then no probe is done. You should
       usually not use extension format guessing because it is not
       reliable enough */
    const char *extensions;
    /* general purpose read only value that the format can use */
    int value;
    /* external error handling function */
    GenericCallbackFn handle_error;
    /* external header probe function */
    GenericCallbackFn custom_probe;
    /* private fields */
    struct AVInputFormat_tag *next;
} AVInputFormat;

#define AV_MOV_MAX_OPTIONAL_DESCRIPTION_SIZE    64


typedef struct tagMOV_Streams_Info{
    BOOL  bVprfExist;

    UINT  uiVisualWidth;
    UINT  uiVisualHeight;
    ULONG ulVisualFlags;
    CHAR  cVideoCodecType[4];// fixed length

    BOOL  bAprfExist;

    BYTE  uiAudioObjectType;
    BYTE  uiDataFormType;
    UINT  uiAudioChannelNumber;
    ULONG ulAudioSampleRate;
    ULONG ulAudioFlags;
    CHAR  cAudioCodecType[4];// fixed length

    BOOL  bMtdtVideoExist; // for video track only
    ULONG ulMetaVPresentationType;
    ULONG ulMetaVData;//track attr, reserved, priority;
    ULONG ulMtdtVideoDataTypeID;

    BOOL  bMtdtAudioExist; // for video track only
    ULONG ulMetaAPresentationType;
    ULONG ulMtdtAudioDataTypeID;

    // Visual Sample Entry fields:
    LONG lVisual_horizresolution;
    LONG lVisual_vertresolution;

    UINT  uiAudioTrackFlags;
    UINT  uiVideoTrackFlags;

    CHAR  cFtypCompatibleBrands[12];

    BYTE AVCProfileIndication;
    BYTE VideoSampleDescriptionNum;
    BYTE AudioSampleDescriptionNum;

    BYTE  AudioDSIChannelConfigurations;

    CHAR  cVideoFormatType[4];
    CHAR  cAudioFormatType[4];

    CHAR AudioObjectTypeIndication;

    ULONG ulSTSDAudioSampleRate; 
    ULONG ulSTSDAudioChannels;
    ULONG ulAudioBitRate;
    BOOL bAudioExist;

    UINT uiVosLen;
    BYTE pbyVosData[AV_MOV_MAX_OPTIONAL_DESCRIPTION_SIZE];  
} MOV_Streams_Info;

typedef struct AVStream_tag {
    int index;    /* stream index in AVFormatContext */
    int id;       /* format specific stream id */
    AVCodecContext codec; /* codec context */
    int r_frame_rate;     /* real frame rate of the stream */
    int r_frame_rate_base;/* real frame rate base of the stream */
    uint64_t time_length; /* real length of the stream in miliseconds */
    void *priv_data;
    /* internal data used in av_find_stream_info() */
    int codec_info_state;     
    int codec_info_nb_repeat_frames;
    int codec_info_nb_real_frames;
    /* PTS generation when outputing stream */
    AVFrac pts;
    /* ffmpeg.c private use */
    int stream_copy; /* if TRUE, just copy stream */
    /* quality, as it has been removed from AVCodecContext and put in AVVideoFrame
     * MN:dunno if thats the right place, for it */
    float quality; 
    int length; /* length of stream in bytes */
    int stream_total_nb_frames;
    int text_sample_id;
	int header_flags;
	uint32 frame_number;
} AVStream;

#define MAX_STREAMS 15 

typedef enum
{
    META_DATA_NONE = 0,
    META_DATA_QIK  = 1,
} AVMETADATATYPE;

typedef struct 
{
    UINT uiFrameSize;
    UINT uiTimeStamp;
    BYTE byPacketType;
    BYTE byLiveCounter[3];
} SQIKDATA, *PSQIKDATA;


/* format I/O context */
typedef struct AVFormatContext {
    /* can only be iformat or oformat, not both at the same time */
    struct AVInputFormat_tag *iformat;
    struct AVOutputFormat *oformat;
    void *priv_data;
    ByteIOContext pb;
    /* Total number of streams */
    int nb_streams;
    AVStream *streams[MAX_STREAMS];
    char filename[1024]; /* input or output filename */
    /* stream info */
    char title[512];
    char author[512];
    char copyright[512];
    char comment[512];
    int flags; /* format specific flags */
    /* private data for pts handling (do not modify directly) */
    int pts_wrap_bits; /* number of bits in pts (used for wrapping control) */
    int pts_num, pts_den; /* value to convert to seconds */
    /* This buffer is only needed when packets were already buffered but
       not decoded, for example to get the codec parameters in mpeg
       streams */
    // :ZZZ: Added keyframe since cannot use the encoding context's field
    int key_frame; /* is the currently written frame a key frame? */
    unsigned int dependency;
    struct AVPacketList *packet_buffer;

    // :ZZZ: added thumbnail image, vendor specific params and user data atom support
    PBYTE  pbyThumbnailImage;
    ULONG  ulThumbnailImageSize;
    ULONG  ulVendorName;            //four characters vendor id
    USHORT uwModel;
    USHORT uwRevision;
    PBYTE  pbyUserAtom;
    PBYTE  pbySkipAtom;
    PBYTE  pbyCustomAtom;
    PUSHORT puwColrAtom;
    ULONG  ulCreTimeStamp;
    ULONG  ulModTimeStamp;
    ULONG  ulTotalVideoFrames;

    USHORT uwGopSize;

	// the file is in special modify mode
    BOOL   bModifyMode;
	// updated file size for micro-step delete frame operation
	offset_t  ulUpdatedFileSize;
	// added for muting audio from the end of the movie
    ULONG  ulMuteEndAudioBytes;
    INT    iAudioSilencePattern; 

	MOV_Streams_Info sMovStreamsInfo;
    ULONG  ulSyntaxElementsFlag;

    int previous_stream_index;
    int first_frame_got; 

	BOOL   bMjpegASupport;

    AVMETADATATYPE eMetaData;

    UINT   uiStream;

} AVFormatContext;

// :ZZZ: Structure to hold stream content description information
typedef struct AVStreamInfo{
    PCHAR title;
    PCHAR author;
    PCHAR copyright;
    PCHAR comment; 
} AVStreamInfo;

typedef struct AVPacketList {
    AVPacket pkt;
    struct AVPacketList *next;
} AVPacketList;


extern AVInputFormat *first_iformat;
extern AVOutputFormat *first_oformat;

/* still image support */
struct AVInputImageContext;
typedef struct AVInputImageContext AVInputImageContext;

typedef struct AVImageInfo {
    PixelFormat pix_fmt; /* requested pixel format */
    int width; /* requested width */
    int height; /* requested height */
    int interleaved; /* image is interleaved (e.g. interleaved GIF) */
    AVPicture pict; /* returned allocated image */
} AVImageInfo;


// QuickTime mov file info
typedef struct AVMOVIEINFO{
	// Video
    ULONG  ulFirstSampleOffset;
    ULONG  ulFirstSampleSize;
    USHORT uwWidth;
    USHORT uwHeight;
    ULONG  ulSampleCount;

	// Audio
    ULONG  ulCodecID; 
    ULONG  ulAudioSampleRate;
    ULONG  ulAudioNumOfChannels;
    ULONG  ulAudioChunkCount;

    // Stream
    UINT uiModifyAtom;

} AVMOVIEINFO, *PAVMOVIEINFO ;

#define MAX_CUSTOM_TAG_NUM		    	4

// MOV Tag configuration type
#define AV_CUSTOM_STREAM_VIDEO_TAG		0x00000001
#define AV_CUSTOM_STREAM_AUDIO_TAG		0x00000002
#define AV_CUSTOM_GENERAL_TAG		    0x00000003

// MOV Atom configuration
#define AV_CUSTOM_TAG_OP_DEL			0x00000001
#define AV_CUSTOM_TAG_OP_INITIAL_VAL	0x00000002
#define AV_CUSTOM_TAG_OP_REGISTER_CB	0x00000004

#define AV_CUSTOM_OP_MAX_VAL			0x00000004

// MOV Tag Id
#define AV_CUSTOM_TAG_ID_STSS			0x00000001
#define AV_CUSTOM_TAG_ID_SKIP			0x00000002
#define AV_CUSTOM_TAG_ID_CUSTOM			0x00000003

// flags for avf commands
#define AV_CMD_DELETE_PACKET_LAST                   0x00000001
#define AV_CMD_DELETE_PACKET_UPDATE_DURATION        0x00000002

// flags for I43_CM_MPEG4_SYNTAX_ELEMENTS (ulSyntaxElementsFlag)
#define MPEG4_SYNTAX_ELEMENT_CSLG    0x00000001
#define MPEG4_SYNTAX_ELEMENT_CTTS    0x00000002
#define MPEG4_SYNTAX_ELEMENT_SDTP    0x00000004
#define MPEG4_SYNTAX_ELEMENT_USMT    0x00000008
#define MPEG4_SYNTAX_ELEMENT_PROF    0x00000010
#define MPEG4_SYNTAX_ELEMENT_EDTS    0x00000020
#define MPEG4_SYNTAX_ELEMENT_HDLR    0x00000040
#define MPEG4_SYNTAX_ELEMENT_DINF    0x00000080
#define MPEG4_SYNTAX_ELEMENT_FREE    0x00000100
#define MPEG4_SYNTAX_ELEMENT_AUD     0x00000200
#define MPEG4_SYNTAX_ELEMENT_BP_SEI  0x00000400
#define MPEG4_SYNTAX_ELEMENT_PT_SEI  0x00000800
#define MPEG4_SYNTAX_ELEMENT_VUI_PS  0x00001000
#define MPEG4_SYNTAX_ELEMENT_RP_SEI  0x00002000
#define MPEG4_SYNTAX_ELEMENT_VUI     0x00004000
#define MPEG4_SYNTAX_ELEMENT_NAL_HDR 0x00008000
#define MPEG4_SYNTAX_ELEMENT_VCL_HDR 0x00010000
#define MPEG4_SYNTAX_ELEMENT_FTYP    0x00040000
#define MPEG4_SYNTAX_ELEMENT_EX_SIZE_REMOVE 0x00080000 // default included
#define MPEG4_SYNTAX_ELEMENT_VUI_VST 0x00100000


//Customed Tag Info
typedef struct AVCustomTagContent
{
    UINT  uiTagId; 	    
    UINT  uiStream; 		
	UINT  uiOpType; 
	UINT  uiSize;
    PVOID pvData;	
	
} AVCustomTagContent, *PAVCustomTagContent;

/* AVImageFormat.flags field constants */
#define AVIMAGE_INTERLEAVED 0x0001 /* image format support interleaved output */

typedef struct AVImageFormat {
    const char *name;
    const char *extensions;
    /* tell if a given file has a chance of being parsing by this format */
    int (*img_probe)(AVProbeData *);
    /* read a whole image. 'alloc_cb' is called when the image size is
       known so that the caller can allocate the image. If 'allo_cb'
       returns non zero, then the parsing is aborted. Return '0' if
       OK. */
    int (*img_read)(ByteIOContext *, 
                    int (*alloc_cb)(void *, AVImageInfo *info), void *);
    /* write the image */
    int supported_pixel_formats; /* mask of supported formats for output */
    int (*img_write)(ByteIOContext *, AVImageInfo *);
    int flags;
    struct AVImageFormat *next;
} AVImageFormat;

void Avf_register_image_format(AVImageFormat *img_fmt);
AVImageFormat *av_probe_image_format(AVProbeData *pd);
AVImageFormat *guess_image_format(const char *filename);
int av_read_image(ByteIOContext *pb, const char *filename,
                  AVImageFormat *fmt,
                  int (*alloc_cb)(void *, AVImageInfo *info), void *opaque);
int av_write_image(ByteIOContext *pb, AVImageFormat *fmt, AVImageInfo *img);

extern AVImageFormat *first_image_format;

extern AVImageFormat pnm_image_format;
extern AVImageFormat pbm_image_format;
extern AVImageFormat pgm_image_format;
extern AVImageFormat ppm_image_format;
extern AVImageFormat pam_image_format;
extern AVImageFormat pgmyuv_image_format;
extern AVImageFormat yuv_image_format;
#ifdef CONFIG_ZLIB
extern AVImageFormat png_image_format;
#endif
#ifdef AVF_JPEG
extern AVImageFormat jpeg_image_format;
#endif
extern AVImageFormat gif_image_format;


AVImageFormat*
Avf_AVGetYuvImageFormat(VOID);

/* XXX: use automatic init with either ELF sections or C file parser */
/* modules */

/* mpeg.c */
int Avf_mpegps_init(void);

/* mpegts.c */
extern AVInputFormat mpegts_demux;
int Avf_mpegts_init(void);

/* rm.c */
int Avf_rm_init(void);

/* crc.c */
int crc_init(void);

/* img.c */
int Avf_img_init(void);

/* asf.c */
int Avf_asf_init(void);

/* avienc.c */
int Avf_avienc_init(void);

/* avidec.c */
int Avf_avidec_init(void);

/* swf.c */
int swf_init(void);

/* movdec.c */
int Avf_movdec_init(void);

/* movenc.c */
int Avf_movenc_init(void);

/* jpeg.c */
int Avf_jpeg_init(void);

// :ZZZ: EXIF file format.
/* exif.c */
int Avf_exif_init(void);

/* gif.c */
int gif_init(void);

/* au.c */
int Avf_au_init(void);

/* wav.c */
int Avf_wav_init(void);

/* raw.c */
int Avf_raw_init(void);

/* ogg.c */
int ogg_init(void);

/* dv.c */
int dv_init(void);

/* ffm.c */
int ffm_init(void);

/* rtsp.c */
extern AVInputFormat redir_demux;
int redir_open(AVFormatContext **ic_ptr, ByteIOContext *f);

#include "rtp.h"

#include "rtsp.h"

/* yuv4mpeg.c */
extern AVOutputFormat yuv4mpegpipe_oformat;

/* utils.c */
void av_register_input_format(AVInputFormat *format);
int Avf_register_input_format_error_callback(void * callback_function);
int Avf_register_input_format_custom_probe_callback(GenericCallbackFn callback_function);
void av_register_output_format(AVOutputFormat *format);
int Avf_register_output_format_callback(void * callback_function);

AVOutputFormat *guess_stream_format(const char *short_name, 
                                    const char *filename, const char *mime_type);
AVOutputFormat *Avf_guess_format(	const char *short_name, 
                             	const char *filename, 
                             	const char *mime_type);

void av_hex_dump(uint8_t *buf, int size);

void av_register_all(void);

typedef struct FifoBuffer {
    uint8_t *buffer;
    uint8_t *rptr, *wptr, *end;
} FifoBuffer;

int fifo_init(FifoBuffer *f, int size);
void fifo_free(FifoBuffer *f);
int fifo_size(FifoBuffer *f, uint8_t *rptr);
int fifo_read(FifoBuffer *f, uint8_t *buf, int buf_size, uint8_t **rptr_ptr);
void fifo_write(FifoBuffer *f, uint8_t *buf, int size, uint8_t **wptr_ptr);

/* media file input */
AVInputFormat *Avf_find_input_format(const char *short_name);
AVInputFormat *av_probe_input_format(AVProbeData *pd, int is_opened);

int Avf_open_input_file(AVFormatContext **ic_ptr, 
							const char *filename, 
							AVInputFormat *fmt,
							int buf_size,
							AVFormatParameters *ap);

int Avf_open_output_file(ByteIOContext *s,
                         const char *filename,
                         int buf_size,
                         int flags);

#define AVERROR_UNKNOWN     (-1)  /* unknown error */
#define AVERROR_IO          (-2)  /* i/o error */
#define AVERROR_NUMEXPECTED (-3)  /* number syntax expected in filename */
#define AVERROR_INVALIDDATA (-4)  /* invalid data found */
#define AVERROR_NOMEM       (-5)  /* not enough memory */
#define AVERROR_NOFMT       (-6)  /* unknown format */

int Avf_find_stream_info(AVFormatContext *ic);
int Avf_read_packet(AVFormatContext *s, AVPacket *pkt);
void Avf_close_input_file(AVFormatContext *s);
AVStream *Avf_new_stream(AVFormatContext *s, int id);
void av_set_pts_info(AVFormatContext *s, int pts_wrap_bits,
                     int pts_num, int pts_den);

// :ZZZ: Seek
int Avf_read_seek(struct AVFormatContext *, int64_t pts);
int Avf_read_seek_idx(	struct AVFormatContext *s, 
							int stream_id, 
							int pos, 
							int whence);
int Avf_read_is_keyframe(AVFormatContext *s, int stream_id, int frame);


/* media file output */
int Avf_set_parameters(AV_PARAMID iParamId, AVFormatContext *s, AVFormatParameters *ap, void *pval);

/* media file output */
int Avf_get_parameters(AV_PARAMID iParamId, AVFormatContext *s, AVFormatParameters *ap, void *pval);


// :ZZZ: Function to set stream content description fields
BOOL Avf_set_stream_content_info(	CONTENT_INFO_TYPE  uiField, 
										PCHAR pcBuffer, 
										UINT uiSize);

int Avf_write_header(AVFormatContext *s);
// :ZZZ: Added key_frame argument
int Avf_write_frame(AVFormatContext *s, 
						int stream_index, 
						const uint8_t *buf, 
						int size, 
						int key_frame);

int Avf_estimate_trailer_size(AVFormatContext *s);

int Avf_write_trailer(AVFormatContext *s);

int Avf_seek_last_chunk_end_pos(AVFormatContext *s);

int Avf_parse_trailer_info(AVFormatContext *si, AVFormatContext *so);

int Avf_delete_packet( AVFormatContext *s , unsigned long ulFlags );

int Avf_add_stream( AVFormatContext *s, AVCodecContext* psCodec );

int Avf_get_stream_frames_number( AVFormatContext* s, int stream_id );

int Avf_read_text_stream( 	AVFormatContext* s,
							int first,
							int total,
							char** text,
							int* indexes,
							int stream_id );

int Avf_video_to_text( AVFormatContext* s, int video, int text_stream );

void Avf_dump_format(	AVFormatContext *ic,
					int index, 
					const char *url,
					int is_output);

int parse_image_size(int *width_ptr, int *height_ptr, const char *str);
int64_t parse_date(const char *datestr, int duration);

int64_t av_gettime(void);


/* ffm specific for ffserver */
#define FFM_PACKET_SIZE 4096
offset_t ffm_read_write_index(int fd);
void ffm_write_write_index(int fd, offset_t pos);
void ffm_set_write_index(AVFormatContext *s, offset_t pos, offset_t file_size);

int find_info_tag(char *arg, int arg_size, const char *tag1, const char *info);

int get_frame_filename(char *buf, int buf_size,
                       const char *path, int number);
int filename_number_test(const char *filename);

/* grab specific */
int video_grab_init(void);
int audio_init(void);

/* DV1394 */
int dv1394_init(void);

#ifdef HAVE_AV_CONFIG_H
int strstart(const char *str, const char *val, const char **ptr);
int stristart(const char *str, const char *val, const char **ptr);
void pstrcpy(char *buf, int buf_size, const char *str);
char *pstrcat(char *buf, int buf_size, const char *s);

struct in_addr;
int resolve_host(struct in_addr *sin_addr, const char *hostname);

void url_split(char *proto, int proto_size,
               char *hostname, int hostname_size,
               int *port_ptr,
               char *path, int path_size,
               const char *url);

int match_ext(const char *filename, const char *extensions);

#endif /* HAVE_AV_CONFIG_H */

// :ZZZ: Setting and getting the current media handle
void Avf_media_set(HMEDIA hMedia);
HMEDIA Avf_media_get( VOID );

// :ZZZ: Setting and getting the current file entry
void Avf_file_entry_set(unsigned short uwEntry);
unsigned short Avf_file_entry_get();

// :ZZZ: Setting/Getting last IO error code
int AvfAvioSetLastError( int iErrorCode );

int AvfAvioGetLastError( VOID );



int Avf_get_total_written_video_frames_number(AVFormatContext *s);

int Avf_get_total_present_video_frames_number(AVFormatContext *s);

BOOL Avf_set_custom_tag_info( AVCustomTagContent *pTagInfo );

BOOL Avf_get_custom_tag_info( AVCustomTagContent *pTagInfo );

void *avf_mallocz_ex(unsigned int size, char* file, int line);
void __avf_freep(void **ptr);
void avf_free_ex(void *ptr, char* file, int line);
void *avf_realloc_ex(void* ptr, unsigned int size, char* file, int line);
void *avf_malloc_ex(unsigned int size, char* file, int line);


extern char *strptime(const char *, const char *, struct tm *);

int Avf_direct_put_data(unsigned char *buf, int size, unsigned int flags);
int Avf_direct_get_data(unsigned int streamidx, unsigned char **buf, int *size, uint64_t *seqnum, unsigned int flags);

#ifdef __cplusplus
}
#endif

#endif /* AVFORMAT_H */
