/*
 * ASF compatible encoder and decoder.
 * Copyright (c) 2000, 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avformat.h"
#include "AvApi.h"
#include "av.h"
#include "avi.h"
#include "mpegaudio.h"

//#define MSDRM10_SUPPORT

#ifdef MSDRM10_SUPPORT

#include "drmcommon.h"
#include "drmtypes.h"

#include "drmcommon.h"
#include "drmutilities.h"
#include "drmhds.h"
#include "drmsecurestore.h"
#include "drmlicense.h"
#include "drmcontextsizes.h"
#include "drmmanager.h"

static const DRM_WCHAR s_szLicense[] = L"i:\\wmdrmpd\\sample.hds";
static const DRM_CONST_STRING s_strLicense = CREATE_DRM_STRING( s_szLicense );

#endif // MSDRM10_SUPPORT

// :ZZZ: change packet size
#define PACKET_SIZE 4125

// #define PACKET_SIZE (32 * 1024)
//#define PACKET_SIZE 3200
#define PACKET_HEADER_SIZE 12
#define FRAME_HEADER_SIZE 17

typedef enum {
    ASFBackward = -1,
    ASFForward = 1    
}  ASFDirection;

typedef struct {
    offset_t position;
    int packet_size_left;
    int packet_flags;
    int packet_property;
    int packet_timestamp;
    int packet_segments;
    int packet_segsizetype;
    int packet_padsize;
    int packet_frag_timestamp;
} ASFIndexEntry;

typedef struct {
    int num;
    int seq;
    /* use for reading */
    AVPacket pkt;
    int frag_offset;
    int timestamp;
    int64_t duration;

    int ds_span;        /* descrambling  */
    int ds_packet_size;
    int ds_chunk_size;
    int ds_data_size;
    int ds_silence_data;

} ASFStream;

typedef struct __attribute__((packed)) {
    uint32_t v1;
    uint16_t v2;
    uint16_t v3;
    uint8_t v4[8];
} GUID;

typedef struct __attribute__((packed)) {
    GUID guid;          // generated by client computer
    uint64_t file_size;     // in bytes
                                // invalid if broadcasting
    uint64_t create_time;   // time of creation, in 100-nanosecond units since 1.1.1601
                                // invalid if broadcasting
    uint64_t packets_count; // how many packets are there in the file
                                // invalid if broadcasting
    uint64_t play_time;     // play time, in 100-nanosecond units
                                // invalid if broadcasting
    uint64_t send_time;     // time to send file, in 100-nanosecond units
                                // invalid if broadcasting (could be ignored)
    uint32_t preroll;       // timestamp of the first packet, in milliseconds
                    // if nonzero - substract from time
    uint32_t ignore;            // preroll is 64bit - but let's just ignore it
    uint32_t flags;     // 0x01 - broadcast
                    // 0x02 - seekable
                                // rest is reserved should be 0
    uint32_t min_pktsize;   // size of a data packet
                                // invalid if broadcasting
    uint32_t max_pktsize;   // shall be the same as for min_pktsize
                                // invalid if broadcasting
    uint32_t max_bitrate;   // bandwith of stream in bps
                    // should be the sum of bitrates of the
                                // individual media streams
} ASFMainHeader;

typedef struct {
    int packet_num;
    short packet_count;

} ASFSimpleIndexEntry;

typedef struct {
    int last_entry;
    int allocated_entries;
    int time_interval; //in milliseconds
    ASFSimpleIndexEntry* entries;

} ASFSimpleIndex;

typedef struct {
    BYTE    pbSecretData[32];
    DWORD   cbSecretData;
    BYTE    pbType[16];
    BYTE    pbKeyID[32];

    WORD cbCEncExData;
    BYTE *pbCEncExData;

    VOID *pJanusContext;
    VOID *pJanusDecryptContext;

    BOOL bHasDRM;
    BOOL bHasJanusDRM;
    BYTE bDRMAuxStreamNum;

    BYTE rgbNextPayloadEnd[15];
    BYTE cbNextPayloadEndValid;

} ASFDrm10Context;

typedef struct {
    int seqno;
    int packet_size;
    int is_streamed;
    int asfid2avid[128];        /* conversion table from asf ID 2 AVStream ID */
    ASFStream streams[128]; /* it's max number and it's not that big */
    /* non streamed additonnal info */
    int64_t nb_packets;
    int64_t duration; /* in 100ns units */
    /* packet filling */
    int packet_size_left;
    int packet_timestamp_start;
    int packet_timestamp_end;
    int packet_nb_frames;
    uint8_t packet_buf[PACKET_SIZE];
    ByteIOContext pb;
    /* only for reading */
    uint64_t data_offset; /* begining of the first data packet */

    ASFMainHeader hdr;

    int packet_flags;
    int packet_property;
    int packet_timestamp;
    int packet_segsizetype;
    int packet_segments;
    int packet_seq;
    int packet_replic_size;
    int packet_key_frame;
    int packet_padsize;
    int packet_frag_offset;
    int packet_frag_size;
    int packet_frag_timestamp;
    int packet_multi_size;
    int packet_obj_size;
    int packet_time_delta;
    int packet_time_start;

    int stream_index;
    ASFStream* asf_st; /* currently decoded stream */

    ASFSimpleIndex index;

    unsigned long long ullEndOfData;

    int local_index_table_size;
    ASFIndexEntry* local_index_table;

    ASFDrm10Context drm10;

} ASFContext;

static const GUID asf_header = {
    0x75B22630, 0x668E, 0x11CF, { 0xA6, 0xD9, 0x00, 0xAA, 0x00, 0x62, 0xCE, 0x6C },
};

static const GUID file_header = {
    0x8CABDCA1, 0xA947, 0x11CF, { 0x8E, 0xE4, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 },
};

static const GUID stream_header = {
    0xB7DC0791, 0xA9B7, 0x11CF, { 0x8E, 0xE6, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 },
};

static const GUID audio_stream = {
    0xF8699E40, 0x5B4D, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B },
};

static const GUID audio_conceal_none = {
    // 0x49f1a440, 0x4ece, 0x11d0, { 0xa3, 0xac, 0x00, 0xa0, 0xc9, 0x03, 0x48, 0xf6 },
    // New value lifted from avifile
    0x20fb5700, 0x5b55, 0x11cf, { 0xa8, 0xfd, 0x00, 0x80, 0x5f, 0x5c, 0x44, 0x2b },
};

static const GUID audio_conceal_spread = {
    0xBFC3CD50, 0x618F, 0x11cf, { 0x8b, 0xb2, 0x00, 0xaa, 0x00, 0xb4, 0xe2, 0x20 },
};

static const GUID video_stream = {
    0xBC19EFC0, 0x5B4D, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B },
};

static const GUID video_conceal_none = {
    0x20FB5700, 0x5B55, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B },
};

static const GUID comment_header = {
    0x75b22633, 0x668e, 0x11cf, { 0xa6, 0xd9, 0x00, 0xaa, 0x00, 0x62, 0xce, 0x6c },
};

static const GUID codec_comment_header = {
    0x86D15240, 0x311D, 0x11D0, { 0xA3, 0xA4, 0x00, 0xA0, 0xC9, 0x03, 0x48, 0xF6 },
};
static const GUID codec_comment1_header = {
    0x86d15241, 0x311d, 0x11d0, { 0xa3, 0xa4, 0x00, 0xa0, 0xc9, 0x03, 0x48, 0xf6 },
};

static const GUID data_header = {
    0x75b22636, 0x668e, 0x11cf, { 0xa6, 0xd9, 0x00, 0xaa, 0x00, 0x62, 0xce, 0x6c },
};

static const GUID index_guid = {
    0x33000890, 0xe5b1, 0x11cf, { 0x89, 0xf4, 0x00, 0xa0, 0xc9, 0x03, 0x49, 0xcb },
};

static const GUID head1_guid = {
    0x5fbf03b5, 0xa92e, 0x11cf, { 0x8e, 0xe3, 0x00, 0xc0, 0x0c, 0x20, 0x53, 0x65 },
};

static const GUID head2_guid = {
    0xabd3d211, 0xa9ba, 0x11cf, { 0x8e, 0xe6, 0x00, 0xc0, 0x0c, 0x20, 0x53, 0x65 },
};

#ifdef MSDRM10_SUPPORT
static const GUID content_encryption = {
    0x2211b3fb, 0xbd23, 0x11d2, { 0xb4, 0xb7, 0x0, 0xa0, 0xc9, 0x55, 0xfc, 0x6e },
};

static const GUID content_encryption_ex = {
    0x298ae614, 0x2622, 0x4c17, { 0xb9, 0x35, 0xda, 0xe0, 0x7e, 0xe9, 0x28, 0x9c },
};

static const GUID drm_aux_stream = {
    0xd5c82c81, 0x8264, 0x4abd, { 0xa7, 0x14, 0xcd, 0x4f, 0x85, 0x54, 0xc7, 0x2e },
};

#endif /* MSDRM10_SUPPORT */

/* I am not a number !!! This GUID is the one found on the PC used to
   generate the stream */
static const GUID my_guid = {
    0, 0, 0, { 0, 0, 0, 0, 0, 0, 0, 0 },
};

static const GUID padding_guid = {
    0x1806d474, 0xcadf, 0x4509, { 0xa4, 0xba, 0x9a, 0xab, 0xcb, 0x96, 0xaa, 0xe8 },
};

static int last_video_pts = 0;

#define ASF_STEP 5
#define ASF_TABLE_SIZE 30
static int asf_actual_max_table_size = ASF_TABLE_SIZE;
//
// Function declarations
//
static BOOL asf_calc_frame_rate(AVFormatContext *s);
static int  asf_read_seek(AVFormatContext *s, int64_t pts);
static void asf_build_local_index_table( AVFormatContext *s, int stream_id, ASFDirection direction, int step, int sper);
static void asf_restore_context( AVFormatContext *s, ASFIndexEntry *entry);
static int asf_find_current_location(AVFormatContext *s, int stream_id, int cur_pos);
static int asf_move_from_current(AVFormatContext *s, int stream_id, int step, int cur_entry);

#ifdef MSDRM10_SUPPORT
static int asf_msdrm10_init(ASFContext *asf);
static void asf_msdrm10_free(ASFContext *asf);
#endif /* MSDRM10_SUPPORT */

static void put_guid(ByteIOContext *s, const GUID *g)
{
    int i;

    put_le32(s, g->v1);
    put_le16(s, g->v2);
    put_le16(s, g->v3);
    for(i=0;i<8;i++)
        put_byte(s, g->v4[i]);
}

static void put_str16(ByteIOContext *s, const char *tag)
{
    int c;

    put_le16(s,strlen(tag) + 1);
    for(;;) {
        c = (uint8_t)*tag++;
        put_le16(s, c);
        if (c == '\0')
            break;
    }
}

static void put_str16_nolen(ByteIOContext *s, const char *tag)
{
    int c;

    for(;;) {
        c = (uint8_t)*tag++;
        put_le16(s, c);
        if (c == '\0')
            break;
    }
}

static int64_t put_header(ByteIOContext *pb, const GUID *g)
{
    int64_t pos;

    pos = url_ftell(pb);
    put_guid(pb, g);
    put_le64(pb, 24);
    return pos;
}

/* update header size */
static void end_header(ByteIOContext *pb, int64_t pos)
{
    int64_t pos1;

    pos1 = url_ftell(pb);
    Avf_url_fseek(pb, pos + 16, SEEK_SET);
    put_le64(pb, pos1 - pos);
    Avf_url_fseek(pb, pos1, SEEK_SET);
}

/* write an asf chunk (only used in streaming case) */
static void put_chunk(AVFormatContext *s, int type, int payload_length, int flags)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int length;

    length = payload_length + 8;
    put_le16(pb, type);
    put_le16(pb, length);
    put_le32(pb, asf->seqno);
    put_le16(pb, flags); /* unknown bytes */
    put_le16(pb, length);
    asf->seqno++;
}

/* convert from unix to windows time */
static int64_t unix_to_file_time(int ti)
{
    int64_t t;

    t = ti * int64_t_C(10000000);
    t += int64_t_C(116444736000000000);
    return t;
}

/* write the header (used two times if non streamed) */
static int asf_write_header1(AVFormatContext *s, int64_t file_size, int64_t data_chunk_size)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int header_size, n, extra_size, extra_size2, wav_extra_size, file_time;
    int has_title;
    AVCodecContext *enc;
    int64_t header_offset, cur_pos, hpos;
    int bit_rate;

    has_title = (s->title[0] || s->author[0] || s->copyright[0] || s->comment[0]);

    bit_rate = 0;
    for(n=0;n<s->nb_streams;n++) {
        enc = &s->streams[n]->codec;

        bit_rate += enc->bit_rate;
    }

    if (asf->is_streamed) {
        put_chunk(s, 0x4824, 0, 0xc00); /* start of stream (length will be patched later) */
    }

    put_guid(pb, &asf_header);
    put_le64(pb, (uint64_t)-1); /* header length, will be patched after */
    put_le32(pb, 3 + has_title + s->nb_streams); /* number of chunks in header */
    put_byte(pb, 1); /* ??? */
    put_byte(pb, 2); /* ??? */

    /* file header */
    header_offset = url_ftell(pb);
    hpos = put_header(pb, &file_header);
    put_guid(pb, &my_guid);
    put_le64(pb, file_size);
    file_time = 0;
    put_le64(pb, unix_to_file_time(file_time));
    put_le64(pb, asf->nb_packets); /* number of packets */
    put_le64(pb, asf->duration); /* end time stamp (in 100ns units) */
    put_le64(pb, asf->duration); /* duration (in 100ns units) */
    put_le32(pb, 0); /* start time stamp */
    put_le32(pb, 0); /* ??? */
    put_le32(pb, asf->is_streamed ? 3 : 2); /* always seekable, sometimes streamed */
    put_le32(pb, asf->packet_size); /* packet size */
    put_le32(pb, asf->packet_size); /* packet size */
    put_le32(pb, bit_rate); /* Nominal data rate in bps */
    end_header(pb, hpos);

    /* unknown headers */
    hpos = put_header(pb, &head1_guid);
    put_guid(pb, &head2_guid);
    put_le32(pb, 6);
    put_le16(pb, 0);
    end_header(pb, hpos);

    /* title and other infos */
    if (has_title) {
        hpos = put_header(pb, &comment_header);
        put_le16(pb, 2 * (strlen(s->title) + 1));
        put_le16(pb, 2 * (strlen(s->author) + 1));
        put_le16(pb, 2 * (strlen(s->copyright) + 1));
        put_le16(pb, 2 * (strlen(s->comment) + 1));
        put_le16(pb, 0);
        put_str16_nolen(pb, s->title);
        put_str16_nolen(pb, s->author);
        put_str16_nolen(pb, s->copyright);
        put_str16_nolen(pb, s->comment);
        end_header(pb, hpos);
    }

    /* stream headers */
    for(n=0;n<s->nb_streams;n++) {
        int64_t es_pos;
        //        ASFStream *stream = &asf->streams[n];

        enc = &s->streams[n]->codec;
        asf->streams[n].num = n + 1;
        asf->streams[n].seq = 0;

        switch(enc->codec_type) {
        case CODEC_TYPE_AUDIO:
            wav_extra_size = 0;
            extra_size = 18 + wav_extra_size;
            extra_size2 = 8; //error correction data length
            break;
        default:
        case CODEC_TYPE_VIDEO:
            wav_extra_size = 0;
            extra_size = 0x33 + enc->extradata_size;
            extra_size2 = 0;
            break;
        }

        hpos = put_header(pb, &stream_header);
        if (enc->codec_type == CODEC_TYPE_AUDIO) {
            put_guid(pb, &audio_stream);
            put_guid(pb, &audio_conceal_spread);
        } else {
            put_guid(pb, &video_stream);
            put_guid(pb, &video_conceal_none);
        }
        put_le64(pb, 0); /* offset */
        es_pos = url_ftell(pb);
        put_le32(pb, extra_size); /* wav header len */
        put_le32(pb, extra_size2); /* error correction data length */
        put_le16(pb, n + 1); /* stream number */
        put_le32(pb, 0); /* reserved */

        if (enc->codec_type == CODEC_TYPE_AUDIO) {
            /* WAVEFORMATEX header */
            int wavsize = put_wav_header(pb, enc);

            if (wavsize < 0)
                return -1;
            if (wavsize != extra_size) {
                cur_pos = url_ftell(pb);
                Avf_url_fseek(pb, es_pos, SEEK_SET);
                put_le32(pb, wavsize); /* wav header len */
                Avf_url_fseek(pb, cur_pos, SEEK_SET);
            }
            
            // Spread audio error correction
            if ( enc->codec_id == CODEC_ID_WMAV8Z && s->nb_streams == 1 )
            {
                // For WMA: Audio only - span 1, Video + Audio - span 2
                put_byte(pb,1);   // Span
            }
            else
            {
                put_byte(pb,2);                     // Can be spread to 2 packets
            }
            put_le16(pb,enc->block_align);
            put_le16(pb,enc->block_align);
            put_le16(pb,1);
            put_byte(pb,0);

        } else {
            put_le32(pb, enc->logical_width);
            put_le32(pb, enc->logical_height);
            put_byte(pb, 2); /* ??? */
            // :ZZZ: bug fix - added extradata_size
            put_le16(pb, 40 + enc->extradata_size); /* size */

            /* BITMAPINFOHEADER header */
            put_bmp_header(pb, enc, codec_bmp_tags, 1);
        }
        end_header(pb, hpos);
    }

    /* media comments */

    hpos = put_header(pb, &codec_comment_header);
    put_guid(pb, &codec_comment1_header);
    put_le32(pb, s->nb_streams);
    for(n=0;n<s->nb_streams;n++) {
        AVCodec *p;

        enc = &s->streams[n]->codec;
        p = AvApi_avcodec_find_encoder(enc->codec_id);

        // :ZZZ: BUG fix       put_le16(pb, asf->streams[n].num);

        /* codec type as specified in asf spec */
        if (enc->codec_type == CODEC_TYPE_VIDEO) {
            put_le16(pb, 0x0001);  /* Video codec */
        } else if (enc->codec_type == CODEC_TYPE_AUDIO) {
            put_le16(pb, 0x0002);  /* Audio codec */
        } else {
            put_le16(pb, 0xFFFF);  /* Unknown codec */
        }            
        put_str16(pb, p ? p->name : enc->codec_name);
        put_le16(pb, 0); /* no parameters */
        /* id */
        if (enc->codec_type == CODEC_TYPE_AUDIO) {
            put_le16(pb, 2);
            put_le16(pb, codec_get_tag(codec_wav_tags, enc->codec_id));
        } else {
            put_le16(pb, 4);
            put_le32(pb, codec_get_tag(codec_bmp_tags, enc->codec_id));
        }
    }
    end_header(pb, hpos);

    /* patch the header size fields */

    cur_pos = url_ftell(pb);
    header_size = cur_pos - header_offset;
    if (asf->is_streamed) {
        header_size += 8 + 30 + 50;

        Avf_url_fseek(pb, header_offset - 10 - 30, SEEK_SET);
        put_le16(pb, header_size);
        Avf_url_fseek(pb, header_offset - 2 - 30, SEEK_SET);
        put_le16(pb, header_size);

        header_size -= 8 + 30 + 50;
    }
    header_size += 24 + 6;
    Avf_url_fseek(pb, header_offset - 14, SEEK_SET);
    put_le64(pb, header_size);
    Avf_url_fseek(pb, cur_pos, SEEK_SET);

    /* movie chunk, followed by packets of packet_size */
    asf->data_offset = cur_pos;
    put_guid(pb, &data_header);
    put_le64(pb, data_chunk_size);
    put_guid(pb, &my_guid);  /* file ID */
    put_le64(pb, asf->nb_packets); /* nb packets */
    put_byte(pb, 1); /* ??? */
    put_byte(pb, 1); /* ??? */
    return 0;
}

static int asf_write_header(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;
    AVStream *st;
    int i;

    av_set_pts_info(s, 32, 1, 1000); /* 32 bit pts in ms */
    
    asf->packet_size = PACKET_SIZE;

    for ( i=0; i < s->nb_streams; i++)
    {
        st = s->streams[i];
        if ( st->codec.codec_type == CODEC_TYPE_AUDIO )
        {
            if ( st->codec.codec_id == CODEC_ID_WMAV8Z )
            {
                asf->packet_size = st->codec.frame_size * st->codec.bit_rate / st->codec.sample_rate +
                                    PACKET_HEADER_SIZE + FRAME_HEADER_SIZE ;
                break;
            }
        }
    }

    asf->nb_packets = 0;

    if (asf_write_header1(s, 0, 50) < 0) {
        //avf_free(asf);
        return -1;
    }

    put_flush_packet(&s->pb);

    asf->packet_nb_frames = 0;
    asf->packet_timestamp_start = -1;
    asf->packet_timestamp_end = -1;
    asf->packet_size_left = asf->packet_size - PACKET_HEADER_SIZE;
    init_put_byte(&asf->pb, asf->packet_buf, asf->packet_size, 1,
                  NULL, NULL, NULL, NULL);

    //init. index
    asf->index.allocated_entries = 0;
    asf->index.last_entry        = 0;
    asf->index.entries           = NULL;
    asf->index.time_interval     = 1000; //1 second

    last_video_pts = 0;
    
    return 0;
}

static int asf_write_stream_header(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;

    asf->is_streamed = 1;

    return asf_write_header(s);
}

/* write a fixed size packet */
static int put_packet(AVFormatContext *s,
                       unsigned int timestamp, unsigned int duration,
                       int nb_frames, int padsize)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int flags;

    if (asf->is_streamed) {
        put_chunk(s, 0x4424, asf->packet_size, 0);
    }

    put_byte(pb, 0x82);
    put_le16(pb, 0);

    flags = 0x01; /* nb segments present */
    if (padsize > 0) {
        if (padsize < 256)
            flags |= 0x08;
        else
            flags |= 0x10;
    }
    //Payload parsing information begins
    put_byte(pb, flags); /* flags */
    put_byte(pb, 0x5d);  /* Property flag */
    if (flags & 0x10)
        put_le16(pb, padsize - 2);
    if (flags & 0x08)
        put_byte(pb, padsize - 1);
    put_le32(pb, timestamp);
    put_le16(pb, duration);
                                                
    //payload parsing information ends
    put_byte(pb, nb_frames | 0x80);

    return PACKET_HEADER_SIZE + ((flags & 0x18) >> 3);
}

static void flush_packet(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;
    int hdr_size, ptr;

    hdr_size = put_packet(s, asf->packet_timestamp_start,
               asf->packet_timestamp_end - asf->packet_timestamp_start,
               asf->packet_nb_frames, asf->packet_size_left);

    /* Correct left size according to actual header size */  
    asf->packet_size_left -= (hdr_size - PACKET_HEADER_SIZE);

    /* Clear out the padding bytes */
    ptr = asf->packet_size - hdr_size - asf->packet_size_left;
    memset(asf->packet_buf + ptr, 0, asf->packet_size_left);

    put_buffer(&s->pb, asf->packet_buf, asf->packet_size - hdr_size);

    put_flush_packet(&s->pb);
    asf->nb_packets++;
    asf->packet_nb_frames = 0;
    asf->packet_timestamp_start = -1;
    asf->packet_timestamp_end = -1;
    asf->packet_size_left = asf->packet_size - PACKET_HEADER_SIZE;
    init_put_byte(&asf->pb, asf->packet_buf, asf->packet_size, 1,
                  NULL, NULL, NULL, NULL);    
}

static void put_frame_header(AVFormatContext *s, ASFStream *stream, int timestamp,
                             int payload_size, int frag_offset, int frag_len)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &asf->pb;
    int val;

    val = stream->num;

    // :ZZZ: Take key_frame info from FormatContext, and not from
    // EncContext. We must separate these variables since encoder is already
    // encoding new frames by the time this frame is written
//    if (s->streams[val - 1]->codec.coded_frame->key_frame /* && frag_offset == 0 */)
    if ( s->key_frame )
        val |= 0x80;
    put_byte(pb, val);
    put_byte(pb, stream->seq);
    put_le32(pb, frag_offset); /* fragment offset */
    put_byte(pb, 0x08); /* flags */
    put_le32(pb, payload_size);
    put_le32(pb, timestamp);
    put_le16(pb, frag_len);
               
    //add a new index entry, if needed
    if ((s->key_frame) &&
         (s->streams[stream->num-1]->codec.codec_type == CODEC_TYPE_VIDEO))    
    {
        if (asf->index.entries != NULL)
        {
            int new_entry_pos = timestamp / asf->index.time_interval; 
            int last_entry_pos = asf->index.last_entry;
            int i;

            // fill the entries from the last entry to the new one with the value of the new time
            for (i = (last_entry_pos + 1); i<=new_entry_pos; i++) 
            {
                //allocate more space for the index if needed
                if (asf->index.allocated_entries <= i)
                {
                    asf->index.entries = avf_realloc(asf->index.entries, 
                                                    asf->index.allocated_entries * 2 * sizeof(ASFSimpleIndexEntry));
                    asf->index.allocated_entries*=2;
                }

                asf->index.entries[i].packet_num   = (int)asf->nb_packets;
                asf->index.entries[i].packet_count = 1; //FIXME: temporary solution...
                //printf("just put index - %d %d\n",(int)asf->nb_packets, (int)asf->index.entries[i].packet_num);
                asf->index.last_entry = i;
            }
        }
        else //empty index
        {
            int first_alloc_size = 10;

            //allocate space
            asf->index.entries = avf_mallocz(first_alloc_size*sizeof(ASFSimpleIndexEntry));
            asf->index.allocated_entries = first_alloc_size;

            asf->index.entries[0].packet_num   = asf->nb_packets;
            asf->index.entries[0].packet_count = 1; //FIXME: temporary solution...

            asf->index.last_entry = 0;
        }
    }
}


/* Output a frame. We suppose that payload_size <= PACKET_SIZE.

   It is there that you understand that the ASF format is really
   crap. They have misread the MPEG Systems spec !
 */
static void put_frame(AVFormatContext *s, ASFStream *stream, int timestamp,
                      uint8_t *buf, int payload_size)
{
    ASFContext *asf = s->priv_data;
    int frag_pos, frag_len, frag_len1;

    frag_pos = 0;
    while (frag_pos < payload_size) {
        frag_len = payload_size - frag_pos;
        frag_len1 = asf->packet_size_left - FRAME_HEADER_SIZE;
        if (frag_len1 > 0) {
            if (frag_len > frag_len1)
                frag_len = frag_len1;

            // Timestamp of Audio is equal to Video
            if (stream->num == 2) //Audio
            {
                timestamp = last_video_pts;
            }
            else
            {                                           
                last_video_pts = timestamp;
            }

            put_frame_header(s, stream, timestamp+1, payload_size, frag_pos, frag_len);
            put_buffer(&asf->pb, buf, frag_len);
            asf->packet_size_left -= (frag_len + FRAME_HEADER_SIZE);
            asf->packet_timestamp_end = timestamp;

            if (asf->packet_timestamp_start == -1)
                asf->packet_timestamp_start = timestamp;
            asf->packet_nb_frames++;
        } else {
            frag_len = 0;
        }
        frag_pos += frag_len;
        buf += frag_len;
        /* output the frame if filled */
        if (asf->packet_size_left <= FRAME_HEADER_SIZE)
            flush_packet(s);
    }
    stream->seq++;
}


static int asf_write_packet(AVFormatContext *s, int stream_index,
                            uint8_t *buf, int size, int64_t timestamp)
{
    ASFContext *asf = s->priv_data;
    ASFStream *stream;
    int64_t duration;
    AVCodecContext *codec;
    int packet_size;

    codec = &s->streams[stream_index]->codec;
    stream = &asf->streams[stream_index];

    if (codec->codec_type == CODEC_TYPE_AUDIO) {
        if ( codec->codec_id == CODEC_ID_WMAV8Z )
        {
            packet_size = codec->frame_size * codec->bit_rate / codec->sample_rate;
            duration = (s->streams[stream_index]->stream_total_nb_frames * packet_size * int64_t_C(1000000)) /
                        codec->bit_rate;
        }
        else
        {
            duration = (s->streams[stream_index]->stream_total_nb_frames * codec->frame_size * int64_t_C(10000000)) /
                        codec->sample_rate;
        }
    } else {
        duration = AvApi_rescale(s->streams[stream_index]->stream_total_nb_frames * codec->frame_rate_base, 10000000, codec->frame_rate);
    }

    if (duration > asf->duration)
        asf->duration = duration;

    //:DANIELK: asf->duration was not updated properly
    asf->duration = timestamp * 10000; // from milli to 100nano

    //  :EREZ: temporary cast time stamp
    put_frame(s, stream, (int)timestamp, buf, size);

    // Update the total number of frames for this stream
    s->streams[stream_index]->stream_total_nb_frames++;

    return 0;
}


static int put_simple_index(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int i;

    // Don't put index if there are no entries
    if (asf->index.entries == NULL)
        return 0;

    // write simple index object header
    put_guid(pb, &index_guid); 
    put_le64(pb, 56 + (asf->index.last_entry+1)*(sizeof(int)+sizeof(short))); //size
    put_guid(pb, &my_guid); //fileID
    put_le64(pb, asf->index.time_interval * 10000); //convert milli to 100-nano
    put_le32(pb, 2); //FIXME
    put_le32(pb, asf->index.last_entry + 1); 

    // write all index entries
    for (i=0; i<=asf->index.last_entry; i++)
    {
        put_le32(pb, asf->index.entries[i].packet_num);
        put_le16(pb, asf->index.entries[i].packet_count);
    }

    avf_free(asf->index.entries);

    return 0;
}

static int asf_write_trailer(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;

    int64_t file_size, file_size_before_index;
    int i;

    // Check if the last packed wasn't flushed
    if ( asf->packet_nb_frames > 0 )
    {
        // The last packet wasn't flushed.
        flush_packet(s);
    }

    // Flush low level (media) context (ByteIOContext) 
    put_flush_packet_end(pb);

    // Recover from file reopen (from burst to regular mode)
    // Force to flush the buffer. 
    pb->must_flush = 1;
    Avf_url_fseek(pb, url_filesize(url_fileno(pb)), SEEK_SET); 

    pb->must_flush = 1;

    file_size_before_index = url_ftell(&s->pb);

    put_simple_index(s);

    if (asf->is_streamed) {
        put_chunk(s, 0x4524, 0, 0); /* end of stream */
    } else {
        /* rewrite an updated header */
        file_size = url_ftell(&s->pb);

        Avf_url_fseek(&s->pb, 0, SEEK_SET);
        asf_write_header1(s, file_size, file_size_before_index - asf->data_offset);
    }

    // :ZZZ: Must flush buffer, since we accumulate frames
    put_flush_packet_end(&s->pb);

    // :ZZZ:
    for(i=0;i<s->nb_streams;i++) 
    {
        AVStream *st = s->streams[i];
        avf_free(st->codec.extradata);
		st->codec.extradata = NULL;
    }
    
    return 0;
}

static int asf_estimate_trailer_size(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;

    return asf->index.last_entry * ( sizeof(LONG) + sizeof(SHORT) );
}


/**********************************/
/* decoding */

//#define DEBUG_ASF

#ifdef DEBUG_ASF
#define PRINT_IF_GUID(g,cmp) \
if (!memcmp(g, &cmp, sizeof(GUID))) \
    printf("(GUID: %s) ", #cmp)

static void print_guid(const GUID *g)
{
    int i;
    PRINT_IF_GUID(g, asf_header);
    else PRINT_IF_GUID(g, file_header);
    else PRINT_IF_GUID(g, stream_header);
    else PRINT_IF_GUID(g, audio_stream);
    else PRINT_IF_GUID(g, audio_conceal_none);
    else PRINT_IF_GUID(g, video_stream);
    else PRINT_IF_GUID(g, video_conceal_none);
    else PRINT_IF_GUID(g, comment_header);
    else PRINT_IF_GUID(g, codec_comment_header);
    else PRINT_IF_GUID(g, codec_comment1_header);
    else PRINT_IF_GUID(g, data_header);
    else PRINT_IF_GUID(g, index_guid);
    else PRINT_IF_GUID(g, head1_guid);
    else PRINT_IF_GUID(g, head2_guid);
    else PRINT_IF_GUID(g, my_guid);
    else PRINT_IF_GUID(g, padding_guid);
    else
        printf("(GUID: unknown) ");
    printf("0x%08x, 0x%04x, 0x%04x, {", g->v1, g->v2, g->v3);
    for(i=0;i<8;i++)
        printf(" 0x%02x,", g->v4[i]);
    printf("}\n");
}
#undef PRINT_IF_GUID

#endif

static void get_guid(ByteIOContext *s, GUID *g)
{
    int i;

    g->v1 = get_le32(s);
    g->v2 = get_le16(s);
    g->v3 = get_le16(s);
    for(i=0;i<8;i++)
        g->v4[i] = get_byte(s);
}

#if 0
static void get_str16(ByteIOContext *pb, char *buf, int buf_size)
{
    int len, c;
    char *q;

    len = get_le16(pb);
    q = buf;
    while (len > 0) {
        c = get_le16(pb);
        if ((q - buf) < buf_size - 1)
            *q++ = c;
        len--;
    }
    *q = '\0';
}
#endif

#ifdef MSDRM10_SUPPORT
static int get_buff32(ByteIOContext *pb, unsigned char* buf)
{
    int len;
    unsigned char *q;
    int b;

    len = get_le32(pb);

    if ( !len )
        return 0;

    q = buf;

    while (len > 0) {

        b = get_byte(pb);

        if ( buf )
        {
            *q = b;
            q++;
        }

        len--;
    }

    return q-buf;
}

static int get_buff32_alloc(ByteIOContext *pb, unsigned char** pbuf)
{
    int len;
    unsigned char *q;

    len = get_le32(pb);

    if ( !len )
        return 0;

    if ( !pbuf )
    {
        return len;
    }

    *pbuf = avf_malloc( len + 1 );

    q = *pbuf;

    while (len > 0) {

        *q = get_byte(pb);;
        q++;

        len--;
    }

    *q = '\0';

    return q-*pbuf;
}
#endif /* MSDRM10_SUPPORT */

static void get_str16_nolen(ByteIOContext *pb, int len, char *buf, int buf_size)
{
    int c;
    char *q;

    q = buf;
    while (len > 0) {
        c = get_le16(pb);
        if ((q - buf) < buf_size - 1)
            *q++ = c;
        len-=2;
    }
    *q = '\0';
}

static int asf_probe(AVProbeData *pd)
{
    GUID g;
    const unsigned char *p;
    int i;

    /* check file header */
    if (pd->buf_size <= 32)
        return 0;
    p = pd->buf;
    g.v1 = p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
    p += 4;
    g.v2 = p[0] | (p[1] << 8);
    p += 2;
    g.v3 = p[0] | (p[1] << 8);
    p += 2;
    for(i=0;i<8;i++)
        g.v4[i] = *p++;

    if (!memcmp(&g, &asf_header, sizeof(GUID)))
        return AVPROBE_SCORE_MAX;
    else
        return 0;
}

static void get_simple_index(AVFormatContext *s, int bLoadIndex)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    GUID g;
    int64_t gsize;
    int i;

    offset_t cur_offs = url_ftell(pb); //save current offset

    Avf_url_fseek(pb, 0, SEEK_SET); //go to the top

    memset(&asf->index, 0, sizeof(ASFSimpleIndex)); //init. index structure

    // Init begining of index to invalid value
    asf->ullEndOfData = (unsigned long long)-1;

    for (;;)
    {
        get_guid(pb, &g);

        if (!memcmp(&g, &index_guid, sizeof(GUID)))
        {
            break; //found ! stop searching...
        }
        else
        {
            offset_t pos,ret;
            gsize = get_le64(pb);        

            if (gsize == 0) //eof
            {               
                // index is not present... the index structure remains empty
                Avf_url_fseek(pb, cur_offs, SEEK_SET); //restore offset
                return;
            }

            //If here , gsize !=0
            pos = url_ftell(pb);
            ret = Avf_url_fseek(pb, gsize - 24, SEEK_CUR);
            // Wanted to move an object further, but stayed in the same place : loop
            if (pos == ret + 24)
            {
                printf("ASF WARNING: file is possibly corrupted\n");
                return;
            }
        }                               
    }

    // index was found

    // Remember the beginning of Index in the file
    asf->ullEndOfData = url_ftell(pb) - sizeof(GUID);

    get_le64(pb); //skip ObjSize
    get_guid(pb, &g); //skip FileID
    asf->index.time_interval = get_le64(pb) / 10000; //convert from 100-nanos to millis
   
    get_le32(pb); //skip MaxPacketCnt
    
    asf->index.last_entry = get_le32(pb) - 1; //read entries number
    
    if (asf->index.last_entry == -1) //no entries
    {
        Avf_url_fseek(pb, cur_offs, SEEK_SET); //restore offset     
        return;
    }
    //allocate enough space                  
    asf->index.entries = avf_malloc((asf->index.last_entry + 1)*sizeof(ASFSimpleIndexEntry));
    asf->index.allocated_entries = asf->index.last_entry + 1;

    for (i=0; i <= asf->index.last_entry; i++)
    {
        asf->index.entries[i].packet_num   = get_le32(pb);
        asf->index.entries[i].packet_count = get_le16(pb);
        
        if (bLoadIndex == 0)
        {
            break;
        }
    }

    Avf_url_fseek(pb, cur_offs, SEEK_SET); //restore offset
}

static int asf_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    ASFContext *asf = s->priv_data;
    GUID g;
    ByteIOContext *pb = &s->pb;
    AVStream *st;
    ASFStream *asf_st;
    int size, i;
    int64_t gsize;
    
    if ( (ap != NULL) && (ap->flags & AV_READ_HEADER_FLAG_NO_INDEX) )
    {
        get_simple_index(s, 0); 
    }
    else
    {
        get_simple_index(s, 1); //fetch the index table from the file, if such is present
    }
    
    av_set_pts_info(s, 32, 1, 1000); /* 32 bit pts in ms */
    
    get_guid(pb, &g);
    if (memcmp(&g, &asf_header, sizeof(GUID)))
        goto fail;
    get_le64(pb);
    get_le32(pb);
    get_byte(pb);
    get_byte(pb);
    memset(&asf->asfid2avid, -1, sizeof(asf->asfid2avid));
    for(;;) {
        get_guid(pb, &g);
        gsize = get_le64(pb);
#ifdef DEBUG_ASF
        printf("%08Lx: ", url_ftell(pb) - 24);
        print_guid(&g);
        printf("  size=0x%Lx\n", gsize);
#endif
        if (gsize < 24)
            goto fail;
        if (!memcmp(&g, &file_header, sizeof(GUID))) {
            get_guid(pb, &asf->hdr.guid);
            asf->hdr.file_size      = get_le64(pb);
            asf->hdr.create_time    = get_le64(pb);
            asf->hdr.packets_count  = get_le64(pb);
            asf->hdr.play_time      = get_le64(pb);
            asf->hdr.send_time      = get_le64(pb);
            asf->hdr.preroll        = get_le32(pb);
            asf->hdr.ignore     = get_le32(pb);
            asf->hdr.flags      = get_le32(pb);
            asf->hdr.min_pktsize    = get_le32(pb);
            asf->hdr.max_pktsize    = get_le32(pb);
            asf->hdr.max_bitrate    = get_le32(pb);
            // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
            printf( "File header :\n" );
            printf( "File size       : %d\n", (unsigned long)(asf->hdr.file_size) );
            printf( "CreateTime      : %d\n", asf->hdr.create_time ); 
            printf( "Packets count   : %d\n", (unsigned long)(asf->hdr.packets_count) ); 
            printf( "Play duration   : %d\n", (unsigned long)(asf->hdr.play_time) ); 
            printf( "Send duration   : %d\n", (unsigned long)(asf->hdr.send_time) ); 
            printf( "Preroll         : %d\n", asf->hdr.preroll ); 
            printf( "Ignore          : %d\n", asf->hdr.ignore ); 
            printf( "Min packet size : %d\n", asf->hdr.min_pktsize ); 
            printf( "Max packet size : %d\n", asf->hdr.max_pktsize ); 
            printf( "Max bit rate    : %d\n", asf->hdr.max_bitrate ); 
#endif
            asf->packet_size = asf->hdr.max_pktsize;
            asf->nb_packets = asf->hdr.packets_count;
        } else if (!memcmp(&g, &stream_header, sizeof(GUID))) {
            int type, total_size, type_specific_size;
            unsigned int tag1;
            int64_t pos1, pos2;
            int drm_aux = 0; 
            
            // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
            printf( "Stream header :\n" );
#endif
            
            pos1 = url_ftell(pb);
            
            st = avf_mallocz(sizeof(AVStream));
            if (!st)
                goto fail;

            AvApi_avcodec_get_context_defaults(&st->codec);
            asf->drm10.bDRMAuxStreamNum = 0;
            
            asf_st = avf_mallocz(sizeof(ASFStream));
            if (!asf_st)
                goto fail;
            st->priv_data = asf_st;
            st->time_length = (asf->hdr.send_time - asf->hdr.preroll) / 10000; //ms
            
            get_guid(pb, &g);

            // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
            printf( "Stream type   : 0x%08X\n", g.v1 );
#endif
            if (!memcmp(&g, &audio_stream, sizeof(GUID))) {
                type = CODEC_TYPE_AUDIO;
            } else if (!memcmp(&g, &video_stream, sizeof(GUID))) {
                type = CODEC_TYPE_VIDEO;
#ifdef MSDRM10_SUPPORT
            } else if (!memcmp(&g, &drm_aux_stream, sizeof(GUID))) {
                drm_aux = 1;
#endif
            } else {
                goto fail;
            }

            get_guid(pb, &g);
            // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
            printf( "Error type    : 0x%08X\n", g.v1 );
#endif
            total_size = get_le64(pb);
            type_specific_size = get_le32(pb);
            get_le32(pb);
            // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
            printf( "Time offset   : %d\n", (unsigned long)total_size );
            printf( "Type specific : %d\n", type_specific_size );
#endif
            st->id = get_le16(pb) & 0x7f; /* stream id */

            if (drm_aux) {
                asf->drm10.bDRMAuxStreamNum = st->id;
            }

            // dummy            
            get_le32(pb);

            if (drm_aux) {
                avf_free( asf_st );
                avf_free( st );
                pos2 = url_ftell(pb);
                url_fskip(pb, gsize - (pos2 - pos1 + 24));
                continue;
            }

            s->streams[s->nb_streams] = st;
            // mapping of asf ID to AV stream ID;
            asf->asfid2avid[st->id] = s->nb_streams++;
            st->codec.codec_type = (enum CodecType)type;
            
            st->codec.frame_rate = 15 * s->pts_den / s->pts_num; // 15 fps default

            if (type == CODEC_TYPE_AUDIO) {
                get_wav_header(pb, &st->codec, type_specific_size);
                /* We have to init the frame size at some point .... */
                pos2 = url_ftell(pb);
                if (gsize > (pos2 + 8 - pos1 + 24)) {
                    asf_st->ds_span = get_byte(pb);
                    asf_st->ds_packet_size = get_le16(pb);
                    asf_st->ds_chunk_size = get_le16(pb);
                    asf_st->ds_data_size = get_le16(pb);
                    asf_st->ds_silence_data = get_byte(pb);
                }
                //printf("Descrambling: ps:%d cs:%d ds:%d s:%d  sd:%d\n",
                //       asf_st->ds_packet_size, asf_st->ds_chunk_size,
                //       asf_st->ds_data_size, asf_st->ds_span, asf_st->ds_silence_data);
                if (asf_st->ds_span > 1) {
                    if (!asf_st->ds_chunk_size
                        || (asf_st->ds_packet_size/asf_st->ds_chunk_size <= 1))
                        asf_st->ds_span = 0; // disable descrambling
                }
                switch (st->codec.codec_id) {
                case CODEC_ID_MP3LAME:
                    st->codec.frame_size = MPA_FRAME_SIZE;
                    break;
                case CODEC_ID_PCM_S16LE:
                case CODEC_ID_PCM_S16BE:
                case CODEC_ID_PCM_U16LE:
                case CODEC_ID_PCM_U16BE:
                case CODEC_ID_PCM_S8:
                case CODEC_ID_PCM_U8:
                case CODEC_ID_PCM_ALAW:
                case CODEC_ID_PCM_MULAW:
                    st->codec.frame_size = 1;
                    break;
                default:
                    /* This is probably wrong, but it prevents a crash later */
                    st->codec.frame_size = 1;
                    break;
                }
            } else {
                // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
                printf( "Video media type : Format data   : 0x%08X\n", g.v1 );
#endif
                get_le32(pb);
                get_le32(pb);
                get_byte(pb);
                get_le16(pb); /* size */
                size = get_le32(pb); /* size */

                st->codec.logical_width = get_le32(pb);
                st->codec.logical_height = get_le32(pb);
                st->codec.physical_width = ( st->codec.logical_width + 15 ) & (~0xF);
                st->codec.physical_height = ( st->codec.logical_height + 15 ) & (~0xF);

                /* not available for asf */
                get_le16(pb); /* panes */
                st->codec.bits_per_sample = get_le16(pb); /* depth */
                tag1 = get_le32(pb);
                // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
                printf( "Width          : %d\n", st->codec.logical_width );
                printf( "Height         : %d\n", st->codec.logical_height );
                printf( "Bpp            : %d\n", st->codec.bits_per_sample );
                printf( "Compression ID : 0x%08X\n", tag1 );
#endif
                url_fskip(pb, 20);
                
                if (size > 40) {
                    st->codec.extradata_size = size - 40;
                    if (st->codec.extradata_size != 0)
                    {
                        st->codec.extradata = avf_mallocz(st->codec.extradata_size);
                        get_buffer(pb, st->codec.extradata, st->codec.extradata_size);
                    }
                }
                st->codec.codec_tag = tag1;
                st->codec.codec_id = codec_get_id(codec_bmp_tags, tag1);
            }
            pos2 = url_ftell(pb);
            url_fskip(pb, gsize - (pos2 - pos1 + 24));
        } else if (!memcmp(&g, &data_header, sizeof(GUID))) {
            break;
        } else if (!memcmp(&g, &comment_header, sizeof(GUID))) {
            int len1, len2, len3, len4, len5;
            
            len1 = get_le16(pb);
            len2 = get_le16(pb);
            len3 = get_le16(pb);
            len4 = get_le16(pb);
            len5 = get_le16(pb);
            get_str16_nolen(pb, len1, s->title, sizeof(s->title));
            get_str16_nolen(pb, len2, s->author, sizeof(s->author));
            get_str16_nolen(pb, len3, s->copyright, sizeof(s->copyright));
            get_str16_nolen(pb, len4, s->comment, sizeof(s->comment));
            url_fskip(pb, len5);
#if 0
        } else if (!memcmp(&g, &head1_guid, sizeof(GUID))) {
            int v1, v2;
            get_guid(pb, &g);
            v1 = get_le32(pb);
            v2 = get_le16(pb);
        } else if (!memcmp(&g, &codec_comment_header, sizeof(GUID))) {
            int len, v1, n, num;
            char str[256], *q;
            char tag[16];
            
            get_guid(pb, &g);
            print_guid(&g);
            
            n = get_le32(pb);
            for(i=0;i<n;i++) {
                num = get_le16(pb); /* stream number */
                get_str16(pb, str, sizeof(str));
                get_str16(pb, str, sizeof(str));
                len = get_le16(pb);
                q = tag;
                while (len > 0) {
                    v1 = get_byte(pb);
                    if ((q - tag) < sizeof(tag) - 1)
                        *q++ = v1;
                    len--;
                }
                *q = '\0';
            }
#endif
#ifdef MSDRM10_SUPPORT
        } else if (!memcmp(&g, &content_encryption, sizeof(GUID))) {
#ifdef DEBUG_ASF
            printf( "content_encryption GUID\n" );
#endif
            memset( asf->drm10.pbSecretData, 0, sizeof(asf->drm10.pbSecretData) );
            asf->drm10.cbSecretData = get_buff32(pb, asf->drm10.pbSecretData );

            memset( asf->drm10.pbType, 0, sizeof(asf->drm10.pbType) );
            get_buff32(pb, asf->drm10.pbType );

            memset( asf->drm10.pbKeyID, 0, sizeof(asf->drm10.pbKeyID) );
            get_buff32(pb, asf->drm10.pbKeyID );

            // ignore the license URL
            get_buff32(pb, NULL );

        } else if (!memcmp(&g, &content_encryption_ex, sizeof(GUID))) {
#ifdef DEBUG_ASF
            printf( "content_encryption_ex GUID\n" );
#endif
            asf->drm10.cbCEncExData = get_buff32_alloc(pb, &asf->drm10.pbCEncExData );
#endif // MSDRM10_SUPPORT
        } 
		else if (url_feof(pb)) 
        {
            
            goto fail;
        } 
		else 
        {
            Avf_url_fseek(pb, gsize - 24, SEEK_CUR);
        }
    }
    get_guid(pb, &g);
    get_le64(pb);
    get_byte(pb);
    get_byte(pb);
    if (url_feof(pb))
        goto fail;
    asf->data_offset = url_ftell(pb);
    
    asf->packet_size_left = 0;

#ifdef MSDRM10_SUPPORT
    if ( (ap == NULL) || !(ap->flags & AV_NO_DRM) )
    { 
        ret = asf_msdrm10_init( asf );
    
#ifdef DEBUG_ASF
        printf( "DRM result: %d (HasDRM=%d, HasJanusDRM=%d)\n", ret, asf->drm10.bHasDRM, asf->drm10.bHasJanusDRM );
#endif

        if ( ret != cWMA_NoErr )
            goto fail;
    }
 
#endif // MSDRM10_SUPPORT
    
    // :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
    printf( "Data header\n" );
    printf( "Data offset %d\n", (unsigned long)(asf->data_offset) );
#endif
    
    if ( asf_calc_frame_rate(s) == TRUE )
    {
        return 0;
    }
    
fail:
    for(i=0;i<s->nb_streams;i++) {
        AVStream *st = s->streams[i];
        if (st) {
            avf_free(st->priv_data);
            avf_free(st->codec.extradata);
        }
        avf_free(st);
    }
    
    avf_free(asf->index.entries);

#ifdef MSDRM10_SUPPORT
    asf_msdrm10_free(asf);
#endif
    
    return -1;
}

#define DO_2BITS(bits, var, defval) \
    switch (bits & 3) \
    { \
    case 3: var = get_le32(pb); rsize += 4; break; \
    case 2: var = get_le16(pb); rsize += 2; break; \
    case 1: var = get_byte(pb); rsize++; break; \
    default: var = defval; break; \
    }

static int asf_get_packet(AVFormatContext *s)
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    uint32_t packet_length, padsize;
    int rsize = 11;
    int c = 0;

// :ZZZ: DEBUG_ASF print
#ifdef DEBUG_ASF
//    printf( "Payload parsing information:\n" );
//    printf( "   Error correction data : 0x%02X\n", c );
#endif                        

    int ec_present = 0;
    
    // :ZZZ: Check if the required packed is beyond the last payload
    if ( ((unsigned)url_ftell(pb)) >= asf->ullEndOfData )
    {
        return -EIO;
    }

    c = get_byte(pb);

    // :ZZZ: Error correction is optional according to spec
    // Determined according to high bit
    ec_present =(c & 0x80) >> 7;

    // :ZZZ: Error correction is optional according to spec
    if( ec_present )
    {
        if (c != 0x82) {
            if (!url_feof(pb))
            {
                printf("ff asf bad header %x  at:%Ld\n", c, url_ftell(pb));
            }
            return -EIO;
        }
        if ((c & 0x0f) == 2) { // always true for now
            if (get_le16(pb) != 0) {
                if (!url_feof(pb))
                {
                    printf("ff asf bad non zero\n");
                }
                return -EIO;
            }
        }

        #ifdef DEBUG_ASF
        printf( "   Error Corr Flags  : 0x%02X\n", c );
        #endif

        asf->packet_flags = get_byte(pb);
    }
    else
    {
        // No ec present, the byte was the flags
        asf->packet_flags = c;

        // the extra data is 3 bytes less than ffmpeg expected
        // Note that EC data length is assumed to be 2!!!
        rsize -= 3;
    }

    asf->packet_property = get_byte(pb);

    #ifdef DEBUG_ASF
    printf( "   Length type flags : 0x%02X\n", asf->packet_flags );
    printf( "   Property flags    : 0x%02X\n", asf->packet_property );
    #endif

    DO_2BITS(asf->packet_flags >> 5, packet_length, asf->packet_size);
    #ifdef DEBUG_ASF
    printf( "       Pck Len Type : 0x%02X\n", (asf->packet_flags >> 5)& 3 );
    printf( "       Pck Len      : 0x%8.8X\n", packet_length );
    #endif

    DO_2BITS(asf->packet_flags >> 1, padsize, 0); // sequence ignored
    #ifdef DEBUG_ASF
    printf( "       SequenceType : 0x%02X\n", (asf->packet_flags >> 1)& 3 );
    printf( "       Sequence     : 0x%8.8X\n", padsize );
    #endif

    DO_2BITS(asf->packet_flags >> 3, padsize, 0); // padding length
    #ifdef DEBUG_ASF
    printf( "       Pad Len Type : 0x%02X\n", (asf->packet_flags >> 3)& 3 );
    printf( "       Padding Len  : 0x%8.8X\n", padsize );
    printf( "       Multiple PL  : 0x%02X\n", (asf->packet_flags)& 1 );
    printf( "       Error Correc : 0x%02X\n", ec_present );
    #endif

    asf->packet_timestamp = get_le32(pb);
    get_le16(pb); /* duration */
    // rsize has at least 11 bytes which have to be present

    #ifdef DEBUG_ASF
    printf( "   Send time         : %d\n", asf->packet_timestamp );
    #endif

    if (asf->packet_flags & 0x01) {
        asf->packet_segsizetype = get_byte(pb); 
        rsize++;
        asf->packet_segments = asf->packet_segsizetype & 0x3f;
        #ifdef DEBUG_ASF
        printf( "   Payload Flags     : 0x%02X\n", asf->packet_segsizetype );
        printf( "      Num of Pay     : 0x%8.8X\n", asf->packet_segments );
        #endif
    } else {
        asf->packet_segments = 1;
        asf->packet_segsizetype = 0x80;
    }
    asf->packet_size_left = packet_length - padsize - rsize;
    if (packet_length < asf->hdr.min_pktsize)
        padsize += asf->hdr.min_pktsize - packet_length;
    asf->packet_padsize = padsize;
#ifdef DEBUG_ASF
    printf("packet: size=%d padsize=%d  left=%d\n", asf->packet_size, asf->packet_padsize, asf->packet_size_left);
#endif
    return 0;
}

int asf_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    ASFContext *asf = s->priv_data;
    ASFStream *asf_st = 0;
    ByteIOContext *pb = &s->pb;

    #ifdef DEBUG_ASF
    static int pc = 0;
    printf( "asf_read_packet\n" );
    #endif

    for (;;) 
    {
        int rsize = 0;
        int st_id = 0;
#ifdef MSDRM10_SUPPORT
        int read = 0;
#endif /* MSDRM10_SUPPORT */
        unsigned char* pktbuf;

        if ( (asf->packet_size_left < FRAME_HEADER_SIZE) || (asf->packet_segments < 1) ) 
        {
            //asf->packet_size_left <= asf->packet_padsize) {
            int ret = asf->packet_size_left + asf->packet_padsize;

            //printf("PacketLeftSize:%d  Pad:%d Pos:%Ld\n", asf->packet_size_left, asf->packet_padsize, url_ftell(pb));
            /* fail safe */
            url_fskip(pb, ret);
            ret = asf_get_packet(s);

            #ifdef DEBUG_ASF
            //printf("READ ASF PACKET  %d   r:%d   c:%d\n", ret, asf->packet_size_left, pc++);
            printf("READ ASF PACKET %d offset 0x%x\n", pc++, pb->buf_ptr - pb->buffer );
            #endif

            if (ret < 0 || url_feof(pb))
            {
                return -EIO;
            }

            asf->packet_time_start = 0;

            continue;
        }

        // :ZZZ: packet_time_start is not zero for payloads other than the first
        // one in the data packet, but we still want to read the stream number for them
        // this is why the following if was commented
        // :ZZZ: if (asf->packet_time_start == 0)
        {
            /* read frame header */
            int num = get_byte(pb);
            st_id = num & 0x7f;

            asf->packet_segments--;
            rsize++;
            asf->packet_key_frame = (num & 0x80) >> 7;
            asf->stream_index = asf->asfid2avid[st_id];

            #ifdef DEBUG_ASF
            printf( "  Stream Num & KeyFr: 0x%02X\n", num );
            printf( "       Stream Number: 0x%02X\n", st_id );
            printf( "       Key Frame Bit: 0x%02X\n", asf->packet_key_frame );

            printf( "   Property flags   : 0x%02X\n", asf->packet_property );
            printf( "       Stream  Type : 0x%02X\n", (asf->packet_property >> 6)& 3 );
            #endif

            // sequence should be ignored!
            DO_2BITS(asf->packet_property >> 4, asf->packet_seq, 0);

            #ifdef DEBUG_ASF
            printf( "       Med ObN Type : 0x%02X\n", (asf->packet_property >> 4)& 3 );
            printf( "       Media Obj Num: 0x%8.8X\n", asf->packet_seq );
            #endif

            DO_2BITS(asf->packet_property >> 2, asf->packet_frag_offset, 0);

            #ifdef DEBUG_ASF
            printf( "       Offset  Type : 0x%02X\n", (asf->packet_property >> 2)& 3 );
            printf( "       Offset Into M: 0x%8.8X\n", asf->packet_frag_offset );
            #endif

            DO_2BITS(asf->packet_property, asf->packet_replic_size, 0);

            #ifdef DEBUG_ASF
            printf( "       Rep Len Type : 0x%02X\n", (asf->packet_property)& 3 );
            printf( "       Replic Len   : 0x%8.8X\n", asf->packet_replic_size );
            #endif

            if (asf->packet_replic_size > 1) 
            {
               // it should be always at least 8 bytes - FIXME validate
                asf->packet_obj_size = get_le32(pb);
                asf->packet_frag_timestamp = get_le32(pb); // timestamp

                #ifdef DEBUG_ASF
                printf( "   ReplData M.O size : 0x%02X\n", asf->packet_obj_size );
                printf( "   ReplData timestamp: 0x%02X\n", asf->packet_frag_timestamp );
                #endif

                if (asf->packet_replic_size > 8)
                {
                    url_fskip(pb, asf->packet_replic_size - 8);
                }

                rsize += asf->packet_replic_size; // FIXME - check validity
            } 
            else 
            {
                #ifdef DEBUG_ASF
                printf( "Setting asf->packet_frag_offset = 0!!!\n" );
                printf( "asf->packet_frag_offset = %d\n", asf->packet_frag_offset );
                printf( "asf->packet_timestamp = %d\n", asf->packet_timestamp );
                #endif

                // multipacket - frag_offset is begining timestamp
                asf->packet_time_start = asf->packet_frag_offset;

                // :ZZZ:
                // packet_frag_offset is the offset of the current payload in the
                // data packet. ffmpeg assumes that there's only one, and cannot
                // decode streams with more than one payload.
                // :ZZZ:  asf->packet_frag_offset = 0;
                asf->packet_frag_timestamp = asf->packet_timestamp;

                if (asf->packet_replic_size == 1) 
                {
                    asf->packet_time_delta = get_byte(pb);
                    rsize++;
                }
            }

            if (asf->packet_flags & 0x01) 
            {
                DO_2BITS(asf->packet_segsizetype >> 6, asf->packet_frag_size, 0); // 0 is illegal

                #ifdef DEBUG_ASF
                printf( "       Pay Len Type : 0x%02X\n", (asf->packet_segsizetype >> 6)& 3 );
                printf( "       Payload Leng : 0x%8.8X\n", asf->packet_frag_size );
                #endif

                //#undef DO_2BITS
                //printf("Fragsize %d\n", asf->packet_frag_size);
            } 
            else
            {
                asf->packet_frag_size = asf->packet_size_left - rsize;
                //printf("Using rest  %d %d %d\n", asf->packet_frag_size, asf->packet_size_left, rsize);
            }

            if (asf->packet_replic_size == 1) 
            {
                asf->packet_multi_size = asf->packet_frag_size;

                if (asf->packet_multi_size > asf->packet_size_left) 
                {
                    asf->packet_segments = 0;
                    continue;
                }
            }

            asf->packet_size_left -= rsize;

            #ifdef DEBUG_ASF
            printf( "Packet size left %d\n", asf->packet_size_left );
            #endif

            //printf("___objsize____  %d   %d    rs:%d\n", asf->packet_obj_size, asf->packet_frag_offset, rsize);

            if (asf->stream_index < 0) 
            {
                asf->packet_time_start = 0;
                /* unhandled packet (should not happen) */
                url_fskip(pb, asf->packet_frag_size);
                asf->packet_size_left -= asf->packet_frag_size;
                printf("ff asf skip %d  %d\n", asf->packet_frag_size, st_id);

                continue;
            }

            asf->asf_st = s->streams[asf->stream_index]->priv_data;
        }

        asf_st = asf->asf_st;

        if ((asf->packet_frag_offset != asf_st->frag_offset
             || (asf->packet_frag_offset
             && asf->packet_seq != asf_st->seq)) // seq should be ignored
           ) 
        {
            /* cannot continue current packet: free it */
            // FIXME better check if packet was already allocated
            printf("ff asf parser skips: %d - %d     o:%d - %d    %d %d   fl:%d\n",
                   asf_st->pkt.size,
                   asf->packet_obj_size,
                   asf->packet_frag_offset, asf_st->frag_offset,
                   asf->packet_seq, asf_st->seq, asf->packet_frag_size);

            if (asf_st->pkt.size)
            {
                Avf_free_packet(&asf_st->pkt);
            }

            asf_st->frag_offset = 0;

            if (asf->packet_frag_offset != 0) 
            {
                url_fskip(pb, asf->packet_frag_size);
                printf("ff asf parser skiping %db\n", asf->packet_frag_size);
                asf->packet_size_left -= asf->packet_frag_size;

                continue;
            }
        }

        if (asf->packet_replic_size == 1) 
        {
            // frag_offset is here used as the begining timestamp
            asf->packet_frag_timestamp = asf->packet_time_start;
            asf->packet_time_start += asf->packet_time_delta;
            asf->packet_obj_size = asf->packet_frag_size = get_byte(pb);
            asf->packet_size_left--;
            asf->packet_multi_size--;

            if (asf->packet_multi_size < asf->packet_obj_size)
            {
                asf->packet_time_start = 0;
                url_fskip(pb, asf->packet_multi_size);
                asf->packet_size_left -= asf->packet_multi_size;

                continue;
            }

            asf->packet_multi_size -= asf->packet_obj_size;
            //printf("COMPRESS size  %d  %d  %d   ms:%d\n", asf->packet_obj_size, asf->packet_frag_timestamp, asf->packet_size_left, asf->packet_multi_size);
        }

        if (asf_st->frag_offset == 0) 
        {
            /* new packet */
            av_new_packet(&asf_st->pkt, asf->packet_obj_size);

            asf_st->seq = asf->packet_seq;
            //:DANIELK: return pts in microseconds - to be compatible with Avi
            asf_st->pkt.pts = (int64_t)(asf->packet_frag_timestamp - asf->hdr.preroll) * 1000; 
            asf_st->pkt.stream_index = asf->stream_index;

            if (asf->packet_key_frame)
            {
                asf_st->pkt.flags |= PKT_FLAG_KEY;
            }
        }

        /* read data */
        #ifdef DEBUG_ASF
        printf("READ PACKET s:%d  os:%d  o:%d,%d  l:%d   DATA:%p\n",
               asf->packet_size, asf_st->pkt.size, asf->packet_frag_offset,
               asf_st->frag_offset, asf->packet_frag_size, asf_st->pkt.data);
        #endif
        asf->packet_size_left -= asf->packet_frag_size;

        if (asf->packet_size_left < 0)
        {
            continue;
        }

        pktbuf = asf_st->pkt.data + asf->packet_frag_offset;
#ifdef MSDRM10_SUPPORT
        read =
#endif /* MSDRM10_SUPPORT */
			get_buffer(pb, pktbuf, asf->packet_frag_size);
        asf_st->frag_offset += asf->packet_frag_size;

#ifdef MSDRM10_SUPPORT
        if ( asf->drm10.bDRMAuxStreamNum == st_id )
        {
            read = MIN( read, (int)sizeof( asf->drm10.rgbNextPayloadEnd ) );

            memset( asf->drm10.rgbNextPayloadEnd, 0, sizeof( asf->drm10.rgbNextPayloadEnd ) );

            if (asf->packet_frag_size == 9)
            {
               asf->drm10.cbNextPayloadEndValid = (BYTE) read - 1;

                memcpy( asf->drm10.rgbNextPayloadEnd + 7, pktbuf+1, read-1 );
            }
            else if (asf->packet_frag_size == 8)
            {
                asf->drm10.cbNextPayloadEndValid = (BYTE) read;
                memcpy( asf->drm10.rgbNextPayloadEnd + 7, pktbuf, read);
            }

            /* test if whole packet is read */
            if (asf_st->frag_offset == asf_st->pkt.size) 
            {
                asf_st->frag_offset = 0;
                asf_st->pkt.size = 0;
                asf_st->pkt.data = 0;
            }

            continue;
        }

        if ( asf->drm10.bHasDRM || asf->drm10.bHasJanusDRM )
        {
            BYTE byLast15[15] = {0};
            BYTE *pLast15 = NULL;
            WORD wDRMOffset = 0;

            if( 0 != asf->drm10.cbNextPayloadEndValid )
            {
                /* We pre-cached the last bytes of this payload - no need to seek / read */
                pLast15 = asf->drm10.rgbNextPayloadEnd;

                /* Move the bytes to the appropriate offset */
                wDRMOffset = asf->packet_frag_size % 8;

                if( ( 0 != wDRMOffset ) && ( 8 == asf->drm10.cbNextPayloadEndValid ) )
                {
                    memmove( pLast15 + 7 - wDRMOffset, pLast15 + 7, 8 );
                    memset( pLast15 + 15 - wDRMOffset, 0, wDRMOffset );
                }
            }
            else
            {
                pLast15 = (PBYTE)byLast15;
                memmove( pLast15, pktbuf + asf->packet_frag_size - 15, 15 );
            }

            if( asf->drm10.bHasJanusDRM )
            {
                DRM_RESULT hr;

                hr = DRM_MGR_InitDecrypt( asf->drm10.pJanusDecryptContext,
                                          pLast15,
                                          asf->packet_frag_size );

                // TODO: Check return value
                if(DRM_SUCCESS != hr) {
                    printf("Can not Init Decrypt\n");
                }

                asf->drm10.cbNextPayloadEndValid = 0;

                hr = DRM_MGR_Decrypt( 
                    asf->drm10.pJanusDecryptContext,
                    pktbuf,
                    asf->packet_frag_size );

                // TODO: Check return value
                if(DRM_SUCCESS != hr) {
                    printf("Can not Decrypt\n");
                }
            }
        }
#endif // MSDRM10_SUPPORT

        /* test if whole packet is read */
        if (asf_st->frag_offset == asf_st->pkt.size) 
        {
            /* return packet */
            if (asf_st->ds_span > 1) 
            {
                /* packet descrambling */
                char* newdata = avf_malloc(asf_st->pkt.size);

                if (newdata) 
                {
                    int offset = 0;

                    while (offset < asf_st->pkt.size) 
                    {
                        int off = offset / asf_st->ds_chunk_size;
                        int row = off / asf_st->ds_span;
                        int col = off % asf_st->ds_span;
                        int idx = row + col * asf_st->ds_packet_size / asf_st->ds_chunk_size;

                        //printf("off:%d  row:%d  col:%d  idx:%d\n", off, row, col, idx);
                        memcpy(newdata + offset,
                               asf_st->pkt.data + idx * asf_st->ds_chunk_size,
                               asf_st->ds_chunk_size);
                        offset += asf_st->ds_chunk_size;
                    }

                    avf_free(asf_st->pkt.data);
                    asf_st->pkt.data = (uint8_t *)newdata;
                }
            }

            asf_st->frag_offset = 0;
            memcpy(pkt, &asf_st->pkt, sizeof(AVPacket));
            //printf("packet %d %d\n", asf_st->pkt.size, asf->packet_frag_size);
            asf_st->pkt.size = 0;
            asf_st->pkt.data = 0;

            break; // packet completed
        }
    }

    #ifdef DEBUG_ASF
    printf("READ PACKET DONE\n\n" );
    #endif

    return 0;
}

static int asf_read_close(AVFormatContext *s)
{
    int i;
    ASFContext *asf = s->priv_data;

    for(i=0;i<s->nb_streams;i++) {
        AVStream *st = s->streams[i];
        avf_free(st->priv_data);
        avf_free(st->codec.extradata);
    }

    avf_free(asf->index.entries);
    avf_free(asf->local_index_table);
    asf_actual_max_table_size = ASF_TABLE_SIZE;

#ifdef MSDRM10_SUPPORT
    asf_msdrm10_free(asf);
#endif

    return 0;
}

static int asf_read_seek(AVFormatContext *s, int64_t pts) //pts is in microseconds
{
    ASFContext *asf = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int requested_entry;
    int seek_offset;
    int rsize, num;
    offset_t packet_start, start_offset;
    int stid;
    
    //is there an index ?
    if ((asf->index.entries == NULL)||
        (asf->index.time_interval == 0))
    {
        printf( "No index (0x%X %d)\n", asf->index.entries, asf->index.time_interval );
        return -1;
    }

    requested_entry = ( pts / 1000 ) / asf->index.time_interval; //convert to millis first
    
    if (requested_entry > asf->index.last_entry)
    {
        printf( "Entry requested bigger than last (%d %d)\n", requested_entry, asf->index.last_entry );
        return -1;
    }
    
    seek_offset = asf->index.entries[requested_entry].packet_num *
                  asf->packet_size +
                  asf->data_offset;

    Avf_url_fseek(pb, seek_offset , SEEK_SET);

    rsize=0;

    packet_start = start_offset = url_ftell(pb);
    
    //read packet header    
    asf_get_packet(s); // return value is not checked

    for(;;)
    {
        //read payload header
        num = get_byte(pb);                                              //Stream and key_frame
        rsize++;

        DO_2BITS(asf->packet_property >> 4, asf->packet_seq, 0);         //Media obj
        DO_2BITS(asf->packet_property >> 2, asf->packet_frag_offset, 0); //Offset into media
        DO_2BITS(asf->packet_property,      asf->packet_replic_size, 0); //Repl. data len

        if (asf->packet_replic_size > 1) 
        {
            asf->packet_obj_size = get_le32(pb);                         //Obj Size
            asf->packet_frag_timestamp = get_le32(pb);                   //Timestamp

            if (asf->packet_replic_size > 8)
            {
                url_fskip(pb, asf->packet_replic_size - 8);
            }
            rsize += asf->packet_replic_size; // FIXME - check validity
        } 
        else 
        {
            if (asf->packet_replic_size == 1) 
            {
                asf->packet_time_delta = get_byte(pb);
                rsize++;
            }
        }
    
        if (asf->packet_flags & 0x01) 
        {
            DO_2BITS(asf->packet_segsizetype >> 6, asf->packet_frag_size, 0); // Payload size
        } 
        else 
        {
            asf->packet_frag_size = asf->packet_size_left - rsize;
        }

        asf->packet_size_left -= rsize;

        stid = asf->asfid2avid[num & 0x7F];

        // Wrong stream number
        if ( stid >= s->nb_streams)
        {
            printf( "Invalid stream id (%d %d %d)\n", stid, s->nb_streams, num );
            return -1;
        }

        if (((num & 0x80) >> 7) &&
            (s->streams[stid]->codec.codec_type == CODEC_TYPE_VIDEO) &&
            (asf->packet_frag_offset == 0) )
        {
            break;
        }

        //Go to the next payload
        url_fskip(pb, asf->packet_frag_size);
        start_offset = url_ftell(pb);
    }

    Avf_url_fseek(pb, start_offset, SEEK_SET);

    if (start_offset == packet_start)
        asf->packet_size_left = 0;    
    else
        asf->packet_size_left = asf->packet_size - (start_offset - packet_start);

    asf->packet_padsize   = 0;
    asf->packet_frag_offset = 0;
    
    return 0;
}

/*
%function

    asf_get_cur_entry

%desc

    searchs the packet previous to the cur_pos timestamp within the local index table.
    returns 0 if the timestamp is earlier then first entry in the table
            entry index if found
            last entry index if timestamp is later then the last entry

%author

    dimar 

*/
static int asf_get_cur_entry(AVFormatContext *s, offset_t cur_pos)
{       
    ASFContext *asf = s->priv_data;
    int i;

    for(i = 0; i < asf->local_index_table_size; i++)
    {
        if(asf->local_index_table[i].packet_timestamp > cur_pos)
        {
            break;
        }    
    }
    i = ( i > 0 ) ? i : i+1;
    return (i-1);
}

/*
%function

    asf_read_seek_key1

%desc

    jump into the file through I-frames
    return -1 if EOF reached, 0 otherwise

%author

    dimar

*/
int asf_read_seek_key1(AVFormatContext *s, int stream_id, int pos, int whence)
{
    int cur_entry;
    int res = 0;
    ASFContext *asf = s->priv_data;
    
    switch ( whence ) {
        case SEEK_CUR:
            cur_entry = asf_find_current_location(s, stream_id, asf->packet_frag_timestamp);
            if(cur_entry < 0)
            {
                break;
            }
            res = asf_move_from_current(s, stream_id, pos, cur_entry);
            break;

        default:

            printf("unsupported option !\n");
    }   
    return res;
}

// :ZZZ: seek according to video frame key
static int asf_read_seek_key(AVFormatContext *s, int stream_id, int pos, int whence)
{
    ASFContext *asf = s->priv_data;
    int64_t new_pts;

    switch ( whence ) {
        case SEEK_CUR: 

            new_pts = (int64_t)asf->packet_frag_timestamp + pos * asf->index.time_interval;

            if (new_pts < 0)
            {
                return -1;
            }

            return asf_read_seek(s, new_pts * 1000);
               
        default:

            printf("usupported option !\n");
    }

    return 0;
}

/**
%function                                                                       
                                                                                
  asf_calc_frame_rate
                                                                                
%desc

   Calculates frame_rate for video stream and sets it into codec-context
   Since there is no explicit frame-rate indication in format,
   and there is no way to know the total number of video packets 
   we fetch two first video packets and look at their pts difference  
*/
static BOOL asf_calc_frame_rate(AVFormatContext *s)
{
    AVPacket packet;
    int pts1 = -1, pts2 = -1;
    BOOL bFoundVideo = FALSE;
    INT n;

    // Check for presence of video stream
    for(n=0; n<s->nb_streams; n++) 
    {
        if (s->streams[n]->codec.codec_type == CODEC_TYPE_VIDEO)
        {
            bFoundVideo = TRUE;
            break;
        }
    }    

    // If it's an audio only clip - don't calculate frame-rate
    if (!bFoundVideo)
        return TRUE;

    while (pts2 == -1)
    {
        if ( asf_read_packet(s, &packet) != 0 )
        {
            return TRUE;
        }

        if (s->streams[packet.stream_index]->codec.codec_type == CODEC_TYPE_VIDEO)        
        {
            if (pts1 == -1) //first video packet
            {
                pts1 = (int)packet.pts;
            }
            else //second video packet
            {
                pts2 = (int)packet.pts;     
                s->streams[packet.stream_index]->codec.frame_rate = (int)ROUND_DIV(1000000, pts2 - pts1);
            }
        }   
    
        packet.destruct(&packet);
    }
                        
    if ( asf_read_seek( s, (int64_t)0 ) != 0) //reset location
    {
        return FALSE;
    }
                        
    return TRUE;
}

/*
%function

    asf_restore_context

%desc

    restores the recieved entry of the local index table

%author

    dimar

*/
static void asf_restore_context( AVFormatContext *s, ASFIndexEntry *entry)
{
    ByteIOContext *pb = &s->pb;
    ASFContext *asf = s->priv_data;

    asf->packet_flags       = entry->packet_flags;
    asf->packet_property    = entry->packet_property;
    asf->packet_segments    = entry->packet_segments;
    asf->packet_segsizetype = entry->packet_segsizetype;
    asf->packet_timestamp   = entry->packet_timestamp;
    asf->packet_size_left   = entry->packet_size_left;
    asf->packet_padsize     = entry->packet_padsize;
    asf->packet_frag_timestamp = entry->packet_frag_timestamp;
    Avf_url_fseek(pb, entry->position, SEEK_SET);
}

/*
%function

    asf_store_context

%desc

    stores the received entry in the local index table

%author

    dimar

*/
static void asf_store_context( AVFormatContext *s, ASFIndexEntry *entry, offset_t pos)
{
    ASFContext *asf = s->priv_data;

    entry->position = pos;
    entry->packet_flags = asf->packet_flags;
    entry->packet_property = asf->packet_property;
    entry->packet_segments = asf->packet_segments;
    entry->packet_segsizetype = asf->packet_segsizetype;
    entry->packet_timestamp = asf->packet_timestamp;
    entry->packet_size_left = asf->packet_size_left;
    entry->packet_padsize = asf->packet_padsize;
    entry->packet_frag_timestamp = asf->packet_frag_timestamp;
}

/*
%function

    asf_get_next_entry

%desc

    goes to the next I-frame in the file

%author

    dimar

*/
static int asf_get_next_entry(AVFormatContext *s, offset_t* start, offset_t* offset)
{
    ByteIOContext *pb = &s->pb;
    ASFContext *asf = s->priv_data;
    offset_t packet_start, segment_start;
    int ret, num, packet_key_frame, stream_index, rsize;
    //
    // start search from the next packet
    //
    ret = asf->packet_size_left + asf->packet_padsize;
    ( ret == 0 ) ? url_fskip(pb, asf->packet_size) : url_fskip(pb, ret);
    //
    // go over packets
    //
    for(;;)
    {           
        rsize = 0;
        //
        // save the packet beggining
        //
        packet_start = url_ftell(pb);
        //
        // get packet
        //
        ret = asf_get_packet(s);
        //
        // check if EOF reached
        //
        if (ret < 0 || url_feof(pb))
        {
            return -1;
        }
        //
        // search I-frame inside the packet
        //
        for(;;)
        {
            if( asf->packet_segments < 1 )
            {
                break;
            }
            asf->packet_segments--;
            //
            // read frame header
            //
            segment_start = url_ftell(pb);
            num = get_byte(pb);            
            #ifdef DEBUG_ASF
                printf( "   Stream Num & KeyFr: 0x%02X\n", num );
                printf( "       Stream Number: 0x%02X\n", num & 0x7f );
                printf( "       Key Frame Bit: 0x%02X\n", asf->packet_key_frame );
                printf( "   Property flags    : 0x%02X\n", asf->packet_property );
                printf( "       Stream  Type : 0x%02X\n", (asf->packet_property >> 6)& 3 );
            #endif

            packet_key_frame = (num & 0x80) >> 7;
            stream_index = asf->asfid2avid[num & 0x7f];          
                
            DO_2BITS(asf->packet_property >> 4, asf->packet_seq, 0);
            #ifdef DEBUG_ASF 
                printf( "       Med ObN Type : 0x%02X\n", (asf->packet_property >> 4)& 3 );
                printf( "       Media Obj Num: 0x%8.8X\n", asf->packet_seq );
            #endif

            DO_2BITS(asf->packet_property >> 2, asf->packet_frag_offset, 0);
            #ifdef DEBUG_ASF
                printf( "       Offset  Type : 0x%02X\n", (asf->packet_property >> 2)& 3 );
                printf( "       Offset Into M: 0x%8.8X\n", asf->packet_frag_offset );
            #endif

            DO_2BITS(asf->packet_property, asf->packet_replic_size, 0);
            #ifdef DEBUG_ASF
                printf( "       Rep Len Type : 0x%02X\n", (asf->packet_property)& 3 );
                printf( "       Replic Len   : 0x%8.8X\n", asf->packet_replic_size );
            #endif
            if (asf->packet_replic_size > 1) 
            {
                asf->packet_obj_size = get_le32(pb);                         //Obj Size
                asf->packet_frag_timestamp = get_le32(pb);                   //Timestamp    

                if (asf->packet_replic_size > 8)
                {
                    url_fskip(pb, asf->packet_replic_size - 8);
                }
            } 
            else 
            {
                if (asf->packet_replic_size == 1) 
                {
                    asf->packet_time_delta = get_byte(pb);
                }
            }
    
            if (asf->packet_flags & 0x01) 
            {
                DO_2BITS(asf->packet_segsizetype >> 6, asf->packet_frag_size, 0); // Payload size
            }
            #ifdef DEBUG_ASF
                printf( "       Pay Len Type : 0x%02X\n", (asf->packet_segsizetype >> 6)& 3 );
                printf( "       Payload Leng : 0x%8.8X\n", asf->packet_frag_size ); 
            #endif
            //
            // if not I-frame of video, pass to the next packet
            //
            if( //stream_index >= s->nb_streams || 
                packet_key_frame != 1         ||
                s->streams[stream_index]->codec.codec_type != CODEC_TYPE_VIDEO )
            {
                url_fskip(pb, asf->packet_frag_size);
                continue;
            }        
            //
            // if not first data of the I-frame, continue
            //        
            if( asf->packet_frag_offset != 0 )
            {
                url_fskip(pb, asf->packet_frag_size);
                continue;
            }
            //
            // I-frame found
            //
            *start = packet_start;
            *offset = segment_start;
            asf->packet_size_left = asf->packet_size - ( segment_start - packet_start ) -
                                    asf->packet_padsize;
            asf->packet_segments++;
            //
            // return to the segment start
            //
            url_fskip(pb, segment_start - url_ftell(pb));
            return 0;
        } //for(;;)
        //
        // go to the next packet
        //
        Avf_url_fseek(pb, packet_start + asf->packet_size, SEEK_SET);
    }
}

/*
%function

    asf_build_local_index_table

%desc

    builds local index table in the received direction.
    jump extra origin index table entries in the oposite direction
    and then goes step+extra origin table entries in the received direction.

%author

    dimar

*/
static void asf_build_local_index_table( AVFormatContext *s,
                                  int stream_id,
                                  ASFDirection direction,
                                  int step,
                                  int extra)
{
    ByteIOContext *pb = &s->pb;
    ASFContext *asf = s->priv_data;
    offset_t packet_start, cur_packet=0, segment_start;
    int index = 0, ret;
    ASFIndexEntry histPos;

    if(asf->local_index_table == 0)
    {
        return;
    }
    //
    // make some sper entries in the table
    //
    if(extra > 0)
    {
        asf_read_seek_key(s, stream_id, -(direction * extra), SEEK_CUR); 
    }
    //
    // get the current packet start
    //
    if(direction == ASFBackward)
    {
        // additional
        ret = asf->packet_size_left + asf->packet_padsize;
        cur_packet = (ret != 0) ? url_ftell(pb) - (asf->packet_size - ret) : url_ftell(pb);        
    }
    else // ASFForward
    {
        asf_store_context(s, &histPos, url_ftell(pb));
    }
    //
    // jump to the previous/next entry in the index table
    //
    do{
        ret = asf_read_seek_key(s, stream_id, ( step + extra ) * direction, SEEK_CUR);
        step--;
    } while( ret < 0 && (step + extra) > 0 );
    //
    // after the seek we are on the I-frame
    //
    if(direction == ASFBackward)
    {
        asf_store_context(s, &(asf->local_index_table[index++]), url_ftell(pb));
    }
    else // ASFForward
    {
        ret = asf->packet_size_left + asf->packet_padsize;
        cur_packet = (ret != 0) ? url_ftell(pb) - (asf->packet_size - ret) : url_ftell(pb);
        asf_restore_context(s, &histPos);
    }
    //
    // go over packets
    //
    for(;;)
    {
        //
        // if EOF reached
        //
        if( asf_get_next_entry(s, &packet_start, &segment_start) < 0 )
        {
            break;
        }
        //
        // check if finished
        //
        if( packet_start >= cur_packet ) 
        {
            break;
        }
        //
        // I-frame, save it
        //
        asf_store_context(s, &(asf->local_index_table[index++]), segment_start);
        if( index >= asf_actual_max_table_size )
        {
            ASFIndexEntry* table =
              avf_malloc( (asf_actual_max_table_size + ASF_TABLE_SIZE) * sizeof(ASFIndexEntry) );
            if( table == 0 )
            {
                break;
            }
            memcpy(table, asf->local_index_table, sizeof(ASFIndexEntry) * index);
            asf_actual_max_table_size += ASF_TABLE_SIZE;
            avf_free(asf->local_index_table);
            asf->local_index_table = table;
        }
    } //for(;;)
    asf->local_index_table_size = index;

/*    {
        int i;
        printf("print out:\n");
        for(i=0; i<local_index_table_size; i++)
        {
            printf("time: %d\n", (asf_local_index_table[i].packet_timestamp));
        }               
    }*/
}

/*
%function

    asf_get_table_range

%desc

    returns the timestamps range of the table

%author

    dimar

*/
static void asf_get_table_range(AVFormatContext *s, int* start, int* end)
{
    ASFContext *asf = s->priv_data;
    offset_t packet, offset;

    *start = asf->local_index_table[0].packet_timestamp;
    asf_restore_context(s, &(asf->local_index_table[asf->local_index_table_size-1]));
    asf_get_next_entry(s, &packet, &offset);
    *end = asf->packet_timestamp;
}

/*
%function

    asf_find_current_location

%desc

    finds current entry in the local index table.
    refreshes the table if needed.

%author

    dimar

*/
static int asf_find_current_location(AVFormatContext *s, int stream_id, int cur_pos)
{
    static ASFDirection direction;
    ASFContext *asf = s->priv_data;
    int distance = 0, start, end, delta = 100;
    ASFIndexEntry entry;
    ByteIOContext *pb = &s->pb;
    
    if(asf->local_index_table_size < 1)
    {
        asf->local_index_table = avf_malloc( ASF_TABLE_SIZE * sizeof(ASFIndexEntry) );
        if( asf->local_index_table == 0 )
        {
            return -1;
        }
        // build table 1 sec in every direction
        asf_build_local_index_table(s, stream_id, ASFBackward, 1, 1);
        return asf_get_cur_entry(s, cur_pos);
    }
    //
    // if the timestamp belongs to index table range
    //
    asf_store_context(s, &entry, url_ftell(pb));
    asf_get_table_range(s, &start, &end);
    if( cur_pos >= start && cur_pos < end )
    {
        return asf_get_cur_entry(s, cur_pos);
    }
    asf_restore_context(s, &entry);
    //
    // calculate distance and direction
    // 
    if( cur_pos < start )
    {
        distance = start - cur_pos;
        direction = ASFBackward;
    } else
    {
        if( cur_pos >= end )
        {
            distance = cur_pos - end;
            direction = ASFForward;
        }
    }
    //
    // if big distance, initialize again
    //
    if( distance > ( ASF_STEP * 1000 - delta ) )
    {
        // build table 1 sec in every direction
        asf_build_local_index_table(s, stream_id, ASFBackward, 1, 1);
    }
    else
    {
        asf_build_local_index_table(s, stream_id, direction, ASF_STEP, 1);
    }
    return asf_get_cur_entry(s, cur_pos);
}

/*
%function

    asf_move_from_current

%desc

    moves inside the index table

%author

    dimar

*/
static int asf_move_from_current(AVFormatContext *s, int stream_id, int step, int cur_entry)
{
    ASFContext *asf = s->priv_data;
    int time, pos;
    
    if( step < 0 )
    {
        while( (cur_entry + step) < 0 )
        {
            time = asf->local_index_table[0].packet_timestamp;
            //
            // move to the start of the table
            //
            step += cur_entry;
            asf_restore_context(s, &(asf->local_index_table[0]));
            pos = asf->packet_frag_timestamp;
            asf_build_local_index_table(s, stream_id, ASFBackward, ASF_STEP, 1);
            //
            // check if EOF
            //
            if(time == asf->local_index_table[0].packet_timestamp)
            {
                asf_restore_context(s, &(asf->local_index_table[0]));
                return -1;
            }
            cur_entry = asf_find_current_location(s, stream_id, pos);
        }
    } else
    if( step > 0 )
    {
        while( (cur_entry + step) >= asf->local_index_table_size )
        {
            time = asf->local_index_table[asf->local_index_table_size - 1].packet_timestamp;
            //
            // update the step
            //
            step -= (asf->local_index_table_size - 1 - cur_entry);
            asf_restore_context(s, &(asf->local_index_table[asf->local_index_table_size - 1]));
            pos = asf->packet_frag_timestamp;
            asf_build_local_index_table(s, stream_id, ASFForward, ASF_STEP, 1);
            //
            // check if EOF
            //
            if(time == asf->local_index_table[asf->local_index_table_size - 1].packet_timestamp)
            {
                asf_restore_context(s, &(asf->local_index_table[asf->local_index_table_size - 1]));
                return -1;
            }
            cur_entry = asf_find_current_location(s, stream_id, pos);
        }
    }
    asf_restore_context(s, &(asf->local_index_table[cur_entry + step]));
    return 0;
}

static AVInputFormat asf_iformat = {
    "asf",
    "asf format",
    sizeof(ASFContext),
    asf_probe,
    asf_read_header,
    asf_read_packet,
    asf_read_close,
    asf_read_seek,
    asf_read_seek_key1,
};

static AVOutputFormat asf_oformat = {
    "asf",
    "asf format",
    "video/x-ms-asf",
    "asf,wmv",
    sizeof(ASFContext),
#ifdef CONFIG_MP3LAME
    CODEC_ID_MP3LAME,
#else
    CODEC_ID_MP2,
#endif
    CODEC_ID_MSMPEG4V3,
    asf_write_header,
    asf_write_packet,
    asf_write_trailer,
    0,
    0,
    0,
    asf_estimate_trailer_size,
};

static AVOutputFormat asf_stream_oformat = {
    "asf_stream",
    "asf format",
    "video/x-ms-asf",
    "asf,wmv",
    sizeof(ASFContext),
#ifdef CONFIG_MP3LAME
    CODEC_ID_MP3LAME,
#else
    CODEC_ID_MP2,
#endif
    CODEC_ID_MSMPEG4V3,
    asf_write_stream_header,
    asf_write_packet,
    asf_write_trailer,
};

// :ZZZ: Add new format
static AVOutputFormat wmv_oformat = {
    "wmv",
    "wmv format",
    "video/x-ms-asf",
    "wmv,wma",
    sizeof(ASFContext),
    CODEC_ID_MP2,
    CODEC_ID_WMV1,
    asf_write_header,
    asf_write_packet,
    asf_write_trailer,
};

int Avf_asf_init(void)
{
    // :ZZZ: Add new format
    av_register_output_format(&wmv_oformat);

    av_register_input_format(&asf_iformat);
    av_register_output_format(&asf_oformat);
    av_register_output_format(&asf_stream_oformat);
    return 0;
}

#ifdef MSDRM10_SUPPORT

static int asf_msdrm10_init(ASFContext *asf)
{
    asf->drm10.bHasDRM = (BOOL) 0;
    asf->drm10.bHasJanusDRM = (BOOL) 0;

    if ( ( asf->drm10.cbSecretData > 0 ) ||
         ( asf->drm10.cbCEncExData > 0 ) )
    {
        DRM_RESULT dr;
        DRM_STRING *rgRights[ 1 ];

        if( asf->drm10.pJanusContext != NULL )
        {
            avf_free( asf->drm10.pJanusContext );
        }

        asf->drm10.pJanusContext = avf_malloc( sizeof( DRM_MANAGER_CONTEXT ) );
        if( NULL == asf->drm10.pJanusContext )
        {
            printf( "Error allocating Janus context\n" );
            return( cWMA_DRMFailed );
        }

        if( asf->drm10.pJanusDecryptContext )
        {
            avf_free( asf->drm10.pJanusDecryptContext );
        }

        asf->drm10.pJanusDecryptContext = avf_malloc( sizeof( DRM_MANAGER_DECRYPT_CONTEXT ) );
        if( NULL == asf->drm10.pJanusDecryptContext )
        {
            printf( "Error allocating decryption context\n" );
            return( cWMA_DRMFailed );
        }

        //
        // Request the playback right; change this if the device is doing
        // anything else with the file.
        //
        rgRights[ 0 ] = (DRM_STRING *) &g_dstrWMDRM_RIGHT_PLAYBACK;

        dr = DRM_MGR_Initialize( asf->drm10.pJanusContext, &s_strLicense );
        if( DRM_SUCCESS != dr )
        {
            printf( "Error in DRM_MGR_Initialize\n" );
            return( cWMA_DRMFailed );
        }

        dr = DRM_SUCCESS;

        do
        {
            if( asf->drm10.cbSecretData > 0 )
            {
                //
                // Tell DRM_MGR about the old header object, if it exists.
                //
                dr = DRM_MGR_SetV1Header( asf->drm10.pJanusContext,
                                          (DRM_BYTE *) asf->drm10.pbKeyID,
                                          32,
                                          (DRM_BYTE *) asf->drm10.pbSecretData,
                                          asf->drm10.cbSecretData,
                                          (DRM_BYTE *) "",
                                          2 );
                if( DRM_SUCCESS != dr )
                {
                    break;
                }
            }

            if( asf->drm10.cbCEncExData > 0 )
            {
                //
                // The DRM_MGR about the new header object, which will exist.
                //
                dr = DRM_MGR_SetV2Header( asf->drm10.pJanusContext,
                                          (DRM_BYTE *) asf->drm10.pbCEncExData,
                                          asf->drm10.cbCEncExData );
                if( DRM_SUCCESS != dr )
                {
                    break;
                }
            }

            //
            // Find the license for this content.
            //
            dr = DRM_MGR_Bind( asf->drm10.pJanusContext,
                               (const DRM_CONST_STRING **) rgRights,
                               1,
                               NULL,
                               NULL,
                               asf->drm10.pJanusDecryptContext );
            if( DRM_SUCCESS == dr )
            {
                //
                // Since we're definitely going to play the file, go ahead
                // and commit right now.
                //
                dr = DRM_MGR_Commit( asf->drm10.pJanusContext );
                if( DRM_SUCCESS != dr )
                {
                    break;
                }
                
                asf->drm10.bHasJanusDRM = (BOOL) 1;
            }
            else
            {
                //
                // OK, just no license. That's fine; try for a PDDRM license.
                //
#ifdef DEBUG_ASF
                printf("DRM_MGR_Bind error 0x%X\n",dr); 
#endif
                dr = DRM_SUCCESS;
            }
        }
        while( FALSE );

        DRM_MGR_Uninitialize( asf->drm10.pJanusContext );
        avf_free( asf->drm10.pJanusContext );
        asf->drm10.pJanusContext = NULL;

        if( DRM_SUCCESS != dr )
        {
            printf( "Error in DRM_MGR_Bind/DRM_MGR_Commit\n" );
            return( cWMA_DRMFailed );
        }
    }

    if( !asf->drm10.bHasJanusDRM && ( asf->drm10.cbSecretData > 0 ) )
    {
        /* only for DRM now */
        CHAR *p = (CHAR *)asf->drm10.pbType;

        if(p[0] == 'D' && p[1] == 'R' && p[2] == 'M' && p[3] == '\0')
        {
            asf->drm10.bHasDRM = (BOOL)( 1 );
        }
        else
        {
            return cWMA_BadDRMType;
        }
    }

    if( ( asf->drm10.cbSecretData > 0 ) ||
        ( asf->drm10.cbCEncExData > 0 ) )
    {
        if( !asf->drm10.bHasDRM && !asf->drm10.bHasJanusDRM )
        {
            //
            // No license.
            //
            printf( "Error - No license to play file\n" );
            return( cWMA_DRMFailed );
        }
    }

    return cWMA_NoErr;
}

static void asf_msdrm10_free(ASFContext *asf)
{
    avf_free( asf->drm10.pbCEncExData );
    avf_free( asf->drm10.pJanusContext );
    avf_free( asf->drm10.pJanusDecryptContext );

    memset( &(asf->drm10), 0, sizeof(asf->drm10) );
}

#endif // MSDRM10_SUPPORT

