/*
 * AVI decoder.
 * Copyright (c) 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avformat.h"
#include "avi.h"
#include "av.h"
#include "AvApi.h"

//#define DEBUG_AVI

// :ZZZ: :TBD:
#define LOAD_FIRST_INDEX 0
#define LOAD_CHECK_FOR_AUDIO_INDEX 1
#define LOAD_FULL_INDEX 2

typedef enum {
    AVI_IDX_GE, 
    AVI_IDX_GT, 
    AVI_IDX_EQ, 
    AVI_IDX_LE, 
    AVI_IDX_LT
} AVI_IDX_RELATION;

// Characteristics of stream position
typedef struct 
{
    // Packet offset in stream
    unsigned int offset;

    // Packet number in stream
    int packet_num;

} AVIStreamPos;

// Structure of the index entry
typedef struct 
{
    // Packet offset in file relative to "movi list" start. Real offset to 
    // seek to
    unsigned int offset;

    // Packet PTS 
    int64_t pts;

    // Stream positions for seek - indexed by stream id, each
    // entry contains specific stream postition for proper PTS 
    // calculation after seek operation
    AVIStreamPos other_pos[MAX_STREAMS];

} AVIKeyIndexEntry;

typedef struct AVIIndex {
    unsigned char tag[4];
    unsigned int flags, pos, len;
    struct AVIIndex *next;
} AVIIndex;

typedef struct {
    int64_t riff_end;
    int64_t movi_end;
    offset_t movi_list;
    AVIIndex *first, *last;
    // :ZZZ:
    unsigned int index_granularity;
} AVIContext;

// :ZZZ: AVI stream context 
typedef struct  {

    // Total number of packets in the stream
    int total_packets;

    // Current packet number (numbering is relative to the stream)
    int curr_packet_num;

    // Total stream length in bytes
    int total_length;

    // Current stream length in bytes
    int curr_length;

    // Keyframe index table
    AVIKeyIndexEntry  *key_index;

    // Keyframe index size
    int index_size;

#define AVI_INDEX_SIZE 1024
    // Allocated index size
    int alloc_index_size;   

} AVIStreamContext;

#ifdef DEBUG_AVI
static void print_tag(const char *str, unsigned int tag, int size)
{
    printf("%s: tag=%c%c%c%c size=0x%x\n",
           str, tag & 0xff,
           (tag >> 8) & 0xff,
           (tag >> 16) & 0xff,
           (tag >> 24) & 0xff,
           size);
}
#endif

static int get_riff(AVIContext *avi, ByteIOContext *pb, GenericCallbackFn pHceErrorCallback)
{
    uint32_t tag; 
    /* check RIFF header */
    tag = get_le32(pb);

    if (tag != MKTAG('R', 'I', 'F', 'F')) {
        //If callback is registered and agrees that tag is errorneous - fail.
        if (pHceErrorCallback != NULL) {
            if (pHceErrorCallback(AVI_RIFF_ERROR_ID, (char *)&tag, 4) == AVI_ERR_FAIL) {
                return -1;
            }
        } 
        else 
        {
			//If callback is not registered - there is nothing else we can do. Fail.
            return -1;
        }
    }
    avi->riff_end = get_le32(pb);   /* RIFF chunk size */
    avi->riff_end += url_ftell(pb); /* RIFF chunk end */
    tag = get_le32(pb);
    if (tag != MKTAG('A', 'V', 'I', ' ') && tag != MKTAG('A', 'V', 'I', 'X')) {
        if (pHceErrorCallback != NULL) {
            if (pHceErrorCallback(AVI_AVI_ERROR_ID, (char *)&tag, 4) == AVI_ERR_FAIL) {
                return -1;
            }
        } 
        else
        {
            return -1;
        }
    }
    
    return 0;
}

static int get_index(AVFormatContext *s, ByteIOContext *pb, int bLoadIndex)
{
    // 
    AVIContext *avi = s->priv_data;
    AVStream *st;
    AVIStreamContext *sc;
    short gopSize=0;
    short prevKeyFrameIndex=-1;
    short videoFrameCount=0;
    unsigned int audioFrameCount=0;

    // AVI file index structure
    AVIIndexEntry avi_index_entry;

    // AVI index internal representation
    AVIKeyIndexEntry *new_key;

    // Total index size
    int index_size = get_le32(pb) / sizeof(AVIIndexEntry);

    // 
    int i, j;

    // This ratio is calculated only once to avoid time-consuming 64bit divisions
    int64_t audio_microsecs_per_byte = 0LL;
   
    // Saves the last time of indexed audio packet
    int64_t aud_last_ankor = 0LL;

    int frame_period = 0;

    unsigned char bfail = 0;

    // Build index per stream  
    for ( i = 0; i < index_size; i++ ) 
    {

        // Stream identification
        unsigned char *ckid = (unsigned char *)&avi_index_entry.ckid;
        get_buffer(pb, (PBYTE)&avi_index_entry, sizeof(avi_index_entry)); 

        // Determine whether belongs to video or audio stream 
        if ( ((ckid[2] == 'd') && (ckid[3] == 'c')) || 
             ((ckid[2] == 'w') && (ckid[3] == 'b'))  ) {

            // Determine which stream the entry belongs to.
            int st_num = (ckid[0] - '0') * 10 + (ckid[1] - '0');

            // Index packet presentation time
            int64_t pts;

            // Whether the packet should be indexed
            int index_packet = 0;

            if ( st_num < 0 || st_num >= s->nb_streams )
            {
                // Eli - How to handle this case?
                assert( s->nb_streams == 2 );

                if ( (ckid[2] == 'd') && (ckid[3] == 'c') )
                {
                    st_num = 0;
                }
                else
                {
                    st_num = 1;
                }
            }

            // Map into relevant stream contexts
            st = s->streams[st_num];
            sc = st->priv_data;

            // Check whether video or audio
            if ( (ckid[2] == 'd') && (ckid[3] == 'c') )  
            {
                videoFrameCount++;
                // Calculate video specific PTS in microsec
                // Assuming video frame rate is constant - calculate only once
                if (frame_period == 0)
                {
                    frame_period = 1000000000000LL / 
                                   (((int64_t)st->codec.frame_rate * 1000000) /
                                   st->codec.frame_rate_base);                
                }
                                
                pts = (int64_t)sc->total_packets * (int64_t)frame_period;

                // Video - discard entries which do not contain key frame 
                if ( (avi_index_entry.flags & AVIIF_KEYFRAME) ) 
                {
                    index_packet = 1;
                    
                    //Calculate the gop: distance between two Key-Frames
                    if (gopSize == 0)
                    {
                        if (prevKeyFrameIndex==-1)
                        {
                            prevKeyFrameIndex=videoFrameCount;
                        }
                        else
                        {
                            gopSize = videoFrameCount-prevKeyFrameIndex;
                            s->uwGopSize = gopSize;
                            //printf("Found GOP size = %d\n",gopSize);
                        }
                    }
                }
            }
            else  { 
                int chunk_duration;

                audioFrameCount++;

                // Make the long calculation (int64 division) only once                
                if (audio_microsecs_per_byte == 0LL)
                {
                    // To maintain accuracy - shift left by 20
                    audio_microsecs_per_byte = ((8 * 1000000LL) << 20) / st->codec.bit_rate;                    
                }
                
                // Calculate audio specific PTS in microsec
                pts = ((int64_t)sc->total_length * audio_microsecs_per_byte) >> 20;

                // Audio - keep only ankor packets
                chunk_duration = ((int64_t)avi_index_entry.chunk_length * audio_microsecs_per_byte) >> 20;

                if (( pts + chunk_duration > aud_last_ankor ) && (pts != 0LL))
                {
                    index_packet = 1;
                    aud_last_ankor += avi->index_granularity;
                }
            }
             
            if ( index_packet ) {
                // Packet is wanted for index - allocate new index entry 
                if ( sc->index_size >= sc->alloc_index_size ) {
                    // Need to realloc - get new allocation size
                    int new_size = sc->alloc_index_size + AVI_INDEX_SIZE;

                    if ( sc->key_index == NULL )
                    {
                        // First allocation
                        sc->key_index = avf_malloc( new_size * sizeof(AVIKeyIndexEntry));
                    }
                    else
                    {
                        sc->key_index = avf_realloc(sc->key_index,
                                        new_size * sizeof(AVIKeyIndexEntry));
                    }

                    if ( !sc->key_index ) {
                        bfail = 1;
                        break;
                    }

                    sc->alloc_index_size = new_size;
                }

                // Fill the key data
                new_key = &sc->key_index[sc->index_size];
                new_key->offset = avi_index_entry.chunk_offset - 4;
                new_key->pts = pts;

                // Store the stream position for all the streams
                for ( j = 0; j < s->nb_streams; j++ ) {
                    AVIStreamContext *other_sc = s->streams[j]->priv_data;

                    new_key->other_pos[j].offset = other_sc->total_length;
                    new_key->other_pos[j].packet_num = other_sc->total_packets;
                    //printf("Stream %d, Other offset %d, packet %d\n", 
                    //        st_num, 
                    //        other_sc->total_length, 
                    //        other_sc->total_packets);
                }

                // Increase index size
                sc->index_size++;
            }       

            // Update total stream length and number of stream packets
            sc->total_length += avi_index_entry.chunk_length;
            sc->total_packets++;

            // Although the index is not needed, for ApsAvGetFirstFrame 
            // purposes, we do need to first video chunk offset,
            // so that seek(0) and first read_packet will work
            // (assumption: first video frame is I-VOP) :TODO:
            // And if bLoadIndex == LOAD_CHECK_FOR_AUDIO_INDEX, check 
            // at least 300 video frames before stopping check for audio index
            if ( ( (bLoadIndex == LOAD_FIRST_INDEX ) && (ckid[2] == 'd') && (ckid[3] == 'c') ) ||
                 ( (bLoadIndex == LOAD_CHECK_FOR_AUDIO_INDEX) && ( videoFrameCount > 300 ) ) )
            {
                break;
            }
        }
    }

    for ( i = 0; i < s->nb_streams; i++ ) {
        sc = s->streams[i]->priv_data;
        //printf("stream %i index size %d\n total length %d\n", 
        //        i, sc->index_size, sc->total_length);
    }

    if ( bLoadIndex == LOAD_CHECK_FOR_AUDIO_INDEX )
    {
        if ( audioFrameCount )
    return 0;
        else
            // there are no audio index entries even though 
            // header indicated the presence of audio
            bfail = 1;
    }

    if ( bfail == 0 ) {
        return 0;
    } else {
    for ( i = 0; i < s->nb_streams; i++ ) {
        sc = s->streams[i]->priv_data;
            if ( sc->key_index != NULL )
        avf_freep(sc->key_index);
    }
    }

    return -1;
}

static int avi_match_key(AVFormatContext *s, 
                         int stream_idx, 
                         AVI_IDX_RELATION relation)
{
    // Map into required stream structures
    AVStream *st = s->streams[stream_idx];
    AVIStreamContext *sc = st->priv_data;

    // Index of the found key.  If the biggest indexed packet number 
    // is smaller then current stream packet number -1 will be returned
    // In case positive index is returned the packet_num contains the 
    // packet number of key packet which is closest to the current packet
    // number of the stream
    int key_idx;

    for ( key_idx = 0; key_idx < sc->index_size; key_idx++ ) {
        // Map into stream position structure of the input stream
        AVIStreamPos *sp = &sc->key_index[key_idx].other_pos[stream_idx];

        switch ( relation )  {
            case AVI_IDX_EQ:
                 if ( sp->packet_num == sc->curr_packet_num )
                    return key_idx;
                else if ( sp->packet_num > sc->curr_packet_num )
                    return -1;
                break;
            case AVI_IDX_GE:
                if ( sp->packet_num >= sc->curr_packet_num )
                    return key_idx;
                break;
            case AVI_IDX_GT:
                if ( sp->packet_num > sc->curr_packet_num )
                    return key_idx;
                break;
            // For AVI_IDX_LE and AVI_IDX_LT if first key packet number
            // greater then current packet number return 0-1=-1
            case AVI_IDX_LE:
                if ( sp->packet_num > sc->curr_packet_num )
                    return (key_idx - 1);
                break;
            case AVI_IDX_LT:
                if ( sp->packet_num >= sc->curr_packet_num )
                    return (key_idx - 1);
                break;
        }
    }

    // No match found
    if ( (relation == AVI_IDX_LE) || (relation == AVI_IDX_LT) )
        return key_idx - 1;
    else 
        return  -1;
}

static int avi_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;
    uint32_t tag, tag1;
    int codec_type, stream_index, frame_period, scale, rate, length;
    int max_bit_rate = 0;
    unsigned int size;
    int i;
    AVStream *st;
    unsigned char check_for_audio_idx = 0;


    // :ZZZ:
    unsigned int avi_flags = 0;
    int total_frames;

    // Default 1000000 microsec
    avi->index_granularity = 1000000;

    if (get_riff(avi, pb, s->iformat->handle_error) < 0)
        return -1;

    /* first list tag */
    stream_index = -1;
    codec_type = -1;
    frame_period = 0;
    for(;;) {
        if (url_feof(pb))
            goto fail;
        tag = get_le32(pb);
        size = get_le32(pb);
#ifdef DEBUG_AVI
        print_tag("tag", tag, size);
#endif

        switch(tag) {
        case MKTAG('L', 'I', 'S', 'T'):
            /* ignored, except when start of video packets */
            tag1 = get_le32(pb);
#ifdef DEBUG_AVI
            print_tag("list", tag1, 0);
#endif
            if (tag1 == MKTAG('m', 'o', 'v', 'i')) {
                if(pb->is_streamed)
                {
                    avi->movi_end = 0x7FFFFFFFFFFF0000ULL;
                }
                else
                {
                    avi->movi_end = url_ftell(pb) + size - 4;
                }
#ifdef DEBUG_AVI
                printf("movi end=%Lx\n", avi->movi_end);
#endif
                // :ZZZ: Read the index
                if ( avi_flags & AVIF_HASINDEX ) {

                    // Save the cirrent iostream position
                    avi->movi_list = url_ftell(pb);

                    // Skip to the end of "movi" chunk
                    url_fskip(pb, size - 4);

                    // Get index tag and size
                    tag1 = get_le32(pb);
                    //print_tag("tag", tag1, 0);

                    // Read index 
                    if ( tag1 == MKTAG('J', 'U', 'N', 'K') )
                    {
                        size = get_le32(pb);
                        printf("Junk (%u) exists after movi\n", size);

                        url_fskip(pb, size);
                        tag1 = get_le32(pb);
                    }

                    if ( tag1 == MKTAG('i', 'd', 'x', '1') ) {

                        if ( (ap != NULL) && (ap->flags & AV_READ_HEADER_FLAG_NO_INDEX) ) {
                            if ( check_for_audio_idx == 1 ) {
                                if ( get_index(s, pb, LOAD_CHECK_FOR_AUDIO_INDEX) == -1 )
                                    return -1;
                            }
                            else
                        {
                                get_index(s, pb, LOAD_FIRST_INDEX);
                            }
                        }
                        else
                        {
                            get_index(s, pb, LOAD_FULL_INDEX);
                        }
                    }

                    else
                    {                        
                        return -1;
                    }
                                                     
                    // Return the stream to "movi" start
                    Avf_url_fseek(pb, avi->movi_list, SEEK_SET);
                }
                // :ZZZ: End index read

                goto end_of_header;
            }
            break;
        case MKTAG('a', 'v', 'i', 'h'):
	    /* avi header */
            /* using frame_period is bad idea */
            frame_period = get_le32(pb);
            //printf("Original frame period is %d\n", frame_period);
            max_bit_rate = get_le32(pb) * 8;

            // :ZZZ: Read flags and total frames
	        // url_fskip(pb, 4 * 4);
            // Padding granularity
            url_fskip(pb, 4);

            // Flags
            avi_flags = get_le32(pb);

            total_frames = get_le32(pb);
            printf("Total play time is %d seconds\n", 
                   (frame_period * total_frames) / 1000000);

            s->ulTotalVideoFrames = total_frames;            
            
            // Skip initial frames :TBD:
            url_fskip(pb, 4);
            // :ZZZ: End Read flags and total frames

            s->nb_streams = get_le32(pb);
            
            // stream counts overflow? 
            if( MAX_STREAMS < s->nb_streams )
            {
                s->nb_streams = 0;
                goto fail;
            }
            
            for(i=0;i<s->nb_streams;i++) 
			{
                // :ZZZ: Allocate AVI specific stream context
                AVStream *st = NULL;
                AVIStreamContext *sc = NULL; 

				st = avf_mallocz(sizeof(AVStream));
				sc = avf_mallocz(sizeof(AVIStreamContext));
				
                if (!st || !sc)
                {
                    goto fail;
                }
				
                AvApi_avcodec_get_context_defaults(&st->codec);

                // :ZZZ: Connect AVI stream structure
                s->streams[i] = st;
                st->priv_data = (AVIStreamContext *)sc;
            }
            url_fskip(pb, size - 7 * 4);
            break;
        case MKTAG('s', 't', 'r', 'h'):
            /* stream header */
            stream_index++;
            tag1 = get_le32(pb);
            switch(tag1) {
            case MKTAG('v', 'i', 'd', 's'):
                codec_type = CODEC_TYPE_VIDEO;

                if (stream_index >= s->nb_streams) {
                    url_fskip(pb, size - 4);
                    break;
                } 

                st = s->streams[stream_index];

                get_le32(pb); /* codec tag */
                get_le32(pb); /* flags */
                get_le16(pb); /* priority */
                get_le16(pb); /* language */
                get_le32(pb); /* XXX: initial frame ? */
                scale= get_le32(pb); /* scale */
                rate= get_le32(pb); /* rate */

                get_le32(pb);   /* start time */
                length = get_le32(pb); /*length*/
    
                s->ulTotalVideoFrames = length;
                st->stream_total_nb_frames = length;
                //printf("scale %d rate %d\n", scale, rate);

                if(scale && rate){                    
                    st->codec.frame_rate = rate;
                    st->codec.frame_rate_base= scale;
                    st->time_length = (int64_t)length * scale * 1000 / rate;
                }else if(frame_period){
                    st->codec.frame_rate = 1000000;
                    st->codec.frame_rate_base= frame_period;
                }else{
                    st->codec.frame_rate = 25;
                    st->codec.frame_rate_base = 1;
                }

                // :ZZZ:
                {
                    printf("Frame rate %u Frame base %u\n", 
                            st->codec.frame_rate,
                            st->codec.frame_rate_base );
                }

                // make the clip bit rate available from this currently unused variable                

                if ( max_bit_rate != 0 )
                {
                    st->codec.bit_rate = max_bit_rate;
                }

                // :End ZZZ:
                
                url_fskip(pb, size - 9 * 4);
                break;
            case MKTAG('a', 'u', 'd', 's'):
                codec_type = CODEC_TYPE_AUDIO;

                st = s->streams[stream_index];
                get_le32(pb); /* codec tag */
                get_le32(pb); /* flags */
                get_le16(pb); /* priority */
                get_le16(pb); /* language */
                get_le32(pb); /* XXX: initial frame ? */
                scale= get_le32(pb); /* scale */
                rate = get_le32(pb); /* rate */
                get_le32(pb);   /* start time */

                length = get_le32(pb); /*length*/
                st->stream_total_nb_frames = length;
                
                if(scale && rate)
                {                    
                    st->time_length = (int64_t)length * scale * 1000 / rate;
                }    
                url_fskip(pb, size - 9 * 4);
                break;
            default:
                goto fail;
            }
            break;
        case MKTAG('s', 't', 'r', 'f'):
            /* stream header */
            if (stream_index >= s->nb_streams) {
                url_fskip(pb, size);
            } else {
                st = s->streams[stream_index];
                switch(codec_type) {
                case CODEC_TYPE_VIDEO:
                    get_le32(pb); /* size */

                    st->codec.logical_width = get_le32(pb);
                    st->codec.logical_height = get_le32(pb);
                    st->codec.physical_width = ( st->codec.logical_width + 15 ) & (~0xF);
                    st->codec.physical_height = ( st->codec.logical_height + 15 ) & (~0xF);

                    get_le16(pb); /* panes */
                    st->codec.bits_per_sample= get_le16(pb); /* depth */
                    tag1 = get_le32(pb);
                    get_le32(pb); /* ImageSize */
                    get_le32(pb); /* XPelsPerMeter */
                    get_le32(pb); /* YPelsPerMeter */
                    get_le32(pb); /* ClrUsed */
                    get_le32(pb); /* ClrImportant */

                    st->codec.extradata_size= size - 10*4;
                    
                    if (st->codec.extradata_size != 0)
                    {
                        st->codec.extradata= avf_malloc(st->codec.extradata_size); //FIXME where should we free this?
                    }

                    get_buffer(pb, st->codec.extradata, st->codec.extradata_size);
                    
                    if(st->codec.extradata_size & 1) //FIXME check if the encoder really did this correctly
                        get_byte(pb);

#ifdef DEBUG_AVI
                    /* ("video", tag1, 0); */ // ?
#endif
                    st->codec.codec_type = CODEC_TYPE_VIDEO;
                    st->codec.codec_tag = tag1;
                    st->codec.codec_id = codec_get_id(codec_bmp_tags, tag1);
//                    url_fskip(pb, size - 5 * 4);
                    break;
                case CODEC_TYPE_AUDIO:
                    get_wav_header(pb, &st->codec, size);

                    st->codec.frame_size = st->codec.block_align;
                    
                    if (size%2) /* 2-aligned (fix for Stargate SG-1 - 3x18 - Shades of Grey.avi) */
                        url_fskip(pb, 1);

                    check_for_audio_idx = 1;

                    printf("Audio: channels %d bps %d codec id %d, sample rate %d bitrate %d length %d\n", 
                           st->codec.channels,
                           st->codec.bits_per_sample,
                           st->codec.codec_id,
                           st->codec.sample_rate,
                           st->codec.bit_rate,
                           st->length);

                    break;
                default:
                    url_fskip(pb, size);
                    break;
                }
            }
            break;
        default:
            /* skip tag */
            size += (size & 1);
            url_fskip(pb, size);
            break;
        }
    }

 end_of_header:
    /* check stream number */
    if (stream_index != s->nb_streams - 1) {
    fail:
        for(i=0;i<s->nb_streams;i++) {
            avf_freep(&s->streams[i]->codec.extradata);
            avf_freep(&s->streams[i]);
        }
        return -1;
    }

    return 0;
}

static int avi_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int n, d[8], size;
    ULONG i;

    // :ZZZ:
    AVIStreamContext *sc;
    AVStream *st;

    memset(d, -1, sizeof(int)*8);    
    //printf("URL offset is %d\n", (unsigned int)url_ftell(pb));

    for(i=url_ftell(pb); !url_feof(pb); i++) {

        int j;

	if (!(pb->is_streamed) && (i >= LO_LONG(avi->movi_end))) { /* Let's see if it's an OpenDML AVI */
	    uint32_t tag, size, tag2;
	    url_fskip(pb, avi->riff_end - url_ftell(pb));
	    if (get_riff(avi, pb, s->iformat->handle_error) < 0)
	        return -1;
	    
	    tag = get_le32(pb);
	    size = get_le32(pb);
	    tag2 = get_le32(pb);
	    if (tag == MKTAG('L','I','S','T') && tag2 == MKTAG('m','o','v','i'))
	        avi->movi_end = url_ftell(pb) + size - 4; 
	    else
	        return -1;
	}

        for(j=0; j<7; j++)
            d[j]= d[j+1];
        d[7]= get_byte(pb);
        
        size= d[4] + (d[5]<<8) + (d[6]<<16) + (d[7]<<24);
        
        //parse ix##
        n= (d[2] - '0') * 10 + (d[3] - '0');
        if(    d[2] >= '0' && d[2] <= '9'
            && d[3] >= '0' && d[3] <= '9'
            && d[0] == 'i' && d[1] == 'x'
            && n < s->nb_streams
            && i + size <= avi->movi_end){
            
            url_fskip(pb, size);
        }
        
        //parse ##dc/##wb
        n= (d[0] - '0') * 10 + (d[1] - '0');
        if(    d[0] >= '0' && d[0] <= '9'
            && d[1] >= '0' && d[1] <= '9'
            &&((d[2] == 'd' && d[3] == 'c') || (d[2] == 'w' && d[3] == 'b') || (d[2] == 'd' && d[3] == 'b') )

            && n < s->nb_streams
            && i + size <= avi->movi_end){
        
            av_new_packet(pkt, size);
            pkt->stream_index = n;

            // :ZZZ: Bookiping of stream data and packet PTS
            st = s->streams[n];
            sc = st->priv_data;

            // Differ between Video and Audio
            if ( d[2] == 'd' && d[3] == 'c' )  {

                // Video - if the packet belogs to the frame index 
                // mark it as key packet and update PTS. Notice that if
                // the AVI file does not have index avi_match_key properly
                // handles it, returning -1
                int key_idx = avi_match_key(s, n, AVI_IDX_EQ);

                //printf("matched key %d\n", key_idx);
                if ( key_idx >= 0 )  {
                    // Yes, mark the packet being returned as key frame,
                    // and update the packet PTS
                    pkt->flags |= PKT_FLAG_KEY;
                    pkt->pts = sc->key_index[key_idx].pts;

                    //printf("Got key packet, packet %d curr_key %d\n", 
                    //       key_idx, sc->curr_packet_num);
                }
                else
                {
                    // Calculate the estimated timestamp according to current packet
                    pkt->pts = ((int64_t)(sc->curr_packet_num * 1000000ULL)) / 
                                   ( st->codec.frame_rate / st->codec.frame_rate_base);                                             
                }
            }
            else  {
                // Audio - always mark as key and update PTS
                pkt->flags |= PKT_FLAG_KEY;
                pkt->pts = ((int64_t)sc->curr_packet_num * 1000000LL * (int64_t)st->codec.frame_size) / 
                             st->codec.sample_rate;
            }

            // Advance the current packet number and current stream length
            sc->curr_length += size;
            sc->curr_packet_num++;
            // :ZZZ: End
            
            get_buffer(pb, pkt->data, pkt->size);

            if (size & 1)
                get_byte(pb);

            return 0;
        }
    }
    
    return -1;
}

static int avi_read_close(AVFormatContext *s)
{
    int i;
    AVIStreamContext *sc;

    for ( i = 0; i < s->nb_streams; i++ ) {
        sc = s->streams[i]->priv_data;
        if ( sc != NULL )  {            
            avf_free(sc->key_index);
            avf_free(sc);
        }
        avf_freep(&s->streams[i]->codec.extradata);
    }

    // s->streams[i], s and s->priv_data will be released by the 
    // caller function (av_close_input_file)

    return 0;
}

static int avi_probe(AVProbeData *p)
{
    /* check file header */
    if (p->buf_size <= 32)
        return 0;
    if (p->buf[0] != 'R' || p->buf[1] != 'I' ||
        p->buf[2] != 'F' || p->buf[3] != 'F')
    {
        if (p->handle_error != NULL) 
        {
            if (p->handle_error(AVI_RIFF_ERROR_ID, p->buf, 4) == AVI_ERR_FAIL) 
            {
                return 0; 
            } 
        } 
        else 
        {
            return 0;
        }
    }        
        
    if (p->buf[8] != 'A' || p->buf[9] != 'V' ||
        p->buf[10] != 'I' || p->buf[11] != ' ')
    {
        if (p->handle_error != NULL) 
        {
            if (p->handle_error(AVI_RIFF_ERROR_ID, p->buf+4, 4) == AVI_ERR_FAIL) 
            {
                return 0; 
            } 
        } 
        else 
        {
            return 0;
        }
    } 
    
    if ( p->custom_probe != NULL )
    {
        if ( p->custom_probe( 0, p->buf, AVPROBE_BUF_SIZE ) == AVI_ERR_FAIL )
        {
            return 0;
        }
    }

    return AVPROBE_SCORE_MAX;
}

// :ZZZ: Seek support
int avi_read_seek(AVFormatContext *s, int64_t pts)
{
    // 
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;
    AVIStreamContext *sc = NULL;
    
    // Stream that will contain the packet whcih offset will be used for seek
    AVIStreamContext *selected_sc = NULL;

    // Packet which offset will be used for seek
    int selected_key = 0;

    // Stream containg the packet with minimum offset

    // Minimum offset of the streams
    unsigned int min_offset = 0xFFFFFFFF;

    //
    int i, j;

    for ( i = 0; i < s->nb_streams; i++ ) {

        // For now search only in Video stream
        if ( s->streams[i]->codec.codec_type != CODEC_TYPE_VIDEO )
            continue;

        sc = s->streams[i]->priv_data;

       // Search for key packet with closest PTS
        for ( j = 0; j < sc->index_size; j++ ) {

            //printf("entry j %d pts %d\n", j, (unsigned int)sc->key_index[j].pts);

            if ( sc->key_index[j].pts >= pts ) {

                //Take the previous I-Frame, not the closest one.
                if (sc->key_index[j].pts > pts)
                {
                    j = j-1;
                }

                if ( sc->key_index[j].offset < min_offset )  {                    
                    selected_sc = sc;
                    selected_key = j;
                    min_offset = sc->key_index[selected_key].offset;
                }
                break;
            }
        }

        // If requested time is bigger then any key frame PTS pick
        // the last frame
        if ( (j > 0) && (j == sc->index_size) ) {
            selected_sc = sc;
            selected_key = j - 1;
            min_offset = sc->key_index[selected_key].offset;
        }
    }

    //printf("Going to seek to %d pts %d stream %d\n",
    //       min_offset, 
    //       (unsigned int)selected_sc->key_index[selected_key].pts, 
    //       selected_st);

    if ( selected_sc != NULL )  
    {

    // Update all the streams current position
    for ( i = 0; i < s->nb_streams; i++ )       
       {        
        AVIKeyIndexEntry *key = &selected_sc->key_index[selected_key];

        sc = s->streams[i]->priv_data;
        sc->curr_packet_num  = key->other_pos[i].packet_num;
        sc->curr_length = key->other_pos[i].offset;

        //printf("Updated stream %d: offset %d, packet %d\n", 
        //        i, sc->curr_length, sc->curr_packet_num);
       }

    // Seek to the smallest offset of all the streams  
        Avf_url_fseek(pb, 
                  avi->movi_list + selected_sc->key_index[selected_key].offset,
                  SEEK_SET);
    }

    return 0;
}

// :ZZZ: seek accroding to video frame key
int avi_read_seek_key(AVFormatContext *s, int stream_id, int pos, int whence)
{
    //  
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;

    // Characteristics of index entry, whcih will serve for seek operation
    AVIStreamContext *seek_sc;
    AVIKeyIndexEntry *seek_key;
    int seek_key_idx = 0;

    //
    int i;

    seek_sc = s->streams[stream_id]->priv_data;

    switch ( whence ) {
        case SEEK_CUR: 
        {
            // Matching criteria for SEEK_CUR option
            // AVI_IDX_RELATION matching_criteria = pos >= 0 ? AVI_IDX_GE : 
            //                                                AVI_IDX_LT;
            seek_key_idx = avi_match_key(s, stream_id, AVI_IDX_LT);
#if 0
            // If we're beyond the last key and going up or before the 
            // first key and going down - no seek operation is performed
            // to avoid actual file seek in direction opposite to the 
            // caller intention, returned -1 indicates the condition.
            if ( seek_key_idx < 0 )
                return -1;                                
#endif
            seek_key_idx += pos;
        }
        break;
        case SEEK_SET:
            seek_key_idx = pos;
            break;
        case SEEK_END:
            seek_key_idx = seek_sc->index_size - 1 + pos;
            break;
    }

    // Keep the key index in the index size range
    if ( (seek_key_idx < 0) || (seek_key_idx >= seek_sc->index_size) )
        return -1;

    //printf("Target key index %d, going to seek to offset %d\n", 
    //       seek_key_idx,
    //       seek_sc->key_index[seek_key_idx].offset);

    // Map into index entry to be used for seek operation
    seek_key = &seek_sc->key_index[seek_key_idx];

    // Actual file seek
    Avf_url_fseek(pb, avi->movi_list + seek_key->offset, SEEK_SET);

    // Update all the streams current position to be consistent with
    // coming file seek operation
    for ( i = 0; i < s->nb_streams; i++ )  { 
        AVIStreamContext *sc = s->streams[i]->priv_data;
        sc->curr_packet_num  = seek_key->other_pos[i].packet_num;
        sc->curr_length      = seek_key->other_pos[i].offset;

        //printf("Updated stream %d: offset %d, packet %d\n", 
        //        i, sc->curr_length, sc->curr_packet_num);
    }

    return seek_key_idx;
}

static int avi_read_is_keyframe(AVFormatContext *s, int stream_id, int frame)
{
    int found = 0;  
    
    if (s != NULL) {
        AVIContext *avi = s->priv_data;
        
        if (avi != NULL) {
            AVIStreamContext *stream = s->streams[stream_id]->priv_data;            

            // If key frame index size is zero then every frame is key frame
            if (stream->index_size == 0)
            {
                found = 1;
            }            
            else
            {
                int idx;

                // :TODO: Improve searching mechanism (use binary search)
                for (idx = 0; idx < stream->index_size; idx++) {
                    // Map into stream position structure of the input stream
                    AVIStreamPos *sp = &stream->key_index[idx].other_pos[stream_id];

                    if (sp->packet_num == frame) {
                        found = 1;
                        break;
                    }

                    if (sp->packet_num > frame) {
                        break;
                    }
                }
            }
        }
    }

    return found;        
}

static AVInputFormat avi_iformat = {
    "avi",
    "avi format",
    sizeof(AVIContext),
    avi_probe,
    avi_read_header,
    avi_read_packet,
    avi_read_close,
    avi_read_seek,
    avi_read_seek_key,
    NULL,
    avi_read_is_keyframe,
};

int Avf_avidec_init(void)
{
    av_register_input_format(&avi_iformat);
    return 0;
}
