/*
 * AVI encoder.
 * Copyright (c) 2000 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *                              
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avformat.h"
#include "avi.h"
#include "av.h"

/*
 * TODO: 
 *  - fill all fields if non streamed (nb_frames for example)
 */

#define INVALID_POSITION     ((UINT) -1UL )

//Chunk header length = 4 chars (usually '##dc' or '##wb') + 4 byte chunk size
#define CHUNK_HEADER_LENGTH  8 

#define IS_ODD(x)   ( x & 1 )

STATIC offset_t first_packet_position = INVALID_POSITION;

typedef struct AVIIentry {
    UINT len    : 21;
    BYTE flags  : 1; 
    BYTE stream_index : 4;
} AVIIentry;

#define AVI_INDEX_CLUSTER_SIZE 16384

#define BLKSIZE 1024


typedef struct AVIIndex {
    offset_t    indx_start;
    int         entry;
    int         ents_allocated;
    AVIIentry** cluster;
} AVIIndex;

typedef struct {
    offset_t riff_start, movi_list, odml_list;
    offset_t frames_hdr_all, frames_hdr_strm[MAX_STREAMS];
    int audio_strm_length[MAX_STREAMS];
    int riff_id;
    AVIIndex common_index;
} AVIContext;

static inline AVIIentry* avi_get_ientry(AVIIndex* idx, int ent_id) 
{
    int cl = ent_id / AVI_INDEX_CLUSTER_SIZE;
    int id = ent_id % AVI_INDEX_CLUSTER_SIZE;

    return &idx->cluster[cl][id];
}

offset_t start_tag(ByteIOContext *pb, const char *tag)
{
    put_tag(pb, tag);
    put_le32(pb, 0);
    return url_ftell(pb);
}

void end_tag(ByteIOContext *pb, offset_t start)
{
    offset_t pos;

    pos = url_ftell(pb);
    Avf_url_fseek(pb, start - 4, SEEK_SET);
    put_le32(pb, (uint32_t)(pos - start));
    Avf_url_fseek(pb, pos, SEEK_SET);
}

/* Note: when encoding, the first matching tag is used, so order is
   important if multiple tags possible for a given codec. */
const CodecTag codec_bmp_tags[] = {
    { CODEC_ID_H263, MKTAG('H', '2', '6', '3') },
    { CODEC_ID_H263P, MKTAG('H', '2', '6', '3') },
    { CODEC_ID_H263I, MKTAG('I', '2', '6', '3') }, /* intel h263 */

    /* added based on MPlayer */
    { CODEC_ID_H263I, MKTAG('i', '2', '6', '3') },
    { CODEC_ID_H263P, MKTAG('U', '2', '6', '3') },
    { CODEC_ID_H263P, MKTAG('h', '2', '6', '3') },
    { CODEC_ID_H263P, MKTAG('v', 'i', 'v', '1') },

    // :ZZZ: 
    { CODEC_ID_MPEG4Z, MKTAG('X', 'V', 'I', 'D'), .invalid_asf = 1 },    
    { CODEC_ID_MPEG4Z, MKTAG('D', 'X', '5', '0'), .invalid_asf = 1 },    

    { CODEC_ID_MPEG4Z, MKTAG('D', 'I', 'V', 'X'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4Z, MKTAG('M', '4', 'S', '2') },
    { CODEC_ID_MPEG4Z, MKTAG('M', 'P', '4', 'S') },
    { CODEC_ID_MPEG4Z, MKTAG('m', '4', 's', '2') },

    { CODEC_ID_MJPEG, MKTAG('M', 'J', 'P', 'G') },
    { CODEC_ID_MPEG4, MKTAG('D', 'I', 'V', 'X'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4, MKTAG('d', 'i', 'v', 'x'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4, MKTAG('D', 'X', '5', '0'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4, MKTAG('X', 'V', 'I', 'D'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4, MKTAG('x', 'v', 'i', 'd'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4, MKTAG('m', 'p', '4', 's'), .invalid_asf = 1 },
    { CODEC_ID_MPEG4, MKTAG('M', 'P', '4', 'S') },
    { CODEC_ID_MPEG4, MKTAG('M', '4', 'S', '2') },
    { CODEC_ID_MPEG4, MKTAG('m', '4', 's', '2') },
    { CODEC_ID_MPEG4, MKTAG(0x04, 0, 0, 0) }, /* some broken avi use this */

    /* added based on MPlayer */
    { CODEC_ID_MPEG4, MKTAG('D', 'I', 'V', '1') },
    { CODEC_ID_MPEG4, MKTAG('d', 'i', 'v', '1') },
    { CODEC_ID_MPEG4, MKTAG('X', 'v', 'i', 'D') },
    { CODEC_ID_MPEG4, MKTAG('B', 'L', 'Z', '0') },
    { CODEC_ID_MPEG4, MKTAG('m', 'p', '4', 'v') },
    { CODEC_ID_MPEG4, MKTAG('U', 'M', 'P', '4') },

    { CODEC_ID_MSMPEG4V3, MKTAG('D', 'I', 'V', '3'), .invalid_asf = 1 }, /* default signature when using MSMPEG4 */
    { CODEC_ID_MSMPEG4V3, MKTAG('d', 'i', 'v', '3'), .invalid_asf = 1 },
    { CODEC_ID_MSMPEG4V3, MKTAG('M', 'P', '4', '3') }, 

    /* added based on MPlayer */
    { CODEC_ID_MSMPEG4V3, MKTAG('M', 'P', 'G', '3') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('m', 'p', 'g', '3') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('m', 'p', '4', '3') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('D', 'I', 'V', '5') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('d', 'i', 'v', '5') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('D', 'I', 'V', '6') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('d', 'i', 'v', '6') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('D', 'I', 'V', '4') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('d', 'i', 'v', '4') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('A', 'P', '4', '1') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('C', 'O', 'L', '1') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('c', 'o', 'l', '1') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('C', 'O', 'L', '0') }, 
    { CODEC_ID_MSMPEG4V3, MKTAG('c', 'o', 'l', '0') }, 

    { CODEC_ID_MSMPEG4V2, MKTAG('M', 'P', '4', '2') }, 

    /* added based on MPlayer */
    { CODEC_ID_MSMPEG4V2, MKTAG('D', 'I', 'V', '2') },
    { CODEC_ID_MSMPEG4V2, MKTAG('d', 'i', 'v', '2') },
    { CODEC_ID_MSMPEG4V2, MKTAG('m', 'p', '4', '2') },
 
    { CODEC_ID_MSMPEG4V1, MKTAG('M', 'P', 'G', '4') }, 

    /* added based on MPlayer */
    { CODEC_ID_MSMPEG4V1, MKTAG('D', 'I', 'V', '4') }, 
    { CODEC_ID_MSMPEG4V1, MKTAG('d', 'i', 'v', '4') }, 
    { CODEC_ID_MSMPEG4V1, MKTAG('m', 'p', 'g', '4') }, 

    // :ZZZ:
    { CODEC_ID_WMVZ,  MKTAG('W', 'M', 'V', 'A') }, 
    { CODEC_ID_WMVMZ, MKTAG('W', 'M', 'V', '3') },
    { CODEC_ID_H264Z, MKTAG('H', '2', '6', '4') },

    { CODEC_ID_WMV1, MKTAG('W', 'M', 'V', '1') }, 

    /* added based on MPlayer */
    { CODEC_ID_WMV1, MKTAG('w', 'm', 'v', '1') }, 

    { CODEC_ID_WMV2, MKTAG('W', 'M', 'V', '2') }, 
    { CODEC_ID_DVVIDEO, MKTAG('d', 'v', 's', 'd') }, 
    { CODEC_ID_DVVIDEO, MKTAG('D', 'V', 'S', 'D') }, 
    { CODEC_ID_DVVIDEO, MKTAG('d', 'v', 'h', 'd') }, 
    { CODEC_ID_DVVIDEO, MKTAG('d', 'v', 's', 'l') }, 
    { CODEC_ID_DVVIDEO, MKTAG('d', 'v', '2', '5') },
    { CODEC_ID_MPEG1VIDEO, MKTAG('m', 'p', 'g', '1') }, 
    { CODEC_ID_MPEG1VIDEO, MKTAG('m', 'p', 'g', '2') }, 
    { CODEC_ID_MPEG1VIDEO, MKTAG('P', 'I', 'M', '1') }, 
    { CODEC_ID_MJPEG, MKTAG('M', 'J', 'P', 'G') },
    { CODEC_ID_HUFFYUV, MKTAG('H', 'F', 'Y', 'U') },
    { CODEC_ID_HUFFYUV, MKTAG('h', 'f', 'y', 'u') },
    { CODEC_ID_CYUV, MKTAG('C', 'Y', 'U', 'V') },
    { CODEC_ID_CYUV, MKTAG('c', 'y', 'u', 'v') },
    // :ZZZ: Raw codec
    { CODEC_ID_RAWZ, MKTAG('Y', 'U', 'Y', '2') },
    { CODEC_ID_RAWZ, MKTAG('Y', '4', '2', '2') },
    { CODEC_ID_RAWVIDEO, MKTAG('Y', '4', '2', '2') },
    { CODEC_ID_RAWZ, MKTAG('I', '4', '2', '0') },
    { CODEC_ID_RAWZ, MKTAG('B', 'Y', 'R', '2') },
    { CODEC_ID_INDEO3, MKTAG('i', 'v', '3', '1') },
    { CODEC_ID_INDEO3, MKTAG('i', 'v', '3', '2') },
    { CODEC_ID_INDEO3, MKTAG('I', 'V', '3', '1') },
    { CODEC_ID_INDEO3, MKTAG('I', 'V', '3', '2') },
    { CODEC_ID_VP3, MKTAG('V', 'P', '3', '1') },
    { 0, 0 },
};

unsigned int codec_get_tag(const CodecTag *tags, int id)
{
    while (tags->id != 0) {
        if (tags->id == id)
            return tags->tag;
        tags++;
    }
    return 0;
}

static unsigned int codec_get_asf_tag(const CodecTag *tags, int id)
{
    while (tags->id != 0) {
        if (!tags->invalid_asf && tags->id == id)
            return tags->tag;
        tags++;
    }
    return 0;
}

enum CodecID codec_get_id(const CodecTag *tags, unsigned int tag)
{
    while (tags->id != 0) {
        if (tags->tag == tag) {
            return (enum CodecID)tags->id;
        }
        tags++;
    }
    return CODEC_ID_NONE;
}

unsigned int codec_get_bmp_tag(int id)
{
    return codec_get_tag(codec_bmp_tags, id);
}

/* BITMAPINFOHEADER header */
void put_bmp_header(ByteIOContext *pb, AVCodecContext *enc, const CodecTag *tags, int for_asf)
{
    put_le32(pb, 40 + enc->extradata_size); /* size */
    put_le32(pb, enc->logical_width);
    put_le32(pb, enc->logical_height);
    put_le16(pb, 1); /* planes */
    
    put_le16(pb, enc->bits_per_sample ? enc->bits_per_sample : 24); /* depth */
    /* compression type */
    put_le32(pb, for_asf ? codec_get_asf_tag(tags, enc->codec_id) : enc->codec_tag);
    put_le32(pb, enc->physical_width * enc->physical_height * 3);
    put_le32(pb, 0);
    put_le32(pb, 0);
    put_le32(pb, 0);
    put_le32(pb, 0);
    
    put_buffer_cached(pb, enc->extradata, enc->extradata_size);


    // :ZZZ: Seems to cause errors with wmv
    if ( for_asf == 0 )
    {
        if (enc->extradata_size & 1)
            put_byte(pb, 0);
    }
}


typedef struct _ConvertTableTagTofccHandler {
    unsigned int codec_tag;
    unsigned int fccHandler;
} SConvertTableTagTofccHandler;

const static SConvertTableTagTofccHandler ConvertTableTagTofccHandler[] =
{   // From                   - To
    {MKTAG('D', 'X', '5', '0'), MKTAG('D', 'I', 'V', 'X')},
    { 0, 0 },
};

static unsigned int get_fccHandler(unsigned int codec_tag)
{
    const SConvertTableTagTofccHandler *p = &ConvertTableTagTofccHandler[0];
    while (p->codec_tag != 0)
    {
        if (p->codec_tag == codec_tag) return p->fccHandler;
        p++;
    }
    return codec_tag;
}

static void parse_specific_params(AVCodecContext *stream, int *au_byterate, int *au_ssize, int *au_scale)
{
    switch(stream->codec_id) {
    case CODEC_ID_PCM_S16LE:
       *au_scale = *au_ssize = 2*stream->channels;
       *au_byterate = *au_ssize * stream->sample_rate;
        break;
    case CODEC_ID_PCM_U8:
    case CODEC_ID_PCM_ALAW:
    case CODEC_ID_PCM_MULAW:
        *au_scale = *au_ssize = stream->channels;
        *au_byterate = *au_ssize * stream->sample_rate;
        break;
    case CODEC_ID_MP2:
        *au_ssize = 1;
        *au_scale = 1;
        *au_byterate = stream->bit_rate / 8;
        break;
    case CODEC_ID_MP3LAME:
        *au_ssize = 1;
        *au_scale = 1;
        *au_byterate = stream->bit_rate / 8;   
        break;
    case CODEC_ID_ADPCM_MS:
    {
        *au_ssize = 1;
        *au_scale = stream->block_align;
        *au_byterate = stream->bit_rate / 8;   
        break;
    } 
    default:
        *au_ssize = 1;
        *au_scale = 1; 
        *au_byterate = stream->bit_rate / 8;
        break;
    }
}

static offset_t avi_start_new_riff(AVIContext *avi, ByteIOContext *pb, 
                                   const char* riff_tag, const char* list_tag)
{
    offset_t loff;
    
    avi->riff_id++;

    avi->common_index.entry = 0;
        
    avi->riff_start = start_tag(pb, "RIFF");
    put_tag(pb, riff_tag);
    loff = start_tag(pb, "LIST");
    put_tag(pb, list_tag);
    return loff;
}

static unsigned char* avi_stream2fourcc(unsigned char* tag, int index, 
                                        enum CodecType type)
{
    tag[0] = '0';
    tag[1] = '0' + index;
    if (type == CODEC_TYPE_VIDEO) {
        tag[2] = 'd';
        tag[3] = 'c';
    } else {
        tag[2] = 'w';
        tag[3] = 'b';
    }
    tag[4] = '\0';
    return tag;
}

static int avi_write_header(AVFormatContext *s)
{
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int bitrate, n, i, nb_frames, au_byterate, au_ssize, au_scale;
    AVCodecContext *stream, *video_enc;
    offset_t list1, list2, strh, strf, strn;

    /* Reset first_packet_position for new file */
    first_packet_position = INVALID_POSITION;

    /* header list */
    avi->riff_id = 0;
    list1 = avi_start_new_riff(avi, pb, "AVI ", "hdrl");

    /* avi header */
    put_tag(pb, "avih");
    put_le32(pb, 14 * 4);
    bitrate = 0;

    video_enc = NULL;
    for(n=0;n<s->nb_streams;n++) {
        stream = &s->streams[n]->codec;

        // bit_rate is set by the recorder in case of MP2, but ADPCM and sets its
        // own parameters.
        if (stream->codec_type == CODEC_TYPE_AUDIO)
        {

            if (stream->codec_id == CODEC_ID_ADPCM_MS)
            {
                 //Initialize codec params (some of them weren't initialized.
                 stream->frame_size = (BLKSIZE - 7 * stream->channels) * 2 / stream->channels + 2;
                 stream->block_align = BLKSIZE;

                 stream->bit_rate =  stream->block_align 
                                    * stream->sample_rate
                                    * 8 
                                    / stream->frame_size;
            }
        }
        bitrate += stream->bit_rate;
        if (stream->codec_type == CODEC_TYPE_VIDEO)
            video_enc = stream;
    }
    
    nb_frames = 0;

    if(video_enc){
        put_le32(pb, (uint32_t)(int64_t_C(1000000) * video_enc->frame_rate_base / video_enc->frame_rate));
    } else {
    put_le32(pb, 0);
    }
    put_le32(pb, bitrate / 8); /* XXX: not quite exact */
    put_le32(pb, 0); /* padding */
    if(pb->is_streamed)
    {
        put_le32(pb, AVIF_TRUSTCKTYPE | AVIF_ISINTERLEAVED); /* flags */
    }
    else
    {
        put_le32(pb, AVIF_TRUSTCKTYPE | AVIF_HASINDEX | AVIF_ISINTERLEAVED); /* flags */
    }
    avi->frames_hdr_all = url_ftell(pb); /* remember this offset to fill later */
    put_le32(pb, nb_frames); /* nb frames, filled later */
    put_le32(pb, 0); /* initial frame */
    put_le32(pb, s->nb_streams); /* nb streams */
    put_le32(pb, 1024 * 1024); /* suggested buffer size */
    if(video_enc){    
    put_le32(pb, video_enc->logical_width);
    put_le32(pb, video_enc->logical_height);
    } else {
    put_le32(pb, 0);
    put_le32(pb, 0);
    }   
    put_le32(pb, 0); /* reserved */
    put_le32(pb, 0); /* reserved */
    put_le32(pb, 0); /* reserved */
    put_le32(pb, 0); /* reserved */
    
    /* stream list */
    for(i=0;i<n;i++) {
        list2 = start_tag(pb, "LIST");
        put_tag(pb, "strl");
    
        stream = &s->streams[i]->codec;

        /* FourCC should really be set by the codec itself */
        if (! stream->codec_tag) {
            stream->codec_tag = codec_get_bmp_tag(stream->codec_id);
        }

        /* stream generic header */
        strh = start_tag(pb, "strh");
        switch(stream->codec_type) {
        case CODEC_TYPE_VIDEO:
            put_tag(pb, "vids");

            put_le32(pb, get_fccHandler(stream->codec_tag));

            put_le32(pb, 0); /* flags */
            put_le16(pb, 0); /* priority */
            put_le16(pb, 0); /* language */
            put_le32(pb, 0); /* initial frame */
            
            put_le32(pb, stream->frame_rate_base); /* scale */
            put_le32(pb, stream->frame_rate); /* rate */

            put_le32(pb, 0); /* start */
            avi->frames_hdr_strm[i] = url_ftell(pb); /* remember this offset to fill later */
            put_le32(pb, nb_frames); /* length, XXX: fill later */
            put_le32(pb, 1024 * 1024); /* suggested buffer size */
            put_le32(pb, (unsigned int)-1); /* quality */
            put_le32(pb, stream->physical_width * stream->physical_height * 3); /* sample size */
            put_le16(pb, 0);
            put_le16(pb, 0);
            put_le16(pb, stream->logical_width);
            put_le16(pb, stream->logical_height);
            break;
        case CODEC_TYPE_AUDIO:
            put_tag(pb, "auds");
            put_le32(pb, 1); /* tag */
            put_le32(pb, 0); /* flags */
            put_le16(pb, 0); /* priority */
            put_le16(pb, 0); /* language */
            put_le32(pb, 0); /* initial frame */
            parse_specific_params(stream, &au_byterate, &au_ssize, &au_scale);
            put_le32(pb, au_scale); /* scale */
            put_le32(pb, au_byterate); /* rate */
            put_le32(pb, 0); /* start */
            avi->frames_hdr_strm[i] = url_ftell(pb); /* remember this offset to fill later */
            put_le32(pb, 0); /* length, XXX: filled later */
            put_le32(pb, 12 * 1024); /* suggested buffer size */
            put_le32(pb, (unsigned int)-1); /* quality */
            put_le32(pb, au_ssize); /* sample size */
            put_le32(pb, 0);
            put_le32(pb, 0);
            break;
        default:
            av_abort();
        }
        end_tag(pb, strh);

        strf = start_tag(pb, "strf");
        switch(stream->codec_type) {
        case CODEC_TYPE_VIDEO:
        put_bmp_header(pb, stream, codec_bmp_tags, 0);
            break;
        case CODEC_TYPE_AUDIO:
            if (put_wav_header(pb, stream) < 0) {
                avf_free(avi);
                return -1;
            }
            break;
        default:
            av_abort();
        }
        end_tag(pb, strf);

        {
            char *data = NULL;

        switch(stream->codec_type) {
        case CODEC_TYPE_VIDEO:
                data = stream->strn_video;
            break;
        case CODEC_TYPE_AUDIO:
                data = stream->strn_audio;
            break;
        default:
            av_abort();
        }

            if (data != NULL)
            {
                strn = start_tag(pb, "strn");
                put_tag(pb, data);
        end_tag(pb, strn);
            }
        }
    
        end_tag(pb, list2);
    }
    
    end_tag(pb, list1);
    
    avi->movi_list = start_tag(pb, "LIST");
    put_tag(pb, "movi");

    put_flush_packet(pb);

    return 0;
}

static int avi_write_ix(AVFormatContext *s)
{
    ByteIOContext *pb = &s->pb;
    AVIContext *avi = s->priv_data;
    unsigned char tag[5];
    unsigned char ix_tag[] = "ix00";
    int j;
    int posn = first_packet_position;
    
    if (avi->riff_id > AVI_MASTER_INDEX_SIZE)
        return -1;
    
   for ( j = 0; j < avi->common_index.entry; ++j ) 
   {
        offset_t ix, pos;
        AVIIentry *ie = avi_get_ientry( &avi->common_index, j );
        
        avi_stream2fourcc(&tag[0], ie->stream_index, s->streams[ie->stream_index]->codec.codec_type);
        ix_tag[3] = '0' + ie->stream_index;
     
        /* Writing AVI OpenDML leaf index chunk */
        ix = url_ftell(pb); 
        put_tag(pb, (const char *)&ix_tag[0]);     /* ix?? */
        put_le32( pb, avi->common_index.entry * 8 + 24 ); /* chunk size */
        put_le16(pb, 2);             /* wLongsPerEntry */
        put_byte(pb, 0);             /* bIndexSubType (0 == frame index) */ 
        put_byte(pb, 1);             /* bIndexType (1 == AVI_INDEX_OF_CHUNKS) */
        // :ZZZ: TODO: This may not be correct 
        put_le32( pb, avi->common_index.entry ); /* nEntriesInUse */
        put_tag(pb, (const char *)&tag[0]);        /* dwChunkId */
        put_le64(pb, avi->movi_list);/* qwBaseOffset */
        put_le32(pb, 0);             /* dwReserved_3 (must be 0) */

        // :ZZZ: Commented out while removing pos field from AVIEntry              
        // put_le32(pb, ie->pos + 8);                      
        put_le32( pb, posn + 8 );     
        posn += ie->len  + 8 + ( IS_ODD(ie->len) ? 1 : 0 );

        put_le32(pb, ((uint32_t)ie->len & ~0x80000000) |
                (ie->flags & 1 /* 0x10 */ ? 0 : 0x80000000));
    
        put_flush_packet(pb);
        pos = url_ftell(pb);
    
        /* Updating one entry in the AVI OpenDML master index */
        Avf_url_fseek( pb, avi->common_index.indx_start - 8, SEEK_SET );
        put_tag(pb, "indx");                 /* enabling this entry */
        url_fskip(pb, 8);
        put_le32(pb, avi->riff_id);          /* nEntriesInUse */
        url_fskip(pb, 16*avi->riff_id);
        put_le64(pb, ix);                    /* qwOffset */
        put_le32(pb, pos - ix);              /* dwSize */
        put_le32( pb, avi->common_index.entry ); /* dwDuration */

        Avf_url_fseek(pb, pos, SEEK_SET);
    }
    return 0;
}

static int avi_write_idx1(AVFormatContext *s)
{
    ByteIOContext *pb = &s->pb;
    AVIContext *avi = s->priv_data;
    offset_t file_size, idx_chunk;
    int  n, nb_frames, au_byterate, au_ssize, au_scale;
    AVCodecContext *stream;
    char tag[5];

    if (!url_is_streamed(pb)) 
    {
        AVIIentry* ie = 0;
        int index_num = 0;
        int posn = first_packet_position;

        idx_chunk = start_tag(pb, "idx1");

        // Skip writing index table if there are no frames
        if( (avi->common_index).entry > 0 )
        {
        do 
        {
            ie = avi_get_ientry( &avi->common_index, index_num );

            index_num++;

            avi_stream2fourcc((unsigned char*)&tag[0], ie->stream_index,
                          s->streams[ie->stream_index]->codec.codec_type); 

            put_tag(pb, &tag[0]);

            put_le32(pb, ie->flags ? 0x10 : 0x00);
            put_le32(pb, posn);                
            put_le32(pb, ie->len);   

            // Incr. posn by (length of chunk + header len. + 0/1 padding bytes)
            posn += ie->len  + CHUNK_HEADER_LENGTH + ( IS_ODD(ie->len) ? 1 : 0 );

        } while ( index_num != avi->common_index.entry );
        }

        end_tag(pb, idx_chunk);

        /* Fill in frame/sample counters */
        file_size = url_ftell(pb);
        nb_frames = 0;
        for( n = 0; n < s->nb_streams; n++ ) 
        {
            if ( avi->frames_hdr_strm[n] != 0 ) 
            {
                stream = &s->streams[n]->codec;
                Avf_url_fseek(pb, avi->frames_hdr_strm[n], SEEK_SET);
                if (stream->codec_type == CODEC_TYPE_VIDEO) 
                {
                    put_le32(pb, s->streams[n]->stream_total_nb_frames); 
                    nb_frames = s->streams[n]->stream_total_nb_frames;
                } 
                else 
                {
                    //Remark: Ignore the case for MP2 to comply with DIVX
                    if (/*stream->codec_id == CODEC_ID_MP2 || */  
                        stream->codec_id == CODEC_ID_MP3LAME ||
                        stream->codec_id == CODEC_ID_ADPCM_MS) 
                    {
                        put_le32(pb, s->streams[n]->stream_total_nb_frames);
                    } 
                    else 
                    {
                        parse_specific_params(stream, &au_byterate, &au_ssize, &au_scale);
                        put_le32(pb, avi->audio_strm_length[n] / au_ssize);
                    }
                }
            }    
        }

        if (avi->frames_hdr_all != 0) 
        {
           Avf_url_fseek(pb, avi->frames_hdr_all, SEEK_SET);
           put_le32(pb, nb_frames); 
        }

        Avf_url_fseek(pb, file_size, SEEK_SET);
    }
    return 0;
}

static int avi_write_packet(AVFormatContext *s, int stream_index,
                            uint8_t *buf, int size, int64_t force_pts)
{
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;
    unsigned char tag[5];
    unsigned int flags;
    AVCodecContext *enc;    

    /* 
       :ZZZ:   Currently the code does not support cases where AVI size is 
               greater than AVI_MAX_RIFF_SIZE.  This case 
               requires multiple RIFFs with AVIX header for 
               additional RIFF structures which is not currently
               supported though following bit of code exists to 
               handle this case! Following code is untested and 
               currently unused 
               (Function avi_write_ix() may not work correctly.)
    */
    if (url_ftell(pb) - avi->riff_start > AVI_MAX_RIFF_SIZE) 
    { 
        avi_write_ix(s);
        end_tag(pb, avi->movi_list);
        
        if (avi->riff_id == 1)
            avi_write_idx1(s);

        end_tag(pb, avi->riff_start);
        avi->movi_list = avi_start_new_riff(avi, pb, "AVIX", "movi");
    }
    
    enc = &s->streams[stream_index]->codec;
    avi_stream2fourcc(&tag[0], stream_index, enc->codec_type);

    if (enc->codec_type == CODEC_TYPE_AUDIO) 
    {
       avi->audio_strm_length[stream_index] += size;
       flags = 1; 
    } 
    else 
    {
        // :ZZZ: Take key_frame info from FormatContext, and not from
        // EncContext. We must separate these variables since encoder is already
        // encoding new frames by the time this frame is written
        // flags = enc->coded_frame->key_frame ? 0x10 : 0x00;
        flags = s->key_frame ? 1 : 0; 
    }

    if (!url_is_streamed(&s->pb)) 
    {
        AVIIndex* idx = &avi->common_index; 
        int cl = idx->entry / AVI_INDEX_CLUSTER_SIZE;
        int id = idx->entry % AVI_INDEX_CLUSTER_SIZE;
   
        if (idx->ents_allocated <= idx->entry) 
        {
            idx->cluster = avf_realloc(idx->cluster, (cl+1) * sizeof(void*)); 
            if (!idx->cluster)
                return -1;
            idx->cluster[cl] = avf_malloc(AVI_INDEX_CLUSTER_SIZE * sizeof(AVIIentry));
            if (!idx->cluster[cl])
                return -1;
            idx->ents_allocated += AVI_INDEX_CLUSTER_SIZE;
        }
        
        idx->cluster[cl][id].flags = flags; 

        idx->cluster[cl][id].len = size;
        
        if ( first_packet_position == INVALID_POSITION )
        {
            first_packet_position =  url_ftell(pb) - avi->movi_list;
        }

        idx->cluster[cl][id].stream_index = stream_index;        

        idx->entry++;
    }
    
    put_buffer_cached(pb, tag, 4);
    put_le32(pb, size);
    put_buffer(pb, buf, size);
    if ( IS_ODD(size) )
        put_byte(pb, 0);        

    put_flush_packet(pb);

    // Update total number of frames for this stream
    s->streams[stream_index]->stream_total_nb_frames++;

    return 0;
}

static int avi_cleanup(AVFormatContext *s)
{
    AVIContext *avi = s->priv_data;
    int res = 0;
    int i;

    if ( avi != NULL )
    {
        for( i = 0; i < avi->common_index.ents_allocated / AVI_INDEX_CLUSTER_SIZE; ++i )
        {
            av_free(avi->common_index.cluster[i]);
        }    

        av_free(avi->common_index.cluster);
        avi->common_index.cluster = NULL;
        avi->common_index.ents_allocated = avi->common_index.entry = 0;
    }

    // :ZZZ:
    for(i=0;i<s->nb_streams;i++) 
    {
        AVStream *st = s->streams[i];
        av_free(st->codec.extradata);
    }
    
    return res;
}

static int avi_write_trailer(AVFormatContext *s)
{
    AVIContext *avi = s->priv_data;
    ByteIOContext *pb = &s->pb;
    int res = 0;
    int i, n, nb_frames;
    offset_t file_size;
    
    if ( pb->opaque == NULL )
    {
        avi_cleanup( s );
        return -1;
    }
    
    put_flush_packet_end(pb);

    // Recover from file reopen (from burst to regular mode)
    Avf_url_fseek(pb, url_filesize(url_fileno(pb)), SEEK_SET); 
    pb->must_flush = 1;

    if (avi->riff_id == 1) 
    {
        end_tag(pb, avi->movi_list);
        res = avi_write_idx1(s);
        end_tag(pb, avi->riff_start);
    }
    else 
    {
        /* 
           :ZZZ:   Currently the code does not support cases where multiple RIFFs 
                   are required though following bit of code exists to 
                   handle this case! Following code is untested. 
                   (Function avi_write_ix() may not work correctly.)
        */
        avi_write_ix(s);
        end_tag(pb, avi->movi_list);
        end_tag(pb, avi->riff_start);

        file_size = url_ftell(pb);
        Avf_url_fseek(pb, avi->odml_list - 8, SEEK_SET);
        put_tag(pb, "LIST"); /* Making this AVI OpenDML one */
        url_fskip(pb, 16);

        for (n=nb_frames=0;n<s->nb_streams;n++) {
             AVCodecContext *stream = &s->streams[n]->codec;
             if (stream->codec_type == CODEC_TYPE_VIDEO) {
                 if (nb_frames < s->streams[n]->stream_total_nb_frames)
                     nb_frames = s->streams[n]->stream_total_nb_frames;
             } else {  
                 //Remark: Ignore the case for MP2 to comply with DIVX
                 if (/*stream->codec_id == CODEC_ID_MP2 || */stream->codec_id == CODEC_ID_MP3LAME) {
                     nb_frames += s->streams[n]->stream_total_nb_frames;
                }
            }
        }
        put_le32(pb, nb_frames);
        Avf_url_fseek(pb, file_size, SEEK_SET);
    }

    // :ZZZ: Must flush buffer, since we accumulate frames
    put_flush_packet_end(pb);

    for( i = 0; i < avi->common_index.ents_allocated / AVI_INDEX_CLUSTER_SIZE; ++i )
    {
        avf_free(avi->common_index.cluster[i]);
    }
    avf_free(avi->common_index.cluster);
    avi->common_index.cluster = NULL;
    avi->common_index.ents_allocated = avi->common_index.entry = 0;

    // :ZZZ:
    for(i=0;i<s->nb_streams;i++) 
    {
        AVStream *st = s->streams[i];
        avf_free(st->codec.extradata);
    }
    
    return res;
}

int avi_estimate_trailer_size(AVFormatContext *s)
{
    AVIContext *avi = s->priv_data;
    return avi->common_index.entry * sizeof(LONG) * 4;
}

static AVOutputFormat avi_oformat = {
    "avi",
    "avi format",
    "video/x-msvideo",
    "avi",
    sizeof(AVIContext),
    CODEC_ID_MP2,
    CODEC_ID_MPEG4,
    avi_write_header,
    avi_write_packet,
    avi_write_trailer,
    0,
    0,
    0,
    avi_estimate_trailer_size,
};

int Avf_avienc_init(void)
{
    av_register_output_format(&avi_oformat);
    return 0;
}


VOID Avf_ModifyDefaultTag(INT iFindTag, UINT uiTag1, UINT uiTag2)
{
    // This function rewrite values to 'const' variable.
    CodecTag *psCodecTags =  (CodecTag *)&codec_bmp_tags[0];

    while (psCodecTags->id != iFindTag) psCodecTags++;

    psCodecTags->tag = uiTag1; psCodecTags++;
    psCodecTags->tag = uiTag2;
}


