/*
 * "Direct queue" protocol.
 * Simply puts packets to a queue specified in the filename.
 */

#include "avformat.h"
#include "Os.h"
#include "Allocator.h"
//#include "Dma.h"

#define DIRECT_PROTOCOL_QUEUE_SIZE 3

//API flags
#define AV_DIRECT_QUEUE_FLAG_BLOCK      0x00000001

//Event flags
#define EV_AV_DIRECT_QUEUE_READY           0x00000001
#define EV_AV_DIRECT_QUEUE_CLOSE_PENDING   0x00000002
#define EV_AV_DIRECT_QUEUE_HANDLE_COMPLETE 0x00000004
#define EV_AV_DIRECT_QUEUE_READER_BLOCKED  0x00000008
#define EV_AV_DIRECT_QUEUE_WRITER_BLOCKED  0x00000010

typedef struct tagSQUEUEPACKET
{
    VOID*   pvData;
    UINT    uiSize;
    UINT64  ullSeqNum;
} SQUEUEPACKET;

typedef struct tagSQUEUECONTEXT
{
    TX_QUEUE     *ptxQueue;
    BYTE         *pbyQueueBuffer;
    UINT          uiStreamIdx;
    UINT          uiBlockMode;
    UINT64        ullPktSeqStamp;
    SQUEUEPACKET  hQueuePacket;
    UINT          uiPktBytesRead;
} SQUEUECONTEXT;

static SQUEUECONTEXT *hDirectQueueContext[2] = {NULL, NULL};
static TX_EVENT_FLAGS_GROUP *ptxEvDirect[2] = {NULL, NULL};

URLProtocol* Avf_AvGetDirectProtocol(void)
{
    if( ptxEvDirect[0] == NULL )
    {
        ptxEvDirect[0] = (TX_EVENT_FLAGS_GROUP*)PoolAlloc( NULL,
                                                sizeof(TX_EVENT_FLAGS_GROUP), 
                                                POOL_MEMORY_CACHED );

        OsEventFlagsCreate(ptxEvDirect[0], "AVFDirectProtocol0");
    }

    if( ptxEvDirect[1] == NULL )
    {
        ptxEvDirect[1] = (TX_EVENT_FLAGS_GROUP*)PoolAlloc( NULL,
                                                sizeof(TX_EVENT_FLAGS_GROUP), 
                                                POOL_MEMORY_CACHED );

        OsEventFlagsCreate(ptxEvDirect[1], "AVFDirectProtocol1");
    }

    return &direct_protocol;
}

int Avf_direct_put_data(unsigned char *buf, int size, unsigned int flags)
{
    SQUEUEPACKET sPkt = {0};
    ULONG ulFlag = TX_NO_WAIT;
    ULONG ulActualFlags = 0;

    OsEventFlagsGet(ptxEvDirect[0],
                    EV_AV_DIRECT_QUEUE_READY|EV_AV_DIRECT_QUEUE_CLOSE_PENDING,
                    TX_OR,
                    &ulActualFlags,
                    TX_WAIT_FOREVER);

    if ( hDirectQueueContext[0] == NULL )
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    if(ulActualFlags & EV_AV_DIRECT_QUEUE_CLOSE_PENDING)
    {
        OsEventFlagsSet(ptxEvDirect[0], 
                        EV_AV_DIRECT_QUEUE_HANDLE_COMPLETE, 
                        TX_OR);
        return 0;        
    }

    if( flags & AV_DIRECT_QUEUE_FLAG_BLOCK )
    {
        ulFlag = TX_WAIT_FOREVER;
    }

    sPkt.pvData = PoolAlloc( NULL, size, POOL_MEMORY_CACHED );

    memcpy( sPkt.pvData, buf, size );

    sPkt.uiSize = (UINT)size;
    sPkt.ullSeqNum = hDirectQueueContext[0]->ullPktSeqStamp++;

    OsEventFlagsSet(ptxEvDirect[0], 
                    EV_AV_DIRECT_QUEUE_WRITER_BLOCKED, 
                    TX_OR);
    
    if ( OsQueueSend(hDirectQueueContext[0]->ptxQueue, &sPkt, ulFlag) != TX_SUCCESS )
    {
        PoolFree(NULL, sPkt.pvData);

        OsEventFlagsSet(ptxEvDirect[0], 
                        ~EV_AV_DIRECT_QUEUE_WRITER_BLOCKED, 
                        TX_AND);
        return -1;
    }
    else
    {
        OsEventFlagsSet(ptxEvDirect[0], 
                        ~EV_AV_DIRECT_QUEUE_WRITER_BLOCKED, 
                        TX_AND);
        return 1;
    }
}


int Avf_direct_get_data(unsigned int streamidx, unsigned char **buf, int *size, uint64_t *seqnum, unsigned int flags)
{
    SQUEUEPACKET sPkt = {0};
    ULONG ulFlag = TX_NO_WAIT;
    ULONG ulActualFlags = 0;
    UINT uiRetval = 0;

    OsEventFlagsGet(ptxEvDirect[streamidx],
                    EV_AV_DIRECT_QUEUE_READY,
                    TX_AND,
                    &ulActualFlags,
                    TX_WAIT_FOREVER);

    if ( hDirectQueueContext[streamidx] == NULL )
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    if( flags & AV_DIRECT_QUEUE_FLAG_BLOCK )
    {
        ulFlag = TX_WAIT_FOREVER;
    }

    OsEventFlagsSet(ptxEvDirect[streamidx],
                    EV_AV_DIRECT_QUEUE_READER_BLOCKED, 
                    TX_OR);

    if ( OsQueueReceive(hDirectQueueContext[streamidx]->ptxQueue, &sPkt, ulFlag) != TX_SUCCESS )
    {
        uiRetval = -1;
    }
    else
    {
        *size = sPkt.uiSize;
        *seqnum = sPkt.ullSeqNum;
        *buf = (unsigned char *)sPkt.pvData;

        uiRetval = 1;
    }

    OsEventFlagsSet(ptxEvDirect[streamidx],
                    ~EV_AV_DIRECT_QUEUE_READER_BLOCKED, 
                    TX_AND);

    OsEventFlagsGet(ptxEvDirect[streamidx],
                EV_AV_DIRECT_QUEUE_CLOSE_PENDING,
                TX_AND,
                &ulActualFlags,
                TX_NO_WAIT);

    if(ulActualFlags & EV_AV_DIRECT_QUEUE_CLOSE_PENDING)
    {
        OsEventFlagsSet(ptxEvDirect[streamidx], 
                        EV_AV_DIRECT_QUEUE_HANDLE_COMPLETE, 
                        TX_OR);
        uiRetval = 0;
    }

    return uiRetval;
}


static int direct_open(URLContext *h, const char *filename, int flags)
{
    SQUEUECONTEXT* hQctx = NULL;
    unsigned int uiStreamIdx;
    char cQueueName[20] = {0};

    printf("direct_open\n");

    h->priv_data = NULL;

    strstart(filename, "direct:", &filename);

    if(strstr(filename, "primary"))
    {
        uiStreamIdx = 0;
    }
    else if(strstr(filename, "secondary"))
    {
        uiStreamIdx = 1;
    }
    else
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    hQctx = (SQUEUECONTEXT*)PoolCalloc( NULL,
                                        1,
                                        sizeof(SQUEUECONTEXT), 
                                        POOL_MEMORY_CACHED );

    hQctx->uiStreamIdx = uiStreamIdx;

    hQctx->ptxQueue = (TX_QUEUE_PTR)PoolAlloc( NULL, 
                                               sizeof( TX_QUEUE ), 
                                               POOL_MEMORY_CACHED );

    hQctx->pbyQueueBuffer = (BYTE*)PoolAlloc( NULL,
                                              DIRECT_PROTOCOL_QUEUE_SIZE * sizeof( SQUEUEPACKET ),
                                              POOL_MEMORY_CACHED );

    sprintf(cQueueName, "%s%d", "AvfDirectQueue", uiStreamIdx);

    OsQueueCreate( hQctx->ptxQueue, 
                   cQueueName,
                   sizeof( SQUEUEPACKET ) / sizeof( ULONG ),
                   hQctx->pbyQueueBuffer,
                   DIRECT_PROTOCOL_QUEUE_SIZE * sizeof( SQUEUEPACKET ) );

    hQctx->ullPktSeqStamp = 0;
    hQctx->uiPktBytesRead = 0;

    if (flags & URL_DIRECT_WR_NOBLOCK)
    {
        hQctx->uiBlockMode = 0;
    }
    else
    {
        hQctx->uiBlockMode = 1;
    }

    h->is_streamed = 1;
    memset(&(hQctx->hQueuePacket), 0, sizeof(SQUEUEPACKET));
    h->priv_data = (VOID*)hQctx;
    hDirectQueueContext[uiStreamIdx] = hQctx;

    OsEventFlagsSet(ptxEvDirect[uiStreamIdx], EV_AV_DIRECT_QUEUE_READY, TX_OR);

    return 0;
}


static int direct_read(URLContext *h, unsigned char *buf, int size)
{
    UINT uiBytesToRead = 0;
    SQUEUECONTEXT *hQctx = (SQUEUECONTEXT*)h->priv_data;
    int iLeft = size;

    if ( hQctx == NULL )
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    do{
        if( hQctx->uiPktBytesRead == 0 )
        {
            if( hQctx->hQueuePacket.pvData != NULL )
            {
                PoolFree(NULL, hQctx->hQueuePacket.pvData);
                hQctx->hQueuePacket.pvData = NULL;
            }

            if(OsQueueReceive( hQctx->ptxQueue, &(hQctx->hQueuePacket), TX_WAIT_FOREVER ) != TX_SUCCESS)
            {
                AvfAvioSetLastError( -EIO );
                return 0;
            }
            
        }

        uiBytesToRead = hQctx->hQueuePacket.uiSize - hQctx->uiPktBytesRead;

        if( uiBytesToRead <= (unsigned int)iLeft )
        {
            memcpy(buf, hQctx->hQueuePacket.pvData + hQctx->uiPktBytesRead, uiBytesToRead);
            hQctx->uiPktBytesRead = 0;
            buf += uiBytesToRead;
            iLeft -= uiBytesToRead;
        }
        else
        {
            memcpy(buf, hQctx->hQueuePacket.pvData + hQctx->uiPktBytesRead, iLeft);
            hQctx->uiPktBytesRead += iLeft;
            iLeft = 0;
        }
    }
    while( iLeft > 0);

    return size;
}


static int direct_write(URLContext *h, unsigned char *buf, int size)
{
    SQUEUEPACKET sPkt;
    SQUEUECONTEXT *hQctx = (SQUEUECONTEXT*)h->priv_data;
    ULONG wait_option;

    if ( hQctx == NULL )
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    sPkt.pvData = PoolAlloc( NULL, size, POOL_MEMORY_CACHED );

    // First 3 SP banks are in use by sensor, so we can only use banks 3+4.
    // As per my testing on C12m 1280*720, DmaCopy(..,3,4) gives only 0.6 extra 
    // MIPS comparing to memcpy(). 
    // Independence from SP banks ownership seems preferable to 0.6 extra MIPS.
    memcpy( sPkt.pvData, buf, size );
    //DmaMemCopyEx( sOutPkt.pvData, buf, size, SCRATCHPAD_ID_BANK3, SCRATCHPAD_ID_BANK4 );

    sPkt.uiSize = (UINT)size;
    sPkt.ullSeqNum = hQctx->ullPktSeqStamp++;

    if(hQctx->uiBlockMode == 1)
    {
        wait_option = TX_WAIT_FOREVER;
    }
    else
    {
        wait_option = TX_NO_WAIT;
    }
    
    while ( OsQueueSend(hQctx->ptxQueue, &sPkt, wait_option) == TX_QUEUE_FULL )
    {
        // If it failed to send, it means nobody get it out so far.
        // Then we just pop out one packet and insert a new one.
        SQUEUEPACKET sTmpPkt;

        OsQueueReceive( hQctx->ptxQueue, &sTmpPkt, TX_WAIT_FOREVER );
        PoolFree( NULL, sTmpPkt.pvData );
    }
    return size;
}


static int direct_close(URLContext *h)
{
    SQUEUECONTEXT *hQctx = (SQUEUECONTEXT*)h->priv_data;
    ULONG ulActualFlags = 0;

    printf("direct_close\n");

    if ( hQctx == NULL )
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    if( hQctx->hQueuePacket.pvData != NULL )
    {
        PoolFree(NULL, hQctx->hQueuePacket.pvData);
    }

    OsEventFlagsSet(ptxEvDirect[hQctx->uiStreamIdx], 
                    EV_AV_DIRECT_QUEUE_CLOSE_PENDING, 
                    TX_OR);

    OsEventFlagsGet(ptxEvDirect[hQctx->uiStreamIdx], 
                    EV_AV_DIRECT_QUEUE_READER_BLOCKED|EV_AV_DIRECT_QUEUE_WRITER_BLOCKED,
                    TX_OR,
                    &ulActualFlags,
                    TX_NO_WAIT);

    //Empty packet to unblock reader
    if(ulActualFlags & EV_AV_DIRECT_QUEUE_READER_BLOCKED)
    {
        SQUEUEPACKET sPkt = {0};
        OsQueueSend(hQctx->ptxQueue, &sPkt, TX_NO_WAIT);

        OsEventFlagsGet(ptxEvDirect[hQctx->uiStreamIdx], 
                        EV_AV_DIRECT_QUEUE_HANDLE_COMPLETE,
                        TX_AND,
                        &ulActualFlags,
                        TX_WAIT_FOREVER);
    }

    //Flush
    while( OsQueueReceive( hQctx->ptxQueue, &(hQctx->hQueuePacket), TX_NO_WAIT ) != TX_QUEUE_EMPTY )
    {
        PoolFree(NULL, hQctx->hQueuePacket.pvData);
    }


    OsEventFlagsSet(ptxEvDirect[hQctx->uiStreamIdx], 
                    0, 
                    TX_AND);

    OsQueueDelete(hQctx->ptxQueue);

    PoolFree( NULL, hQctx->pbyQueueBuffer );
    PoolFree( NULL, hQctx->ptxQueue );
    PoolFree( NULL, hQctx );

    h->priv_data = NULL;

    hDirectQueueContext[hQctx->uiStreamIdx] = NULL;

    return 0;
}


uint64_t direct_size(URLContext *h)
{
    SQUEUECONTEXT *hQctx = (SQUEUECONTEXT*)h->priv_data;

    if ( hQctx == NULL )
    {
        AvfAvioSetLastError( -ENOENT );
        return 0;
    }

    return 0x7FFFFFFFFFFF0000ULL;
}


URLProtocol direct_protocol = {
    "direct",
    direct_open,
    direct_read,
    direct_write,
    NULL, // seek
    direct_close,
    direct_size, // filesize
};

