/*
 * Buffered file io for ffmpeg system
 * Copyright (c) 2001 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avformat.h"
//#include <fcntl.h>
//#include <io.h>

#ifdef ZAVF
#include "AvFx.h"
#include "Allocator.h"
#endif

//#define DEBUG

static int giLastAvioError = 0;

int
AvfAvioSetLastError( int iErrorCode )
{
    giLastAvioError = iErrorCode;
    return giLastAvioError;
}

int
AvfAvioGetLastError()
{
    return giLastAvioError;
}


#ifdef ZAVF
// Move AvSetMedia and av_media_set out of garbage section
// Avoid a section type conflict, move AvSetMedia and av_media_set to different section
STATIC HMEDIA hAvMedia SECTION(".primaryboot");	
void Avf_av_media_set(HMEDIA hMedia) SECTION(".primaryboot2");

void Avf_media_set(HMEDIA hMedia)
{
    hAvMedia = hMedia;
}

HMEDIA Avf_media_get()
{
    return hAvMedia;
}

#endif 

#ifdef ZAVF

static unsigned short uwAvfEntry;

void Avf_file_entry_set(unsigned short uwEntry)
{
    uwAvfEntry = uwEntry;
}

unsigned short Avf_file_entry_get()
{
    return uwAvfEntry;
}

#endif 


/* standard file protocol */

static int file_open(URLContext *h, const char *filename, int flags)
{
#ifdef ZAVF
    int access;
    AvFile *hFile = NULL;
    UINT uiStatus;

    // Reset error state on every open
    AvfAvioSetLastError( 0 );
    h->priv_data = NULL;

    hFile = AvFxFileInitHandle();
    if (!hFile) {
        AvfAvioSetLastError( -EIO );
        return -EIO;
    }
    strstart(filename, "file:", &filename);

    if (AvFxFileDetermineOpenAccess(hAvMedia, hFile, h, filename, flags, &access))
        {
        AvFxFileDestroyHandle (hFile);
                AvfAvioSetLastError( -EIO );
                return -EIO;
            }

    if( (flags & URL_RDONLY ) && ( uwAvfEntry != 0 ) ) 
    {
        // come from AvInputOpenFile    
        uiStatus = AvFxFileOpenByEntry( hAvMedia,
                               hFile,
                               (CHAR*)filename,
                               access,
                               uwAvfEntry,
                               NULL );

        uwAvfEntry = 0 ;        
    }
    else
    {
        // come from AvUrlOpen or AvInputOpenFile
        uiStatus = AvFxFileOpen( hAvMedia,
                               hFile,
                               (CHAR*)filename,
                               access );
    }
    
    if ( uiStatus )
    {
        AvFxFileDestroyHandle (hFile);
        AvfAvioSetLastError( -ENOENT );

        return -ENOENT;
    }

    h->priv_data = (void *)hFile;
#endif

    return 0;
}

static int file_read(URLContext *h, unsigned char *buf, int size)
{
#ifdef ZAVF
    ULONG ulSize;

    UINT uiStatus;

    AvFile *hFile = (AvFile *)h->priv_data;
    

    if ( hFile == NULL )
    {
        AvfAvioSetLastError( -EIO );
        return 0;
    }

    // We don't considere "end of file" us failure.
    if ( AvFxFileRead( hFile, (PBYTE)buf, size, &ulSize ))
    {
        AvfAvioSetLastError( -EIO );
        return 0;
    }

    return ulSize;
#else
    return 0;
#endif
}

static int file_write(URLContext *h, unsigned char *buf, int size)
{
#ifdef ZAVF
    AvFile *hFile = (AvFile*)h->priv_data;

    if ( hFile == NULL )
    {
        AvfAvioSetLastError( -EIO );
        return -EIO;
    }

    if ( AvFxFileWrite (hFile,
                      buf,
                       size ))
    {
        AvfAvioSetLastError( -EIO );
        return -EIO;
    }

    return size;
#else
    return 0;
#endif
}

/* XXX: use llseek */
static offset_t file_seek(URLContext *h, offset_t pos, int whence)
{
#ifdef ZAVF
    AvFile* hFile = (AvFile*)h->priv_data;

    if (AvFxFileSeek (hFile, pos, whence)) {
        AvfAvioSetLastError( -EIO );
        return -EIO;
    }

    return AvFxFileCurrentPosition(hFile);

#else
    return 0;
#endif
}

static int file_close(URLContext *h)
{
#ifdef ZAVF
    AvFile* hFile = (AvFile*)h->priv_data;

    if ( hFile == NULL )
    {
        AvfAvioSetLastError( -EIO );
        return -EIO;
    }
    if (AvFxFileClose (hAvMedia, hFile)) {
        AvfAvioSetLastError( -EIO );
    }    

    AvFxFileDestroyHandle ( hFile );
    h->priv_data = NULL;

    return AvfAvioGetLastError();
#else
    return 0;
#endif
}

static uint64_t file_size(URLContext *h)
{
#ifdef ZAVF
    AvFile* hFile = (AvFile*)h->priv_data;

    if ( hFile == NULL )
    {
        AvfAvioSetLastError( -EIO );
        return 0;
    }

    return AvFxFileSize(hFile);
#else
    return 0;
#endif

}


URLProtocol file_protocol = {
    "file",
    file_open,
    file_read,
    file_write,
    file_seek,
    file_close,
    file_size,
};

URLProtocol*
Avf_AvGetFileProtocol(VOID)
{
	return &file_protocol;
}
