/*
 * MOV decoder.
 * Copyright (c) 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "av.h"
#include "mov.h"
#include "avformat.h"
#include "AvApi.h"
#include "Timing.h"

#ifdef CONFIG_ZLIB
#include <zlib.h>
#endif

/*
 * First version by Francois Revol revol@free.fr
 *
 * Features and limitations:
 * - reads most of the QT files I have (at least the structure),
 *   the exceptions are .mov with zlib compressed headers ('cmov' section). It shouldn't be hard to implement.
 *   FIXED, Francois Revol, 07/17/2002
 * - ffmpeg has nearly none of the usual QuickTime codecs,
 *   although I succesfully dumped raw and mp3 audio tracks off .mov files.
 *   Sample QuickTime files with mp3 audio can be found at: http://www.3ivx.com/showcase.html
 * - .mp4 parsing is still hazardous, although the format really is QuickTime with some minor changes
 *   (to make .mov parser crash maybe ?), despite what they say in the MPEG FAQ at
 *   http://mpeg.telecomitalialab.com/faq.htm
 * - the code is quite ugly... maybe I won't do it recursive next time :-)
 *
 * Funny I didn't know about http://sourceforge.net/projects/qt-ffmpeg/
 * when coding this :) (it's a writer anyway)
 *
 * Reference documents:
 * http://www.geocities.com/xhelmboyx/quicktime/formats/qtm-layout.txt
 * Apple:
 *  http://developer.apple.com/techpubs/quicktime/qtdevdocs/QTFF/qtff.html
 *  http://developer.apple.com/techpubs/quicktime/qtdevdocs/PDF/QTFileFormat.pdf
 * QuickTime is a trademark of Apple (AFAIK :))
 */

//#define DEBUG
#undef DEBUG // this is pretty hopeless - fcntl.h is not present and the below uses printf's

#ifdef DEBUG
#include <stdio.h>
#include <fcntl.h>
#endif
 
#ifdef DEBUG
/*
 * XXX: static sux, even more in a multithreaded environment...
 * Avoid them. This is here just to help debugging.
 */
static int debug_indent = 0;
void print_atom(const char *str, MOV_atom_t atom)
{
    unsigned int tag, i;
    tag = (unsigned int) atom.type;
    i=debug_indent;
    if(tag == 0) tag = MKTAG('N', 'U', 'L', 'L');
    while(i--)
        printf("|");
    printf("parse:");
    printf(" %s: tag=%c%c%c%c offset=0x%x size=0x%x\n",
           str, tag & 0xff,
           (tag >> 8) & 0xff,
           (tag >> 16) & 0xff,
           (tag >> 24) & 0xff,
           (unsigned int)atom.offset,
       (unsigned int)atom.size);
    assert((unsigned int)atom.size < 0x7fffffff);// catching errors
}
#else
#define print_atom(a,b) 
#endif

//
static int bDropAudioOnPlayback = 0;

/* the flag indicates whether mov_read_seek is called from mov_read_seek_key 
   TODO: this must be reworked for Mainline as well as calcualtion of audio frame 
         number in mov_calcualte_audio and mov_read_seek functions. 
         Please DO NOT integrate this change to Mainline "as is"! */
static int bSeekKey = 0; 

#define AV_SYNC_LAG_THRESHOLD    1000000

static int mov_read_leaf(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("leaf", atom);

    if (atom.size>1)
        url_fskip(pb, atom.size);
/*        url_seek(pb, atom_offset+atom.size, SEEK_SET); */
    return 0;
}

static int mov_read_default(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    int64_t total_size = 0;
    MOV_atom_t a;
    int i;
    int err = 0;

    a.offset = atom.offset;

    if (atom.size < 0)
        atom.size = 0x7fffffffffffffffLL;

    while(((total_size + 8) < atom.size) && !url_feof(pb) && !err) 
    {
        a.size = atom.size;
        a.type=0L;
    
        if(atom.size >= 8) 
        {
            a.size = get_be32(pb);
            a.type = get_le32(pb);
        }
        total_size += 8;
        a.offset += 8;

        if (a.size == 1) 
        { /* 64 bit extended size */
            a.size = get_be64(pb) - 8;
            a.offset += 8;
            total_size += 8;
        }
        if (a.size == 0) 
        {
            a.size = atom.size - total_size;
            if (a.size <= 8)
                    break;
        }
    
//      printf("__RL__: mov_read_default:  read Atom type = %08X (%c%c%c%c)\n",
//          a.type,
//          (a.type)& 0xff,
//          (a.type >> 8)& 0xff,
//          (a.type >> 16)& 0xff,
//          (a.type >> 24)& 0xff);

     

        for (i = 0; c->parse_table[i].type != 0L && c->parse_table[i].type != a.type; i++)
        /* empty */;

        a.size -= 8;

        if (c->parse_table[i].type == 0) 
        { /* skip leaf atoms data */
            url_fskip(pb, a.size);
        } 
        else 
        {
            err = (c->parse_table[i].func)(c, pb, a);

        }

        a.offset += a.size;
        total_size += a.size;

        // Sometimes the atom is not fully read (i.e. when bLoadIndex=0)
        Avf_url_fseek(pb, a.offset, SEEK_SET);
    }

    if (!err && total_size < atom.size && atom.size < 0x7ffff) 
    {
        url_fskip(pb, atom.size - total_size);
    }


    return err;
}

static int mov_read_ctab(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    unsigned int len;
    MOV_ctab_t *t;
    //url_fskip(pb, atom.size); // for now
    c->ctab = avf_realloc(c->ctab, ++c->ctab_size);
    t = c->ctab[c->ctab_size];
    t->seed = get_be32(pb);
    t->flags = get_be16(pb);
    t->size = get_be16(pb) + 1;
    len = 2 * t->size * 4;
    if (len > 0) {
    t->clrs = avf_malloc(len); // 16bit A R G B
    if (t->clrs)
        get_buffer(pb, t->clrs, len);
    }

    return 0;
}

static int mov_read_hdlr(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    int len = 0;
    uint32_t type;
    uint32_t ctype;

    print_atom("hdlr", atom);

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    /* component type */
    ctype = get_le32(pb);
    type = get_le32(pb); /* component subtype */

#ifdef DEBUG
    printf("ctype= %c%c%c%c (0x%08lx)\n", *((char *)&ctype), ((char *)&ctype)[1], ((char *)&ctype)[2], ((char *)&ctype)[3], (long) ctype);
    printf("stype= %c%c%c%c\n", *((char *)&type), ((char *)&type)[1], ((char *)&type)[2], ((char *)&type)[3]);
#endif
#ifdef DEBUG
/* XXX: yeah this is ugly... */
    if(ctype == MKTAG('m', 'h', 'l', 'r')) { /* MOV */
        if(type == MKTAG('v', 'i', 'd', 'e'))
            puts("hdlr: vide");
        else if(type == MKTAG('s', 'o', 'u', 'n'))
            puts("hdlr: soun");
    } else if(ctype == 0) { /* MP4 */
        if(type == MKTAG('v', 'i', 'd', 'e'))
            puts("hdlr: vide");
        else if(type == MKTAG('s', 'o', 'u', 'n'))
            puts("hdlr: soun");
        else if(type == MKTAG('o', 'd', 's', 'm'))
            puts("hdlr: odsm");
        else if(type == MKTAG('s', 'd', 's', 'm'))
            puts("hdlr: sdsm");
    } else puts("hdlr: meta");
#endif

    if(ctype == MKTAG('m', 'h', 'l', 'r')) { /* MOV */
        /* helps parsing the string hereafter... */
        c->mp4 = 0;
        if(type == MKTAG('v', 'i', 'd', 'e'))
            st->codec.codec_type = CODEC_TYPE_VIDEO;
        else if(type == MKTAG('s', 'o', 'u', 'n'))
            st->codec.codec_type = CODEC_TYPE_AUDIO;
        else if(type == MKTAG('t', 'e', 'x', 't'))
            st->codec.codec_type = CODEC_TYPE_TEXT;
    } else if(ctype == 0) { /* MP4 */
        /* helps parsing the string hereafter... */
        c->mp4 = 1;
        if(type == MKTAG('v', 'i', 'd', 'e'))
            st->codec.codec_type = CODEC_TYPE_VIDEO;
        else if(type == MKTAG('s', 'o', 'u', 'n'))
            st->codec.codec_type = CODEC_TYPE_AUDIO;
        else if(type == MKTAG('t', 'e', 'x', 't'))
            st->codec.codec_type = CODEC_TYPE_TEXT;
    }
    get_be32(pb); /* component  manufacture */
    get_be32(pb); /* component flags */
    get_be32(pb); /* component flags mask */

    if(atom.size <= 24)
        return 0; /* nothing left to read */
    /* XXX: MP4 uses a C string, not a pascal one */
    /* component name */

    if(c->mp4) {
        /* .mp4: C string */
        while(get_byte(pb) && (++len < (atom.size - 24)));
    } else {
        /* .mov: PASCAL string */
#ifdef DEBUG
        char* buf;
#endif
        len = get_byte(pb);
#ifdef DEBUG
    buf = (uint8_t*) avf_malloc(len+1);
    if (buf) {
        get_buffer(pb, buf, len);
        buf[len] = '\0';
        printf("**buf='%s'\n", buf);
        avf_free(buf);
    } else
#endif
        url_fskip(pb, len);
    }

    return 0;
}

static int mov_mp4_read_descr_len(ByteIOContext *pb)
{
    int len = 0;
    int count = 4;
    while (count--) {
        int c = get_byte(pb);
    len = (len << 7) | (c & 0x7f);
    if (!(c & 0x80))
        break;
    }
    return len;
}

static int mov_mp4_read_descr(ByteIOContext *pb, int *tag)
{
    int len;
    *tag = get_byte(pb);
    len = mov_mp4_read_descr_len(pb);
#ifdef DEBUG
    printf("MPEG4 description: tag=0x%02x len=%d\n", *tag, len);
#endif
    return len;
}

static inline unsigned int get_be24(ByteIOContext *s)
{
    unsigned int val;
    val = get_byte(s) << 16;
    val |= get_byte(s) << 8;
    val |= get_byte(s);
    return val;
}

static int mov_read_esds(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    offset_t start_pos = url_ftell(pb);
    int tag, len;

    print_atom("esds", atom);

    /* Well, broken but suffisant for some MP4 streams */
    get_be32(pb); /* version + flags */
    len = mov_mp4_read_descr(pb, &tag);
    if (tag == MP4ESDescrTag) {
    get_be16(pb); /* ID */
    get_byte(pb); /* priority */
    } else
    get_be16(pb); /* ID */

    len = mov_mp4_read_descr(pb, &tag);
    if (tag == MP4DecConfigDescrTag) {
    sc->esds.object_type_id = get_byte(pb);
    sc->esds.stream_type = get_byte(pb);
    sc->esds.buffer_size_db = get_be24(pb);
    sc->esds.max_bitrate = get_be32(pb);
    sc->esds.avg_bitrate = get_be32(pb);

    len = mov_mp4_read_descr(pb, &tag);
    //printf("LEN %d  TAG %d  m:%d a:%d\n", len, tag, sc->esds.max_bitrate, sc->esds.avg_bitrate);
    if (tag == MP4DecSpecificDescrTag) {
#ifdef DEBUG
        printf("Specific MPEG4 header len=%d\n", len);
#endif
        //in case that the atom is met twice
        if (st->codec.extradata)
        {
            avf_freep(&st->codec.extradata);
        }

        st->codec.extradata = (uint8_t*) avf_mallocz(len);
        if (st->codec.extradata) {
            get_buffer(pb, st->codec.extradata, len);
            st->codec.extradata_size = len;
        }
    }
    }
    /* in any case, skip garbage */
    url_fskip(pb, atom.size - ((url_ftell(pb) - start_pos)));
    return 0;
}


/* Get/Put unsigned int from memory buffer: */
static inline void mov_dec_put32(uint8_t *buf, int val)
{
    buf[0] = (val >> 24) & 0xFF;
    buf[1] = (val >> 16) & 0xFF;
    buf[2] = (val >> 8 ) & 0xFF;
    buf[3] =  val & 0xFF;
}

static inline unsigned int mov_dec_get32(uint8_t *buf)
{
    return  ((unsigned int)buf[0] << 24) |
            ((unsigned int)buf[1] << 16) |
            ((unsigned int)buf[2] << 8)  |
             (unsigned int)buf[3];
}



/* Update H.264 NAL units: change from QT start code format to standart one*/
static int mov_read_update_h264_data(uint8_t *buf, int buf_size)
{
    uint8_t *curr_buf = buf;
    int curr_buf_size = buf_size;
    int curr_size = (int) mov_dec_get32(buf);

    if(buf_size < 5){
        return -1;
    }

    // In the case that the buffer size is greater than 130 and smaller that
    // 515 the first four bytes of buf will be 00 00 01 and wrogly recognised
    // as NAL start code. In order to avoid this eventuality a check against
    // the buffer size is executed in the first 4 bytes of buf.
    if( curr_size != buf_size - 4 )
    {
    // check for NAL startcode,return if found,no need to update
    if(  buf[0] == 0 && buf[1] == 0  &&
       ((buf[2] == 0 && buf[3] == 1) ||
         buf[2] == 1 ) ){
        return -1;
    }
    }

    // Put correct NAL Start Code:
    while(curr_buf_size > 0){
        int NaLSize = mov_dec_get32(curr_buf);

        mov_dec_put32(curr_buf, 0x1);

        NaLSize += 4;

        curr_buf_size -= NaLSize;
        curr_buf += NaLSize;
    }

    return curr_buf_size;//return 0 - if Ok; else - if something wrong..
}


static int mov_read_avcc(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int64_t start_pos = url_ftell(pb);
    int len; // GabiD - the code below is disabled, so should this

    uint8_t* write_ptr;
    INT data_size = atom.size;

    print_atom("avcC", atom);

     /* prepare copy of the data */
    if (data_size <= AV_MOV_MAX_OPTIONAL_DESCRIPTION_SIZE) {
        get_buffer(pb, c->fc->sMovStreamsInfo.pbyVosData, data_size);
        c->fc->sMovStreamsInfo.uiVosLen = data_size;
        Avf_url_fseek(pb, start_pos, SEEK_SET);
    }    

    get_byte(pb);     // version = 1;
    c->fc->sMovStreamsInfo.AVCProfileIndication = get_byte(pb);     // profile
    get_byte(pb);     // profile compat
    get_byte(pb);           // level
    get_byte(pb);           // = 0xff
    get_byte(pb);           // = 0xe1

    if( data_size > 0 )
    {
        // Read SPS and PPS H264 headers:

        if (sc->header_data)
            avf_free(sc->header_data);

        sc->header_data = (uint8_t*) avf_mallocz(data_size + 4);

        if (sc->header_data)
        {
            write_ptr = sc->header_data;

            //Read SPS:
            len = get_be16(pb);
            mov_dec_put32(write_ptr, len);
            write_ptr += 4;
            get_buffer(pb, write_ptr, len);

            write_ptr += len;

            sc->header_len = len + 4;


            get_byte(pb); // pps number;


            // Read PPS:
            len = get_be16(pb);
            mov_dec_put32(write_ptr, len);
            write_ptr += 4;
            get_buffer(pb, write_ptr, len);

            sc->header_len += len + 4;

            st->codec.extradata = (uint8_t*) avf_mallocz(sc->header_len);
            memcpy(st->codec.extradata, sc->header_data, sc->header_len);
            st->codec.extradata_size = sc->header_len;

        }
    }

    /* in any case, skip garbage */
    url_fskip(pb, atom.size - ((url_ftell(pb) - start_pos)));

    return 0;
}

/* this atom contains actual media data */
static int mov_read_mdat(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("mdat", atom);

    if(atom.size == 0) /* wrong one (MP4) */
        return 0;
    c->found_mdat=1;
    c->mdat_offset = atom.offset;
    c->mdat_size = atom.size;
    if(c->found_moov)
        return 1; /* found both, just go */
    url_fskip(pb, atom.size);
    return 0; /* now go for moov */
}

/* this atom should contain all header atoms */
static int mov_read_moov(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    // return value is never checked

    print_atom("moov", atom);

    mov_read_default(c, pb, atom);
    /* we parsed the 'moov' atom, we can terminate the parsing as soon as we find the 'mdat' */
    /* so we don't parse the whole file if over a network */
    c->found_moov=1;
    if(c->found_mdat)
        return 1; /* found both, just go */
    return 0; /* now go for mdat */
}

/* this atom should contain the user data */
static int mov_read_udta(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    ULONG ulSize = (long)atom.size;

    print_atom("udta", atom);

    c->fc->pbyUserAtom = avf_malloc(ulSize + 8);
    
//    c->fc->pbyUserAtom = avf_malloc(ulSize + 4);

//    ((PLONG)c->fc->pbyUserAtom)[0] = SWAP_WORDS(ulSize);
    ((PLONG)c->fc->pbyUserAtom)[0] = SWAP_WORDS( (ulSize + 8) );
    ((PCHAR)c->fc->pbyUserAtom)[4] = 'u';
    ((PCHAR)c->fc->pbyUserAtom)[5] = 'd';
    ((PCHAR)c->fc->pbyUserAtom)[6] = 't';
    ((PCHAR)c->fc->pbyUserAtom)[7] = 'a';

    get_buffer(pb, c->fc->pbyUserAtom + 8, ulSize);

//    get_buffer(pb, c->fc->pbyUserAtom + 4, ulSize);

    return 0; 
}

static int mov_read_skip(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    ULONG ulSize = (long)atom.size;

    print_atom("skip", atom);

    c->fc->pbySkipAtom = avf_malloc(ulSize + 4);

//    ((PLONG)c->fc->pbySkipAtom)[0] = SWAP_WORDS(ulSize);
    ((PLONG)c->fc->pbySkipAtom)[0] = SWAP_WORDS( (ulSize + 4) );

    get_buffer(pb, c->fc->pbySkipAtom + 4, ulSize);

    return 0; 
}

static int mov_read_mdhd(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("mdhd", atom);

    get_byte(pb); /* version */

    get_byte(pb); get_byte(pb);
    get_byte(pb); /* flags */

    get_be32(pb); /* creation time */
    get_be32(pb); /* modification time */

    c->streams[c->fc->nb_streams-1]->time_scale = get_be32(pb);

#ifdef DEBUG
    printf("track[%i].time_scale = %i\n", c->fc->nb_streams-1, c->streams[c->fc->nb_streams-1]->time_scale); /* time scale */
#endif
    get_be32(pb); /* duration */

    get_be16(pb); /* language */
    get_be16(pb); /* quality */

    return 0;
}

static int mov_read_mvhd(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("mvhd", atom);

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    c->fc->ulCreTimeStamp = get_be32(pb); /* creation time */
    c->fc->ulModTimeStamp = get_be32(pb); /* modification time */

    c->time_scale = get_be32(pb); /* time scale */
#ifdef DEBUG
    printf("time scale = %i\n", c->time_scale);
#endif
    c->duration = get_be32(pb); /* duration */
    get_be32(pb); /* preferred scale */

    get_be16(pb); /* preferred volume */

    url_fskip(pb, 10); /* reserved */

    url_fskip(pb, 36); /* display matrix */

    get_be32(pb); /* preview time */
    get_be32(pb); /* preview duration */
    get_be32(pb); /* poster time */
    get_be32(pb); /* selection time */
    get_be32(pb); /* selection duration */
    get_be32(pb); /* current time */
    get_be32(pb); /* next track ID */

    return 0;
}

static int mov_read_smi(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];

    // currently SVQ3 decoder expect full STSD header - so let's fake it
    // this should be fixed and just SMI header should be passed
    avf_freep(&st->codec.extradata);
    st->codec.extradata_size = 0x5a + atom.size;
    if (st->codec.extradata_size !=0)
    {
        st->codec.extradata = (uint8_t*) avf_mallocz(st->codec.extradata_size);
    }

    if (st->codec.extradata) {
        strcpy(st->codec.extradata, "SVQ3"); // fake
        get_buffer(pb, (unsigned char *)st->codec.extradata + 0x5a, atom.size);
        //printf("Reading SMI %Ld  %s\n", atom.size, (char*)st->codec.extradata + 0x5a);
    } else
    url_fskip(pb, atom.size);

    return 0;
}

static int mov_read_stco(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int entries, i;

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    entries = get_be32(pb);
    sc->chunk_count = entries;
    sc->chunk_offsets = (offset_t*) avf_malloc(entries * sizeof(offset_t));
    //sc->chunk_offsets = (uint32_t*) avf_malloc(entries * sizeof(uint32_t));
    if (!sc->chunk_offsets)
        return -1;
    if (atom.type == MKTAG('s', 't', 'c', 'o')) {
        print_atom("stco", atom);
        for(i=0; i<entries; i++) {
            
            // Although the index is not needed, for ApsAvGetFirstFrame 
            // purposes, we do need to load two first chunk offsets,
            // so that seek(0) and first read_packet will work
            // (assumption: first video frame is I-VOP) :TODO:
            if ( (c->eLoadIndex == MOV_HEADER_INDEX_FIRST_ENTRY_ONLY) && (i > 1) )
            {
                break;
            }

            sc->chunk_offsets[i] = get_be32(pb);
        }
    } else if (atom.type == MKTAG('c', 'o', '6', '4')) {
        print_atom("co64", atom);
        for(i=0; i<entries; i++) {
/* undead code */
            // Although the index is not needed, for ApsAvGetFirstFrame 
            // purposes, we do need to load two first chunk offsets,
            // so that seek(0) and first read_packet will work
            // (assumption: first video frame is I-VOP) :TODO:
            if ( (c->eLoadIndex == MOV_HEADER_INDEX_FIRST_ENTRY_ONLY) && (i > 1) )
            {
                break;
            }

            sc->chunk_offsets[i] = get_be64(pb);
/* 0 */
        }
    } else
        return -1;
#ifdef DEBUG
/*
    for(i=0; i<entries; i++) {
        printf("chunk offset=0x%Lx\n", sc->chunk_offsets[i]);
    }
*/
#endif
    return 0;
}

static int mov_read_stsd(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    //MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int entries, sample_descr_num;
    AVFormatContext *fc = c->fc;
    uint32_t format;
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    
    print_atom("stsd", atom);
    
    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */
    
    entries = get_be32(pb);
    
    sample_descr_num = entries;

    while(entries--) { //Parsing Sample description table
        enum CodecID id;
        offset_t description_start = url_ftell(pb);
        
        int size = get_be32(pb); /* size */
        format = get_le32(pb); /* data format */
        
        get_be32(pb); /* reserved */
        get_be16(pb); /* reserved */
        get_be16(pb); /* index */
        
        size -= (4+4+4+2+2);
        /* for MPEG4: set codec type by looking for it */
        id = codec_get_id(mov_video_tags, format);
        if (id > 0) {
            AVCodec *codec;
            codec = AvApi_avcodec_find_decoder(id);
            
            if (codec)
            {
                st->codec.codec_type = codec->type;
            }
            else // mjpeg support
            {
                st->codec.codec_type = CODEC_TYPE_VIDEO;
            }
        }
        
#ifdef DEBUG
        printf("size=%d 4CC= %c%c%c%c codec_type=%d\n",
            size,
            (format >> 0) & 0xff,
            (format >> 8) & 0xff,
            (format >> 16) & 0xff,
            (format >> 24) & 0xff,
            st->codec.codec_type);
#endif

        st->codec.codec_tag = format;
        if( st->codec.codec_type==CODEC_TYPE_VIDEO ) {
            MOV_atom_t a = { 0, 0, 0 };
            st->codec.codec_id = id;
            
            fc->sMovStreamsInfo.cVideoFormatType[0] = format & 0xff;
            fc->sMovStreamsInfo.cVideoFormatType[1] = (format >> 8) & 0xff;
            fc->sMovStreamsInfo.cVideoFormatType[2] = (format >> 16) & 0xff;
            fc->sMovStreamsInfo.cVideoFormatType[3] = (format >> 24) & 0xff;

            fc->uwModel      = get_be16(pb); /* version */
            fc->uwRevision   = get_be16(pb); /* revision level */
            fc->ulVendorName = get_be32(pb); /* vendor */
            
            get_be32(pb); /* temporal quality */
            get_be32(pb); /* spacial quality */
            st->codec.logical_width = get_be16(pb); /* width */
            st->codec.logical_height = get_be16(pb); /* height */

            st->codec.physical_width = ( st->codec.logical_width + 15 ) & (~0xF);
            st->codec.physical_height = ( st->codec.logical_height + 15 ) & (~0xF);
#if 1
            if (st->codec.codec_id == CODEC_ID_MPEG4) {
            /* in some MPEG4 the width/height are not correct, so
                we ignore this info */
                st->codec.logical_width = 0;
                st->codec.logical_height = 0;

                st->codec.physical_width = 0;
                st->codec.physical_height = 0;
            }
#endif
            fc->sMovStreamsInfo.lVisual_horizresolution = get_be32(pb); /* horiz resolution */
            fc->sMovStreamsInfo.lVisual_vertresolution = get_be32(pb); /* vert resolution */
            get_be32(pb); /* data size, always 0 */
#ifdef DEBUG
            frames_per_sample =
#endif /* DEBUG */
            get_be16(pb); /* frames per samples */
#ifdef DEBUG
            printf("frames/samples = %d\n", frames_per_sample);
#endif
            get_buffer(pb, (uint8_t *)st->codec.codec_name, 32); /* codec name */
            
            st->codec.bits_per_sample = get_be16(pb); /* depth */
            st->codec.color_table_id = get_be16(pb); /* colortable id */
            
            st->codec.frame_rate      = 25;
            st->codec.frame_rate_base = 1;
            
            size -= (2+2+4 +4+4 +2+2 +4+4+4+2 +32+2+2);
#if 0
            while (size >= 8) {
                MOV_atom_t a;
                int64_t start_pos;
                
                a.size = get_be32(pb);
                a.type = get_le32(pb);
                size -= 8;
#ifdef DEBUG
                printf("VIDEO: atom_type=%c%c%c%c atom.size=%Ld size_left=%d\n",
                    (a.type >> 0) & 0xff,
                    (a.type >> 8) & 0xff,
                    (a.type >> 16) & 0xff,
                    (a.type >> 24) & 0xff,
                    a.size, size);
#endif
                start_pos = url_ftell(pb);
                
                switch(a.type) {
                case MKTAG('e', 's', 'd', 's'):
                    {
                        int tag, len;
                        /* Well, broken but suffisant for some MP4 streams */
                        get_be32(pb); /* version + flags */
                        len = mov_mp4_read_descr(pb, &tag);
                        if (tag == 0x03) {
                            /* MP4ESDescrTag */
                            get_be16(pb); /* ID */
                            get_byte(pb); /* priority */
                            len = mov_mp4_read_descr(pb, &tag);
                            if (tag != 0x04)
                                goto fail;
                            /* MP4DecConfigDescrTag */
                            get_byte(pb); /* objectTypeId */
                            get_be32(pb); /* streamType + buffer size */
                            get_be32(pb); /* max bit rate */
                            get_be32(pb); /* avg bit rate */
                            len = mov_mp4_read_descr(pb, &tag);
                            if (tag != 0x05)
                                goto fail;
                            /* MP4DecSpecificDescrTag */
#ifdef DEBUG
                            printf("Specific MPEG4 header len=%d\n", len);
#endif
                            sc->header_data = avf_mallocz(len);
                            if (sc->header_data) {
                                get_buffer(pb, sc->header_data, len);
                                sc->header_len = len;
                            }
                        }
                        /* in any case, skip garbage */
                    }
                    break;
                default:
                    break;
                }
fail:
                printf("ATOMENEWSIZE %Ld   %d\n", atom.size, url_ftell(pb) - start_pos);
                if (atom.size > 8) {
                    url_fskip(pb, (atom.size - 8) -
                        ((url_ftell(pb) - start_pos)));
                    size -= atom.size - 8;
                }
            }
            if (size > 0) {
                /* unknown extension */
                url_fskip(pb, size);
            }
#else
            a.size = size;
            a.offset = url_ftell(pb);
            mov_read_default(c, pb, a);
            st->codec.bit_rate = sc->esds.max_bitrate;  
#endif
        } 
        // :ZZZ: Support other then audio or video streams
        else if ( st->codec.codec_type == CODEC_TYPE_AUDIO ) 
        {
            MOV_atom_t a = { 0, 0, 0 };

            fc->sMovStreamsInfo.cAudioFormatType[0] = format & 0xff;
            fc->sMovStreamsInfo.cAudioFormatType[1] = (format >> 8) & 0xff;
            fc->sMovStreamsInfo.cAudioFormatType[2] = (format >> 16) & 0xff;
            fc->sMovStreamsInfo.cAudioFormatType[3] = (format >> 24) & 0xff;

            fc->sMovStreamsInfo.bAudioExist = TRUE;

            st->codec.codec_id = codec_get_id(mov_audio_tags, format);
            if(st->codec.codec_id==CODEC_ID_AMR_NB) //from TS26.244
            {
#ifdef DEBUG
                printf("AMR-NB audio identified!!\n");
#endif
                get_be32(pb);get_be32(pb); //Reserved_8
                get_be16(pb);//Reserved_2
                get_be16(pb);//Reserved_2
                get_be32(pb);//Reserved_4
                get_be16(pb);//TimeScale
                get_be16(pb);//Reserved_2
            
                size -= (4+4+2+2+4+2+2);
                //AMRSpecificBox.(10 bytes)
                // :ZZZ: Add missing parenthesis for compilation with
                // DEBUG defined
                {                
#ifdef DEBUG
                    int damr_size=
#endif
                    get_be32(pb); //size
#ifdef DEBUG
                    int damr_type=
#endif
                    get_be32(pb); //type=='damr'
#ifdef DEBUG
                    int damr_vendor=
#endif
                    get_be32(pb); //vendor
                    get_byte(pb); //decoder version
                    get_be16(pb); //mode_set
                    get_byte(pb); //mode_change_period
                    get_byte(pb); //frames_per_sample
                    size -= (4+1+2+1+1);
                
#ifdef DEBUG
                    printf("Audio: damr_type=%c%c%c%c damr_size=%Ld damr_vendor=%c%c%c%c\n",
                        (damr_type >> 24) & 0xff,
                        (damr_type >> 16) & 0xff,
                        (damr_type >> 8) & 0xff,
                        (damr_type >> 0) & 0xff,
                        damr_size, 
                        (damr_vendor >> 24) & 0xff,
                        (damr_vendor >> 16) & 0xff,
                        (damr_vendor >> 8) & 0xff,
                        (damr_vendor >> 0) & 0xff
                        );
#endif
                    // :ZZZ: Add missing parenthesis for compilation with
                    // DEBUG defined
                }
            
                //st->time_length=0;//Not possible to get from this info, must count number of AMR frames
                st->codec.sample_rate=8000;
                st->codec.channels=1;
                st->codec.bits_per_sample=16;
                st->codec.bit_rate=0; /*It is not possible to tell this before we have 
                an audio frame and even then every frame can be different*/
                fc->sMovStreamsInfo.ulSTSDAudioSampleRate = st->codec.sample_rate;  
                fc->sMovStreamsInfo.ulSTSDAudioChannels = st->codec.channels;
            }
            else
            {
                unsigned int version = 0;
            
                version = get_be16(pb); /* version */
                get_be16(pb); /* revision level */
                get_be32(pb); /* vendor */
            
                st->codec.channels = get_be16(pb);      /* channel count */
                st->codec.bits_per_sample = get_be16(pb);   /* sample size */
            
                /* handle specific s8 codec */
                get_be16(pb); /* compression id = 0*/
                get_be16(pb); /* packet size = 0 */
            
                st->codec.sample_rate = ((get_be32(pb) >> 16));
                printf("CODECID %d  %d  %.4s\n", st->codec.codec_id, CODEC_ID_PCM_S16BE, (char*)&format);
            
                fc->sMovStreamsInfo.ulSTSDAudioSampleRate = st->codec.sample_rate;  
                fc->sMovStreamsInfo.ulSTSDAudioChannels = st->codec.channels;
                
                size -= (2+2+4+2+2+2+2+4);
                st->codec.bit_rate = st->codec.sample_rate * st->codec.bits_per_sample;
                switch (st->codec.codec_id) {
                case CODEC_ID_PCM_S16BE:
                    if (st->codec.bits_per_sample == 8)
                    {
                        st->codec.codec_id = CODEC_ID_PCM_S8;
                    }
                    break;
                case CODEC_ID_PCM_U8:
                    if (st->codec.bits_per_sample == 16)
                    {
                        st->codec.codec_id = CODEC_ID_PCM_S16BE;
                    }
                    break;
                
                case CODEC_ID_AACZ:
                case CODEC_ID_AAC:
                    st->codec.codec_id = CODEC_ID_AACZ;
                    // Do not hard code bit rate
                    // st->codec.bit_rate = 800000;
                    break;

                case CODEC_ID_PCM_MULAW:
                    st->codec.bit_rate = st->codec.sample_rate * 8;
                    break;

                default:
                    ;
                }
            
                if (version == 1)
                {
                    int spp = get_be32(pb); /* samples per packet */
                    int bpp = get_be32(pb); /* bytes per packet (for one channel)*/
                
                    st->codec.frame_size  = spp; 
                    st->codec.block_align = bpp*st->codec.channels;
                
                    // frame_size can be set to zero by encoder
                    if (spp > 0)
                    {
                        st->codec.bit_rate    = (int64_t)bpp * st->codec.sample_rate * st->codec.channels * 8 / spp;
                    }

                    // skip 2 long from header;
                    get_be32(pb);
                    get_be32(pb);
                    size -= (4+4+4+4);
                }    

                if (st->codec.codec_id == CODEC_ID_AACZ)
                {
                    int i;

                    a.size = get_be32(pb) - 8;
                    a.type = get_le32(pb);
                    a.offset = url_ftell(pb);

                    // Parse here to prevent jump back
                    for (i = 0; c->parse_table[i].type != 0L && c->parse_table[i].type != a.type; i++);
                    if (c->parse_table[i].type == 0) 
                    { /* skip leaf atoms data */
                        url_fskip(pb, a.size);
                    } 
                    else 
                    {
                        (c->parse_table[i].func)(c, pb, a);
                    }

                    if (st->codec.extradata && (st->codec.extradata_size == 2)){
                        // parse stream header to get actual sample rate:
                        static int sample_rates[] = {
                                96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000,
                                12000, 11025, 8000 };

                        st->codec.sample_rate = sample_rates[((((uint8_t*)st->codec.extradata)[0] << 1)|(((uint8_t*)st->codec.extradata)[1] >> 7)) & 0xf];

                        fc->sMovStreamsInfo.AudioDSIChannelConfigurations = (((uint8_t*)st->codec.extradata)[1] >> 3) & 0xf;

                        fc->sMovStreamsInfo.AudioObjectTypeIndication = sc->esds.object_type_id;
                    }

                    st->codec.bit_rate = sc->esds.avg_bitrate;
                    // TODO: Remove hardcoding to 1024
                    st->codec.frame_size  = 1024; //sc->esds.buffer_size_db;
                }    
            
                Avf_url_fseek(pb, description_start + size, SEEK_SET);
            }

            fc->sMovStreamsInfo.ulAudioBitRate = st->codec.bit_rate;
        }
        // :ZZZ: Support for other then audio or video streams
        else
        {
            MOV_atom_t a = { format, url_ftell(pb), size - (16) };
        
            if ( format != MKTAG('r', 't', 'p', ' ') )
                mov_read_default(c, pb, a); 
            else
                mov_read_leaf(c, pb, a);
        }
    }
    
    if(st->codec.codec_type == CODEC_TYPE_VIDEO)
    {
        fc->sMovStreamsInfo.VideoSampleDescriptionNum = sample_descr_num;
    }
    else if(st->codec.codec_type == CODEC_TYPE_AUDIO)
    {
        fc->sMovStreamsInfo.AudioSampleDescriptionNum = sample_descr_num;
    }

    return 0;
}

static int mov_read_stsc(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int entries, i;

    print_atom("stsc", atom);

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    entries = get_be32(pb);
#ifdef DEBUG
printf("track[%i].stsc.entries = %i\n", c->fc->nb_streams-1, entries);
#endif

    if (sc->sample_to_chunk_sz)
        avf_free(sc->sample_to_chunk);        

    sc->sample_to_chunk_sz = entries;
    sc->sample_to_chunk = (MOV_sample_to_chunk_tbl*) avf_malloc(entries * sizeof(MOV_sample_to_chunk_tbl));
    if (!sc->sample_to_chunk)
    {
        return -1;
    }
    for(i=0; i<entries; i++) {
        sc->sample_to_chunk[i].first = get_be32(pb);
        sc->sample_to_chunk[i].count = get_be32(pb);
        sc->sample_to_chunk[i].id = get_be32(pb);
#ifdef DEBUG
/*        printf("sample_to_chunk first=%ld count=%ld, id=%ld\n", sc->sample_to_chunk[i].first, sc->sample_to_chunk[i].count, sc->sample_to_chunk[i].id); */
#endif
    }
    return 0;
}

static int mov_read_stsz(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int entries, i;

    print_atom("stsz", atom);

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    sc->sample_size = get_be32(pb);
    entries = get_be32(pb);
    sc->sample_count = entries;

    if (st->codec.codec_type==CODEC_TYPE_VIDEO){
        c->fc->ulTotalVideoFrames = entries;
    }

    if ( c->eLoadIndex == MOV_HEADER_NO_INDEX )
    {
        return 0;
    }

#ifdef DEBUG
    printf("sample_size = %ld sample_count = %ld\n", sc->sample_size, sc->sample_count);
#endif
    if(sc->sample_size)
        return 0; /* there isn't any table following */
    sc->sample_sizes = (long*) avf_malloc(entries * sizeof(long));
    if (!sc->sample_sizes)
        return -1;
    for(i=0; i<entries; i++) {
        sc->sample_sizes[i] = get_be32(pb);
#ifdef DEBUG
/*        printf("sample_sizes[]=%ld\n", sc->sample_sizes[i]); */
#endif
        if( c->eLoadIndex == MOV_HEADER_INDEX_FIRST_ENTRY_ONLY )
        {
            break;
        }
    }
    return 0;
}

static int mov_convert_stream_duration( MOVStreamContext *sc,
                                        int entries,
                                        int total,
                                        long* counts )
{
    long* durations = sc->sample_duration;

    int i, j=0, count = 0, dur = 0;

    // allocate tables
    sc->sample_duration = (long*) avf_malloc(total * sizeof(long));

    for( i = 0; i < total; i++ )
    {
        if( count == 0 )
        {
            dur = durations[j];
            count = counts[j];
            j++;
        }

        sc->sample_duration[i] = dur;
        count--;
    }

    avf_free( durations );

    return 0;
}

static int  mov_convert_text_stream_duration( MOVStreamContext *sc,
                                              int entries,
                                              int video_duration )
{
    int i;
    int prev_dur, dur;

    // First sample is always associated with first video frame
    prev_dur = sc->sample_duration[0];
    sc->sample_duration[0] = 0;

    for( i = 1; i < entries; i++ )
    {
        dur = sc->sample_duration[i];

        sc->sample_duration[i] = sc->sample_duration[i-1] +
                                 prev_dur / video_duration;

        prev_dur = dur;

    }

    return 0;
}


static int mov_read_stts(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int entries, i;
    int duration=0;
    int total_sample_count=0;
    int video_duration = 0;
    long *sample_duration_count;


    print_atom("stts", atom);

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */
    entries = get_be32(pb);

    if ( c->eLoadIndex == MOV_HEADER_NO_INDEX && entries > 1 )
    {
        return 0;
    }

    if (sc->sample_duration_num)
        avf_free(sc->sample_duration);

    // allocate tables
    sc->sample_duration_num = entries;
    sc->sample_duration = (long*) avf_malloc(entries * sizeof(long));
    sample_duration_count = (long*) avf_malloc(entries * sizeof(long));

    // find sample duration of video frame
    // the timesacle is the same for text and video
    if( st->codec.codec_type == CODEC_TYPE_TEXT )
    {
        for( i = 0; i < c->fc->nb_streams ; i++ )
        {
            if( c->fc->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO )
            {
                video_duration = c->streams[i]->sample_duration[0];                
                break;
            }
        }
    }

    // read the table
    for( i = 0; i < entries; i++ )
    {
        int sample_duration;
        int sample_count;

        // get table entry
        sample_count    = get_be32(pb);
        sample_duration = get_be32(pb);

        // fill index table
        sample_duration_count[i] = sample_count;        

        sc->sample_duration[i]       = sample_duration;

        // update duration
        duration += ( sample_duration * sample_count );
        total_sample_count += sample_count;
    }

    if( st->codec.codec_type == CODEC_TYPE_TEXT )
    {
        // Ungrouping of sample durations
        mov_convert_stream_duration( sc,
                                     entries,
                                     total_sample_count,
                                     sample_duration_count );

        mov_convert_text_stream_duration( sc, total_sample_count, video_duration );
        sc->sample_duration_num = total_sample_count;
    }

    else if(st->codec.codec_type==CODEC_TYPE_VIDEO)
    {
        //Only support constant frame rate. But lets calculate the average. Needed by .mp4-files created with nec e606 3g phone.
        //Hmm, lets set base to 1000, in order to have at least 3 decimal digits precision on the frame rate.
        int64_t temp = 0;


        if(st->codec.codec_id == CODEC_ID_MJPEG)
        {
            st->codec.frame_rate_base = 1;
        }
        else
        {
        if ( ( entries == 1 ) &&
             ( sc->sample_duration[0] != 1000 ) )
        {
            st->codec.frame_rate_base = sc->sample_duration[0]; 
        }
        else
        {
            st->codec.frame_rate_base = 1000; 
        } 
        } 

        temp = st->codec.frame_rate_base;
        temp *= c->streams[c->total_streams]->time_scale;
        temp *= total_sample_count;
        temp /= duration;

        st->codec.frame_rate = (int)(temp);

#ifdef DEBUG
        printf( "VIDEO FRAME RATE average= %2.3f (tot sample count= %i ,tot dur= %i)\n", 
                (float)st->codec.frame_rate/(float)st->codec.frame_rate_base, total_sample_count, duration );
#endif
    }

    avf_free( sample_duration_count );
    return 0;
}

static int mov_read_trak(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st;
    MOVStreamContext *sc;
    int64_t duration = 0;
    int res = 0;
    print_atom("trak", atom);

    st = Avf_new_stream(c->fc, c->fc->nb_streams);
    if (!st) return -2;
    sc = (MOVStreamContext*) avf_mallocz(sizeof(MOVStreamContext));
    if (!sc) {
    avf_free(st);
        return -1;
    }

    sc->sample_to_chunk_index = -1;
    st->priv_data = sc;
    st->codec.codec_type = CODEC_TYPE_MOV_OTHER;
    duration = ((int64_t)(c->duration)) * 1000;
    st->time_length = duration / ((int64_t)(c->time_scale)); // time in miliseconds
    st->header_flags = 0;
    c->streams[c->fc->nb_streams-1] = sc;

    res = mov_read_default(c, pb, atom);

    return res;
}

static int mov_read_tkhd(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st;
    int64_t duration = 0;

    print_atom("tkhd", atom);

    st = c->fc->streams[c->fc->nb_streams-1];

    get_byte(pb); /* version */

    /* flags */
    st->header_flags = (int)get_be24(pb);
    /*
    MOV_TRACK_ENABLED 0x0001
    MOV_TRACK_IN_MOVIE 0x0002
    MOV_TRACK_IN_PREVIEW 0x0004
    MOV_TRACK_IN_POSTER 0x0008
    */

    get_be32(pb); /* creation time */
    get_be32(pb); /* modification time */
    st->id = (int)get_be32(pb); /* track id (NOT 0 !)*/
    get_be32(pb); /* reserved */
    duration = (int64_t)get_be32(pb);
    st->time_length = (duration * 1000) / ((int64_t)(c->time_scale)); /* duration in milliseconds*/

    get_be32(pb); /* reserved */
    get_be32(pb); /* reserved */

    get_be16(pb); /* layer */
    get_be16(pb); /* alternate group */
    get_be16(pb); /* volume */
    get_be16(pb); /* reserved */

    url_fskip(pb, 36); /* display matrix */

    /* those are fixed-point */
    st->codec.logical_width = get_be32(pb) >> 16; /* track width */
    st->codec.logical_height = get_be32(pb) >> 16; /* track height */

    st->codec.physical_width = ( st->codec.logical_width + 15 ) & (~0xF);
    st->codec.physical_height = ( st->codec.logical_height + 15 ) & (~0xF);

    return 0;
}

static int mov_read_stss(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int entries_num, i;
    int gop_size = 0;

    if ( c->eLoadIndex == MOV_HEADER_NO_INDEX)
    {
        return 0;
    }

    get_byte(pb); /* version */

    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */
                                              
    entries_num = sc->sync_samples_count = get_be32(pb); 
    
    sc->sync_samples = (long*) avf_mallocz(entries_num * sizeof(long));
    
    if (!sc->sync_samples)
        return -1;

    for(i=0; i < entries_num; i++)
    {
        sc->sync_samples[i] = get_be32(pb);
        
        if ( gop_size == 0 && i > 1 )
        {
            gop_size = sc->sync_samples[i] - sc->sync_samples[i - 1];
        }
    }
        
    /* recovering in case of lack of data in sync table */
    if ( (gop_size != 0) && 
         (sc->sync_samples[entries_num - 1] + gop_size < sc->sample_count) )
    {
        int expected_entries_num = (sc->sample_count / gop_size) + 1;
           
        printf( "error in stss atom. recovering... %d -> %d\n",
                entries_num,
                expected_entries_num );
        
        sc->sync_samples_count = expected_entries_num;
        sc->sync_samples = (long*)avf_realloc(sc->sync_samples, 
                                                  expected_entries_num * sizeof(long));
        
        for(i=entries_num; i<expected_entries_num; i++)
        {
            /* it's assumed that clip has constant GOP size. If it's not true 
                    then decoding error will occur */
           sc->sync_samples[i] = sc->sync_samples[i-1] + gop_size;
        }
    }

    // save the gop size to the AvFormatContext for future reference
    c->fc->uwGopSize = gop_size;

    return 0;
}

/* this atom should be null (from specs), but some buggy files put the 'moov' atom inside it... */
/* like the files created with Adobe Premiere 5.0, for samples see */
/* http://graphics.tudelft.nl/~wouter/publications/soundtests/ */
static int mov_read_wide(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    int err;

#ifdef DEBUG
    print_atom("wide", atom);
    debug_indent++;
#endif
    if (atom.size < 8)
        return 0; /* continue */
    if (get_be32(pb) != 0) { /* 0 sized mdat atom... use the 'wide' atom size */
        url_fskip(pb, atom.size - 4);
        return 0;
    }
    atom.type = get_le32(pb);
    atom.offset += 8;
    atom.size -= 8;
    if (atom.type != MKTAG('m', 'd', 'a', 't')) {
        url_fskip(pb, atom.size);
        return 0;
    }
    err = mov_read_mdat(c, pb, atom);
#ifdef DEBUG
    debug_indent--;
#endif
    return err;
}


#ifdef CONFIG_ZLIB
static int null_read_packet(void *opaque, uint8_t *buf, int buf_size)
{
    return -1;
}

static int mov_read_cmov(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    ByteIOContext ctx;
    uint8_t *cmov_data;
    uint8_t *moov_data; /* uncompressed data */
    long cmov_len, moov_len;
    int ret;

    print_atom("cmov", atom);

    get_be32(pb); /* dcom atom */
    if (get_le32(pb) != MKTAG( 'd', 'c', 'o', 'm' ))
        return -1;
    if (get_le32(pb) != MKTAG( 'z', 'l', 'i', 'b' )) {
        dprintf("unknown compression for cmov atom !");
        return -1;
    }
    get_be32(pb); /* cmvd atom */
    if (get_le32(pb) != MKTAG( 'c', 'm', 'v', 'd' ))
        return -1;
    moov_len = get_be32(pb); /* uncompressed size */
    cmov_len = atom.size - 6 * 4;

    cmov_data = (uint8_t *) avf_malloc(cmov_len);
    if (!cmov_data)
        return -1;
    moov_data = (uint8_t *) avf_malloc(moov_len);
    if (!moov_data) {
        avf_free(cmov_data);
        return -1;
    }
    get_buffer(pb, cmov_data, cmov_len);
    if(uncompress (moov_data, (uLongf *) &moov_len, (const Bytef *)cmov_data, cmov_len) != Z_OK)
        return -1;
    if(init_put_byte(&ctx, moov_data, moov_len, 0, NULL, null_read_packet, NULL, NULL) != 0)
        return -1;
    ctx.buf_end = ctx.buffer + moov_len;
    atom.type = MKTAG( 'm', 'o', 'o', 'v' );
    atom.offset = 0;
    atom.size = moov_len;
#ifdef DEBUG
    { int fd = open("/tmp/uncompheader.mov", O_WRONLY | O_CREAT); write(fd, moov_data, moov_len); close(fd); }
#endif
    ret = mov_read_default(c, &ctx, atom);
    avf_free(moov_data);
    avf_free(cmov_data);

    
    return ret;
}
#endif

static void mov_read_uuid_prof( MOVDecodeContext *c, 
                                ByteIOContext *pb, 
                                AVFormatContext *fc, 
                                AVStream *st, 
                                int size )
{
    unsigned long Tag;
    int EntrySize = 0;

    // uuid prof tag found
    while(size >= 8)
    {
        EntrySize = get_be32(pb);
        Tag = get_be32(pb); // entry type;

        if(size < EntrySize)
        {
            // wrong size
            return;
        }

        size -= EntrySize;
        EntrySize -= 8;

        switch(Tag)
        {
            case 0x41505246:
            {
                // APRF:

                if(EntrySize < 36)
                {
                    // wrong size
                    break;
                }

                fc->sMovStreamsInfo.bAprfExist = TRUE;

                fc->sMovStreamsInfo.ulAudioFlags = get_be32(pb) & 0xffffff;

                url_fskip(pb, 4);

                get_buffer(pb, (PBYTE)fc->sMovStreamsInfo.cAudioCodecType, 4);

                fc->sMovStreamsInfo.uiDataFormType = get_byte(pb);

                fc->sMovStreamsInfo.uiAudioObjectType = get_byte(pb); // skip
                fc->sMovStreamsInfo.uiAudioObjectType = get_byte(pb) & 0x1f;

                url_fskip(pb, 13);

                fc->sMovStreamsInfo.ulAudioSampleRate = get_be32(pb);
                fc->sMovStreamsInfo.uiAudioChannelNumber = get_be32(pb);

                // Sum of skiped data
                EntrySize -= 36;

                break;
            }

            case 0x56505246:
            {
                // VPRF:

                if(EntrySize < 44){
                    // wrong size
                    break;
                }

                fc->sMovStreamsInfo.bVprfExist = TRUE;

                fc->sMovStreamsInfo.ulVisualFlags = get_be32(pb) & 0xffffff;

                url_fskip(pb, 4); // TID;

                get_buffer(pb, (PBYTE)fc->sMovStreamsInfo.cVideoCodecType, 4);

                url_fskip(pb, 4+4); // codec specific info & Video attr.

                c->video_bitrate = get_be32(pb); // average video bitrate

                url_fskip(pb, 4+4+4); // max bitrate; average fps, max fps

                fc->sMovStreamsInfo.uiVisualWidth = get_be16(pb);
                fc->sMovStreamsInfo.uiVisualHeight = get_be16(pb);

                url_fskip(pb, 4); // Aspect ratio;

                // Sum of skiped data
                EntrySize -= 44;

                break;
            }

            case 0x4d544454:
            {
                // MTDT:

                if((EntrySize < (4 + 16)) || (st == NULL))
                {
                     // wrong size or st == NULL
                     break;
                }

                url_fskip(pb, 4);
                EntrySize -= 4;

                if(st->codec.codec_type == CODEC_TYPE_VIDEO)
                {
                    fc->sMovStreamsInfo.bMtdtVideoExist = TRUE;

                    fc->sMovStreamsInfo.ulMtdtVideoDataTypeID = get_be32(pb);

                    url_fskip(pb, 4);

                    fc->sMovStreamsInfo.ulMetaVPresentationType = get_be32(pb);
                    fc->sMovStreamsInfo.ulMetaVData = get_be32(pb);
                }
                else if(st->codec.codec_type == CODEC_TYPE_AUDIO)
                {
                    fc->sMovStreamsInfo.bMtdtAudioExist = TRUE;

                    fc->sMovStreamsInfo.ulMtdtAudioDataTypeID = get_be32(pb);

                    url_fskip(pb, 4);

                    fc->sMovStreamsInfo.ulMetaAPresentationType = get_be32(pb);

                    url_fskip(pb, 4); //Meta Data audio
                }
                else
                {
                    url_fskip(pb, 8 + 8);
                }

                EntrySize -= 16;

                break;
            }
        }

        if(EntrySize > 0)
        {
            // skip unknown tag
            url_fskip(pb, EntrySize);
            EntrySize = 0;
        }
    }
}

static int mov_read_uuid(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    AVFormatContext *fc = c->fc;
    unsigned long Tag;
    int size = atom.size;
    offset_t AtomStartPosition;

    AtomStartPosition = url_ftell( pb );

    print_atom("uuid", atom);

    if( fc == NULL )
    {
        url_fskip(pb, atom.size);
        return 0;
    }

    Tag = get_be32(pb);

    switch(Tag) 
    {
        case 0x50524f46:
        {
            Tag = get_be32(pb);
            if (Tag != 0x21D24FCE)
            {
                break;
            }

            Tag = get_be32(pb);
            if (Tag != 0xBB88695C)
            {
                break;
            }

            Tag = get_be32(pb);
            if (Tag != 0xFAC9C740)
            {
                break;
            }

            size -= 16;

            Tag = get_be32(pb);
            if (Tag != 0x0) 
            {
                break;
            }

            Tag = get_be32(pb);
            if (Tag == 0) 
            {
                break;
            }
            size -= 8;
            mov_read_uuid_prof( c, pb, fc, st, size );

            break;
        }

        case 0x55534D54:
        {
            Tag = get_be32(pb);
            if (Tag != 0x21D24FCE)
            {
                break;
            }

            Tag = get_be32(pb);
            if (Tag != 0xBB88695C)
            {
                break;
            }

            Tag = get_be32(pb);
            if (Tag != 0xFAC9C740)
            {
                break;
            }

            size -= 16;
            mov_read_uuid_prof( c, pb, fc, st, size );
            break;
        }

        // undefined uuid - we will skip past it
        default:
        {
            break;
        }
    }

    url_fskip(pb, atom.size - ((url_ftell(pb) - AtomStartPosition)));
    return 0; 
}


static int mov_read_ftyp(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("ftyp", atom);

    url_fskip(pb, 8); //

    if( atom.size - 8 < 12)
    {
        get_buffer(pb, (PBYTE)c->fc->sMovStreamsInfo.cFtypCompatibleBrands, atom.size - 8);//for test
    }
    else
    {
        get_buffer(pb, (PBYTE)c->fc->sMovStreamsInfo.cFtypCompatibleBrands, 12);//for test
    }

    return 0;
}

static int mov_read_edts(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("ftyp", atom);

    /* the flag is needed for Modify mode */
    c->fc->ulSyntaxElementsFlag |= MPEG4_SYNTAX_ELEMENT_EDTS;

    return mov_read_default(c, pb, atom);
}

static int mov_read_dinf(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("dinf", atom);

    /* the flag is needed for Modify mode */
    c->fc->ulSyntaxElementsFlag |= MPEG4_SYNTAX_ELEMENT_DINF;

    return mov_read_default(c, pb, atom);
}

/* Composition Shift Least GreatestAtom */
static int mov_read_cslg(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    print_atom("cslg", atom);

    /* the flag is needed for Modify mode */
    c->fc->ulSyntaxElementsFlag |= MPEG4_SYNTAX_ELEMENT_CSLG;

    return mov_read_default(c, pb, atom);
}

static int mov_read_ctts(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    AVStream *st = c->fc->streams[c->fc->nb_streams-1];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    int i;

    if ( c->eLoadIndex == MOV_HEADER_NO_INDEX )
    {
        return 0;
    }

    get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    sc->ctts_entry_num = get_be32(pb); 

    sc->ctts_table = (MOV_ctts_tbl *) avf_malloc(sc->ctts_entry_num * sizeof(MOV_ctts_tbl));

    if (!sc->ctts_entry_num)
        return -1;

    for(i=0; i < sc->ctts_entry_num; i++)
    {
        sc->ctts_table[i].sample_count = get_be32(pb);
        sc->ctts_table[i].sample_offset = get_be32(pb);
    }

    return 0;
}

/* Sample Dependency Atom */
static int mov_read_sdtp(MOVDecodeContext *c, ByteIOContext *pb, MOV_atom_t atom)
{
    int i;

    print_atom("sdtp", atom);

	get_byte(pb); /* version */
    get_byte(pb); get_byte(pb); get_byte(pb); /* flags */

    /* the flag is needed for Modify mode */
    c->fc->ulSyntaxElementsFlag |= MPEG4_SYNTAX_ELEMENT_SDTP;

    c->dependency = (unsigned char*) avf_malloc(c->ctts_entries * sizeof(unsigned char));
    for (i=0; i<c->ctts_entries; i++) {
		c->dependency[i] = get_byte(pb);
    }

	return 0;
}



static const MOVParseTableEntry mov_default_parse_table[] = {
/* mp4 atoms */
{ MKTAG( 's', 'k', 'i', 'p' ), mov_read_skip },
{ MKTAG( 'c', 'o', '6', '4' ), mov_read_stco },
{ MKTAG( 'c', 'p', 'r', 't' ), mov_read_default },
{ MKTAG( 'c', 'r', 'h', 'd' ), mov_read_default },
{ MKTAG( 'c', 's', 'l', 'g' ), mov_read_cslg },
{ MKTAG( 'c', 't', 't', 's' ), mov_read_ctts }, /* composition time to sample */
{ MKTAG( 's', 'd', 't', 'p' ), mov_read_sdtp },
{ MKTAG( 'd', 'i', 'n', 'f' ), mov_read_dinf }, /* data information */
{ MKTAG( 'd', 'p', 'n', 'd' ), mov_read_leaf },
{ MKTAG( 'd', 'r', 'e', 'f' ), mov_read_leaf },
{ MKTAG( 'e', 'd', 't', 's' ), mov_read_edts },
{ MKTAG( 'e', 'l', 's', 't' ), mov_read_leaf },
{ MKTAG( 'f', 'r', 'e', 'e' ), mov_read_leaf },
{ MKTAG( 'h', 'd', 'l', 'r' ), mov_read_hdlr }, 
{ MKTAG( 'h', 'i', 'n', 't' ), mov_read_leaf },
{ MKTAG( 'h', 'm', 'h', 'd' ), mov_read_leaf },
{ MKTAG( 'i', 'o', 'd', 's' ), mov_read_leaf },
{ MKTAG( 'm', 'd', 'a', 't' ), mov_read_mdat },
{ MKTAG( 'm', 'd', 'h', 'd' ), mov_read_mdhd },
{ MKTAG( 'm', 'd', 'i', 'a' ), mov_read_default },
{ MKTAG( 'm', 'i', 'n', 'f' ), mov_read_default },
{ MKTAG( 'm', 'o', 'o', 'v' ), mov_read_moov },
{ MKTAG( 'u', 'd', 't', 'a' ), mov_read_udta },
{ MKTAG( 'm', 'p', '4', 'a' ), mov_read_default }, 
{ MKTAG( 'm', 'p', '4', 's' ), mov_read_default },
{ MKTAG( 'm', 'p', '4', 'v' ), mov_read_default },
{ MKTAG( 'm', 'p', 'o', 'd' ), mov_read_leaf },
{ MKTAG( 'm', 'v', 'h', 'd' ), mov_read_mvhd },
{ MKTAG( 'n', 'm', 'h', 'd' ), mov_read_leaf },
{ MKTAG( 'o', 'd', 'h', 'd' ), mov_read_default },
{ MKTAG( 's', 'd', 'h', 'd' ), mov_read_default },
{ MKTAG( 's', 'k', 'i', 'p' ), mov_read_default },
{ MKTAG( 's', 'm', 'h', 'd' ), mov_read_leaf }, /* sound media info header */
{ MKTAG( 'S', 'M', 'I', ' ' ), mov_read_smi }, /* Sorrenson extension ??? */ 
{ MKTAG( 's', 't', 'b', 'l' ), mov_read_default },
{ MKTAG( 's', 't', 'c', 'o' ), mov_read_stco },
{ MKTAG( 's', 't', 'd', 'p' ), mov_read_default },
{ MKTAG( 's', 't', 's', 'c' ), mov_read_stsc },
{ MKTAG( 's', 't', 's', 'd' ), mov_read_stsd }, /* sample description */
{ MKTAG( 's', 't', 's', 'h' ), mov_read_default },
{ MKTAG( 's', 't', 's', 's' ), mov_read_stss }, /* sync sample */
{ MKTAG( 's', 't', 's', 'z' ), mov_read_stsz }, /* sample size */
{ MKTAG( 's', 't', 't', 's' ), mov_read_stts },
{ MKTAG( 't', 'k', 'h', 'd' ), mov_read_tkhd }, /* track header */               
{ MKTAG( 't', 'r', 'a', 'k' ), mov_read_trak },
{ MKTAG( 't', 'r', 'e', 'f' ), mov_read_default }, /* not really */
{ MKTAG( 'u', 'd', 't', 'a' ), mov_read_leaf },
{ MKTAG( 'u', 'r', 'l', ' ' ), mov_read_leaf },
{ MKTAG( 'u', 'r', 'n', ' ' ), mov_read_leaf },
{ MKTAG( 'u', 'u', 'i', 'd' ), mov_read_uuid }, //mov_read_default },
{ MKTAG( 'v', 'm', 'h', 'd' ), mov_read_leaf }, /* video media info header */
{ MKTAG( 'w', 'a', 'v', 'e' ), mov_read_default },
/* extra mp4 */
{ MKTAG( 'M', 'D', 'E', 'S' ), mov_read_leaf },
/* QT atoms */
{ MKTAG( 'c', 'h', 'a', 'p' ), mov_read_leaf },
{ MKTAG( 'c', 'l', 'i', 'p' ), mov_read_default },
{ MKTAG( 'c', 'r', 'g', 'n' ), mov_read_leaf },
{ MKTAG( 'c', 't', 'a', 'b' ), mov_read_ctab },
{ MKTAG( 'e', 's', 'd', 's' ), mov_read_esds },
{ MKTAG( 'k', 'm', 'a', 't' ), mov_read_leaf },
{ MKTAG( 'm', 'a', 't', 't' ), mov_read_default },
{ MKTAG( 'r', 'd', 'r', 'f' ), mov_read_leaf },
{ MKTAG( 'r', 'm', 'd', 'a' ), mov_read_default },
{ MKTAG( 'r', 'm', 'd', 'r' ), mov_read_leaf },
{ MKTAG( 'r', 'm', 'r', 'a' ), mov_read_default },
{ MKTAG( 's', 'c', 'p', 't' ), mov_read_leaf },
{ MKTAG( 's', 's', 'r', 'c' ), mov_read_leaf },
{ MKTAG( 's', 'y', 'n', 'c' ), mov_read_leaf },
{ MKTAG( 't', 'c', 'm', 'd' ), mov_read_leaf },
{ MKTAG( 'a', 'v', 'c', 'C' ), mov_read_avcc },
{ MKTAG( 'f', 't', 'y', 'p' ), mov_read_ftyp },
{ MKTAG( 'w', 'i', 'd', 'e' ), mov_read_wide }, /* place holder */
//{ MKTAG( 'r', 'm', 'q', 'u' ), mov_read_leaf },
#ifdef CONFIG_ZLIB
{ MKTAG( 'c', 'm', 'o', 'v' ), mov_read_cmov },
#else
{ MKTAG( 'c', 'm', 'o', 'v' ), mov_read_leaf },
#endif
{ 0L, mov_read_leaf }
};

static void mov_free_stream_context(MOVStreamContext *sc)
{
    if(sc) {
        avf_free(sc->chunk_offsets);
        avf_free(sc->sample_to_chunk);
        avf_free(sc->sample_sizes);
        avf_free(sc->header_data);
        avf_free(sc->sync_samples);
        avf_free(sc->sample_duration);
        avf_free(sc->ctts_table);
        avf_free(sc);
    }
}

static inline uint32_t mov_to_tag(uint8_t *buf)
{
    return MKTAG(buf[0], buf[1], buf[2], buf[3]);
}

static inline uint32_t to_be32(uint8_t *buf)
{
    return (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
}

/* XXX: is it sufficient ? */
static int mov_probe(AVProbeData *p)
{
    //int is enough since we will never get beyond the beginning of mdat
    unsigned int offset;
    uint32_t tag;
    int size = 0;

    /* check file header */
    if (p->buf_size <= 12)
        return 0;
    offset = 0;
    for(;;) {
        /* ignore invalid offset */
        if ((offset + 8) > (unsigned int)p->buf_size)
            return 0;
        tag = mov_to_tag(p->buf + offset + 4);
        printf("tag %X\n", tag);
        switch(tag) {
        case MKTAG( 'm', 'o', 'o', 'v' ):
        case MKTAG( 'w', 'i', 'd', 'e' ):
        case MKTAG( 'f', 'r', 'e', 'e' ):
        case MKTAG( 'm', 'd', 'a', 't' ):
        case MKTAG( 'p', 'n', 'o', 't' ): /* detect movs with preview pics like ew.mov and april.mov */
            return AVPROBE_SCORE_MAX;

        case MKTAG( 's', 'k', 'i', 'p' ):
        {
            size = to_be32(p->buf + offset);

            //:TODO: add ftyp major brands fourcc parsing here for more confidence
            if(size > 12)
                return AVPROBE_SCORE_MAX/2;
            else if (size == 0)
                return 0;

            offset += size;
        }
        break;

        case MKTAG( 'u', 'u', 'i', 'd' ):
        {
            size = to_be32(p->buf + offset);
            if (offset + 24 < (unsigned int)p->buf_size)
            {
				/* Thumbnail uuid 4AB03B0F-618D-4075-82B2-D9FACED35FF5 */
                if ( (to_be32(p->buf + offset + 8)  == 0x4AB03B0F) &&
                     (to_be32(p->buf + offset + 12) == 0x618D4075) &&
                     (to_be32(p->buf + offset + 16) == 0x82B2D9FA) &&
                     (to_be32(p->buf + offset + 20) == 0xCED35FF5) )
                {
                    return AVPROBE_SCORE_MAX;
                }
            }
            offset += size;
            break;
        }

        case MKTAG( 'f', 't', 'y', 'p' ):
        default:
        {
            // :ZZZ: fix ffmpeg bug
            size = to_be32(p->buf + offset);

            // Atom size cannot be "0"
            if (size == 0)
                return 0;

            offset += size;                  

            break;
        }
        }
    }

}


int mov_read_drop_audio(void)
{
    // will audio be dropped during playback?
    return bDropAudioOnPlayback;
}

static int mov_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    MOVDecodeContext *mov = (MOVDecodeContext *) s->priv_data;
    ByteIOContext *pb = &s->pb;
    int i, j, err;
    MOV_atom_t atom = { 0, 0, 0 };

    if ( (ap != NULL) && (ap->flags & AV_READ_HEADER_FLAG_NO_INDEX) )
    {
        mov->eLoadIndex = MOV_HEADER_NO_INDEX;
    }
    else if ( (ap != NULL) && (ap->flags & AV_READ_HEADER_FLAG_INDEX_FIRST_ENTRY_ONLY) )
    {
        mov->eLoadIndex = MOV_HEADER_INDEX_FIRST_ENTRY_ONLY;
    }
    else
    {
        mov->eLoadIndex = MOV_HEADER_FULL_INDEX;
    }

    mov->fc = s;
    mov->parse_table = mov_default_parse_table;

    if(!url_is_streamed(pb)) /* .mov and .mp4 aren't streamable anyway (only progressive download if moov is before mdat) */
        atom.size = url_filesize(url_fileno(pb));
    else
        atom.size = 0x7FFFFFFFFFFFFFFFLL;

    memset(&(s->sMovStreamsInfo), 0, sizeof(MOV_Streams_Info));

    /* set uwGopSize to default (1 for iFrame clip) before mov_read_stss() setting the uwGopSize to correct value based on video content */
    s->uwGopSize = 1;

    /* check MOV header */
    err = mov_read_default(mov, pb, atom);

    if (err<0 || !mov->found_moov || !mov->found_mdat)
    {
/*
        // no stdio allowed!
        fprintf(stderr, "mov: header not found !!! (err:%d, moov:%d, mdat:%d) pos:%Ld\n",err,
                mov->found_moov, mov->found_mdat, url_ftell(pb));
*/
        return -1;
    }

    /* some cleanup : make sure we are on the mdat atom */
    if(!url_is_streamed(pb) && (url_ftell(pb) != mov->mdat_offset))
        Avf_url_fseek(pb, mov->mdat_offset, SEEK_SET);

    mov->next_chunk_offset = mov->mdat_offset; /* initialise reading */

    mov->total_streams = s->nb_streams;

    for(i=0; i<s->nb_streams;) 
    {
        if(s->streams[i]->codec.codec_type == CODEC_TYPE_AUDIO)
        {
            s->sMovStreamsInfo.uiAudioTrackFlags = s->streams[i]->header_flags;
        }
        else if(s->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO)
        {
            s->sMovStreamsInfo.uiVideoTrackFlags = s->streams[i]->header_flags;
        }
        
        if((s->streams[i]->codec.codec_type == CODEC_TYPE_MOV_OTHER) ||
            ( mov_read_drop_audio() && (s->streams[i]->codec.codec_type == CODEC_TYPE_AUDIO)))
        {/* not audio, or audio should be dropped, not video, delete */
            if ( s->streams[i]->codec.extradata != NULL )
            {
                avf_freep( &s->streams[i]->codec.extradata );
            }
            avf_free(s->streams[i]);

            for(j=i+1; j<s->nb_streams; j++)
                s->streams[j-1] = s->streams[j];
            
            s->nb_streams--;
        } else
            i++;
    }

    for(i=0; i<s->nb_streams;i++) {
        MOVStreamContext *sc;
        sc = (MOVStreamContext *)s->streams[i]->priv_data;
        sc->ffindex = i;
        sc->is_ff_stream = 1;

        // mjpeg support
        s->streams[i]->stream_total_nb_frames = sc->sample_count;

        if((s->streams[i]->codec.bit_rate == 0) && (s->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO))
        {
            s->streams[i]->codec.bit_rate = mov->video_bitrate * 1000;
        }
    }
    
    return 0;
}

int   mov_is_key_frame(MOVStreamContext *sc, int iFrameNumber)
{
    int iFound = 0;
    int iBeginPoint = 0;
    int iEndPoint = sc->sync_samples_count - 1;
    int iMiddlePoint = (iBeginPoint + iEndPoint) / 2;

    if (sc->sync_samples_count <= 0) 
    {
        // if there is no sync table , every frame should be key frame
        return 1;
    }

    //If every frame is Key frame, we can find the position directly 
    if (iFrameNumber <= sc->sync_samples_count -1)
    {
        if (sc->sync_samples[iFrameNumber] -1 == iFrameNumber)
        {
            return 1;
        }
    }
    
    //binary search in key frame table
    if ((sc->sync_samples[iBeginPoint] -1 == iFrameNumber) ||
        (sc->sync_samples[iEndPoint] -1 == iFrameNumber))
    {
        return 1;
    }
    
    while ((iFound == 0) && (iMiddlePoint != iBeginPoint))
    {        
        if (sc->sync_samples[iMiddlePoint] -1 == iFrameNumber)
        {
            iFound = 1;
            break;
        }
        else if(sc->sync_samples[iMiddlePoint] -1 > iFrameNumber)
        {
            iEndPoint = iMiddlePoint;
        }
        else
        {
            iBeginPoint = iMiddlePoint; 
        }        
        
        iMiddlePoint = (iBeginPoint + iEndPoint) / 2;
    }

    return iFound;
}

static int mov_read_is_keyframe(AVFormatContext *s, int stream_id, int frame)
{
    int is_key = 0;
    
    if (s != NULL) {
        MOVDecodeContext* mov = (MOVDecodeContext*) s->priv_data;
        
        if (mov != NULL) {
            MOVStreamContext* msc = mov->streams[stream_id];
            
            is_key = mov_is_key_frame(msc, frame);
        }
    }

    return is_key;
}

/* Yes, this is ugly... I didn't write the specs of QT :p */
/* XXX:remove useless commented code sometime */
static int mov_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    MOVDecodeContext *mov = (MOVDecodeContext *) s->priv_data;
    MOVStreamContext *sc;
    offset_t offset = 0x0FFFFFFFFFFFFFFFLL;
    int i;
    int size = 0x0FFFFFFF;
    int cur_stream_num = 0;

    if( AvfAvioGetLastError() < 0 )
    {
        return -1;
    }

    //if more samples are present in the current chunk
    if (mov->partial) {

        int idx;

        sc = mov->partial;
  
        cur_stream_num = mov->partial_stream_num;

        idx = sc->sample_to_chunk_index;

        if (idx < 0) return 0;
        
        if(sc->sample_sizes != NULL)
        {
            size = sc->sample_sizes[sc->current_sample];
        }
        else
        {
            size = sc->sample_size;
        }

        sc->current_sample++;
        sc->left_in_chunk--;

        if (sc->left_in_chunk <= 0)
            mov->partial = 0;

        offset = mov->next_chunk_offset;
        /* extract the sample */

        goto readchunk;
    }

again:
    sc = 0;

    //find the stream - the one with the smallest next chunk offset
    for(i=0; i<mov->total_streams; i++) {
        
        MOVStreamContext *msc = mov->streams[i];

        if ((msc->next_chunk < msc->chunk_count) && msc->next_chunk >= 0
            && (msc->chunk_offsets[msc->next_chunk] < offset)) 
        {
                sc = msc;
                offset = msc->chunk_offsets[msc->next_chunk]; 
                cur_stream_num = i;         
        }
    }      

    if (!sc || offset==0x0FFFFFFFFFFFFFFFLL)
    {
        return -1;
    }

    sc->next_chunk++;
    if(mov->next_chunk_offset < offset) 
    {   
        printf("skipping meta\n");
        /* some meta data */
        url_fskip(&s->pb, (offset - mov->next_chunk_offset));
        mov->next_chunk_offset = offset;
    }

    if(!sc->is_ff_stream) 
    {
        printf("skipping non-ffmpeg stream\n");
        url_fskip(&s->pb, (offset - mov->next_chunk_offset));
        mov->next_chunk_offset = offset;
        offset = 0x0FFFFFFFFFFFFFFFLL;
        goto again;
    }

    /* split chunks into samples */
    if( (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_ADPCM_MS) ||
        (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_ADPCM_IMA_WAV) )
    {
        size = mov->fc->streams[cur_stream_num]->codec.block_align;
    }
    else if( (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_MULAW) ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_ALAW)  ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_S16LE) ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_S16BE) ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_U16LE) ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_U16BE) ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_S8) ||
             (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_U8) )
    {
        // Find current idx from - mdia\minf\stbl\stsc
        int idx = sc->sample_to_chunk_index + 1;
        if ((idx < sc->sample_to_chunk_sz)
            && (sc->next_chunk >= sc->sample_to_chunk[idx].first))
        {
            sc->sample_to_chunk_index = idx;
        }

        // Current size = SampleSize * SamplesPerChunk
        size = sc->sample_size * sc->sample_to_chunk[sc->sample_to_chunk_index].count;
    }
    else
    {
        int idx = sc->sample_to_chunk_index;

        //move to the next entry in sample_to_chunk, if needed
        if ((idx + 1 < sc->sample_to_chunk_sz)
            && (sc->next_chunk >= sc->sample_to_chunk[idx + 1].first))
        {
           idx++;
        }

        sc->sample_to_chunk_index = idx;
        
        if (idx >= 0 && sc->sample_to_chunk[idx].count != 1) 
        {
            mov->partial = sc;
            mov->partial_stream_num = cur_stream_num;

            /* we'll have to get those samples before next chunk */
            sc->left_in_chunk = sc->sample_to_chunk[idx].count - 1;
        }

        if(sc->sample_sizes != NULL)
        {
            size = sc->sample_sizes[sc->current_sample];
        }
        else
        {
            size = sc->sample_size;
        }
    }

    sc->current_sample++;

readchunk:

    if(size == 0x0FFFFFFF)
    {
        size = mov->mdat_size + mov->mdat_offset - offset;
    }

    if(size < 0)    
        return -1;

// mjpeg allows empty frames
//    if(size == 0)
//        return -1;
    
    if ( mov->fc->streams[cur_stream_num]->codec.codec_type == CODEC_TYPE_AUDIO )
    {
        TimingMarkF(TIMING_GROUP_BOOT, "%d %llx", __LINE__, offset );
    }

    Avf_url_fseek(&s->pb, offset, SEEK_SET);

    if (sc->header_len > 0) 
    {
        if ( (s->flags & AVFMT_MODIFY_APPEND) && (mov->fc->streams[cur_stream_num]->codec.codec_type == CODEC_TYPE_VIDEO) )
        {
            av_new_packet(pkt, size);
            get_buffer(&s->pb, pkt->data, pkt->size);
        }
        else
        {
            av_new_packet(pkt, size + sc->header_len);
            memcpy(pkt->data, sc->header_data, sc->header_len);
            get_buffer(&s->pb, pkt->data + sc->header_len, size);
        }
        
        /* free header */
        avf_freep(&sc->header_data);
        sc->header_len = 0;
    } 
    else 
    {
        av_new_packet(pkt, size);
        get_buffer(&s->pb, pkt->data, pkt->size);
    }             

    pkt->stream_index = sc->ffindex;
    
    mov->next_chunk_offset = offset + size;

    if (mov->fc->streams[cur_stream_num]->codec.codec_type == CODEC_TYPE_VIDEO)
    {
        if (mov_is_key_frame(sc, sc->current_sample - 1)  == 1)
        {            
            pkt->flags = PKT_FLAG_KEY;
        }

        if (mov->fc->streams[cur_stream_num]->codec.frame_rate == 0)
        {
           printf("frame rate == 0 for stream number %d\n", cur_stream_num);
           return -1;
        }

        //return timestamp in micro-secs
        pkt->pts = ((int64_t)(sc->current_sample - 1)* 1000000) * 
                    mov->fc->streams[cur_stream_num]->codec.frame_rate_base / 
                    mov->fc->streams[cur_stream_num]->codec.frame_rate; 
    }
    else if( mov->fc->streams[cur_stream_num]->codec.codec_type == CODEC_TYPE_AUDIO )
    {    
        if( mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_AACZ ||
            mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_AAC )
        {
            // calculate pts with minimal error due to truncation
            pkt->pts = (sc->current_sample - 1) * sc->sample_duration[0] *
                       (int64_t)1000000 / mov->streams[cur_stream_num]->time_scale;
        }
        else if( (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_ADPCM_MS) ||
                 (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_ADPCM_IMA_WAV) )
        {
            pkt->pts = (sc->current_sample - 1) * mov->fc->streams[cur_stream_num]->codec.block_align * 
                       (int64_t)1000000 / mov->streams[cur_stream_num]->time_scale;
        }
        else if( (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_MULAW) ||
                 (mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_PCM_ALAW) )
        {
            pkt->pts = (sc->current_sample - 1) * sc->sample_to_chunk[0].count *
                       (int64_t)1000000 / mov->streams[cur_stream_num]->time_scale;
        }
        
        pkt->flags = PKT_FLAG_KEY;
    }

    if( AvfAvioGetLastError() < 0 )
    {
        pkt->destruct( pkt );                                    
        return -1;
    }

    if(mov->fc->streams[cur_stream_num]->codec.codec_id == CODEC_ID_H264Z)
    {
        mov_read_update_h264_data(pkt->data, pkt->size);
    }

    return 0;
}

static int mov_read_close(AVFormatContext *s)
{
    int i;
    MOVDecodeContext *mov = (MOVDecodeContext *) s->priv_data;

    for(i=0; i<mov->total_streams; i++)
    {
        mov_free_stream_context(mov->streams[i]);
    }

    for(i=0; i<s->nb_streams; i++)
    {
        avf_freep(&s->streams[i]->codec.extradata);
        avf_freep(&s->streams[i]);
    }
    /* free color tabs */
    for(i=0; i<mov->ctab_size; i++)
    avf_freep(&mov->ctab[i]);
    avf_freep(&mov->ctab);
    avf_freep(&s->pbyUserAtom);
    avf_freep(&s->pbySkipAtom);
    avf_freep(&mov->dependency);
    return 0;
}

int mov_calcualte_audio(MOVDecodeContext *mov,
                        MOVStreamContext *sc_aud,
                        int video_chunk_num, 
                        int audio_stream_num,
                        int64_t pts)
{
    int chunk_table_index = 0;
    int chunks_summed = 0;
    int64_t audio_pts; 

    sc_aud->current_sample = 0;
    
    while ( chunks_summed < video_chunk_num - 1 )
    {
        int chunks_within_index;
       
        if ( chunk_table_index+1 >= sc_aud->sample_to_chunk_sz ||
             video_chunk_num <= sc_aud->sample_to_chunk[chunk_table_index+1].first )
        {
            // on last table entry, add only as many samples as needed
            chunks_within_index = ( video_chunk_num - 1 - chunks_summed );

            sc_aud->current_sample += sc_aud->sample_to_chunk[chunk_table_index].count *
                                      chunks_within_index;
            chunks_summed += chunks_within_index ;
        }
        else
        {   
            chunks_within_index = ( sc_aud->sample_to_chunk[chunk_table_index+1].first -
                                    sc_aud->sample_to_chunk[chunk_table_index].first );

            sc_aud->current_sample += sc_aud->sample_to_chunk[chunk_table_index].count *
                                      chunks_within_index;

            chunks_summed += chunks_within_index;

            audio_pts = sc_aud->current_sample * sc_aud->sample_duration[0] * 
                       (int64_t)1000000 / mov->streams[audio_stream_num]->time_scale;

            if ( ABS(audio_pts - pts) <= 
                    sc_aud->current_sample * sc_aud->sample_duration[0] * chunks_within_index )
            {
                break;
            }

            chunk_table_index++;
        }    
    }

    sc_aud->next_chunk = chunks_summed;
    sc_aud->sample_to_chunk_index = chunk_table_index;
    sc_aud->left_in_chunk = 0; 

    return 1;
}

static int mov_read_seek(AVFormatContext *s, int64_t pts) //pts is in microseconds
{
    MOVDecodeContext *mov = (MOVDecodeContext *) s->priv_data;
    MOVStreamContext *sc = 0, *sc_aud = 0;
    int i;
    int frame_rate = 0, frame_rate_base = 1;
    int req_sample, seek_sample;
    int64_t offset;
    int chunk_number;
    int temp_samples;
    int cur_stream_num = 0;
    int audio_sream_num = 0;

    //find frame-rate
    for (i=0; i<mov->fc->nb_streams; i++)
    {
        if (mov->fc->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO)
        {
            frame_rate = mov->fc->streams[i]->codec.frame_rate;   
            frame_rate_base = mov->fc->streams[i]->codec.frame_rate_base;                
            sc = (MOVStreamContext*) mov->fc->streams[i]->priv_data;
            cur_stream_num = i;
        }
        else 
        if (mov->fc->streams[i]->codec.codec_type == CODEC_TYPE_AUDIO)
        {
            sc_aud = (MOVStreamContext*) mov->fc->streams[i]->priv_data;
            audio_sream_num = i;
        }
    }
                            
    if (frame_rate == 0)
        return -1;

    // Calculate a sample number (round to the closest). 
    req_sample = (pts * frame_rate / frame_rate_base + 500000) / 1000000;  

    // just check it hasn't overflowed.  Seeking beyond end is OK.
    if (req_sample < 0) 
        return -1;

    // Assume that every frame is a key-frame
    seek_sample = req_sample;    

    // If there is no sync index - assume that every frame is a key-frame
    // In sync table, first sample is number "1" - need to adjust.
    if (sc->sync_samples_count > 0)
    {
        //       Flow of searching for nearest key-frame from left side:
        // Sample movie structure with GOP size 15 and total 22 frames can be 
        // described as follows (I - key-frame, O - not key-frame)
        //  1  2  3  4  5  6  7  8  9 10 11 12 14 14 15 16 17 18 19 20 21 22 EOF
        //  I  O  O  O  O  O  O  O  O  O  O  O  O  O  I  O  O  O  O  O  O  O EOF
        //  |-Pos_1              |- Pos_2                   |- Pos_3
        
        // Find next key frame
        for (i=0; i<sc->sync_samples_count; i++)
        {
            if (sc->sync_samples[i] - 1 >= req_sample)
                break;
        }
    
        // Case when requested frame is situated at Pos_3
        if (i == sc->sync_samples_count)
        {
            // Use the last key-frame
            seek_sample = sc->sync_samples[sc->sync_samples_count - 1] - 1;
        }
        else
        {
            // Case when requested frame is located at Pos_1
            if (i==0) 
            {
                // Use the first key-frame
                seek_sample = sc->sync_samples[0] - 1;
            }
            else
            {
                // Case when requested frame is situated at Pos_2 and the frame
                // is not a key-frame
                if ( (sc->sync_samples[i] - 1) != req_sample )
                {
                    // Use nearest key-frame from left side
                    seek_sample = sc->sync_samples[i - 1] - 1;
                }
            }
        }
    }            

    // Find number of chunk that holds the required key sample
    temp_samples = 0;
    for (i=0; i < sc->sample_to_chunk_sz; i++)
    {
        // Summarize entries until temp_samples exceeds seek_sample
        if (i == sc->sample_to_chunk_sz - 1) //last entry
            break;

        temp_samples += (sc->sample_to_chunk[i+1].first - sc->sample_to_chunk[i].first) *
                sc->sample_to_chunk[i].count;

        if (temp_samples > seek_sample)
        {
            // Roll back
            temp_samples -= (sc->sample_to_chunk[i+1].first - sc->sample_to_chunk[i].first) *
                sc->sample_to_chunk[i].count;
            break;
        }    
    }

    // Add the remaining samples
    chunk_number = sc->sample_to_chunk[i].first +
                   (seek_sample - temp_samples) / sc->sample_to_chunk[i].count; //first chunk is chunk number "1"

    temp_samples += (chunk_number - sc->sample_to_chunk[i].first) * sc->sample_to_chunk[i].count; 

    sc->sample_to_chunk_index = i;

    offset = sc->chunk_offsets[chunk_number - 1];


    // Go thru the samples inside the found chunk, and add their sizes to offset
    // until needed sample is reached.
    for (i = temp_samples; i < seek_sample; i++)
    {
        // Is there a table of different sample sizes or the size is fixed ?
        if (sc->sample_size == 0)
            offset += sc->sample_sizes[i];
        else
            offset += sc->sample_size;
    }

    // Update parameters

    sc->left_in_chunk = sc->sample_to_chunk[sc->sample_to_chunk_index].count - 
                        ( seek_sample - temp_samples);

    sc->left_in_chunk %= sc->sample_to_chunk[sc->sample_to_chunk_index].count;
    
    if (sc->left_in_chunk != 0)
    {
        mov->partial = sc;
        mov->partial_stream_num = cur_stream_num;
        sc->next_chunk = chunk_number;
    }  
    else
    {
        mov->partial = 0;
        sc->next_chunk = chunk_number - 1;
    }
    
    mov->next_chunk_offset = offset;
    sc->current_sample = seek_sample;

    Avf_url_fseek(&s->pb, offset, SEEK_SET);

    //update next_chunk for audio stream
    if (sc_aud != 0)
    {
        if ( sc_aud->sample_to_chunk_sz == 1 )
        {
        for (i=0; i < sc_aud->chunk_count; i++)
        {
            if (sc_aud->chunk_offsets[i] >= sc->chunk_offsets[sc->next_chunk])
                break;
        }
        
        sc_aud->next_chunk = i;
            sc_aud->current_sample =  i; //sc_aud->sample_to_chunk[0].count * ( chunk_number - 1 );
            sc_aud->sample_to_chunk_index = 0;
            sc_aud->left_in_chunk = 0;
        }
        else
        {
            int chunk_table_index = 0;
            int chunks_summed = 0;
            int64_t audio_pts = 0;
        
            sc_aud->current_sample = 0;

            while ( chunks_summed < chunk_number - 1 )
            {
                int chunks_within_index;
               
                if ( chunk_table_index+1 >= sc_aud->sample_to_chunk_sz ||
                     chunk_number <= sc_aud->sample_to_chunk[chunk_table_index+1].first )
                {
                    // on last table entry, add only as many samples as needed
                    chunks_within_index = ( chunk_number - 1 - chunks_summed );

                    sc_aud->current_sample += sc_aud->sample_to_chunk[chunk_table_index].count *
                                              chunks_within_index;
                    chunks_summed += chunks_within_index ;
    }
                else
                {   
                    chunks_within_index =  ( sc_aud->sample_to_chunk[chunk_table_index+1].first -
                                             sc_aud->sample_to_chunk[chunk_table_index].first );

                    sc_aud->current_sample += sc_aud->sample_to_chunk[chunk_table_index].count *
                                              chunks_within_index;
    
                    chunks_summed += chunks_within_index;
                    chunk_table_index++;
                }
            }
            
            sc_aud->next_chunk = chunk_number - 1;
            sc_aud->sample_to_chunk_index = chunk_table_index;
            sc_aud->left_in_chunk = 0;
            audio_pts = sc_aud->current_sample * sc_aud->sample_duration[0] * 
                        (int64_t)1000000 / mov->streams[audio_sream_num]->time_scale;
            /* recalculate number of audio frame if its timestamp doesn't 
                        correspond to video  */
            if ( ABS(audio_pts - pts) > AV_SYNC_LAG_THRESHOLD && bSeekKey == 0 )             
            {
                mov_calcualte_audio(mov,sc_aud,chunk_number,audio_sream_num,pts);
            }
        }
    }

    return 0;
}

int mov_read_seek_key(AVFormatContext *s, int stream_id, int pos, int whence)
{    
    MOVDecodeContext *mov = (MOVDecodeContext *) s->priv_data;
    MOVStreamContext *sc = 0;
    int i, seek_sample;
    int frame_rate = 0;
    int frame_rate_base = 1;
    int ret = 0;

    switch ( whence ) {
    
        case SEEK_CUR: 

            for (i=0; i<mov->fc->nb_streams; i++)
            {
                if (mov->fc->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO)
                {
                    sc = (MOVStreamContext*) mov->fc->streams[i]->priv_data;
                    frame_rate = mov->fc->streams[i]->codec.frame_rate;        
                    frame_rate_base = mov->fc->streams[i]->codec.frame_rate_base;
                }
            }

            // If there is no sync index - assume that every frame is a key-frame
            if ( sc->sync_samples_count <= 0 )
            {
                seek_sample = (sc->current_sample-1) + pos;
            }
            else
            {
                // Find closest key frame (from left)
                for (i=0; i<sc->sync_samples_count; i++)
                {
                    if (sc->sync_samples[i] -1 >= sc->current_sample)
                        break;
                }

                i += (pos-1);

                if ((i < 0) || (i >= sc->sync_samples_count))
                    return -1;

                seek_sample = sc->sync_samples[i] -1;
            }
            bSeekKey = 1;
            // Calculate timestamp and seek to it
            ret = mov_read_seek(s, (int64_t)seek_sample * 1000000 * frame_rate_base / frame_rate);
            bSeekKey = 0;

            break;

        default:
            printf("usupported option !\n");
    }

    return ret;
}

int mov_read_text_stream(struct AVFormatContext *s, 
                         int first, 
                         int total,
                         char** text,
                         int* indexes,
                         int stream_id )
{
    int i;
    AVStream *st = s->streams[stream_id];
    MOVStreamContext *sc = (MOVStreamContext *)st->priv_data;
    offset_t pos;
    int size;
    CHAR buf[256] = {0};
    int str_size;


    for( i = 0; i < total; i++ )
    {
        pos = sc->chunk_offsets[first+i];
        Avf_url_fseek(&s->pb, pos, SEEK_SET);
        if( sc->sample_size == 0 )
        {
            size = sc->sample_sizes[first+i];
        }
        else
        {
            size = sc->sample_size;
        }
        get_buffer( &s->pb, (uint8_t *)buf, size );

        str_size = (int)(buf[0]<<8) + (int)(buf[1]<<0) + 2;
        text[i] = avf_mallocz( str_size + 1 );
        memcpy( text[i], buf, str_size );
        
        indexes[i] = sc->sample_duration[first+i];
    }

    return 0;
}



typedef struct AACContext {
    int packet_obj_size;
} AACContext;

static int aac_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    AVStream *st;
    AACContext *aac = s->priv_data;

    aac->packet_obj_size = 8192;

    st = avf_mallocz(sizeof(AVStream));
    if (!st)
        return -1;

    AvApi_avcodec_get_context_defaults(&st->codec);
    s->streams[s->nb_streams] = st;   
    st->codec.codec_type = CODEC_TYPE_AUDIO;
    st->codec.codec_id = CODEC_ID_AACZ;
    st->codec.frame_size = 1;
    st->codec.frame_rate = 15 * s->pts_den / s->pts_num; // 15 fps default
    
    s->nb_streams++;

    return 0;
}

static int aac_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    AACContext *aac = s->priv_data;
    ByteIOContext *pb = &s->pb;

    /* new packet */
    av_new_packet( pkt, aac->packet_obj_size);

    pkt->size = get_buffer( pb, pkt->data, aac->packet_obj_size );

    if ( pkt->size == 0 )
    {
        return -1;
    }

    return 0;
}

static int aac_read_close(AVFormatContext *s)
{
    return 0;
}


AVInputFormat mov_iformat = {
    "mov",
    "QuickTime/MPEG4 format",
    sizeof(MOVDecodeContext),
    mov_probe,
    mov_read_header,
    mov_read_packet,
    mov_read_close,
    mov_read_seek,
    mov_read_seek_key,
    mov_read_text_stream,
    mov_read_is_keyframe,
};

AVInputFormat aac_iformat = {
    "aac",
    "application/mp4",
    sizeof(AACContext),
    NULL,
    aac_read_header,
    aac_read_packet,
    aac_read_close,
    .extensions = "aac",
    .value = CODEC_ID_AACZ,
};

int Avf_movdec_init(void)
{
    av_register_input_format(&mov_iformat);
    av_register_input_format(&aac_iformat);

    return 0;
}

void mov_set_drop_audio(int bDrop)
{
    // will audio be dropped during playback?
    bDropAudioOnPlayback = bDrop;
}


