/* 
 * RAW encoder and decoder
 * Copyright (c) 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avformat.h"
#include "av.h"

#define MPEG_HEADER_CHECK_SIZE 1792*32      //1792 = MPA_MAX_CODED_FRAME_SIZE
/* simple formats */
static int raw_write_header(struct AVFormatContext *s)
{
    return 0;
}

static int raw_write_packet(struct AVFormatContext *s, int stream_index,
                unsigned char *buf, int size, int64_t force_pts)
{
    put_buffer(&s->pb, buf, size);
    put_flush_packet(&s->pb);
    return 0;
}

static int raw_write_trailer(struct AVFormatContext *s)
{
    // :ZZZ: Must flush buffer, since we accumulate frames
    put_flush_packet_end(&s->pb);

    return 0;
}

/* raw input */
static int raw_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    AVStream *st;
    int id;

    st = Avf_new_stream(s, 0);
    if (!st)
        return AVERROR_NOMEM;
    if (ap) {
        id = s->iformat->value;
        if (id == CODEC_ID_RAWVIDEO) {
            st->codec.codec_type = CODEC_TYPE_VIDEO;
        } else {
            st->codec.codec_type = CODEC_TYPE_AUDIO;
        }
        st->codec.codec_id = (enum CodecID)id;

        switch(st->codec.codec_type) {
        case CODEC_TYPE_AUDIO:
            st->codec.sample_rate = ap->sample_rate;
            st->codec.channels = ap->channels;
            break;
        case CODEC_TYPE_VIDEO:
            st->codec.frame_rate      = ap->frame_rate;
            st->codec.frame_rate_base = ap->frame_rate_base;
            st->codec.physical_width = ap->width;
            st->codec.physical_height = ap->height;
        st->codec.pix_fmt = ap->pix_fmt;
            break;
        default:
            return -1;
        }
    } else {
        return -1;
    }
    return 0;
}

/* yuv422 input */
static int yuv422_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    AVStream *st;

    st = Avf_new_stream(s, 0);

    if (!st)
        return AVERROR_NOMEM;

    st->codec.codec_type        = CODEC_TYPE_VIDEO;
    st->codec.frame_rate        = 30;
    st->codec.frame_rate_base   = 1;
    st->codec.physical_width             = 160;
    st->codec.physical_height            = 128;
    st->codec.pix_fmt           = PIX_FMT_YUV422Z;
    st->codec.codec_id          = (enum CodecID)s->iformat->value;

    return 0;
}

static int yuv422_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    int packet_size, ret, width, height;
    AVStream *st = s->streams[0];

    width = st->codec.physical_width;
    height = st->codec.physical_height;

    packet_size = avpicture_get_size(st->codec.pix_fmt, width, height);

    if (packet_size < 0)
        av_abort();

    if (av_new_packet(pkt, packet_size) < 0)
        return -EIO;

    pkt->stream_index = 0;
    pkt->flags |= PKT_FLAG_KEY;

#if 0
    /* bypass buffered I/O */
    ret = url_read(url_fileno(&s->pb), pkt->data, pkt->size);
#else
    ret = get_buffer(&s->pb, pkt->data, pkt->size);

#endif
    if (ret != pkt->size) {
        Avf_free_packet(pkt);
        return -EIO;
    } else {
        return 0;
    }
}

/* byr2 input */
static int byr2_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    AVStream *st;

    st = Avf_new_stream(s, 0);

    if (!st)
        return AVERROR_NOMEM;

    st->codec.codec_type        = CODEC_TYPE_VIDEO;

    st->codec.frame_rate        = ap->frame_rate;
    st->codec.frame_rate_base   = ap->frame_rate_base;
    st->codec.physical_width    = ap->width;
    st->codec.physical_height   = ap->height;
	st->codec.pix_fmt           = PIX_FMT_RAW16Z;
    st->codec.codec_tag         = MKTAG('B', 'Y', 'R', '2');

    st->codec.codec_id          = (enum CodecID)s->iformat->value;

    return 0;
}

#define RAW_PACKET_SIZE 1024

static int raw_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    int ret, size;
    AVStream *st = s->streams[0];
    
    size= RAW_PACKET_SIZE;

    if (av_new_packet(pkt, size) < 0)
        return -EIO;

    // :PATCH: In order to allow to finish the clip playback 
    if ( st->codec.codec_id == CODEC_ID_H264Z )
    {
        pkt->pts = 1;
    }

    pkt->stream_index = 0;
    ret = get_buffer(&s->pb, pkt->data, size);

    // TODO: not all packet is key
    pkt->flags |= PKT_FLAG_KEY;

    if (ret <= 0) {
        Avf_free_packet(pkt);
        return -EIO;
    }
    /* note: we need to modify the packet size here to handle the last
       packet */
    pkt->size = ret;
    return ret;
}

static int raw_read_close(AVFormatContext *s)
{
    return 0;
}

static int mp3_is_valid_header( int mpegVer,
                                int layer,
                                int bitrate_idx,
                                int sample_idx,
                                int flags )
{
    int iSupportMpeg25 = (flags & AV_READ_HEADER_FLAG_MP3_SUPPORT_MPEG25);

    if ( (mpegVer == 1) ||                          // mpeg ver reserved
         ((mpegVer == 0) && (iSupportMpeg25==0)) )  // mpeg ver 2.5 not supported
    {
        return FALSE;
    }
        

    if ( (layer == 0)         || // layer reserved
         (bitrate_idx == 0xF) || // bitrate reserved
         (bitrate_idx == 0x0) ||    //means variable bitrate.
         (sample_idx  == 3) )    // sample rate reserved
    {
        return FALSE;
    }

    return TRUE;
}              
              

static int mp3_get_data_stream_pos( PBYTE pbyBuffer, 
                                    UINT  uiSize )
{
    UINT uiId3Size = 0;

    // Check if ID3 ver >= 2 presents
    if ( strncmp( (char *)pbyBuffer, "ID3", 3 ) != 0 )
    {
        // No ID3 Tag
        return 0;
    }

    // Check the version. Has to be >= 2
    if ( ((UINT)pbyBuffer[3]) < 2 )
    {
        // ID3 Version 1 is located at the end of the file
        return 0;
    }

    // Calculate the ID3 tag size
    // The size is located in bytes 6~9 of the ID3 header and
    // it's compressed: each byte contains 7 bit info
    uiId3Size = (pbyBuffer[6] << (24-3)) |
                (pbyBuffer[7] << (16-2)) |
                (pbyBuffer[8] <<  (8-1)) |
                 pbyBuffer[9];

    // Consider the ID3 header size
    uiId3Size += 10;

    return uiId3Size;
}

static int mp3_get_frame_size(unsigned long header)
{
    INT sample_rate = 0;
    INT frame_size = 0;
    INT mpeg25 = 0;
    INT padding= 0;
    INT sample_rate_index =0;
    INT bitrate_index =0;
    BYTE blsf = 0;
    BYTE blayer = 0;

    USHORT usMpaBitrateTab[2][3][15] = {
        { {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448 },
          {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384 },
          {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320 } },
        { {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
          {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
          {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
        }
    };

    USHORT usMpaFreqTab[3] = { 44100, 48000, 32000 };
    
    if (header & (1<<20)) 
    {
        blsf = (header & (1<<19)) ? 0 : 1;
        mpeg25 = 0;
    } 
    else 
    {
        blsf = 1;
        mpeg25 = 1;
    }
    
    blayer = 4 - ((header >> 17) & 3);
    
    /* extract frequency */
    sample_rate_index = (header >> 10) & 3;
    if (sample_rate_index >= (INT)COUNT(usMpaFreqTab))
    {
        OsSysHalt("Invalid sample rate index from header");
    }
    sample_rate = usMpaFreqTab[sample_rate_index] >> (blsf + mpeg25);
    sample_rate_index += 3 * (blsf + mpeg25);

    bitrate_index = (header >> 12) & 0xf;
    padding = (header >> 9) & 1;
    
    if (bitrate_index != 0) 
    {
        frame_size = usMpaBitrateTab[blsf][blayer - 1][bitrate_index];
        switch(blayer) 
        {
            case 1:
            {
                frame_size = (frame_size * 12000) / sample_rate;
                frame_size = (frame_size + padding) * 4;
                break;
            }
            case 2:
            {
                frame_size = (frame_size * 144000) / sample_rate;
                frame_size += padding;
                break;
            }
            case 3:
            default:
            {
                frame_size = (frame_size * 144000) / (sample_rate << blsf);
                frame_size += padding;
                break;
            }
        }
    } 
    
    return frame_size;
}

//wayneh for issue 18
static int mp3_probe(AVProbeData *p)
{
            const uint8_t *d;

            LONG header;
            BYTE sample_idx = 0;
            BYTE channels = 0;
            BYTE mpegVer = 0;
            BYTE layer = 0;
            BYTE bitrate_idx = 0;

            int flags = 0;
            int pos = 0;

            // If ID3 presents, change the initial position to data that 
            // follows ID3 tag
            pos = mp3_get_data_stream_pos(  p->buf, p->buf_size );

//            printf("[18] mp3_probe: pos=%d\n", pos);

            //ID3 tags found, recognize as mp3 file.
            if(pos > 0) return AVPROBE_SCORE_MAX - 1;

            //ID3 not found, try to find FF FA - MP3 = MPEG 1 Layer 
            if (p->buf_size < pos+4)
                return 0;
            
            d = p->buf;

            if( (d[0] == 0xFF) && ( (d[1] & 0xFE) == 0xFA ) )        
            {
                        // Don't have to read index here
                    flags |= AV_READ_HEADER_FLAG_NO_INDEX;
#if 0
                    // Support MPEG2.5 in MP3
                    if ( gbMp3SupportMpeg25 )
                    {
                        flags |= AV_READ_HEADER_FLAG_MP3_SUPPORT_MPEG25;
                    }
#endif
                    flags |= AV_FAST_FILE_OPEN;
                 flags |= AV_NO_DRM;
             
                    header = (d[0] << 24) |(d[1] << 16) | (d[2] << 8) | d[3];

                  sample_idx = (header >> 10) & 3;
                  channels = (header >> 6) & 3;
                  mpegVer = (header >> 19) & 3;
                    layer   = (header >> 17) & 3;
                    bitrate_idx = (header >> 12) & 0xF;    

//printf("[18] mp3_probe: sample_idx=%x, channels=%x, mpegVer=%x\n", sample_idx, channels, mpegVer);
//printf("[18] mp3_probe: layer=%x, bitrate_idx=%x\n", layer, bitrate_idx);

                    if ( mp3_is_valid_header(mpegVer, 
                                              layer, 
                                              bitrate_idx, 
                                              sample_idx,
                                              flags) == TRUE )
                            return AVPROBE_SCORE_MAX - 1;
        }    

    return 0;
}


/* mp3 read */
static int mp3_read_header(AVFormatContext *s,
                           AVFormatParameters *ap)
{
    AVStream *st;
    int pos = 0;

    LONG header;
    BYTE sample_idx = 0;
    BYTE sample_idx_check = 0;
    BYTE channels = 0;
    BYTE channels_check = 0;
    BYTE mpegVer = 0;
    BYTE mpegVer_check = 0;
    BYTE layer = 0;
    BYTE layer_check = 0;
    BYTE bitrate_idx = 0;
    BYTE bitrate_idx_check = 0;
    INT size = MPEG_HEADER_CHECK_SIZE;
    UINT uwCheckNum = 0;    // For avoid false frame header we have to find three continuouse same frame header

    // If ID3 presents, change the initial position to data that 
    // follows ID3 tag
    pos = mp3_get_data_stream_pos( s->pb.buffer, size );

    /* looking for 11111111 111MMLLC - MPEG synchronization tag
    MM: 00 - MPEG-2.5, 10 - MPEG-2, 11 - MPEG-1
    LL: 11 - Layer I, 10 - Layer II, 01 - Layer III
    XXX: this code does not read more bytes from file 
    so if ID3 (or other stuff) length > IO_BUFFER_SIZE it fails back to CODEC_ID_MP2 */

    size += pos;        //increase the search buffer

    while ( pos < (size - 3) )
    {
        char buf[4];

        //get the buffer from file
        Avf_url_fseek(&s->pb, pos, SEEK_SET);
        get_buffer(&s->pb, (unsigned char *)buf, 4);
        
        if( (buf[0] == 0xFF) && (buf[1] & 0xE0) == 0xE0 )        
        {
            header = (buf[0] << 24) |(buf[1] << 16) | (buf[2] << 8) | buf[3];

            sample_idx = (header >> 10) & 3;
            channels = (header >> 6) & 3;
            mpegVer = (header >> 19) & 3;
            layer   = (header >> 17) & 3;
            bitrate_idx = (header >> 12) & 0xF;    
                    
            if ( mp3_is_valid_header( mpegVer, 
                                      layer, 
                                      bitrate_idx, 
                                      sample_idx,
                                      ap->flags ) == TRUE )
            {
                if( uwCheckNum == 0)
                {
                    uwCheckNum = 1;

                    sample_idx_check = sample_idx;
                    channels_check = channels;
                    mpegVer_check= mpegVer;
                    layer_check = layer;
                    bitrate_idx_check = bitrate_idx;
                }
                else
                {
                    if(uwCheckNum >= 3)
                    {
                // found the mp3 header 
                break;
            }
            else
            {
                        if( (sample_idx_check == sample_idx) && 
                            (channels_check == channels) &&     
                            (mpegVer == mpegVer_check) &&           
                            (layer == layer_check)   &&                         
                            (bitrate_idx == bitrate_idx_check) )
                        {
                            //check counter increse when the frame hear is same as previouse one
                            uwCheckNum ++;                          
                        }
                        else
                        {
                            uwCheckNum = 0;
                        }
                        
                        pos += mp3_get_frame_size( header);
                    }
                }
                // found the mp3 header 
                //break;
            }
            else
            {
                // this is not a valid mp3 header or we cannot play this 
                // format, keep looking
                pos++;
            }
        }
        else
        {
        pos++;
    }
    }

    if ( pos >= (size - 3) )
    {
        return AVERROR_NOFMT;
    }   
    
    st = Avf_new_stream(s, 0);
    if (!st)
    {
        return AVERROR_NOMEM;
    }

    st->codec.codec_type = CODEC_TYPE_AUDIO;
    st->codec.codec_id = CODEC_ID_MP2;        
        
    if( (s->pb.buffer[pos+1] & 6) == 2 )
    {
        st->codec.codec_id = CODEC_ID_MP3LAME;
    }
        
    if(channels == 3)
    {   
        st->codec.channels = 1;
    }
    else                
    {   
        st->codec.channels = 2;
    }

    if(mpegVer == 0)    
    {
        if(sample_idx == 0)
            st->codec.sample_rate = 11025;
        else if(sample_idx == 1)
            st->codec.sample_rate = 12000;
        else if(sample_idx == 2)
            st->codec.sample_rate = 8000;
    }
    else if(mpegVer == 2)   
    {
        if(sample_idx == 0)
            st->codec.sample_rate = 22050;
        else if(sample_idx == 1)
            st->codec.sample_rate = 24000;
        else if(sample_idx == 2)
            st->codec.sample_rate = 16000;
    }
    else if(mpegVer == 3)   
    {
        if(sample_idx == 0)
            st->codec.sample_rate = 44100;
        else if(sample_idx == 1)
            st->codec.sample_rate = 48000;
        else if(sample_idx == 2)
            st->codec.sample_rate = 32000;
    }
    
    
    //add assign mp3 bitrate
    //wayne for issue 18 
    {
      int  bitrate_table[2][16] = {{0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0},
                                   {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0}      };
       //only for CBR
            if(mpegVer == 0)    
        {
                st->codec.bit_rate = (int)bitrate_table[1][bitrate_idx]*1000;
        }
        else if(mpegVer == 2)   
        {
                st->codec.bit_rate = (int)bitrate_table[1][bitrate_idx]*1000;
        }
        else if(mpegVer == 3)   
        {
                st->codec.bit_rate = bitrate_table[0][bitrate_idx]*1000;
        }
    }


    return 0;
}

/* ac3 read */
static int ac3_read_header(AVFormatContext *s,
                           AVFormatParameters *ap)
{
    AVStream *st;

    st = Avf_new_stream(s, 0);
    if (!st)
        return AVERROR_NOMEM;

    st->codec.codec_type = CODEC_TYPE_AUDIO;
    st->codec.codec_id = CODEC_ID_AC3;
    /* the parameters will be extracted from the compressed bitstream */
    return 0;
}

/* mpeg1/h263 input */
static int video_read_header(AVFormatContext *s,
                             AVFormatParameters *ap)
{
    AVStream *st;

    st = Avf_new_stream(s, 0);
    if (!st)
        return AVERROR_NOMEM;

    st->codec.codec_type = CODEC_TYPE_VIDEO;
    st->codec.codec_id = (enum CodecID)s->iformat->value;
    /* for mjpeg, specify frame rate */
    /* for mpeg4 specify it too (most mpeg4 streams dont have the fixed_vop_rate set ...)*/
    if (st->codec.codec_id == CODEC_ID_MJPEG || st->codec.codec_id == CODEC_ID_MPEG4) {
        if (ap) {
            st->codec.frame_rate      = ap->frame_rate;
            st->codec.frame_rate_base = ap->frame_rate_base;
        } else {
            st->codec.frame_rate      = 25;
            st->codec.frame_rate_base = 1;
        }
    }
    else if ( st->codec.codec_id == CODEC_ID_H264Z )
    {
        // Set width & height to max size
        st->codec.physical_width = 1280;
        st->codec.physical_height = 720;
        st->codec.flags |= CODEC_FLAG_TRUNCATED;
    }

    return 0;
}

#define SEQ_START_CODE      0x000001b3
#define GOP_START_CODE      0x000001b8
#define PICTURE_START_CODE  0x00000100

/* XXX: improve that by looking at several start codes */
static int mpegvideo_probe(AVProbeData *p)
{
    int code;
    const uint8_t *d;

    /* we search the first start code. If it is a sequence, gop or
       picture start code then we decide it is an mpeg video
       stream. We do not send highest value to give a chance to mpegts */
    /* NOTE: the search range was restricted to avoid too many false
       detections */

    if (p->buf_size < 6)
        return 0;
    d = p->buf;
    code = (d[0] << 24) | (d[1] << 16) | (d[2] << 8) | (d[3]);
    if ((code & 0xffffff00) == 0x100) {
        if (code == SEQ_START_CODE ||
            code == GOP_START_CODE ||
            code == PICTURE_START_CODE)
            return 50 - 1;
        else
            return 0;
    }
    return 0;
}

AVInputFormat mp3_iformat = {
    "mp3",
    "MPEG audio",
    0,
    mp3_probe/*NULL*/,
    mp3_read_header,
    raw_read_packet,
    raw_read_close,
    .extensions = "mp2,mp3", /* XXX: use probe */
};

AVOutputFormat mp2_oformat = {
    "mp2",
    "MPEG audio layer 2",
    "audio/x-mpeg",
#ifdef CONFIG_MP3LAME
    "mp2",
#else
    "mp2,mp3",
#endif
    0,
    CODEC_ID_MP2,
    CODEC_ID_NONE,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

#ifdef CONFIG_MP3LAME
AVOutputFormat mp3_oformat = {
    "mp3",
    "MPEG audio layer 3",
    "audio/x-mpeg",
    "mp3",
    0,
    CODEC_ID_MP3LAME,
    0,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};
#endif

AVInputFormat ac3_iformat = {
    "ac3",
    "raw ac3",
    0,
    NULL,
    ac3_read_header,
    raw_read_packet,
    raw_read_close,
    .extensions = "ac3",
};

AVOutputFormat ac3_oformat = {
    "ac3",
    "raw ac3",
    "audio/x-ac3", 
    "ac3",
    0,
    CODEC_ID_AC3,
    CODEC_ID_NONE,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVOutputFormat h263_oformat = {
    "h263",
    "raw h263",
    "video/x-h263",
    "h263",
    0,
    CODEC_ID_NONE,
    CODEC_ID_H263,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVInputFormat m4v_iformat = {
    "m4v",
    "raw MPEG4 video format",
    0,
    NULL /*mpegvideo_probe*/,
    video_read_header,
    raw_read_packet,
    raw_read_close,
    .extensions = "m4v", //FIXME remove after writing mpeg4_probe
    .value = CODEC_ID_MPEG4,
};

AVOutputFormat m4v_oformat = {
    "m4v",
    "raw MPEG4 video format",
    NULL,
    "m4v",
    0,
    CODEC_ID_NONE,
    CODEC_ID_MPEG4,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVInputFormat h264_iformat = {
    "h264",
    "raw H264 video format",
    0,
    NULL /*mpegvideo_probe*/,
    video_read_header,
    raw_read_packet,
    raw_read_close,
    .extensions = "h26l,h264,264", //FIXME remove after writing mpeg4_probe
    .value = CODEC_ID_H264Z,
};


AVOutputFormat h264_oformat = {
    "h264",
    "raw H264 video format",
    NULL,
    "h264,264",
    0,
    CODEC_ID_NONE,
    CODEC_ID_H264Z,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVInputFormat mpeg1_iformat = {
    "mpeg",
    "raw MPEG4 video format",
    0,
    NULL /*mpegvideo_probe*/,
    video_read_header,
    raw_read_packet,
    raw_read_close,
    .extensions = "mpg,mpeg", //FIXME remove after writing mpeg4_probe
    .value = CODEC_ID_MPEG1Z,
};


AVOutputFormat mpeg1_oformat = {
    "mpg",
    "raw mpeg1 video format",
    NULL,
    "mpg,mpeg",
    0,
    CODEC_ID_NONE,
    CODEC_ID_MPEG1Z,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVInputFormat mpegvideo_iformat = {
    "mpegvideo",
    "MPEG video",
    0,
    mpegvideo_probe,
    video_read_header,
    raw_read_packet,
    raw_read_close,
    .value = CODEC_ID_MPEG1VIDEO,
};

AVOutputFormat mpeg1video_oformat = {
    "mpeg1video",
    "MPEG video",
    "video/x-mpeg",
    "mpg,mpeg",
    0,
    CODEC_ID_NONE,
    CODEC_ID_MPEG1VIDEO,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVInputFormat mjpeg_iformat = {
    "mjpeg",
    "MJPEG video",
    0,
    NULL,
    video_read_header,
    raw_read_packet,
    raw_read_close,
    .extensions = "mjpg,mjpeg",
    .value = CODEC_ID_MJPEG,
};

AVOutputFormat mjpeg_oformat = {
    "mjpeg",
    "MJPEG video",
    "video/x-mjpeg",
    "mjpg,mjpeg",
    0,
    CODEC_ID_NONE,
    CODEC_ID_MJPEG,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

/* pcm formats */

#define PCMDEF(name, long_name, ext, codec) \
AVInputFormat pcm_ ## name ## _iformat = {\
    #name,\
    long_name,\
    0,\
    NULL,\
    raw_read_header,\
    raw_read_packet,\
    raw_read_close,\
    .extensions = ext,\
    .value = codec,\
};\
\
AVOutputFormat pcm_ ## name ## _oformat = {\
    #name,\
    long_name,\
    NULL,\
    ext,\
    0,\
    codec,\
    (enum CodecID)0,\
    raw_write_header,\
    raw_write_packet,\
    raw_write_trailer,\
};

#ifdef WORDS_BIGENDIAN
#define BE_DEF(s) s
#define LE_DEF(s) NULL
#else
#define BE_DEF(s) NULL
#define LE_DEF(s) s
#endif


PCMDEF(s16le, "pcm signed 16 bit little endian format", 
       LE_DEF("sw"), CODEC_ID_PCM_S16LE)

PCMDEF(s16be, "pcm signed 16 bit big endian format", 
       BE_DEF("sw"), CODEC_ID_PCM_S16BE)

PCMDEF(u16le, "pcm unsigned 16 bit little endian format", 
       LE_DEF("uw"), CODEC_ID_PCM_U16LE)

PCMDEF(u16be, "pcm unsigned 16 bit big endian format", 
       BE_DEF("uw"), CODEC_ID_PCM_U16BE)

PCMDEF(s8, "pcm signed 8 bit format", 
       "sb", CODEC_ID_PCM_S8)

PCMDEF(u8, "pcm unsigned 8 bit format", 
       "ub", CODEC_ID_PCM_U8)

PCMDEF(mulaw, "pcm mu law format", 
       "ul", CODEC_ID_PCM_MULAW)

PCMDEF(alaw, "pcm A law format", 
       "al", CODEC_ID_PCM_ALAW)

static int rawvideo_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    int packet_size, ret, width, height;
    AVStream *st = s->streams[0];

    width = st->codec.physical_width;
    height = st->codec.physical_height;

    packet_size = avpicture_get_size(st->codec.pix_fmt, width, height);
    if (packet_size < 0)
        av_abort();

    if (av_new_packet(pkt, packet_size) < 0)
        return -EIO;

    pkt->stream_index = 0;
#if 0
    /* bypass buffered I/O */
    ret = url_read(url_fileno(&s->pb), pkt->data, pkt->size);
#else
    ret = get_buffer(&s->pb, pkt->data, pkt->size);
#endif
    if (ret != pkt->size) 
    {
        Avf_free_packet(pkt);
        return -EIO;
    } 
    else 
    {
        return 0;
    }
}

AVInputFormat yuv422video_iformat = {
    "422",
    "raw video format",
    0,
    NULL,
    yuv422_read_header,
    yuv422_read_packet,
    raw_read_close,
    .extensions = "422",
    .value = CODEC_ID_RAWZ,
};

AVInputFormat byr2video_iformat = {
    "BYR2",
    "Bayer16 raw video format",
    0,
    NULL,
    byr2_read_header,
    yuv422_read_packet,
    raw_read_close,
    .extensions = "BYR2",
    .value = CODEC_ID_RAWZ,
};

AVOutputFormat byr2video_oformat = {
    "BYR2",
    "Bayer16 raw video format",
    NULL,
    "BYR2",
    0,
    CODEC_ID_NONE,
    CODEC_ID_RAWZ,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer
};

AVInputFormat rawvideo_iformat = {
    "rawvideo",
    "raw video format",
    0,
    NULL,
    raw_read_header,
    rawvideo_read_packet,
    raw_read_close,
    .extensions = "yuv",
    .value = CODEC_ID_RAWVIDEO,
};

AVOutputFormat rawvideo_oformat = {
    "rawvideo",
    "raw video format",
    NULL,
    "yuv",
    0,
    CODEC_ID_NONE,
    CODEC_ID_RAWVIDEO,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

AVOutputFormat raw_oformat = {
    "raw",
    "raw format",
    NULL,
    "raw",
    0,
    CODEC_ID_NONE,
    CODEC_ID_NONE,
    raw_write_header,
    raw_write_packet,
    raw_write_trailer,
};

static int null_write_packet(struct AVFormatContext *s, 
                             int stream_index,
                             unsigned char *buf, int size, int64_t force_pts)
{
    return 0;
}

AVOutputFormat null_oformat = {
    "null",
    "null video format",
    NULL,
    NULL,
    0,
#ifdef WORDS_BIGENDIAN
    CODEC_ID_PCM_S16BE,
#else
    CODEC_ID_PCM_S16LE,
#endif
    CODEC_ID_RAWVIDEO,
    raw_write_header,
    null_write_packet,
    raw_write_trailer,
    .flags = AVFMT_NOFILE | AVFMT_RAWPICTURE,
};

int Avf_raw_init(void)
{
    av_register_input_format(&mp3_iformat);
    av_register_output_format(&mp2_oformat);
#ifdef CONFIG_MP3LAME
    av_register_output_format(&mp3_oformat);
#endif    
    av_register_input_format(&ac3_iformat);
    av_register_output_format(&ac3_oformat);

    av_register_output_format(&h263_oformat);
    
    av_register_input_format(&m4v_iformat);
    av_register_output_format(&m4v_oformat);
    
    av_register_input_format(&h264_iformat);
    av_register_output_format(&h264_oformat);

    av_register_input_format(&mpegvideo_iformat);
    av_register_output_format(&mpeg1video_oformat);

    av_register_input_format(&mjpeg_iformat);
    av_register_output_format(&mjpeg_oformat);

    av_register_input_format(&pcm_s16le_iformat);
    av_register_output_format(&pcm_s16le_oformat);
    av_register_input_format(&pcm_s16be_iformat);
    av_register_output_format(&pcm_s16be_oformat);
    av_register_input_format(&pcm_u16le_iformat);
    av_register_output_format(&pcm_u16le_oformat);
    av_register_input_format(&pcm_u16be_iformat);
    av_register_output_format(&pcm_u16be_oformat);
    av_register_input_format(&pcm_s8_iformat);
    av_register_output_format(&pcm_s8_oformat);
    av_register_input_format(&pcm_u8_iformat);
    av_register_output_format(&pcm_u8_oformat);
    av_register_input_format(&pcm_mulaw_iformat);
    av_register_output_format(&pcm_mulaw_oformat);
    av_register_input_format(&pcm_alaw_iformat);
    av_register_output_format(&pcm_alaw_oformat);

    av_register_input_format(&rawvideo_iformat);
    av_register_output_format(&rawvideo_oformat);

    av_register_input_format(&yuv422video_iformat);
    
    av_register_input_format(&byr2video_iformat);
    av_register_output_format(&byr2video_oformat);

    av_register_output_format(&raw_oformat);

    av_register_output_format(&null_oformat);
    return 0;
}

