/*
 * Various utilities for ffmpeg system
 * Copyright (c) 2000, 2001, 2002 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "Allocator.h"
#include "avformat.h"
#include "av.h"
#include <ctype.h>
#ifdef CONFIG_WIN32
#define strcasecmp _stricmp
#include <sys/types.h>
#include <sys/timeb.h>
#elif defined(CONFIG_OS2)
#include <string.h>
#define strcasecmp stricmp
#include <sys/time.h>
#else
#if !defined(__CC_ARM) /* unsupported by ARM */
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#endif /* __CC_ARM */
#endif
#include <time.h>

#include "AvApi.h"
#include "avcodec.h"
#include "AvFormat.h"
#include "mov.h"

#include "misc.h"

// :ZZZ: Added NULL
AVInputFormat *first_iformat = NULL;
AVOutputFormat *first_oformat = NULL;
AVImageFormat *first_image_format = NULL;

// Structure that holds custom content description information 
AVStreamInfo custom_stream_info = {0};

// Structure that holds custom tag information. This custom_tag_content[0]
// is unused and for initialization purpose
static AVCustomTagContent custom_tag_content[MAX_CUSTOM_TAG_NUM] = {{0}};

// This function pointer holds the Cb function registered by the host for
// update the skip atom content
static PBYTE pbCustomSkipAtomCbFuncPtr = NULL; 

// This function pointer holds the Cb function registered by the host for
// update the custom atom content
static PBYTE pbCustomAtomCbFuncPtr = NULL;


/* cannot call it directly because of 'void **' casting is not automatic */
void __avf_freep(void **ptr)
{
    if (NULL != (*ptr))
    {
        avf_free(*ptr);
        *ptr = NULL;
    }
}

void avf_free_ex(void *ptr, char* file, int line)
{
    PoolFreeEx( NULL, ptr, file, line );
}

void *avf_malloc_ex(unsigned int size, char* file, int line)
{
    // :ZZZ: Should be fixed in the component
    if ( size == 0 )
    {
        size = 1;
    }
    return PoolAllocEx( NULL, 
                        size, 
                        POOL_MEMORY_CACHED | 
                        POOL_WAIT_FOREVER | 
                        POOL_DISABLE_FAILURE_PRINT, 
                        file, 
                        line );
}

void *avf_mallocz_ex(unsigned int size, char* file, int line)
{
    void *ptr;
    
    // :ZZZ: Should be fixed in the component
    if ( size == 0 )
    {
        size = 1;
    }

    ptr = avf_malloc_ex(size, file, line);
    if (!ptr)
        return NULL;
    memset(ptr, 0, size);
	//fprintf(stderr, "avf_mallocz_ex: 0x%08X\n", ptr);
    return ptr;
}

void *avf_realloc_ex(void* ptr, unsigned int size, char* file, int line)
{
    return PoolReallocEx( NULL, ptr, size, POOL_MEMORY_CACHED, file, line );
}

void av_register_input_format(AVInputFormat *format)
{
    AVInputFormat **p;
    p = &first_iformat;
    while (*p != NULL) {
		p = &(*p)->next;
    }
    *p = format;
    format->next = NULL;
}

int Avf_register_input_format_error_callback(void * callback_function)
{
    AVInputFormat *p;
    p = first_iformat;
    while ( p != NULL ) {
        p->handle_error = (GenericCallbackFn)callback_function;
        p = p->next;
    }
    return 0;
}

int Avf_register_input_format_custom_probe_callback(GenericCallbackFn callback_function)
{
    AVInputFormat *p;
    p = first_iformat;
    while ( p != NULL ) {
        p->custom_probe = callback_function;
        p = p->next;
    }
    return 0;
}

void av_register_output_format(AVOutputFormat *format)
{
    AVOutputFormat **p;
    p = &first_oformat;
    while (*p != NULL) p = &(*p)->next;
    *p = format;
    format->next = NULL;
}

int Avf_register_output_format_callback(void * callback_function)
{
    AVOutputFormat *p;
    p = first_oformat;
    while ( p != NULL ) {
        p->generic_callback = (GenericCallbackFn)callback_function;
        p = p->next;
    }
    return 0;
}

int match_ext(const char *filename, const char *extensions)
{
    const char *ext, *p;
    char ext1[32], *q;

    ext = strrchr(filename, '.');
    if (ext) {
        ext++;
        p = extensions;
        for(;;) {
            q = ext1;
            while (*p != '\0' && *p != ',') 
                *q++ = *p++;
            *q = '\0';
            if (!strcasecmp(ext1, ext)) 
                return 1;
            if (*p == '\0') 
                break;
            p++;
        }
    }
    return 0;
}

AVOutputFormat *Avf_guess_format(	const char *short_name, 
								const char *filename, 
                             	const char *mime_type)
{
    AVOutputFormat *fmt, *fmt_found;
    int score_max, score;

    /* specific test for image sequences */
    if (!short_name && filename && 
        filename_number_test(filename) >= 0 &&
        guess_image_format(filename)) 
    {
        return Avf_guess_format("image", NULL, NULL);
    }

    /* find the proper file type */
    fmt_found = NULL;
    score_max = 0;
    fmt = first_oformat;
    while (fmt != NULL) 
	{
        score = 0;
        if (fmt->name && short_name && !strcmp(fmt->name, short_name))
            score += 100;
        if (fmt->mime_type && mime_type && !strcmp(fmt->mime_type, mime_type))
            score += 10;
        if (filename && fmt->extensions && 
            match_ext(filename, fmt->extensions)) {
            score += 5;
        }
        if (score > score_max) {
            score_max = score;
            fmt_found = fmt;
        }
        fmt = fmt->next;
    }
    return fmt_found;
}   

AVOutputFormat *guess_stream_format(const char *short_name, const char *filename, 
                             const char *mime_type)
{
    AVOutputFormat *fmt = Avf_guess_format(short_name, filename, mime_type);

    if (fmt) {
        AVOutputFormat *stream_fmt;
        char stream_format_name[64];

        snprintf(stream_format_name, sizeof(stream_format_name), "%s_stream", fmt->name);
        stream_fmt = Avf_guess_format(stream_format_name, NULL, NULL);

        if (stream_fmt)
            fmt = stream_fmt;
    }

    return fmt;
}

AVInputFormat *Avf_find_input_format(const char *short_name)
{
    AVInputFormat *fmt;
    for(fmt = first_iformat; fmt != NULL; fmt = fmt->next) 
	{
        if (!strcmp(fmt->name, short_name))
        {
            return fmt;
        }
    }
    return NULL;
}

/* memory handling */

/**
 * Default packet destructor 
 */
static void av_destruct_packet(AVPacket *pkt)
{
    avf_free(pkt->data);
    pkt->data = NULL; pkt->size = 0;
}

/**
 * Allocate the payload of a packet and intialized its fields to default values.
 *
 * @param pkt packet
 * @param size wanted payload size
 * @return 0 if OK. AVERROR_xxx otherwise.
 */
int av_new_packet(AVPacket *pkt, int size)
{
    void *data = avf_malloc(size + FF_INPUT_BUFFER_PADDING_SIZE);
    if (!data)
        return AVERROR_NOMEM;
    memset( (char *)data + size, 0, FF_INPUT_BUFFER_PADDING_SIZE);

    av_init_packet(pkt);
    pkt->data = data; 
    pkt->size = size;
    pkt->destruct = av_destruct_packet;
    return 0;
}

void Avf_free_packet(AVPacket *pkt)
{
    if (pkt != NULL && pkt->destruct != NULL) {
        pkt->destruct(pkt);
    }        
}

/* fifo handling */

int fifo_init(FifoBuffer *f, int size)
{
    f->buffer = avf_malloc(size);
    if (!f->buffer)
        return -1;
    f->end = f->buffer + size;
    f->wptr = f->rptr = f->buffer;
    return 0;
}

void fifo_free(FifoBuffer *f)
{
    avf_free(f->buffer);
}

int fifo_size(FifoBuffer *f, uint8_t *rptr)
{
    int size;

    if (f->wptr >= rptr) {
        size = f->wptr - rptr;
    } else {
        size = (f->end - rptr) + (f->wptr - f->buffer);
    }
    return size;
}

/* get data from the fifo (return -1 if not enough data) */
int fifo_read(FifoBuffer *f, uint8_t *buf, int buf_size, uint8_t **rptr_ptr)
{
    uint8_t *rptr = *rptr_ptr;
    int size, len;

    if (f->wptr >= rptr) {
        size = f->wptr - rptr;
    } else {
        size = (f->end - rptr) + (f->wptr - f->buffer);
    }
    
    if (size < buf_size)
        return -1;
    while (buf_size > 0) {
        len = f->end - rptr;
        if (len > buf_size)
            len = buf_size;
        memcpy(buf, rptr, len);
        buf += len;
        rptr += len;
        if (rptr >= f->end)
            rptr = f->buffer;
        buf_size -= len;
    }
    *rptr_ptr = rptr;
    return 0;
}

void fifo_write(FifoBuffer *f, uint8_t *buf, int size, uint8_t **wptr_ptr)
{
    int len;
    uint8_t *wptr;
    wptr = *wptr_ptr;
    while (size > 0) {
        len = f->end - wptr;
        if (len > size)
            len = size;
        memcpy(wptr, buf, len);
        wptr += len;
        if (wptr >= f->end)
            wptr = f->buffer;
        buf += len;
        size -= len;
    }
    *wptr_ptr = wptr;
}

int filename_number_test(const char *filename)
{
    char buf[1024];
    return get_frame_filename(buf, sizeof(buf), filename, 1);
}

/* guess file format */
AVInputFormat *av_probe_input_format(AVProbeData *pd, int is_opened)
{
    AVInputFormat *fmt1, *fmt;
    int score, score_max;

    fmt = NULL;
    score_max = 0;
    for(fmt1 = first_iformat; fmt1 != NULL; fmt1 = fmt1->next) {
        if (!is_opened && !(fmt1->flags & AVFMT_NOFILE))
            continue;
        score = 0;
        if (fmt1->read_probe) {
            // Specify which error handler and custom probe callback has to be used in case of error
            pd->handle_error = fmt1->handle_error; 
            pd->custom_probe = fmt1->custom_probe; 
            score = fmt1->read_probe(pd);
        } else if (fmt1->extensions) {
            if (match_ext(pd->filename, fmt1->extensions)) {
                score = 50;
            }
        } 
        if (score > score_max) {
            score_max = score;
            fmt = fmt1;
        }
    }
    return fmt;
}

/************************************************************/
/* input media file */

/**
 * Open a media file as input. The codec are not opened. Only the file
 * header (if present) is read.
 *
 * @param ic_ptr the opened media file handle is put here
 * @param filename filename to open.
 * @param fmt if non NULL, force the file format to use
 * @param buf_size optional buffer size (zero if default is OK)
 * @param ap additionnal parameters needed when opening the file (NULL if default)
 * @return 0 if OK. AVERROR_xxx otherwise.
 */
int Avf_open_input_file(AVFormatContext **ic_ptr, 
							const char *filename, 
							AVInputFormat *fmt,
							int buf_size,
							AVFormatParameters *ap)
{
    AVFormatContext *ic = NULL;
    int err;
    char* pBuf = NULL; 
    AVProbeData probe_data, *pd = &probe_data;
    int buffer_flag = 0;
    int tmp_buf_size = 0;

    pBuf = avf_malloc(AVPROBE_BUF_SIZE);

    if (buf_size > 0)
    {
      buffer_flag = (URL_RD_BUFFERED | URL_RD_FAST_OPEN);
    }

    ic = avf_mallocz(sizeof(AVFormatContext));
    if (!ic) {
        err = AVERROR_NOMEM;
        goto fail;
    }
    pstrcpy(ic->filename, sizeof(ic->filename), filename);
    pd->filename = ic->filename;
    pd->buf = (unsigned char *)pBuf;
    pd->buf_size = 0;

    if (!fmt) {
        /* guess format if no file can be opened  */
        fmt = av_probe_input_format(pd, 0);
    }

    if (!fmt || !(fmt->flags & AVFMT_NOFILE)) {
        int flags = URL_RDONLY | buffer_flag;
        if ( ( ap != NULL ) && (ap->flags & AV_FAST_FILE_OPEN) )
        {
            flags |= URL_RD_FAST_OPEN;
        }
        /* if no file needed do not try to open one */
        if (Avf_url_fopen(&ic->pb, filename, flags) < 0) {
            err = AVERROR_IO;
            goto fail;
        }

        //if buffer size was provided, then just use it
        if (buf_size > 0)
        {
            url_setbufsize(&ic->pb, buf_size);
        }

        //let's work with a small buffer and switch to a bigger one later
        if (ic->pb.buffer_size > IO_BUFFER_SIZE_SMALL)
        {
           tmp_buf_size = ic->pb.buffer_size;
           url_setbufsize(&ic->pb, IO_BUFFER_SIZE_SMALL);
        }

        if (!fmt) {
            /* read probe data */
            pd->buf_size = get_buffer(&ic->pb, (UCHAR_PTR)pBuf, AVPROBE_BUF_SIZE);
            Avf_url_fseek(&ic->pb, 0, SEEK_SET);
        }
    }
    
    /* guess file format */
    if (!fmt) {
        fmt = av_probe_input_format(pd, 1);
    }

    /* if still no format found, error */
    if (!fmt) {
        err = AVERROR_NOFMT;
        goto fail1;
    }
        
    /* XXX: suppress this hack for redirectors */
#ifdef CONFIG_NETWORK
    if (fmt == &redir_demux) {
        err = redir_open(ic_ptr, &ic->pb);
        Avf_url_fclose(&ic->pb);
        avf_free(ic);
        avf_free(pBuf);
        return err;
    }
#endif

    ic->iformat = fmt;

    /* check filename in case of an image number is expected */
    if (ic->iformat->flags & AVFMT_NEEDNUMBER) {
        if (filename_number_test(ic->filename) < 0) { 
            err = AVERROR_NUMEXPECTED;
            goto fail1;
        }
    }
    
    /* allocate private data */
    if (fmt->priv_data_size > 0) {
        ic->priv_data = avf_mallocz(fmt->priv_data_size);
        if (!ic->priv_data) {
            err = AVERROR_NOMEM;
        goto fail1;
        }
    } else
        ic->priv_data = NULL;

    /* default pts settings is MPEG like */
    av_set_pts_info(ic, 33, 1, 90000);

    err = ic->iformat->read_header(ic, ap);
    if (err < 0)
        goto fail1;
    *ic_ptr = ic;
    avf_free(pBuf);
    
    //Cache the number of frames - we might need it later 
    Avf_get_total_present_video_frames_number(ic);
    
    if (tmp_buf_size != 0)
    {
       url_changebufsize(&ic->pb, tmp_buf_size);
    }

    return 0;
 fail1:
    if (!fmt || !(fmt->flags & AVFMT_NOFILE)) {
        Avf_url_fclose(&ic->pb);
    }
 fail:
    if (ic) {
        avf_freep(&ic->priv_data);
    }
    avf_free(ic);
    *ic_ptr = NULL;
    avf_free(pBuf);
    return err;
}


int Avf_open_output_file(ByteIOContext *s, const char *filename, int buf_size, int flags)
{
   int retVal = 0;

   retVal = Avf_url_fopen(s, filename, flags);

   if ( (0 == retVal) &&
        (buf_size > 0) )
   {
       url_setbufsize(s, buf_size);
   }

   return retVal;
}

/**
 * Read a packet from a media file
 * @param s media file handle
 * @param pkt is filled 
 * @return 0 if OK. AVERROR_xxx if error.
 */
int Avf_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    AVPacketList *pktl;

    pktl = s->packet_buffer;
    if (pktl) {
        /* read packet from packet buffer, if there is data */
        *pkt = pktl->pkt;
        s->packet_buffer = pktl->next;
        // :ZZZ: Added av_free_packet, according to an ffmpeg e-mail
        printf( "Avf_free_packet17\n" );                              
        Avf_free_packet(&pktl->pkt);
        avf_free(pktl);
        return 0;
    } else {
        return s->iformat->read_packet(s, pkt);
    }
}

/**
 * Seek to a time
 * @param s media file handle
 * @param pts - timestamp 
 * @return 0 if OK. -1 if seek could not be performed.
 */
int Avf_read_seek(AVFormatContext *s, int64_t pts)
{
    if ( !s->iformat->read_seek )
        return -1;

    return s->iformat->read_seek(s, pts);
}


/**
 * Seek by index
 * @param s media file handle
 * @param stream_id - id of the stream to seek
 * @param pos - number of index packets to skip
 * @param whence - SEEK_CURR, SEEK_START, SEEK_END,
 * @return 0 if OK. -1 if seek could not be performed.
 */
int Avf_read_seek_idx(	struct AVFormatContext *s, 
							int stream_id, 
							int pos, 
							int whence)
{
    if ( !s->iformat->read_seek_idx )
        return -1;

    return s->iformat->read_seek_idx(s, stream_id, pos, whence);
}

/**
 * Check if the given frame is Key frame
 * @param s - media file handle
 * @param stream_id - id of the stream
 * @param frame - number of frame to check
 * @return 1 if the frame is Key, 0 - otherwise.
 */
int Avf_read_is_keyframe(AVFormatContext *s, int stream_id, int frame)
{
    if ( !s->iformat->read_is_key_frame )
        return 0;

    return s->iformat->read_is_key_frame(s, stream_id, frame);
}

/* state for codec information */
#define CSTATE_NOTFOUND    0
#define CSTATE_DECODING    1
#define CSTATE_FOUND       2

static int has_codec_parameters(AVCodecContext *enc)
{
    int val;
    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        val = enc->sample_rate;
        break;
    case CODEC_TYPE_VIDEO:
        val = enc->physical_width;
        break;
    default:
        val = 1;
        break;
    }
    return (val != 0);
}

/**
 * Read the beginning of a media file to get stream information. This
 * is useful for file formats with no headers such as MPEG. This
 * function also compute the real frame rate in case of mpeg2 repeat
 * frame mode.
 *
 * @param ic media file handle
 * @return >=0 if OK. AVERROR_xxx if error.  
 */
int Avf_find_stream_info(AVFormatContext *ic)
{
    int i, count, ret, got_picture, size, read_size;
    AVCodec *codec;
    AVStream *st;
    AVPacket *pkt;
    AVFrame picture;
    AVPacketList *pktl=NULL, **ppktl;
    short samples[AVCODEC_MAX_AUDIO_FRAME_SIZE / 2];
    uint8_t *ptr;
    int min_read_size, max_read_size;

    /* typical mpeg ts rate is 40 Mbits. DVD rate is about 10
       Mbits. We read at most 0.2 second of file to find all streams */

    /* XXX: base it on stream bitrate when possible */
    if (ic->iformat == &mpegts_demux) {
        /* maximum number of bytes we accept to read to find all the streams
           in a file */
        min_read_size = 6000000;
    } else {
        min_read_size = 250000;
    }
    /* max read size is 2 seconds of video max */
    max_read_size = min_read_size * 10;

    /* set initial codec state */
    for(i=0;i<ic->nb_streams;i++) {
        st = ic->streams[i];
        if (has_codec_parameters(&st->codec))
            st->codec_info_state = CSTATE_FOUND;
        else
            st->codec_info_state = CSTATE_NOTFOUND;
        st->codec_info_nb_repeat_frames = 0;
        st->codec_info_nb_real_frames = 0;
    }

    count = 0;
    read_size = 0;
    ppktl = &ic->packet_buffer;
    for(;;) {
        /* check if one codec still needs to be handled */
        for(i=0;i<ic->nb_streams;i++) {
            st = ic->streams[i];
            if (st->codec_info_state != CSTATE_FOUND)
                break;
        }
        if (i == ic->nb_streams) {
            /* NOTE: if the format has no header, then we need to read
               some packets to get most of the streams, so we cannot
               stop here */
            if (!(ic->iformat->flags & AVFMT_NOHEADER) ||
                read_size >= min_read_size) {
                /* if we found the info for all the codecs, we can stop */
                ret = count;
                break;
            }
        } else {
            /* we did not get all the codec info, but we read too much data */
            if (read_size >= max_read_size) {
                ret = count;
                break;
            }
        }

        pktl = av_mallocz(sizeof(AVPacketList));
        if (!pktl) {
            ret = AVERROR_NOMEM;
            break;
        }

        /* add the packet in the buffered packet list */
        *ppktl = pktl;
        ppktl = &pktl->next;

        /* NOTE: a new stream can be added there if no header in file
           (AVFMT_NOHEADER) */
        pkt = &pktl->pkt;
        if (ic->iformat->read_packet(ic, pkt) < 0) {
            /* EOF or error */
            ret = -1; /* we could not have all the codec parameters before EOF */
            if ((ic->iformat->flags & AVFMT_NOHEADER) &&
                i == ic->nb_streams)
                ret = 0;
            break;
        }
        read_size += pkt->size;

        /* open new codecs */
        for(i=0;i<ic->nb_streams;i++) {
            st = ic->streams[i];
            if (st->codec_info_state == CSTATE_NOTFOUND) {
                /* set to found in case of error */
                st->codec_info_state = CSTATE_FOUND; 
                codec = avcodec_find_decoder(st->codec.codec_id);
                if (codec) {
                    if(codec->capabilities & CODEC_CAP_TRUNCATED)
                    {        
                        st->codec.flags |= CODEC_FLAG_TRUNCATED;
                    }

                    ret = avcodec_open(&st->codec, codec);
                    if (ret >= 0)
                        st->codec_info_state = CSTATE_DECODING;
                }
            }
        }

        st = ic->streams[pkt->stream_index];
        if (st->codec_info_state == CSTATE_DECODING) {
            /* decode the data and update codec parameters */
            ptr = pkt->data;
            size = pkt->size;
            while (size > 0) {
                switch(st->codec.codec_type) {
                case CODEC_TYPE_VIDEO:
                    ret = avcodec_decode_video(&st->codec, &picture, 
                                               &got_picture, ptr, size);
                    break;
                case CODEC_TYPE_AUDIO:
                    ret = avcodec_decode_audio(&st->codec, samples, 
                                               &got_picture, ptr, size);
                    break;
                default:
                    ret = -1;
                    break;
                }
                if (ret < 0) {
                    /* if error, simply ignore because another packet
                       may be OK */
                    break;
                }
                if (got_picture) {
                    /* we got the parameters - now we can stop
                       examining this stream */
                    /* XXX: add a codec info so that we can decide if
                       the codec can repeat frames */
//                    if (st->codec.codec_id == CODEC_ID_MPEG1VIDEO && 
//                        ic->iformat != &mpegts_demux &&
//                        st->codec.sub_id == 2) {
//                        /* for mpeg2 video, we want to know the real
//                           frame rate, so we decode 40 frames. In mpeg
//                           TS case we do not do it because it would be
//                           too long */
//                        st->codec_info_nb_real_frames++;
//                        st->codec_info_nb_repeat_frames += st->codec.coded_frame->repeat_pict;
//#if 0
//                        /* XXX: testing */
//                        if ((st->codec_info_nb_real_frames % 24) == 23) {
//                            st->codec_info_nb_repeat_frames += 2;
//                        }
//#endif
//                        /* stop after 40 frames */
//                        if (st->codec_info_nb_real_frames >= 40) {
//                            av_reduce(
//                                &st->r_frame_rate,
//                                &st->r_frame_rate_base,
//                                (int64_t)st->codec.frame_rate * st->codec_info_nb_real_frames,
//                                st->codec_info_nb_real_frames + (st->codec_info_nb_repeat_frames >> 1),
//                                1<<30);
//                            goto close_codec;
//                        }
//                    } else {
//                    close_codec:
                        st->codec_info_state = CSTATE_FOUND;
                        avcodec_close(&st->codec);
                        break;
//                    }
                }
                ptr += ret;
                size -= ret;
            }
        }
        count++;
    }

    /* close each codec if there are opened */
    for(i=0;i<ic->nb_streams;i++) {
        st = ic->streams[i];
        if (st->codec_info_state == CSTATE_DECODING)
            avcodec_close(&st->codec);
    }

    /* set real frame rate info */
    for(i=0;i<ic->nb_streams;i++) {
        st = ic->streams[i];
        if (st->codec.codec_type == CODEC_TYPE_VIDEO) {
            if (!st->r_frame_rate){
                st->r_frame_rate      = st->codec.frame_rate;
                st->r_frame_rate_base = st->codec.frame_rate_base;
            }
        }
    }

    return ret;
}

/**
 * Close a media file (but not its codecs)
 *
 * @param s media file handle
 */
void Avf_close_input_file(AVFormatContext *s)
{
    int i;

    if (s->iformat->read_close)
        s->iformat->read_close(s);
    for(i=0;i<s->nb_streams;i++) 
	{
        avf_free(s->streams[i]);
    }
    if (s->packet_buffer) 
	{
        AVPacketList *p, *p1;
        p = s->packet_buffer;
        while (p != NULL) 
		{
            p1 = p->next;
            printf( "Avf_free_packet17\n" );                                  
            Avf_free_packet(&p->pkt);
            avf_free(p);
            p = p1;
        }
        s->packet_buffer = NULL;
    }
    if (!(s->iformat->flags & AVFMT_NOFILE)) 
	{
        Avf_url_fclose(&s->pb);
    }
    avf_freep(&s->priv_data);
    avf_free(s);
}

/**
 * Add a new stream to a media file. Can only be called in the
 * read_header function. If the flag AVFMT_NOHEADER is in the format
 * description, then new streams can be added in read_packet too.
 *
 *
 * @param s media file handle
 * @param id file format dependent stream id
 */
AVStream* Avf_new_stream(AVFormatContext *s, int id)
{
    AVStream *st;
    
    if (s->nb_streams >= MAX_STREAMS)
        return NULL;

    st = avf_mallocz(sizeof(AVStream));
    if (!st)
        return NULL;

    AvApi_avcodec_get_context_defaults(&st->codec);

    st->index = s->nb_streams;
    st->id = id;
    s->streams[s->nb_streams++] = st;

    return st;
}

/************************************************************/
/* output media file */

int Avf_set_parameters(	AV_PARAMID iParamId, 	
							AVFormatContext *s, 
							AVFormatParameters *ap, 
							void *pval)
{
    switch ( iParamId )
    {
        case AV_PARAMID_CONTEXT:
        {
            if (s->oformat->priv_data_size > 0) {
                s->priv_data = avf_mallocz(s->oformat->priv_data_size);
                if (!s->priv_data)
                    return AVERROR_NOMEM;
            } else
                s->priv_data = NULL;

            // If customized stream content information is present, copy it
            // Title
            if ( custom_stream_info.title != NULL )
            {
                strcpy( s->title, custom_stream_info.title );
            }

            // Author
            if ( custom_stream_info.author != NULL )
            {
                strcpy( s->author, custom_stream_info.author );
            }

            // Copyright
            if ( custom_stream_info.copyright != NULL )
            {
                strcpy( s->copyright, custom_stream_info.copyright );
            }

            // Comments
            if ( custom_stream_info.comment != NULL )
            {
                strcpy( s->comment, custom_stream_info.comment );
            }
    
            if (s->oformat->set_parameters) 
            {
                int ret = s->oformat->set_parameters(s, ap, NULL);

                if( ret < 0 )
                    return ret;
            }

            break;
        }

        case AV_PARAMID_RESET_IO:
        {
            AvfAvioSetLastError( 0 );
            break;
        }

    	case AV_PARAMID_MOV_ATOM_MODIFY:
		{
			if ( s->oformat->set_parameters ) 
            {
				s->oformat->set_parameters(s, NULL, pval);
			}
			break;
		}

        case AV_PARAMID_SET_CONTEXT_FLAGS:
        {
            s->flags = *(int *)pval ;
            break;
        }

        default:
            return -1;
    }

    return 0;
}

/************************************************************/
/* output media file */

int Avf_get_parameters(	AV_PARAMID iParamId, 
						AVFormatContext *s, 
						AVFormatParameters *ap, 
						void *pval)
{
    switch ( iParamId )
    {
        case AV_PARAMID_CURRENT_POS:
        {
            ByteIOContext *bs = &s->pb;
            offset_t *offset1 = (offset_t *) pval;
            offset_t offset2 = 0;

            if ( bs->write_flag )
            {
                offset2 = AV_SEEK_CURR_POS_WRITE(bs); //bs->pos + (bs->buf_ptr - bs->buffer);
            }else
            {
                offset2 = AV_SEEK_CURR_POS_READ(bs); //bs->pos - (bs->buf_end - bs->buffer) + (bs->buf_ptr - bs->buffer);
            }
            *offset1 = offset2;
            break;
        }

        default:
            return -1;
    }

    return 0;
}

/**
 * allocate the stream private data and write the stream header to an
 * output media file
 *
 * @param s media file handle
 * @return 0 if OK. AVERROR_xxx if error.  
 */
int Avf_write_header(AVFormatContext *s)
{
    int i;
    AVStream *st;

    /* default pts settings is MPEG like */
    av_set_pts_info(s, 33, 1, 90000);

    s->oformat->write_header(s);

    if (AvfAvioGetLastError() < 0)
    {
        return AvfAvioGetLastError();
    }

    /* init PTS generation */
    for(i=0;i<s->nb_streams;i++) {
        st = s->streams[i];

        switch (st->codec.codec_type) {
        case CODEC_TYPE_AUDIO:
            if ( st->codec.codec_id == CODEC_ID_WMAV8Z )
            {
                av_frac_init(&st->pts, 0, 0, 
                         (int64_t)s->pts_num * st->codec.bit_rate);
            }
            else
            {
                av_frac_init(&st->pts, 0, 0, 
                         (int64_t)s->pts_num * st->codec.sample_rate);
            }
            break;
        case CODEC_TYPE_VIDEO:
            av_frac_init(&st->pts, 0, 0, 
                         (int64_t)s->pts_num * st->codec.frame_rate);
            break;
        default:
            break;
        }
    }
    return 0;
}

/**
 * Write a packet to an output media file. The packet shall contain
 * one audio or video frame.
 *
 * @param s media file handle
 * @param stream_index stream index
 * @param buf buffer containing the frame data
 * @param size size of buffer
 * @return < 0 if error, = 0 if OK, 1 if end of stream wanted.
 */
// :ZZZ: Added key_frame argument
int Avf_write_frame(AVFormatContext *s, int stream_index, const uint8_t *buf, 
                   int size, int key_frame)
{
    AVStream *st;
    int64_t pts_mask;
    int frame_size;
    int packet_size;

    st = s->streams[stream_index];
    pts_mask = (1LL << s->pts_wrap_bits) - 1;
    // :ZZZ: Update key_frame
    s->key_frame = key_frame;
    if( st->codec.codec_type != CODEC_TYPE_TEXT )
    {
        s->oformat->write_packet(s, stream_index, (uint8_t *)buf, size, 
                                   st->pts.val & pts_mask);
    }
    else
    {
        s->oformat->write_packet( s, stream_index, (uint8_t *)buf, size, 
                                   st->text_sample_id );
    }

    if (AvfAvioGetLastError() < 0)
    {
        return AvfAvioGetLastError();
    }

    /* update pts */
    switch (st->codec.codec_type) {
    case CODEC_TYPE_AUDIO:
        if (st->codec.frame_size <= 1) {
            frame_size = size / st->codec.channels;
            /* specific hack for pcm codecs because no frame size is provided */
            switch(st->codec.codec_id) {
            case CODEC_ID_PCM_S16LE:
            case CODEC_ID_PCM_S16BE:
            case CODEC_ID_PCM_U16LE:
            case CODEC_ID_PCM_U16BE:
                frame_size >>= 1;
                break;
            default:
                break;
            }
        } else {
            frame_size = st->codec.frame_size;
        }
        if ( st->codec.codec_id == CODEC_ID_WMAV8Z )
        {
            packet_size = st->codec.frame_size * st->codec.bit_rate / st->codec.sample_rate;
            av_frac_add(&st->pts, 
                        (int64_t)s->pts_den * packet_size * 8);
        }
        else
        {
            av_frac_add(&st->pts, 
                        (int64_t)s->pts_den * frame_size);
        }
        break;
    case CODEC_TYPE_VIDEO:
        av_frac_add(&st->pts, 
                    (int64_t)s->pts_den * st->codec.frame_rate_base);
        break;
    default:
        break;
    }

    return AvfAvioGetLastError();
}

int Avf_estimate_trailer_size(AVFormatContext *s)
{
    if ( s->oformat->estimate_trailer_size )
        return s->oformat->estimate_trailer_size(s);
    else
        return 0;
}

/**
 * write the stream trailer to an output media file and and free the
 * file private data.
 *
 * @param s media file handle
 * @return 0 if OK. AVERROR_xxx if error.  */
int Avf_write_trailer(AVFormatContext *s)
{
    s->oformat->write_trailer(s);
    avf_freep(&s->priv_data);

    return AvfAvioGetLastError();
}

/************************************************************/
/* Parsing clip trailer infos                                */
/************************************************************/
int Avf_parse_trailer_info(AVFormatContext *s_in, AVFormatContext *s_out)
{
	int i;

	s_out->nb_streams = s_in->nb_streams;
    
    strcpy( s_out->filename,  s_in->filename);
    strcpy( s_out->title,     s_in->title);
    strcpy( s_out->author,    s_in->author);
    strcpy( s_out->copyright, s_in->copyright);
    strcpy( s_out->comment,   s_in->comment);

//    s_out->flags     = 0; // currently not in use
    s_out->key_frame = 0;  // will be updated when writing frame

    // currently is not used
    s_out->packet_buffer = s_in->packet_buffer;

    // to check what happens if there is thumbnail
    s_out->pbyThumbnailImage    = NULL;
    s_out->ulThumbnailImageSize = 0;

    s_out->ulVendorName     = s_in->ulVendorName;
    s_out->uwModel          = s_in->uwModel;
    s_out->uwRevision       = s_in->uwRevision;

    // 
    s_out->pbyUserAtom      = s_in->pbyUserAtom;
    s_out->pbySkipAtom      = NULL;//s_in->pbySkipAtom;
    s_out->pbyCustomAtom    = NULL;//s_in->pbyCustomAtom;

    s_out->ulCreTimeStamp        = s_in->ulCreTimeStamp;
    s_out->ulModTimeStamp        = s_in->ulModTimeStamp;
    s_out->ulTotalVideoFrames    = s_in->ulTotalVideoFrames;

    if( ( s_out->flags & AVFMT_MODIFY_APPEND ) == 0 )
    {
        s_out->bModifyMode          = TRUE;
    }
    s_out->ulUpdatedFileSize    = 0;
    s_out->ulMuteEndAudioBytes  = 0;
    s_out->iAudioSilencePattern = 0;

    // Init stream settings
	for( i = 0; i < s_out->nb_streams; i++ )
    {        
        s_out->streams[i] = avf_mallocz(sizeof(AVStream));

        // all the codak pointers are also copied,
        // so the input codak destroy operation must NOT release pointers
        *(s_out->streams[i]) = *(s_in->streams[i]);

        s_out->streams[i]->pts.den = 0;
        s_out->streams[i]->pts.num = 0;
        s_out->streams[i]->pts.val = 0;
        
        if( s_out->streams[i]->codec.extradata_size > 0 )
        {
            s_out->streams[i]->codec.extradata =
                        avf_mallocz( s_out->streams[i]->codec.extradata_size );

            memcpy( s_out->streams[i]->codec.extradata,
                    s_in->streams[i]->codec.extradata,
                    s_out->streams[i]->codec.extradata_size );
        }
    }
    	
	// Reset IO in format level
    Avf_set_parameters( AV_PARAMID_RESET_IO, NULL, NULL, NULL );
    
    // Init. priv_data
    Avf_set_parameters( AV_PARAMID_CONTEXT, s_out, NULL, NULL ); 

    if( ( s_out->flags & AVFMT_MODIFY_APPEND ) == 0 )
    {
        Avf_write_header( s_out );
    }
	
	if( s_out->oformat->parse_private_context != NULL )
    {
        s_out->oformat->parse_private_context( s_out, s_in );
    }
		
    return 0;
}

/************************************************************/
/* Delete the last chunk in movie                           */
/************************************************************/
int Avf_delete_packet( AVFormatContext *s , unsigned long ulFlags )
{
	if( s->oformat->delete_packet != NULL )
    {
        s->oformat->delete_packet( s, ulFlags );
    }
		
    return 0;	
}


void avfcodec_string(char *buf, int buf_size, AVCodecContext *enc, int encode)
{
    const char *codec_name;
    AVCodec *p;
    char buf1[32];
    char channels_str[100];
    int bitrate = 0;

    if (encode)
	{
        p = AvApi_avcodec_find_encoder(enc->codec_id);
	}
    else
	{
        p = AvApi_avcodec_find_decoder(enc->codec_id);
	}

    if (p) {
        codec_name = p->name;
    } else if (enc->codec_name[0] != '\0') {
        codec_name = enc->codec_name;
    } else {
        /* output avi tags */
        if (enc->codec_type == CODEC_TYPE_VIDEO) {
            snprintf(buf1, sizeof(buf1), "%c%c%c%c", 
                     enc->codec_tag & 0xff,
                     (enc->codec_tag >> 8) & 0xff,
                     (enc->codec_tag >> 16) & 0xff,
                     (enc->codec_tag >> 24) & 0xff);
        } else {
            snprintf(buf1, sizeof(buf1), "0x%04x", enc->codec_tag);
        }
        codec_name = buf1;
    }

    switch(enc->codec_type) {
    case CODEC_TYPE_VIDEO:
        snprintf(buf, buf_size,
                 "Video: %s",
                 codec_name );
        if (enc->physical_width) {
            snprintf(buf + strlen(buf), buf_size - strlen(buf),
                     ", %dx%d, %0.2f fps",
                     enc->physical_width, enc->physical_height, 
                     (float)enc->frame_rate / enc->frame_rate_base);
        }
        bitrate = enc->bit_rate;
        break;
    case CODEC_TYPE_AUDIO:
        snprintf(buf, buf_size,
                 "Audio: %s",
                 codec_name);
        switch (enc->channels) {
            case 1:
                strcpy(channels_str, "mono");
                break;
            case 2:
                strcpy(channels_str, "stereo");
                break;
            case 6:
                strcpy(channels_str, "5:1");
                break;
            default:
                sprintf(channels_str, "%d channels", enc->channels);
                break;
        }
        if (enc->sample_rate) {
            snprintf(buf + strlen(buf), buf_size - strlen(buf),
                     ", %d Hz, %s",
                     enc->sample_rate,
                     channels_str);
        }
        
        /* for PCM codecs, compute bitrate directly */
        switch(enc->codec_id) {
        default:
            bitrate = enc->bit_rate;
            break;
        }
        break;
    default:
        av_abort();
    }
    if (bitrate != 0) {
        snprintf(buf + strlen(buf), buf_size - strlen(buf), 
                 ", %d kb/s", bitrate / 1000);
    }
}

/* "user interface" functions */

void Avf_dump_format(	AVFormatContext *ic,
					int index, 
					const char *url,
					int is_output)
{
    int i, flags;
    char buf[256];

    fprintf(stderr, "%s #%d, %s, %s '%s':\n", 
            is_output ? "Output" : "Input",
            index, 
            is_output ? ic->oformat->name : ic->iformat->name, 
            is_output ? "to" : "from", url);
    for(i=0;i<ic->nb_streams;i++) {
        AVStream *st = ic->streams[i];
        avfcodec_string(buf, sizeof(buf), &st->codec, is_output);
        fprintf(stderr, "  Stream #%d.%d", index, i);
        /* the pid is an important information, so we display it */
        /* XXX: add a generic system */
        if (is_output)
            flags = ic->oformat->flags;
        else
            flags = ic->iformat->flags;
        if (flags & AVFMT_SHOW_IDS) {
            fprintf(stderr, "[0x%x]", st->id);
        }
        fprintf(stderr, ": %s\n", buf);
    }
}

typedef struct {
    const char *str;
    int width, height;
} SizeEntry;

static SizeEntry sizes[] = {
    { "sqcif", 128, 96 },
    { "qcif", 176, 144 },
    { "cif", 352, 288 },
    { "4cif", 704, 576 },
};
    
int parse_image_size(int *width_ptr, int *height_ptr, const char *str)
{
    int i;
    int n = sizeof(sizes) / sizeof(SizeEntry);
    const char *p;
    int frame_width = 0, frame_height = 0;

    for(i=0;i<n;i++) {
        if (!strcmp(sizes[i].str, str)) {
            frame_width = sizes[i].width;
            frame_height = sizes[i].height;
            break;
        }
    }
    if (i == n) {
        p = str;
        frame_width = strtol(p, (char **)&p, 10);
        if (*p)
            p++;
        frame_height = strtol(p, (char **)&p, 10);
    }
    if (frame_width <= 0 || frame_height <= 0)
        return -1;
    *width_ptr = frame_width;
    *height_ptr = frame_height;
    return 0;
}

#if !defined(__CC_ARM) /* unsupported by ARM */
int64_t av_gettime(void)
{
#ifdef CONFIG_WIN32
    struct _timeb tb;
    _ftime(&tb);
    return ((int64_t)tb.time * int64_t_C(1000) + (int64_t)tb.millitm) * int64_t_C(1000);
#else
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return (int64_t)tv.tv_sec * 1000000 + tv.tv_usec;
#endif
}
#endif /* __CC_ARM */

static time_t mktimegm(struct tm *tm)
{
    time_t t;

    int y = tm->tm_year + 1900, m = tm->tm_mon + 1, d = tm->tm_mday;

    if (m < 3) {
        m += 12;
        y--;
    }

    t = 86400 * 
        (d + (153 * m - 457) / 5 + 365 * y + y / 4 - y / 100 + y / 400 - 719469);

    t += 3600 * tm->tm_hour + 60 * tm->tm_min + tm->tm_sec;

    return t;
}

/* Syntax:
 * - If not a duration:
 *  [{YYYY-MM-DD|YYYYMMDD}]{T| }{HH[:MM[:SS[.m...]]][Z]|HH[MM[SS[.m...]]][Z]}
 * Time is localtime unless Z is suffixed to the end. In this case GMT
 * Return the date in micro seconds since 1970 
 * - If duration:
 *  HH[:MM[:SS[.m...]]]
 *  S+[.m...]
 */
int64_t parse_date(const char *datestr, int duration)
{
    const char *p;
    int64_t t;
    struct tm dt;
    unsigned int i;
    static const char *date_fmt[] = {
        "%Y-%m-%d",
        "%Y%m%d",
    };
    static const char *time_fmt[] = {
        "%H:%M:%S",
        "%H%M%S",
    };
    const char *q;
    int is_utc, len;
    char lastch;
    time_t now = time(0);

    len = strlen(datestr);
    if (len > 0)
        lastch = datestr[len - 1];
    else
        lastch = '\0';
    is_utc = (lastch == 'z' || lastch == 'Z');

    memset(&dt, 0, sizeof(dt));

    p = datestr;
    q = NULL;
    if (duration == 0) {
        for (i = 0; i < sizeof(date_fmt) / sizeof(date_fmt[0]); i++) {
            q = strptime(p, date_fmt[i], &dt);
            if (q) {
                break;
            }
        }

        if (!q) {
            if (is_utc) {
                dt = *gmtime(&now);
            } else {
                dt = *localtime(&now);
            }
            dt.tm_hour = dt.tm_min = dt.tm_sec = 0;
        } else {
            p = q;
        }

        if (*p == 'T' || *p == 't' || *p == ' ')
            p++;

        for (i = 0; i < sizeof(time_fmt) / sizeof(time_fmt[0]); i++) {
            q = strptime(p, time_fmt[i], &dt);
            if (q) {
                break;
            }
        }
    } else {
        q = strptime(p, time_fmt[0], &dt);
        if (!q) {
            dt.tm_sec = strtol(p, (char **)&q, 10);
            dt.tm_min = 0;
            dt.tm_hour = 0;
        }
    }

    /* Now we have all the fields that we can get */
    if (!q) {
        if (duration)
            return 0;
        else
            return now * int64_t_C(1000000);
    }

    if (duration) {
        t = dt.tm_hour * 3600 + dt.tm_min * 60 + dt.tm_sec;
    } else {
        dt.tm_isdst = -1;       /* unknown */
        if (is_utc) {
            t = mktimegm(&dt);
        } else {
            t = mktime(&dt);
        }
    }

    t *= 1000000;

    if (*q == '.') {
        int val, n;
        q++;
        for (val = 0, n = 100000; n >= 1; n /= 10, q++) {
            if (!isdigit(*q)) 
                break;
            val += n * (*q - '0');
        }
        t += val;
    }
    return t;
}

/* syntax: '?tag1=val1&tag2=val2...'. Little URL decoding is done. Return
   1 if found */
int find_info_tag(char *arg, int arg_size, const char *tag1, const char *info)
{
    const char *p;
    char tag[128], *q;

    p = info;
    if (*p == '?')
        p++;
    for(;;) {
        q = tag;
        while (*p != '\0' && *p != '=' && *p != '&') {
            if ((q - tag) < (char)(sizeof(tag) - 1))
                *q++ = *p;
            p++;
        }
        *q = '\0';
        q = arg;
        if (*p == '=') {
            p++;
            while (*p != '&' && *p != '\0') {
                if ((q - arg) < arg_size - 1) {
                    if (*p == '+')
                        *q++ = ' ';
                    else
                        *q++ = *p;
                }
                p++;
            }
            *q = '\0';
        }
        if (!strcmp(tag, tag1)) 
            return 1;
        if (*p != '&')
            break;
        p++;
    }
    return 0;
}

/* Return in 'buf' the path with '%d' replaced by number. Also handles
   the '%0nd' format where 'n' is the total number of digits and
   '%%'. Return 0 if OK, and -1 if format error */
int get_frame_filename(char *buf, int buf_size,
                       const char *path, int number)
{
    const char *p;
    char *q, buf1[20];
    int nd, len, c, percentd_found;

    q = buf;
    p = path;
    percentd_found = 0;
    for(;;) {
        c = *p++;
        if (c == '\0')
            break;
        if (c == '%') {
            do {
                nd = 0;
                while (isdigit(*p)) {
                    nd = nd * 10 + *p++ - '0';
                }
                c = *p++;
                if (c == '*' && nd > 0) {
                    // The nd field is actually the modulus
                    number = number % nd;
                    c = *p++;
                    nd = 0;
                }
            } while (isdigit(c));

            switch(c) {
            case '%':
                goto addchar;
            case 'd':
                if (percentd_found)
                    goto fail;
                percentd_found = 1;
                snprintf(buf1, sizeof(buf1), "%0*d", nd, number);
                len = strlen(buf1);
                if ((q - buf + len) > buf_size - 1)
                    goto fail;
                memcpy(q, buf1, len);
                q += len;
                break;
            default:
                goto fail;
            }
        } else {
        addchar:
            if ((q - buf) < buf_size - 1)
                *q++ = c;
        }
    }
    if (!percentd_found)
        goto fail;
    *q = '\0';
    return 0;
 fail:
    *q = '\0';
    return -1;
}

/**
 *
 * Print on stdout a nice hexa dump of a buffer
 * @param buf buffer
 * @param size buffer size
 */
void av_hex_dump(uint8_t *buf, int size)
{
    int len, i, j, c;

    for(i=0;i<size;i+=16) {
        len = size - i;
        if (len > 16)
            len = 16;
        printf("%08x ", i);
        for(j=0;j<16;j++) {
            if (j < len)
                printf(" %02x", buf[i+j]);
            else
                printf("   ");
        }
        printf(" ");
        for(j=0;j<len;j++) {
            c = buf[i+j];
            if (c < ' ' || c > '~')
                c = '.';
            printf("%c", c);
        }
        printf("\n");
    }
}

void url_split(char *proto, int proto_size,
               char *hostname, int hostname_size,
               int *port_ptr,
               char *path, int path_size,
               const char *url)
{
    const char *p;
    char *q;
    int port;

    port = -1;

    p = url;
    q = proto;
    while (*p != ':' && *p != '\0') {
        if ((q - proto) < proto_size - 1)
            *q++ = *p;
        p++;
    }
    if (proto_size > 0)
        *q = '\0';
    if (*p == '\0') {
        if (proto_size > 0)
            proto[0] = '\0';
        if (hostname_size > 0)
            hostname[0] = '\0';
        p = url;
    } else {
        p++;
        if (*p == '/')
            p++;
        if (*p == '/')
            p++;
        q = hostname;
        while (*p != ':' && *p != '/' && *p != '?' && *p != '\0') {
            if ((q - hostname) < hostname_size - 1)
                *q++ = *p;
            p++;
        }
        if (hostname_size > 0)
            *q = '\0';
        if (*p == ':') {
            p++;
            port = strtoul(p, (char **)&p, 10);
        }
    }
    if (port_ptr)
        *port_ptr = port;
    pstrcpy(path, path_size, p);
}

/**
 * Set the pts for a given stream
 * @param s stream 
 * @param pts_wrap_bits number of bits effectively used by the pts
 *        (used for wrap control, 33 is the value for MPEG) 
 * @param pts_num numerator to convert to seconds (MPEG: 1) 
 * @param pts_den denominator to convert to seconds (MPEG: 90000)
 */
void av_set_pts_info(AVFormatContext *s, int pts_wrap_bits,
                     int pts_num, int pts_den)
{
    s->pts_wrap_bits = pts_wrap_bits;
    s->pts_num = pts_num;
    s->pts_den = pts_den;
}

/* fraction handling */

/**
 * f = val + (num / den) + 0.5. 'num' is normalized so that it is such
 * as 0 <= num < den.
 *
 * @param f fractional number
 * @param val integer value
 * @param num must be >= 0
 * @param den must be >= 1 
 */
void av_frac_init(AVFrac *f, int64_t val, int64_t num, int64_t den)
{
    num += (den >> 1);
    if (num >= den) {
        val += num / den;
        num = num % den;
    }
    f->val = val;
    f->num = num;
    f->den = den;
}

/* set f to (val + 0.5) */
void av_frac_set(AVFrac *f, int64_t val)
{
    f->val = val;
    f->num = f->den >> 1;
}

/**
 * Fractionnal addition to f: f = f + (incr / f->den)
 *
 * @param f fractional number
 * @param incr increment, can be positive or negative
 */
void av_frac_add(AVFrac *f, int64_t incr)
{
    int64_t num, den;

    num = f->num + incr;
    den = f->den;
    if (num < 0) {
        f->val += num / den;
        num = num % den;
        if (num < 0) {
            num += den;
            f->val--;
        }
    } else if (num >= den) {
        f->val += num / den;
        num = num % den;
    }
    f->num = num;
}

/**
 * register a new image format
 * @param img_fmt Image format descriptor
 */
void Avf_register_image_format(AVImageFormat *img_fmt)
{
    AVImageFormat **p;

    p = &first_image_format;
    while (*p != NULL) p = &(*p)->next;
    *p = img_fmt;
    img_fmt->next = NULL;
}

/* guess image format */
AVImageFormat *av_probe_image_format(AVProbeData *pd)
{
    AVImageFormat *fmt1, *fmt;
    int score, score_max;

    fmt = NULL;
    score_max = 0;
    for(fmt1 = first_image_format; fmt1 != NULL; fmt1 = fmt1->next) {
        if (fmt1->img_probe) {
            score = fmt1->img_probe(pd);
            if (score > score_max) {
                score_max = score;
                fmt = fmt1;
            }
        }
    }
    return fmt;
}

AVImageFormat *guess_image_format(const char *filename)
{
    AVImageFormat *fmt1;

    for(fmt1 = first_image_format; fmt1 != NULL; fmt1 = fmt1->next) {
        if (fmt1->extensions && match_ext(filename, fmt1->extensions))
            return fmt1;
    }
    return NULL;
}

/**
 * Read an image from a stream. 
 * @param gb byte stream containing the image
 * @param fmt image format, NULL if probing is required
 */
int av_read_image(ByteIOContext *pb, const char *filename,
                  AVImageFormat *fmt,
                  int (*alloc_cb)(void *, AVImageInfo *info), void *opaque)
{
    char buf[AVPROBE_BUF_SIZE];
    AVProbeData probe_data, *pd = &probe_data;
    offset_t pos;
    int ret;

    if (!fmt) {
        pd->filename = filename;
        pd->buf = (unsigned char *)buf;
        pos = url_ftell(pb);
        pd->buf_size = get_buffer(pb, (UCHAR_PTR)buf, AVPROBE_BUF_SIZE);
        Avf_url_fseek(pb, pos, SEEK_SET);
        fmt = av_probe_image_format(pd);
    }
    if (!fmt)
        return AVERROR_NOFMT;
    ret = fmt->img_read(pb, alloc_cb, opaque);
    return ret;
}

/**
 * Write an image to a stream.
 * @param pb byte stream for the image output
 * @param fmt image format
 * @param img image data and informations
 */
int av_write_image(ByteIOContext *pb, AVImageFormat *fmt, AVImageInfo *img)
{
    return fmt->img_write(pb, img);
}

int Avf_get_total_written_video_frames_number(AVFormatContext *s)
{
    int f_num=-1;

    if (s!=NULL &&
        s->nb_streams > 0 && 
        s->streams[0]!= NULL && 
        s->streams[0]->codec.codec_type == CODEC_TYPE_VIDEO)
    {
        f_num =  s->streams[0]->stream_total_nb_frames;
    }

    if (s!=NULL && 
        s->nb_streams > 1 && 
        s->streams[1]!= NULL && 
        s->streams[1]->codec.codec_type == CODEC_TYPE_VIDEO)
    {
        f_num =  s->streams[1]->stream_total_nb_frames;
    }
   
    return f_num;
}
        
int Avf_get_total_present_video_frames_number(AVFormatContext *s)
{
    static int f_num=-1;

    if (s!=NULL)                         
    {
        f_num = s->ulTotalVideoFrames;
    }
                                              
    return f_num;
}


/**
 * Set customized stream content description information. 
 * Fields that can be set  are title, author, copyright & comments.
 * These form a part of the Content Description object in ASF header object.  
 * May not be relevant for other formats.
 *
 * @param field - Content field to be set
 * @param pcBuffer - Buffer to set the chosen content field
 * @param uiSize - Size of buffer
 *
 * @return - TRUE if success
 *         - FALSE if failure
 */
BOOL Avf_set_stream_content_info( 	CONTENT_INFO_TYPE  field, 
                                 	PCHAR              pcBuffer, 
                                 	UINT               uiSize )
{    
    // Pointer to custom content field selected to be set 
    PCHAR *content_field_chosen = NULL;

    if ( uiSize > 512 )
    {
        printf("Custom content description field cannot be longer than 512 chars\n");
        return FALSE;  
    }

    switch( field )
    {
        case CONTENT_INFO_TITLE:
        {
            content_field_chosen = &(custom_stream_info.title); 
            break;
        }

        case CONTENT_INFO_AUTHOR:
        {
            content_field_chosen = &(custom_stream_info.author);
            break;  
        }

        case CONTENT_INFO_COPYRIGHT :
        {
            content_field_chosen = &(custom_stream_info.copyright);
            break;
        }

        case CONTENT_INFO_COMMENTS : 
        {
            content_field_chosen = &(custom_stream_info.comment);
            break;
        }
        default:
            return FALSE;
    }
    
    // Allocate memory to hold custom field information
    *content_field_chosen = avf_realloc( *content_field_chosen, uiSize );

    // Copy given custom string over selected field
    strcpy( *content_field_chosen, pcBuffer );

    return TRUE;
}

/**
 * Set customized tag content information. 
 *
 * @return - TRUE if success
 *         - FALSE if failure
 */
BOOL Avf_set_custom_tag_info( PAVCustomTagContent psTagInfo )
{    
	switch( psTagInfo->uiOpType )
	{
		case AV_CUSTOM_TAG_OP_DEL:
		{
			// we now only support this op for stss atom
			if( psTagInfo->uiTagId != AV_CUSTOM_TAG_ID_STSS )
			{
				break;
			}

			memcpy( (PBYTE)&custom_tag_content[psTagInfo->uiTagId],
				    psTagInfo,
				    sizeof(AVCustomTagContent) );
			break;
		}	

		case AV_CUSTOM_TAG_OP_REGISTER_CB:
		{
			// we now only support this op for skip and custom atom
			if( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_SKIP )
			{
				pbCustomSkipAtomCbFuncPtr = (PBYTE)psTagInfo->pvData;
			}
			else if( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_CUSTOM )
			{
				pbCustomAtomCbFuncPtr = (PBYTE)psTagInfo->pvData;
			}
			break;
		}
		
		case AV_CUSTOM_TAG_OP_INITIAL_VAL:
		{
			// we now only support this op for skip and custom atom
			if ( ( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_SKIP   ) ||
			     ( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_CUSTOM ) )
			{
    			memcpy( (PBYTE)&custom_tag_content[psTagInfo->uiTagId],
    				    psTagInfo,
    				    sizeof(AVCustomTagContent) );
			}
			break;
		}

	}

    return TRUE;
}

/**
 * Set customized tag content information. 
 *
 * @return - TRUE if success
 *         - FALSE if failure
 */
BOOL Avf_get_custom_tag_info( PAVCustomTagContent psTagInfo )
{    
	// if the uiTagId that holds in custom_tag_content[] is not corresponding 
	// to the request, means that this atom is not configured. Return false
	// and do nothing
	if( custom_tag_content[psTagInfo->uiTagId].uiTagId != psTagInfo->uiTagId )
	{
		return FALSE;
	}
			
	switch( psTagInfo->uiOpType )
	{
		case AV_CUSTOM_TAG_OP_DEL:
		{
			// we now only support this op for stss atom
			if( psTagInfo->uiTagId != AV_CUSTOM_TAG_ID_STSS )
			{
				return FALSE;
			}
			memcpy( psTagInfo,
					(PBYTE)&custom_tag_content[psTagInfo->uiTagId],
			    	sizeof(AVCustomTagContent) );
			break;
		}	

		case AV_CUSTOM_TAG_OP_REGISTER_CB:
		{
			// we now only support this op for skip ant custom atom
			if( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_SKIP )
			{
                psTagInfo->pvData = (PVOID)pbCustomSkipAtomCbFuncPtr;
			}
			else if( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_CUSTOM )
			{
                psTagInfo->pvData = (PVOID)pbCustomAtomCbFuncPtr;
			}
			else
			{
			    return FALSE;
			}

			break;
		}	
		
		case AV_CUSTOM_TAG_OP_INITIAL_VAL:
		{
			// we now only support this op for skip and custom atom
			if ( ( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_SKIP   ) ||
			     ( psTagInfo->uiTagId == AV_CUSTOM_TAG_ID_CUSTOM ) )
			{
    			memcpy( psTagInfo,
    					(PBYTE)&custom_tag_content[psTagInfo->uiTagId],
    			    	sizeof(AVCustomTagContent) );
			}
			else
			{
				return FALSE;
			}

			break;
		}	

	}

    return TRUE;
}

int Avf_add_stream( AVFormatContext *s, AVCodecContext* psCodec )
{
    AVStream *st = Avf_new_stream( s, s->nb_streams );

    if( psCodec != 0 )
    {
        st->codec = *psCodec;
    } 
    
    if( s->oformat->add_stream != NULL )
    {
        s->oformat->add_stream( s, st->id );
    }

    return 0;
}

int Avf_seek_last_chunk_end_pos(AVFormatContext *s)
{
    return 0;
}


int Avf_get_stream_frames_number( AVFormatContext* s, int stream_id )
{    
    return s->streams[stream_id]->stream_total_nb_frames;
}


int Avf_read_text_stream( 	AVFormatContext* s,
							int first,
							int total,
							char** text,
							int* indexes,
							int stream_id )                          
{
    int pos = url_ftell( &s->pb );

    if( s->streams[stream_id]->codec.codec_type != CODEC_TYPE_TEXT )
    {
        return 1;
    }

    if( s->iformat != NULL && s->iformat->read_text_stream != NULL )
    {
        s->iformat->read_text_stream( s, first, total, text, indexes, stream_id );
    }
    else if( s->oformat != NULL && s->oformat->read_text_stream != NULL )
    {
        s->oformat->read_text_stream( s, first, total, text, indexes, stream_id );
    }
    
    Avf_url_fseek(&s->pb, pos, SEEK_SET);

    return 0;
}

int Avf_video_to_text( AVFormatContext* s, int video, int text_stream )
{
    if( s->oformat->video_to_text != NULL )
    {
        return s->oformat->video_to_text( s, video, text_stream );
    }
    return -1;
}
