/*
 * Buffered file io for ffmpeg system
 * Copyright (c) 2001 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef BUILD_VERILOG 

#include "avformat.h"
#if !defined(__CC_ARM) /* unsupported by ARM */
#include <fcntl.h>
//#include <io.h>
#endif /* __CC_ARM */
#include "Types.h"
#include "Allocator.h"
#include "Register.h"

#define NOA_DEBUG_PRINT
/* standard verilog protocol */

PBYTE pbyVerilogBuffer = NULL;
UINT  uiVerilogSize = 0;
UINT  uiCurOffset = 0;
UINT  uiMaxOffset = 0;

UINT uiFileSize;
extern BYTE byRaw[];

STATIC UINT uiFlags;

static int verilog_open(URLContext *h, const char *filename, int flags)
{
    uiFlags = flags;

    if ( flags & URL_APPEND )
    {
        printf ("verilog_open with  URL_APPEND\n");
        return 0;
    }

    if ( flags == URL_RDONLY )
    {
        printf ("verilog_open with  URL_RDONLY\n");

        // Read
        pbyVerilogBuffer = byRaw;
        uiVerilogSize    = uiFileSize;
        uiMaxOffset      = uiVerilogSize;

        uiCurOffset = 0;
    }
    else if ( flags & URL_WRONLY )
    {
        printf ("verilog_open with  URL_WRONLY\n");

        // Write
        uiVerilogSize = 0x100000;
        uiMaxOffset   = 0;

        uiCurOffset = 0;

        PoolFree( NULL, pbyVerilogBuffer );

        pbyVerilogBuffer = PoolAlloc( NULL, uiVerilogSize, POOL_MEMORY_CACHED );        
    }
                
    printf( "Verilog open flags %x\n", flags );
    printf( "Verilog buffer 0x%x\n", pbyVerilogBuffer );

    return 0;
}

static int verilog_read(URLContext *h, unsigned char *buf, int size)
{
    UINT i;
    UINT uiSize = size;

    for ( i = 0; ( i < uiSize ) && ( uiCurOffset < uiVerilogSize ); i++ )
    {
        buf[i] = pbyVerilogBuffer[uiCurOffset];
        uiCurOffset++;
    }

#ifdef NOA_DEBUG_PRINT
    printf("uiVerilogSize = %d size = %d\n",uiVerilogSize,size);
    printf( "Verilog read %d , i = %d\n", uiCurOffset , i );
#endif

    if ( i < uiSize )
    {
        printf( "verilog_read - End of buffer reached\n" );
    }

    uiMaxOffset = MAX( uiMaxOffset, uiCurOffset );  

    return i;
}

static int verilog_write(URLContext *h, unsigned char *buf, int size)
{
    UINT i;
    UINT uiSize = size;

    printf("verilog_write: Size %d, CurrOffset %d, VerilogSize %d \n", 
            uiSize, uiCurOffset, uiVerilogSize );

    for ( i = 0; ( i < uiSize ) && ( uiCurOffset < uiVerilogSize ); i++ )
    {
        pbyVerilogBuffer[uiCurOffset] = buf[i];
        uiCurOffset++;
    }

    // Check that the verilog buffer is big enough.
    assert( uiCurOffset < uiVerilogSize );

#ifdef NOA_DEBUG_PRINT
    printf( "Verilog write %d\n", uiCurOffset );
#endif

    if ( i < uiSize )
    {
        printf( "verilog_write - End of buffer reached !!!\n" );
    }

    uiMaxOffset = MAX( uiMaxOffset, uiCurOffset );  

    return i;
}

static offset_t verilog_seek(URLContext *h, offset_t pos, int whence)
{
    printf("verilog_seek: pos %d, whence %d ", (ULONG)pos, whence );
    printf("CurrOffset %d ==> ", uiCurOffset );

    switch ( whence )
    {
        case SEEK_SET:
        {
            // Beginning of file
            uiCurOffset = pos;
            break;
        }

        case SEEK_CUR:
        {
            // Current position of file pointer
            uiCurOffset += pos;
            break;
        }

        case SEEK_END:
        {
            // End of file
            uiCurOffset = uiMaxOffset - pos;
            break;
        }

        default:
        {
            printf( "verilog_seek unknown option %d\n", whence );
        }
    }

    printf("%d\n", uiCurOffset );
 
#ifdef NOA_DEBUG_PRINT
    printf( "Verilog seek %d\n", uiCurOffset );
#endif

    uiMaxOffset = MAX( uiMaxOffset, uiCurOffset );  

    return uiCurOffset;
}

char hexa_chars[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };

static int verilog_close(URLContext *h)
{
    UINT i;
    PBYTE pbyCur = pbyVerilogBuffer;

    printf( "verilog_close curr %d max %d\n", uiCurOffset, uiMaxOffset );

    if ( ( uiFlags & URL_WRONLY ) != 0 )
    {        
        // Go over all reached area
        for ( i = 0; i < uiMaxOffset; i++ )
        {
            // Is this a new line?
            if ( i % 16 == 0 )
            {
    #ifdef BUILD_VERILOG
                WRITE_REG( VERILOG_ADDR_MESSAGE_4, 'V' );
                WRITE_REG( VERILOG_ADDR_MESSAGE_4, ':' );
                WRITE_REG( VERILOG_ADDR_MESSAGE_4, ' ' );
    #else
                printf("V: ");
    #endif
            }

    #ifdef BUILD_VERILOG
            // Write the char in hexa digits            
            WRITE_REG( VERILOG_ADDR_MESSAGE_4, hexa_chars[ (*pbyCur>>4)&0xF ] );
            WRITE_REG( VERILOG_ADDR_MESSAGE_4, hexa_chars[ (*pbyCur   )&0xF ] );
            WRITE_REG( VERILOG_ADDR_MESSAGE_4, ' ' );
    #else
            printf("%c%c ", (BYTE)hexa_chars[ (*pbyCur>>4)&0xF ], (BYTE)hexa_chars[ (*pbyCur   )&0xF ] );
    #endif

            pbyCur++;

            // Is this the last char in line?
            if ( i % 16 == 15 )
            {        
    #ifdef BUILD_VERILOG
                // Add carriage return for the verilog system to recognize
                // a new line, and clear the 100-byte buffer
                WRITE_REG( VERILOG_ADDR_MESSAGE_4, '\r' );
    #else
                printf("\n" );
    #endif
            }
        }
    }        

#ifdef BUILD_VERILOG
    // Add carriage return for the verilog system to recognize
    // a new line, and clear the 100-byte buffer
    WRITE_REG( VERILOG_ADDR_MESSAGE_4, '\r' );
#else
    printf("\n" );
#endif

    printf( "Verilog closed\n" );

    return 0;
}

URLProtocol verilog_protocol = {
    "verilog",
    verilog_open,
    verilog_read,
    verilog_write,
    verilog_seek,
    verilog_close,
};

#endif /* BUILD_VERILOG */



